import datetime
import logging
import os
import sys
import typing

if typing.TYPE_CHECKING:
    from submit2gm.worker_job import WorkerJob
    from PySide2 import QtWidgets

# Return codes
RETURN_CODE_SUCCESS = 0
RETURN_CODE_FAILED = 1

SUBMIT2GM_ALREADY_RUNNING_MSG = "{0} is already running, redirecting request to existing server..."
SUBMIT2GM_NOT_ALREADY_RUNNING_MSG = "{0} is not already running, creating a new {0} server..."
SUBMIT2GM_NOT_ALREADY_RUNNING_MSG_2 = "{0} is not already running, processing command..."

# Create logging folder if id does not already exist
LOGGING_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logs')
LOGGING_FILE = os.path.join(LOGGING_DIR, datetime.datetime.now().strftime("%Y_%m_%d-%I_%M_%S_%p.log"))

if not os.path.exists(LOGGING_DIR):
    os.makedirs(LOGGING_DIR, exist_ok=True)

logging.basicConfig(filename=LOGGING_FILE,
                    level=logging.DEBUG,
                    format='%(asctime)s - p%(process)s - %(threadName)s - {%(pathname)s:%(lineno)d} - %(levelname)s - %(message)s',
                    datefmt='%m/%d/%Y %I:%M:%S %p')

SUBMIT2GM_PWY_LOGGER = logging.getLogger('submit2gm-pyw')
SUBMIT2GM_PWY_LOGGER.setLevel(logging.INFO)

# set up logging to stdout
__stream_handler = logging.StreamHandler(sys.stdout)
__stream_handler.setLevel(logging.INFO)
__stream_handler.setFormatter(logging.Formatter('%(message)s'))
SUBMIT2GM_PWY_LOGGER.addHandler(__stream_handler)


def _setup_libs() -> None:
    """ Make sure the submit2gm module and all lib dependencies are available for import and correctly added to sys.path """

    if getattr(sys, 'frozen', False):
        # we are running in a bundle
        bundle_dir = sys._MEIPASS
    else:
        # we are running in a normal Python environment
        bundle_dir = os.path.dirname(os.path.abspath(__file__))

    # Set sys.path so modules can be imported
    parent_folder = bundle_dir
    libs_folder = os.path.join(parent_folder, 'libs')

    # Add the parent folder so that the submit2gm module can be found
    if parent_folder not in sys.path:
        sys.path.insert(0, parent_folder)

    # Add the libs folder so that any modules in libs can be found (PySide2, shiboken2, etc...)
    if libs_folder not in sys.path:
        sys.path.insert(0, libs_folder)


def _create_application_instance() -> 'QtWidgets.QApplication':
    from PySide2 import QtCore
    from PySide2 import QtWidgets

    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

    app = QtWidgets.QApplication([])
    app.setOrganizationName("GridMarkets")
    app.setOrganizationDomain("gridmarkets.com")
    app.setApplicationName("Submit2gm")
    return app


def main() -> int:
    from submit2gm import constants
    from submit2gm.job_processor import JobProcessor
    from submit2gm.worker_job import WorkerJob

    # Add a stream handler when not using the daemon
    stream_logging_handler = logging.StreamHandler()
    stream_log_format = '%(message)s'
    stream_formatter = logging.Formatter(stream_log_format)
    stream_logging_handler.setFormatter(stream_formatter)
    stream_logging_handler.setLevel(logging.INFO)
    stream_logging_handler.addFilter(logging.Filter(constants.CLI_LOGGER_ID))
    logging.getLogger().addHandler(stream_logging_handler)

    from PySide2 import QtWidgets
    if not QtWidgets.QApplication.instance():
        _create_application_instance()

    job_processor = JobProcessor()
    job = WorkerJob(sys.argv)
    job_processor.process(job)

    return RETURN_CODE_SUCCESS


if __name__ == '__main__':
    _setup_libs()

    try:
        main()
        SUBMIT2GM_PWY_LOGGER.info("Submit2gm stopped gracefully.")
    except Exception as e:
        # Catch and log any unhandled exceptions
        SUBMIT2GM_PWY_LOGGER.exception(str(e), exc_info=True)
        sys.exit(RETURN_CODE_FAILED)
