__all__ = 'APIClient'

import PySide2
from PySide2 import QtCore

import json
import pathlib
import platform
import sys
import typing

from . import constants, errors, utils
from .cached_value import CachedValue
from .attribute_types import EnumItem
from .machine_option import MachineOption
from .plugin_accessor import PluginAccessor
from .product import Product
from .user import User
from .user_info import UserInfo
from .xml_api_schema_parser import XMLAPISchemaParser
from .errors import *

import gridmarkets
from gridmarkets.http_client import HttpClient
from gridmarkets.errors import AuthenticationError, APIError, InvalidRequestError, InsufficientCreditsError
from gridmarkets.version import VERSION as ENVOY_PYTHON_CLIENT_VERSION

if typing.TYPE_CHECKING:
    from .api_schema import APISchema
    from .dynamic_job_preset import DynamicJobPreset
    from .attribute_types import EnumAttributeType
    from .factory_collection import FactoryCollection
    from .logging_coordinator import LoggingCoordinator
    from .project import Project
    from .packed_project import PackedProject
    from .submit2gm_plugin import Submit2gmPlugin
    from .static_job_preset import StaticJobPreset
    from .widgets.node_editor.node_widget import DependencyNode


class EnvoyProject:
    def __init__(self,
                 name: str,
                 storage: str,
                 deleting: bool,
                 num_files: int,
                 total_size: int,
                 auto_downloading: bool,
                 watched_list):
        self._name = name
        self._storage = storage
        self._deleting = deleting
        self._num_files = num_files
        self._total_size = total_size
        self._auto_downloading = auto_downloading
        self._watched_list = watched_list

    def get_name(self) -> str:
        return self._name

    def get_storage(self) -> str:
        return self._storage

    def is_deleting(self) -> bool:
        return self._deleting

    def get_file_count(self) -> int:
        return self._num_files

    def get_total_size(self) -> int:
        return self._total_size

    def is_auto_downloading(self) -> bool:
        return self._auto_downloading

    def get_watched_list(self):
        return self._watched_list


class EnvoyFile:
    def __init__(self, check_sum, last_modified, name, key, size, fmt_name):
        self._check_sum = check_sum
        self._last_modified = last_modified
        self._name = name
        self._key = key
        self._size = size
        self._fmt_name = fmt_name

    def get_check_sum(self):
        return self._check_sum

    def get_last_modified(self):
        return self._last_modified

    def get_name(self):
        return self._name

    def get_key(self):
        return self._key

    def get_size(self):
        return self._size

    def get_fmt_name(self):
        return self._fmt_name


class APIClient(PluginAccessor, QtCore.QObject):

    onRemoteProjectsChanged = QtCore.Signal()

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 logging_coordinator: 'LoggingCoordinator',
                 factory_collection: 'FactoryCollection'):

        PluginAccessor.__init__(self, plugin)
        QtCore.QObject.__init__(self)
        self._modified_api_cache = False
        self._signed_in_user = None

        self._factory_collection = factory_collection
        self._signed_in = False
        self._envoy_client = None
        self._log = logging_coordinator.get_logger(u"GridMarketsAPIClient")

        # used for caching projects
        self._api_schema_cache = CachedValue['APISchema']()
        self._remote_projects_cache = CachedValue[typing.List['EnvoyProject']]([])
        self._project_files_dictionary_cache = {}
        self._product_resolver = CachedValue['gridmarkets.Resolver'](None)
        self._products_cache = CachedValue([])
        self._cached_machine_options = CachedValue({})
        self._cached_user_info = CachedValue['UserInfo']()

    def clear_all_caches(self) -> None:
        self._log.info(u"Clearing cached API data")
        self._remote_projects_cache.reset_and_clear_cache()
        self._project_files_dictionary_cache.clear()
        self._product_resolver.reset_and_clear_cache()
        self._products_cache.reset_and_clear_cache()
        self._cached_machine_options.reset_and_clear_cache()
        self._cached_user_info.reset_and_clear_cache()

    def sign_in(self, user: 'User', skip_validation: 'bool' = False) -> None:
        self._log.info(u"Signing in as " + user.get_auth_email())

        if not skip_validation:
            self.validate_credentials(user.get_auth_email())

        self._signed_in_user = user

        # add the user to the saved profiles list if not already added
        user_container = self.get_plugin().get_preferences_container().get_user_container()
        profiles = user_container.get_all()

        for profile in profiles:
            if profile.get_auth_email() == user.get_auth_email():
                return None

        user_container.append(user)

        self._envoy_client = gridmarkets.EnvoyClient()

        # refresh caches
        self.clear_all_caches()
        self.get_root_directories(ignore_cache=True)
        self.get_api_schema(ignore_cache=True)
        self.get_user_info(ignore_cache=True)
        self._signed_in = True

    def sign_in_with_envoy(self) -> None:
        self._log.info("Signing in with envoy")
        http_client = HttpClient()

        method = "get"
        url = gridmarkets.envoy_client.API_BASE + "/auth"

        try:
            resp = http_client.request(method, url)
        except gridmarkets.errors.APIError as e:
            e_msg = e.message

            # pretty error message for common errors
            if e_msg.startswith(u"Unexpected error communicating with Envoy.Please start Envoy"):
                e_msg = u"Unable to connect to Envoy."

            raise errors.APIClientError(message=e_msg)
        except Exception as e:
            raise errors.APIClientError(message=str(e))

        # Check that the user is signed into Envoy
        try:
            content = resp.json()
        except json.decoder.JSONDecodeError:
            raise errors.NotSignedInError(message=u"Please sign-in to Envoy.")

        user_name = content['Username']

        if not user_name:
            raise errors.InvalidUserError(u"Please sign-in to Envoy.")

        # create user
        user = User(user_name)

        self.sign_in(user, skip_validation=True)

    def sign_out(self) -> None:
        if self.is_user_signed_in():
            self._log.info(u"Signing out user " + self.get_signed_in_user().get_auth_email())

        self._signed_in_user = None
        self._signed_in = False

    def is_user_signed_in(self) -> bool:
        return self._signed_in

    def connected(self) -> bool:
        raise NotImplementedError

    def get_signed_in_user(self) -> typing.Optional['User']:
        if self.is_user_signed_in():
            return self._signed_in_user
        return None

    def get_cached_api_schema(self) -> typing.Optional['APISchema']:
        return self._api_schema_cache.get_value()

    def get_api_schema(self,
                       factory_collection: 'FactoryCollection' = None,
                       ignore_cache: bool = False) -> 'APISchema':

        self._log.info(u"Getting api schema. ignore_cache=" + str(ignore_cache))

        if factory_collection is None:
            factory_collection = self._factory_collection

        if ignore_cache or not self._api_schema_cache.is_cached():
            self._log.info(u"Fetching api schema from XML source.")

            try:
                api_schema = XMLAPISchemaParser.parse(self.get_plugin(), factory_collection)
                self._api_schema_cache.set_value(api_schema)
            except Exception as e:
                msg = u"Could not fetch api schema." + utils.get_exception_with_traceback(e)
                self._log.error(msg, exc_info=True)
                raise errors.APIClientError(msg)

        api_schema = self._api_schema_cache.get_value()

        if constants.JOB_DEFINITION_ID_FLAG in sys.argv and (not self._modified_api_cache or ignore_cache):
            self._modified_api_cache = True

        return api_schema

    @staticmethod
    def validate_credentials(auth_email: str) -> bool:

        # check an email address has been entered
        has_email = isinstance(auth_email, str) and len(auth_email) > 0

        if has_email:
            try:
                client = gridmarkets.EnvoyClient()
                client.validate_auth()
            except AuthenticationError as e:
                raise InvalidUserError(message=str(e))
            except APIError as e:
                e_msg = str(e)

                # pretty error message for common errors
                if e_msg.startswith(u"Unexpected error communicating with Envoy.Please start Envoy"):
                    e_msg = u"Unable to connect to Envoy."

                raise APIClientError(message=e_msg)

        elif not has_email:
            message = u"No GridMarkets email address provided."
            raise InvalidEmailError(message=message)

        return True

    def upload_project(self,
                       packed_project: 'PackedProject',
                       submission_name: str,
                       submission_comment: str,
                       delete_local_files_after_upload: bool = False) -> None:

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to upload a project.")

        epc_project = self.get_epc_project(packed_project, submission_name, submission_comment)

        try:
            self._generate_summary_contents(epc_project)
            self._envoy_client.upload_project_files(epc_project)
        except InsufficientCreditsError as e:
            raise CreditsError(str(e))
        except APIError as e:
            raise APIClientError(str(e))

        self._log.info(u"Uploaded project")

    def _get_gm_job(self, static_job_preset: 'StaticJobPreset') -> 'gridmarkets.Job':
        self._log.info(u"Calculating Job parameters...")

        job = gridmarkets.Job(**static_job_preset.get_job_settings())

        # Handle credits budget attributes
        if (static_job_preset.contains_job_attribute_with_id(constants.API_KEYS.CREDITS_BUDGET) and
                static_job_preset.contains_job_attribute_with_id(constants.API_KEYS.CREDITS_BUDGET_ACTION)):

            credits_budget_attribute = static_job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.CREDITS_BUDGET)
            credits_budget_action_attribute = static_job_preset.get_job_preset_attribute_by_key(
                constants.API_KEYS.CREDITS_BUDGET_ACTION)

            if credits_budget_attribute.has_value() and credits_budget_action_attribute.has_value():
                credits_budget = credits_budget_attribute.get_value()
                credits_budget_action = credits_budget_action_attribute.get_value()

                if credits_budget_action == constants.JOB_ACTIONS.CREDITS_BUDGET_ACTION_ALERT:
                    credits_budget_action = gridmarkets.Job.CREDITS_BUDGET_ACTION_ALERT
                elif credits_budget_action == constants.JOB_ACTIONS.CREDITS_BUDGET_ACTION_STOP:
                    credits_budget_action = gridmarkets.Job.CREDITS_BUDGET_ACTION_STOP
                elif credits_budget_action == constants.JOB_ACTIONS.CREDITS_BUDGET_ACTION_SUSPEND:
                    credits_budget_action = gridmarkets.Job.CREDITS_BUDGET_ACTION_SUSPEND
                else:
                    raise Exception("{} is not a known job action key".format(credits_budget_action))

                job.set_credits_budget(credits_budget, credits_budget_action)

        # Check the app is supported
        if job.app not in self.get_product_app_types():
            msg = "App \"" + job.app + "\" is not supported for your account. Please contact GridMarkets support for " \
                                       "details."
            self._log.error(msg)
            raise UnsupportedApplicationError(msg)

        # Check the version is supported
        if job.app_version not in self.get_product_versions(job.app):
            msg = "Version \"" + job.app_version + "\" for app_type \"" + job.app + "\" is not supported for your " \
                  "account. Please contact GridMarkets support for details."

            self._log.error(msg)
            raise UnsupportedApplicationError(msg)

        return job

    def _set_watch_patterns(self, project: 'Project', gm_project: 'gridmarkets.Project') -> None:

        watch_patterns = project.get_watch_patterns()

        # The python envoy clients skip_auto_download attribute determines if any watch files are set for the request
        # If we have no watch files then do not attempt to download anything
        gm_project.skip_auto_download = len(watch_patterns) <= 0

        self._log.info(u"Setting watch patterns...")
        for watch_pattern in watch_patterns:
            self._log.info(u"{0} : {1}".format(watch_pattern.get_remote_watch_pattern(),
                                               watch_pattern.get_local_download_path()))

            gm_project.add_watch_files(gridmarkets.WatchFile(watch_pattern.get_remote_watch_pattern(),
                                                             watch_pattern.get_local_download_path()))

        gm_project.remote_output_folder_name = project.get_name()
        gm_project.remote_output_folder = project.get_name()

    def _add_files_to_upload(self, gm_project: 'gridmarkets.Project', packed_project: 'PackedProject') -> None:
        for project_file in packed_project.get_files():
            file_to_upload = str(project_file)

            self._log.info(u"Marking \"" + file_to_upload + u"\" for upload.")
            gm_project.add_files(file_to_upload)

    def _get_gm_project(self,
                        project: 'Project',
                        submission_name: str,
                        submission_comment: str) -> 'gridmarkets.Project':

        project_name = project.get_name()
        project_dir = str(project.get_root_dir())

        plugin = self.get_plugin()

        sys_info = {
            'os': platform.platform(),
            'envoy_python_client_version': ENVOY_PYTHON_CLIENT_VERSION,
            'submit2gm_version': plugin.get_version().to_string(),
            'pyside2_version': PySide2.__version__,
        }

        try:
            cps = plugin.get_active_command_port_server()
            sys_info['cps_info'] = {
                'address': cps.get_address(),
                'product': cps.get_product(),
                'meta_info': cps.get_meta_info()
            }
        except ValueError:
            pass

        return gridmarkets.Project(project_dir,
                                   project_name,
                                   submission_name=submission_name,
                                   submission_comment=submission_comment,
                                   sys_info=sys_info)

    def _generate_summary_contents(self, gm_project: 'gridmarkets.Project') -> None:
        summary = """
PROJECT NAME:\t\t""" + gm_project.name + """ 
SUBMISSION NAME:\t""" + str(gm_project.submission_name) + """

PYTHON ENVOY CLIENT VERSION:\t""" + gridmarkets.version.VERSION + """

SYS INFO:
""" + self.get_plugin().get_plugin_utils().get_sys_info() + """
"""

        self._log.debug(json.dumps(gm_project.serialize, indent=3))

        gm_project.params['summary'] = summary

    def get_epc_project(self,
                        packed_project: 'Project',
                        submission_name: str,
                        submission_comment: str,
                        static_job_presets: typing.Union[typing.List['StaticJobPreset'], None] = None,
                        dependency_graph: typing.Union[typing.Dict[str, 'DependencyNode'], None] = None) -> 'gridmarkets.Project':
        """
        Given a Submit2gm PackedProject and an optional StaticJobPreset return am Envoy Python Client Project.
        """

        gm_project = self._get_gm_project(packed_project, submission_name, submission_comment)

        self._add_files_to_upload(gm_project, packed_project)

        if static_job_presets:
            # Create gm jobs
            gm_jobs = dict()
            for static_job_preset in static_job_presets:
                job = self._get_gm_job(static_job_preset)
                gm_jobs[static_job_preset.get_id()] = job

                # add job to project
                gm_project.add_jobs(job)

            # Setup Dependencies
            for static_job_preset in static_job_presets:
                job = gm_jobs[static_job_preset.get_id()]
                dependencies = dependency_graph[static_job_preset.get_id()]
                dependant_job_preset_ids = dependencies.get_inputs()

                for dependant_job_preset_id in dependant_job_preset_ids:
                    job.add_dependencies(gm_jobs[dependant_job_preset_id])

            # Finally set watch patterns
            self._set_watch_patterns(packed_project, gm_project)

        return gm_project

    def submit_new_project(self,
                           packed_project: 'PackedProject',
                           submission_name: str,
                           submission_comment: str,
                           dynamic_job_presets: typing.List['DynamicJobPreset'],
                           dependency_graph: typing.Dict[str, 'DependencyNode'],
                           delete_local_files_after_upload: bool = False) -> None:

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to Submit a job.")

        self._log.info(u"Preparing to submit current scene as project \"" + packed_project.get_name() +
                       u"\" with " + str(len(dynamic_job_presets)) + u" jobs.")

        # Convert the dynamic job preset to a static job preset
        static_job_presets = []
        for dynamic_job_preset in dynamic_job_presets:
            static_job_preset = dynamic_job_preset.get_static_job_preset(packed_project)
            static_job_presets.append(static_job_preset)

        epc_project = self.get_epc_project(packed_project,
                                           submission_name,
                                           submission_comment,
                                           static_job_presets=static_job_presets,
                                           dependency_graph=dependency_graph)

        # submit the job to the remote project
        self._log.info("Submitting current scene with job \"" + epc_project.jobs[0].name + "\" to project \"" +
                       packed_project.get_name() + "\".")

        try:
            self._generate_summary_contents(epc_project)
            self._envoy_client.submit_project(epc_project, skip_upload=False)

        except InvalidRequestError as e:
            e_msg = "Submission Failed.\n\n{}".format(utils.format_invalid_request_error(e))
            raise APIClientError(e_msg)
        except InsufficientCreditsError as e:
            print("InsufficientCreditsError", e)
            raise CreditsError(str(e))
        except APIError as e:
            print("APIError", e)
            raise APIClientError(str(e))

        self._log.info(u"Job submitted - See Render Manager for details")

    def submit_to_remote_project(self,
                                 remote_project: 'Project',
                                 submission_name: str,
                                 submission_comment: str,
                                 static_job_presets: typing.List['StaticJobPreset'],
                                 dependency_graph: typing.Dict[str, 'DependencyNode']) -> None:

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to Submit a job.")

        self._log.info(u"Preparing " + str(len(static_job_presets)) + u" jobs for submission to remote project " +
                       str(remote_project.get_name()))

        epc_project = self.get_epc_project(remote_project,
                                           submission_name,
                                           submission_comment,
                                           static_job_presets=static_job_presets,
                                           dependency_graph = dependency_graph)

        # submit the job to the remote project
        self._log.info(u"Submitting " + str(len(static_job_presets)) + u" jobs to existing remote project " +
                       str(remote_project.get_name()))

        try:
            self._generate_summary_contents(epc_project)
            self._envoy_client.submit_project(epc_project, skip_upload=True)

        except InvalidRequestError as e:
            e_msg = "Submission Failed.\n\n{}".format(utils.format_invalid_request_error(e))
            raise APIClientError(e_msg)
        except InsufficientCreditsError as e:
            raise CreditsError(str(e))
        except APIError as e:
            raise APIClientError(str(e))

        self._log.info(u"Job submitted - See Render Manager for details")

    def get_cached_root_directories(self) -> typing.List['str']:
        return list(map(lambda x: x.get_name(), self._remote_projects_cache.get_value()))

    def get_root_directories(self, ignore_cache: bool = False) -> typing.List['str']:
        self._log.info(u"Getting root directories. ignore_cache=" + str(ignore_cache))

        if ignore_cache or not self._remote_projects_cache.is_cached():
            self.get_remote_projects(ignore_cache=True)

        return self.get_cached_root_directories()

    def get_remote_projects(self, ignore_cache: bool = False) -> typing.List['EnvoyProject']:
        self._log.info(u"Getting Envoy Projects. ignore_cache=" + str(ignore_cache))

        if ignore_cache or not self._remote_projects_cache.is_cached():
            url = self._envoy_client.url + u'/projects'
            self._log.info(u"Fetching existing remote project directories from \"" + url + u"\"")

            try:
                import requests
                r = requests.get(url)
                json = r.json()
                projects = json.get(u"projects", [])

                envoy_projects = []

                for project in projects:
                    envoy_projects.append(EnvoyProject(project.get('name', ''),
                                                       project.get('storage', ''),
                                                       project.get('deleting', False),
                                                       project.get('num_files', 0),
                                                       project.get('total_size', 0),
                                                       project.get('auto_downloading', False),
                                                       project.get('watched_list', [])))

                projects = sorted(envoy_projects, key=lambda s: s.get_name().lower())

            except Exception as e:
                self._log.error(u"Could not make request to envoy API. " + str(e))
                self.sign_out()
                projects = []

            self._remote_projects_cache.set_value(projects)
            self.onRemoteProjectsChanged.emit()

        remote_projects = self._remote_projects_cache.get_value()
        self._log.info(u"Returning " + str(len(remote_projects)) + u" Envoy Project(s).")
        return remote_projects

    def delete_remote_project(self, project_name: str) -> None:
        envoy_projects = self.get_remote_projects()

        for envoy_project in envoy_projects:
            if envoy_project.get_name() == project_name:
                storage = envoy_project.get_storage()
                break
        else:
            raise ValueError("Project '{}' not found".format(project_name))

        body = {'storage': storage}

        url = '{0}/project/{1}/delete'.format(self._envoy_client.url, project_name)

        headers = {'content-type': 'application/json'}

        try:
            resp = self._envoy_client.http_client.request('post', url, headers, post_data=body)
        except Exception as e:
            raise gridmarkets.errors.APIError(str(e))
        else:
            if 200 <= resp.status_code < 300:
                self._log.info("Project '{}' deleted.".format(project_name))

            elif resp.status_code == 404:
                raise APIClientError("404: {0} not found".format(url))

            elif resp.status_code == 400:
                raise APIClientError("Invalid request: {}".format(resp.text))

            elif resp.status_code in (500, 503):
                raise APIClientError("{0} {1}".format(resp.status_code, resp.text))

    def get_product_versions(self, product: str, ignore_cache: bool = False) -> typing.List[str]:
        if ignore_cache or not self._product_resolver.is_cached():
            # get product resolver
            resolver = self._envoy_client.get_product_resolver()

            # cache all products
            self._product_resolver.set_value(resolver.get_all_types())

        # sorted by latest version to oldest version
        return sorted(self._product_resolver.get_value().get(product, {}).get(u"versions", []),
                      key=lambda s: s.lower(), reverse=True)

    def get_closest_matching_product_version(self,
                                             product: str,
                                             product_version: str,
                                             ignore_cache: bool = False) -> typing.Optional[str]:
        product_versions = self.get_product_versions(product, ignore_cache=ignore_cache)

        if not len(product_versions):
            return None

        # first check to see if the provided version exactly matches a version option
        for version_option in product_versions:
            if version_option == product_version:
                return version_option

        # if not then we are down to heuristics
        import difflib
        return str(difflib.get_close_matches(product_version, product_versions)[0])

    def get_remote_project_files(self, project_name: str, ignore_cache: bool = False) -> typing.List['EnvoyFile']:
        """ Returns a list of Envoy File objects, if project does not exist it returns an empty list. """

        if ignore_cache or project_name not in self._project_files_dictionary_cache:
            import requests

            try:
                r = requests.get(self._envoy_client.url + u'/project/' + project_name + u'/Files')
            except Exception as e:
                raise APIClientError(str(e))

            json = r.json()
            all_files = json.get(u'project_files', {}).get(u'all_files', [])

            envoy_files = []
            for file in all_files:
                envoy_files.append(EnvoyFile(
                    file.get(u'CheckSum', u''),
                    file.get(u'LastModified', u''),
                    file.get(u'Name', u''),
                    file.get(u'Key', u''),
                    file.get(u'Size', 0),
                    file.get(u'FmtName', u''),
                ))

            self._project_files_dictionary_cache[project_name] = envoy_files

        return self._project_files_dictionary_cache[project_name]

    def clear_project_files_cache(self) -> None:
        self._project_files_dictionary_cache.clear()

    def get_products(self, ignore_cache: bool = False) -> typing.List['Product']:
        self._log.info(u"Getting products list...")

        if not self.is_user_signed_in():
            msg = u"Must be signed-in to get products list."
            self._log.error(msg)
            raise NotSignedInError(msg)

        if ignore_cache or not self._products_cache.is_cached():
            url = u'{0}/products'.format(self._envoy_client.url)

            self._log.info(u"Fetching products list from " + url)

            try:
                resp = self._envoy_client.http_client.request('get', url)
            except Exception as e:
                msg = u"Could not fetch products list: " + utils.get_exception_with_traceback(e)
                self._log.error(msg)
                raise APIClientError(msg)
            else:
                if resp.status_code == 200:
                    content = resp.json()

                    def parse_product(json):
                        return Product(json.get(u'id', -1),
                                       json.get(u'name', u''),
                                       json.get(u'app_type', u''),
                                       json.get(u'version', u''))

                    self._log.info(u"Caching fetched products...")
                    self._products_cache.set_value(list(map(lambda x: parse_product(x), content)))

                if resp.status_code == 404:
                    msg = u"404: {0} not found".format(url)
                    self._log.error(msg)
                    raise APIClientError(msg)

        cached_products = self._products_cache.get_value()
        self._log.info(u"Returning " + str(len(cached_products)) + u" cached products...")
        return cached_products

    def get_products_with_app_type(self, app_type: str, ignore_cache: bool = False) -> typing.List['Product']:
        products = self.get_products(ignore_cache=ignore_cache)
        return list(filter(lambda x: x.get_app_type() == app_type, products))

    def get_product_app_types(self, ignore_cache: bool = False) -> typing.List['Product']:
        return sorted(list(set(map(lambda x: x.get_app_type(), self.get_products(ignore_cache=ignore_cache)))),
                      key=lambda s: s.lower(), reverse=True)

    def get_user_info(self, ignore_cache: bool = False) -> 'UserInfo':
        self._log.info(u"Getting User info. ignore_cache=" + str(ignore_cache))

        if ignore_cache or not self._cached_user_info.is_cached():
            self._log.info(u"Fetching user info")

            try:
                user_info_raw = self._envoy_client.get_user_info()

                business_verticals = user_info_raw.get('business_verticals', [])
                created = user_info_raw.get('created', '')
                credits_available = user_info_raw.get('credits_available', 0)
                credits_consumed = user_info_raw.get('credits_consumed', 0)
                email_address = user_info_raw.get('email_address', '')
                job_throughput_limit = user_info_raw.get('job_throughput_limit', 30)
                max_throughput = user_info_raw.get('max_throughput', 9999)
                name = user_info_raw.get('name', '')
                plugin_versions = user_info_raw.get('plugin_versions', {})
                service_plan = user_info_raw.get('service_plan', '')
                username = user_info_raw.get('username', '')

                user_info = UserInfo(business_verticals,
                                     created,
                                     credits_available,
                                     credits_consumed,
                                     email_address,
                                     job_throughput_limit,
                                     max_throughput,
                                     name,
                                     plugin_versions,
                                     service_plan,
                                     username)

                self._log.info(u"Caching user info...")
                self._cached_user_info.set_value(user_info)

            except Exception as e:
                msg = u"Could not fetch user info: " + utils.get_exception_with_traceback(e)
                self._log.error(msg)
                raise APIClientError(msg)

        self._log.info(u"Returning User info.")
        return self._cached_user_info.get_value()

    def get_machines(self,
                     app: str,
                     operation: str,
                     plugins: typing.List[str] = None,
                     ignore_cache: bool = False) -> typing.List['MachineOption']:

        cached_machine_options = self._cached_machine_options.get_value()

        if plugins is None:
            plugins = []

        _app_cache_id = app + ''.join(plugins)

        if _app_cache_id in cached_machine_options:
            app_machine_options = cached_machine_options[_app_cache_id]
            if operation in app_machine_options:
                operation_machine_options = app_machine_options[operation]

                if not ignore_cache:
                    return operation_machine_options

        try:
            self._log.info(u"Fetching machine types for product \"" + app + u"\" (operation=\"" + operation +
                           u"\" plugins=[" + ','.join(plugins) + u"])")

            machine_options = self._envoy_client.get_machines(app,
                                                              operation,
                                                              gridmarkets.EnvoyClient.MACHINE_CATEGORY_ALL,
                                                              *plugins)

        except APIError as e:
            msg = str(e)
            self._log.error(msg)
            raise APIClientError(message=msg)

        def parse_machine_option(machine_option):
            return MachineOption(machine_option[u'id'], machine_option[u'name'], machine_option[u'is_default'],
                                 machine_option[u'gpu'])

        combined_machine_options = list(map(lambda x: parse_machine_option(x), machine_options))

        # It is possible no machine options were returned, in that case we can still let users choose between their
        # default machine options (cpu vs gpu)
        if len(combined_machine_options) <= 0:
            combined_machine_options.append(MachineOption(u"default_machine_cpu",
                                                          u"Default CPU Machine (If Supported)",
                                                          True,
                                                          False))
            combined_machine_options.append(MachineOption(u"default_machine_gpu",
                                                          u"Default GPU Machine (If Supported)",
                                                          True,
                                                          True))

        app_machine_options = cached_machine_options.get(_app_cache_id, {})
        app_machine_options[operation] = combined_machine_options
        cached_machine_options[_app_cache_id] = app_machine_options

        self._cached_machine_options.set_value(cached_machine_options)

        return combined_machine_options

    def get_machines_options(self,
                             app: str,
                             operation: str,
                             plugins: typing.Optional[typing.List[str]] = None,
                             ignore_cache: bool = False,
                             use_gpu_default: bool = False,
                             filter_cpu: bool = False,
                             filter_gpu: bool = False) -> typing.Tuple[typing.List['EnumItem'], int]:

        items = []

        machines = self.get_machines(app, operation, plugins=plugins)

        index = 0
        default = 0
        default_machines = []
        for machine in machines:

            if filter_cpu and not machine.is_gpu_machine():
                continue

            if filter_gpu and machine.is_gpu_machine():
                continue

            items.append(EnumItem(machine.get_id(), machine.get_name(), u""))

            if machine.is_default():
                default_machines.append((machine, index))

            index = index + 1

        # if at least one machine was marked as default
        if default_machines:

            # use the first default machine who's is_gpu_machine() getter matches the use_gpu_default key word argument
            for default_machine in default_machines:
                if default_machine[0].is_gpu_machine() == use_gpu_default:
                    default = default_machine[1]
                    break
            else:  # if no mach is found then use the first default machine in the list
                default = default_machines[0][1]

        return items, default

    def _is_gpu_enabled_machine(self, machine_id: str) -> bool:

        # get all cached machine types
        cached_machine_options_dict = self._cached_machine_options.get_value()

        cached_machine_options = set()
        for app, operations in cached_machine_options_dict.items():
            for operation, machine_options in operations.items():
                cached_machine_options.update(machine_options)

        for machine_option in cached_machine_options:
            if machine_option.get_id() == machine_id:
                return machine_option.is_gpu_machine()

        raise ValueError(u"Could not find cached machine option with id '" + str(machine_id) + u"'")
