__all__ = ['StringAttributeType', 'StringSubtype',
           'EnumAttributeType', 'EnumItem', 'EnumSubtype',
           'NullAttributeType',
           'BooleanAttributeType',
           'IntegerAttributeType', 'IntegerSubtype',
           'FloatAttributeType',
           'ListAttributeType', 'ListSubtype']

import enum
import typing

from .attribute import Attribute, AttributeType
from .errors import InvalidAttributeError
from .inference_source import InferenceSource


class StringSubtype(enum.Enum):
    NONE = "NONE"
    FRAME_RANGES = "FRAME_RANGES"
    PATH = "PATH"
    JOB_EVENT_HANDLERS = "JOB_EVENT_HANDLERS"
    PLUGINS = "PLUGINS"


class StringAttributeType(Attribute):

    def __init__(self,
                 key: str,
                 display_name: str,
                 description: str,
                 subtype_kwargs: typing.Dict,
                 default_value: typing.Optional[str] = None,
                 subtype: 'StringSubtype' = StringSubtype.NONE,
                 placeholder: typing.Optional[str] = None,
                 max_length: typing.Optional[int] = None,
                 min_length: typing.Optional[int] = None):

        Attribute.__init__(self, key, display_name, description, AttributeType.STRING, subtype_kwargs)

        if default_value is None:
            default_value = ""

        if subtype is None:
            subtype = StringSubtype.NONE

        self._default_value = default_value
        self._subtype = subtype
        self._placeholder = placeholder
        self._max_length = max_length
        self._min_length = min_length

    def get_default_value(self) -> typing.Optional[str]:
        return self._default_value

    def get_subtype(self) -> 'StringSubtype':
        return self._subtype

    def get_placeholder(self) -> typing.Optional[str]:
        return self._placeholder

    def get_max_length(self) -> typing.Optional[int]:
        return self._max_length

    def get_min_length(self) -> typing.Optional[int]:
        return self._min_length

    def validate_value(self, value: any) -> None:
        if value is None:
            raise InvalidAttributeError("value must not be None")

        if type(value) != str:
            raise InvalidAttributeError("value must be string type")

        if self._max_length is not None and len(value) > self._max_length:
            raise InvalidAttributeError("value be be a maximum of " + str(self._max_length) + " characters long")

        if self._min_length is not None and len(value) < self._min_length:
            raise InvalidAttributeError("value be be a minimum of " + str(self._min_length) + " characters long")


class EnumItem:

    def __init__(self, key: str, display_name: str, description: str):
        self._key = key
        self._display_name = display_name
        self._description = description

    def get_key(self) -> str:
        return self._key

    def get_display_name(self) -> str:
        return self._display_name

    def get_description(self) -> str:
        return self._description


class EnumSubtype(enum.Enum):
    NONE = "NONE"
    APP_INFERRED = "APP_INFERRED"
    PRODUCT_VERSIONS = "PRODUCT_VERSIONS"
    MACHINE_TYPE = "MACHINE_TYPE"


class EnumAttributeType(Attribute):

    DYNAMIC_ENUM_SUBTYPES = [EnumSubtype.PRODUCT_VERSIONS,
                             EnumSubtype.MACHINE_TYPE,
                             EnumSubtype.APP_INFERRED]

    def __init__(self,
                 key: str,
                 display_name: str,
                 description: str,
                 subtype_kwargs: typing.Dict,
                 items: typing.List['EnumItem'],
                 default_item_key: typing.Optional['str'] = None,
                 subtype: 'EnumSubtype' = EnumSubtype.NONE):

        Attribute.__init__(self, key, display_name, description, AttributeType.ENUM, subtype_kwargs)

        if len(items) <= 0 and subtype not in self.DYNAMIC_ENUM_SUBTYPES:
            raise ValueError("Enum '" + key + "' must have at least one possible value")

        self._items = items
        self._subtype = subtype

        if subtype in self.DYNAMIC_ENUM_SUBTYPES:
            # items and default value do not matter for PRODUCT_VERSIONS subtype
            self._default_item_key = None
        else:
            self._default_item_key = default_item_key

    def get_default_value(self) -> typing.Optional['str']:
        return self.get_default_item_key()

    def get_default_item_key(self) -> typing.Optional['str']:
        return self._default_item_key

    def get_default_enum_item(self) -> typing.Optional['EnumItem']:
        for item in self._items:
            if item.get_key() == self.get_default_item_key():
                return item
        return None

    def get_subtype(self) -> 'EnumSubtype':
        return self._subtype

    def get_items(self) -> typing.List['EnumItem']:
        return self._items

    def validate_value(self, value: any) -> None:
        if value is None:
            raise InvalidAttributeError("value must not be None")

        if type(value) != str:
            raise InvalidAttributeError("value must be string type")

        # don't compare items for dynamic enums
        if self.get_subtype() in self.DYNAMIC_ENUM_SUBTYPES:
            return

        for item in self.get_items():
            if item.get_key() == value:
                return
        else:
            raise InvalidAttributeError("value must be one of the possible enum values")


class NullAttributeType(Attribute):

    def __init__(self, key: str, display_name: str, description: str, subtype_kwargs: typing.Dict):
        Attribute.__init__(self, key, display_name, description, AttributeType.NULL, subtype_kwargs)
        self._default_value = None

    def get_default_value(self) -> None:
        return self._default_value

    def validate_value(self, value: any):
        if value is not None:
            raise InvalidAttributeError("value must be None")


class BooleanAttributeType(Attribute):

    def __init__(self,
                 key: str,
                 display_name: str,
                 description: str,
                 subtype_kwargs: typing.Dict,
                 default_value: typing.Optional[bool] = True):

        Attribute.__init__(self, key, display_name, description, AttributeType.BOOLEAN, subtype_kwargs)

        if default_value is None:
            default_value = False

        self._default_value = default_value

    def get_default_value(self) -> bool:
        return self._default_value

    def validate_value(self, value: any) -> None:
        if value is None:
            raise InvalidAttributeError("value must not be None")

        if type(value) != bool:
            raise InvalidAttributeError("value must be bool type")


class IntegerSubtype(enum.Enum):
    NONE = "NONE"
    INSTANCES = "INSTANCES"


class IntegerAttributeType(Attribute):

    def __init__(self,
                 key: str,
                 display_name: str,
                 description: str,
                 subtype_kwargs: typing.Dict,
                 default_value: typing.Optional[int] = 0,
                 subtype: 'IntegerSubtype' = IntegerSubtype.NONE):

        Attribute.__init__(self, key, display_name, description, AttributeType.INTEGER, subtype_kwargs)

        if default_value is None:
            default_value = 0

        self._default_value = default_value
        self._subtype = subtype

    def get_default_value(self) -> int:
        return self._default_value

    def get_subtype(self) -> IntegerSubtype:
        return self._subtype

    def validate_value(self, value: any) -> None:
        if value is None:
            raise InvalidAttributeError("value must not be None")

        if type(value) != int:
            raise InvalidAttributeError("value must be int type")


class FloatAttributeType(Attribute):

    def __init__(self,
                 key: str,
                 display_name: str,
                 description: str,
                 subtype_kwargs: typing.Dict,
                 default_value: typing.Optional[float] = 0):

        Attribute.__init__(self, key, display_name, description, AttributeType.FLOAT, subtype_kwargs)

        if default_value is None:
            default_value = 0

        self._default_value = default_value

    def get_default_value(self) -> float:
        return self._default_value

    def validate_value(self, value: any) -> None:
        if value is None:
            raise InvalidAttributeError("value must not be None")

        if type(value) != float:
            raise InvalidAttributeError("value must be float type")


class ListSubtype(enum.Enum):
    NONE = "NONE"
    RADIO = "RADIO"


class ListAttributeType(Attribute):

    def __init__(self,
                 key: str,
                 display_name: str,
                 description: str,
                 subtype_kwargs: typing.Dict,
                 items_inference_source: 'InferenceSource',
                 items: typing.List[EnumItem],
                 list_type: 'AttributeType',
                 default_value: typing.Optional[typing.List[any]] = None,
                 subtype: 'ListSubtype' = ListSubtype.NONE):

        Attribute.__init__(self, key, display_name, description, AttributeType.LIST, subtype_kwargs)

        if default_value is None:
            default_value = []

        self._items_inference_source = items_inference_source
        self._items = items
        self._default_value = default_value
        self._list_type = list_type
        self._subtype = subtype

    def get_items_inference_source(self) -> 'InferenceSource':
        return self._items_inference_source

    def get_items(self) -> typing.List[EnumItem]:
        return self._items

    def get_default_value(self) -> typing.List[any]:
        return self._default_value

    def get_list_type(self) -> 'AttributeType':
        return self._list_type

    def get_subtype(self) -> 'ListSubtype':
        return self._subtype

    def validate_value(self, value: any) -> None:
        if value is None:
            raise InvalidAttributeError("value must not be None")

        if isinstance(value, list):
            raise InvalidAttributeError("value must be list type")
