__all__ = ['DynamicJobPresetAttribute']

from PySide2 import QtCore

import typing

from .inference_source import InferenceSource
from .job_preset_attribute import JobPresetAttribute
from .errors.cli_prompt_not_enabled_error import CLIPromptNotEnabledError

if typing.TYPE_CHECKING:
    from .job_attribute import JobAttribute
    from .dynamic_job_preset import DynamicJobPreset
    from .project import Project
    from .submit2gm_plugin import Submit2gmPlugin


class DynamicJobPresetAttribute(JobPresetAttribute):

    inference_source_changed = QtCore.Signal()
    enabled_state_changed = QtCore.Signal(bool)

    user_input_value_changed = QtCore.Signal()

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 parent_preset: 'DynamicJobPreset',
                 job_attribute: 'JobAttribute',
                 inference_source: 'InferenceSource' = None,
                 value: 'any' = None):

        id = parent_preset.get_id() + '_' + job_attribute.get_attribute().get_key()
        JobPresetAttribute.__init__(self, plugin, id, job_attribute)

        self._is_eval_locked = False
        self._parent_preset = parent_preset
        self._job_attribute = job_attribute

        if inference_source is not None:
            if inference_source not in job_attribute.get_inference_sources():
                raise ValueError("Inference source must be set to a valid inference source.")
        else:
            inference_source = job_attribute.get_default_inference_source()

        self._inference_source = inference_source

        if value is None:
            value = job_attribute.get_attribute().get_default_value()

        self._user_defined_value = value
        self._constant_value = value
        self._project_defined_value = value
        self._application_defined_value = value
        self._python_defined_value = self.eval_python_defined_value() if InferenceSource.get_python_inference_source() in job_attribute.get_inference_sources() else value
        self._argv_defined_value = value

        self._is_enabled = job_attribute.default_enabled_state()

        """ Create Connections """
        self.inference_source_changed.connect(lambda: parent_preset.dynamic_job_preset_attribute_value_changed.emit(job_attribute.get_id()))
        self.user_input_value_changed.connect(lambda: parent_preset.dynamic_job_preset_attribute_value_changed.emit(job_attribute.get_id()))


    def get_parent_preset(self) -> 'DynamicJobPreset':
        return self._parent_preset

    # -- User defined
    def get_user_defined_value(self) -> any:
        return self._user_defined_value

    def set_user_defined_value(self, value: any) -> None:
        if self.get_user_defined_value() != value:
            self._user_defined_value = value
            self.user_input_value_changed.emit()

    def eval_user_defined_value(self, fail_on_eval_lock=True) -> any:
        if self._is_eval_locked:
            if fail_on_eval_lock:
                raise ValueError("Attempted to evaluate job preset attribute ('{}') value while evaluation lock is "
                                 "enabled".format(self.get_id()))
            return self.get_user_defined_value()

        return self.get_user_defined_value()

    # -- Constant
    def get_constant_value(self) -> any:
        return self._constant_value

    def set_constant_value(self, value: any) -> None:
        self._constant_value = value

    def eval_constant_value(self, fail_on_eval_lock=True) -> any:
        return self.get_constant_value()

    # -- Project
    def get_project_defined_value(self) -> any:
        return self._project_defined_value

    def set_project_defined_value(self, value: any) -> None:
        self._project_defined_value = value

    def eval_project_defined_value(self, source_project: 'Project', fail_on_eval_lock=True) -> any:
        if self._is_eval_locked:
            if fail_on_eval_lock:
                raise ValueError("Attempted to evaluate job preset attribute ('{}') value while evaluation lock is "
                                 "enabled".format(self.get_id()))
            return self.get_project_defined_value()

        attribute = self.get_job_attribute().get_attribute()
        project_defined_value = source_project.get_attribute(attribute.get_key())
        self.set_project_defined_value(project_defined_value)
        return self.get_project_defined_value()

    # -- Application
    def get_application_defined_value(self) -> any:
        return self._application_defined_value

    def set_application_defined_value(self, value: any) -> None:
        self._application_defined_value = value

    def eval_application_defined_value(self, fail_on_eval_lock: bool = True) -> any:
        if self._is_eval_locked:
            if fail_on_eval_lock:
                raise ValueError("Attempted to evaluate job preset attribute ('{}') value while evaluation lock is "
                                 "enabled".format(self.get_id()))
            return self.get_application_defined_value()

        application_attribute_source = self.get_plugin().get_application_attribute_source()
        app_defined_value = application_attribute_source.get_attribute_value(None, None, self.get_key())
        self.set_application_defined_value(app_defined_value)
        return self.get_application_defined_value()

    # -- Python
    def get_python_defined_value(self) -> any:
        return self._python_defined_value

    def set_python_defined_value(self, value: any) -> None:
        self._python_defined_value = value

    def eval_python_defined_value(self, fail_on_eval_lock: bool = True) -> any:
        if self._is_eval_locked:
            if fail_on_eval_lock:
                raise ValueError("Attempted to evaluate job preset attribute ('{}') value while evaluation lock is "
                                 "enabled".format(self.get_id()))
            return self.get_python_defined_value()

        key = self.get_key()

        """
        # Currently only supports attributes with the key "submission_name"
        if key == constants.API_KEYS.SUBMISSION_NAME:
            timestamp = datetime.datetime.now()
            submission_name_timestamp = timestamp.strftime("%Y-%b-%d %H:%M:%S %p")
            path_safe_submission_name = submission_name_timestamp.replace(' ', '_').replace(':', '-').replace('.', '_')
            self.set_python_defined_value(path_safe_submission_name)
            return self.get_python_defined_value()
        """

        raise NotImplementedError

    # -- Argv
    def get_argv_defined_value(self) -> any:
        return self._user_defined_value

    def set_argv_defined_value(self, value: any) -> None:
        self._user_defined_value = value

    def eval_argv_defined_value(self, fail_on_eval_lock=True) -> any:
        if self._is_eval_locked:
            if fail_on_eval_lock:
                raise ValueError("Attempted to evaluate job preset attribute ('{}') value while evaluation lock is "
                                 "enabled".format(self.get_id()))
            return self.get_value()

        attribute = self.get_job_attribute().get_attribute()

        try:
            argv_defined_value = attribute.get_argv_value(self.get_plugin(), job_preset=self.get_parent_preset())
        except CLIPromptNotEnabledError:
            argv_defined_value = attribute.get_default_value()

        self.set_argv_defined_value(argv_defined_value)
        return self.get_argv_defined_value()

    def get_value(self) -> any:
        inference_source = self.get_inference_source()

        if inference_source == InferenceSource.get_constant_inference_source():
            return self.get_constant_value()

        if inference_source == InferenceSource.get_application_inference_source():
            return self.get_application_defined_value()

        if inference_source == InferenceSource.get_user_defined_inference_source():
            return self.get_user_defined_value()

        if inference_source == InferenceSource.get_argv_inference_source():
            return self.get_argv_defined_value()

        if inference_source == InferenceSource.get_python_inference_source():
            return self.get_python_defined_value()

        if inference_source == InferenceSource.get_project_inference_source():
            return self.get_project_defined_value()

        raise ValueError("Unsupported inference source type: " + inference_source.get_display_name())

    def get_inference_source(self) -> 'InferenceSource':
        return self._inference_source

    def set_inference_source(self, inference_source: 'InferenceSource'):

        if inference_source not in self.get_job_attribute().get_inference_sources():
            raise ValueError("Inference source must be set to a valid inference source.")

        self._inference_source = inference_source
        self.inference_source_changed.emit()

    def get_dependencies(self) -> typing.List['DynamicJobPresetAttribute']:
        """
        A JobPresetAttribute's dependencies are the list of JobPresetAttributes it's pre condition evaluator depends on
        """
        dependencies = []
        parent_preset = self.get_parent_preset()

        for dependency_id in self.get_job_attribute().get_dependencies_ids():
            dependencies.append(parent_preset.get_job_preset_attribute_with_job_attribute_id(dependency_id))

        return dependencies

    def is_active(self) -> bool:
        """
        A job preset attribute is considered to be 'active' if it's precondition evaluator returns True.
        In-active job preset attributes are not visible and are not used when submitting.
        """
        job_attribute = self.get_job_attribute()
        return job_attribute.get_precondition_evaluator()(self, self.get_parent_preset())

    def is_enabled(self) -> bool:
        """
        A job preset attribute can only be disabled when the JobAttribute that it instantiates is optional
        """
        try:
            optional_parent_attribute = self.get_optional_parent()
            is_enabled = optional_parent_attribute.is_enabled()
        except ValueError:
            is_enabled = self._is_enabled

        return is_enabled

    @QtCore.Slot(bool)
    def set_enabled(self, enabled: bool) -> None:
        self._is_enabled = enabled
        self.enabled_state_changed.emit(self.is_enabled())

    def eval(self, source_project: typing.Optional['Project'] = None, fail_on_eval_lock: bool = True):

        inference_source = self.get_inference_source()
        attribute = self._job_attribute.get_attribute()

        # application source
        if inference_source == InferenceSource.get_application_inference_source():
            return self.eval_application_defined_value(fail_on_eval_lock=fail_on_eval_lock)

        # project source
        elif inference_source == InferenceSource.get_project_inference_source():
            if source_project is None:
                raise ValueError("Must provide source project for attribute '" + attribute.get_display_name() + "'")

            return self.eval_project_defined_value(source_project, fail_on_eval_lock=fail_on_eval_lock)

        # constant source
        elif inference_source == InferenceSource.get_constant_inference_source():
            return self.eval_constant_value(fail_on_eval_lock=fail_on_eval_lock)

        # user defined source
        elif inference_source == InferenceSource.get_user_defined_inference_source():
            return self.eval_user_defined_value(fail_on_eval_lock=fail_on_eval_lock)

        # argv source
        elif inference_source == InferenceSource.get_argv_inference_source():
            return self.eval_argv_defined_value(fail_on_eval_lock=fail_on_eval_lock)

        # python source
        elif inference_source == InferenceSource.get_python_inference_source():
            return self.eval_python_defined_value(fail_on_eval_lock=fail_on_eval_lock)

        else:
            raise ValueError("Unknown inference source: " + str(inference_source))

    def is_eval_locked(self) -> bool:
        return self._is_eval_locked

    def set_eval_locked(self, value: bool) -> None:
        self._is_eval_locked = value

    def should_skip(self) -> bool:

        if self.get_job_attribute().should_skip():
            return True

        if self.get_job_attribute().is_optional():
            return not self.is_enabled()

        return False
