__all__ = 'InferenceSource'

import enum
import typing


class InferenceSource:
    """
    An InferenceSource is the source from which to obtain the value for an attribute. JobPreset Attributes can be
    derived from the JobPreset it self as pre-defined or user-defined constants or from the application or project.
    """

    # InferenceSource id constants
    class Type(enum.Enum):
        USER_DEFINED = "USER_DEFINED"
        CONSTANT = "CONSTANT"
        PROJECT = "PROJECT"
        APPLICATION = "APPLICATION"
        PYTHON = "PYTHON"
        ARGV = "ARGV"

    # descriptions
    USER_DEFINED_DESCRIPTION = "Use the value specified by the user in the form field"
    CONSTANT_DESCRIPTION = "Use the pre-defined constant value for this attribute"
    PROJECT_DESCRIPTION = "Read the value from the project's attributes when you submit"
    APPLICATION_DESCRIPTION = "Use the value specified by the application"
    PYTHON_DESCRIPTION = "Evaluate a python expression to determine the value"
    ARGV_DESCRIPTION = "read the value from sys.argv"

    _user_defined = None
    _constant = None
    _project = None
    _application = None
    _python = None
    _argv = None

    def __init__(self, inference_source_id: 'InferenceSource.Type', display_name: str, description: str):
        self._inference_source_id = inference_source_id
        self._display_name = display_name
        self._description = description

    def get_id(self) -> 'InferenceSource.Type':
        return self._inference_source_id

    def get_display_name(self) -> str:
        return self._display_name

    def get_description(self) -> str:
        return self._description

    def __eq__(self, other: 'InferenceSource'):
        return self.get_id() == other.get_id()

    def __ne__(self, other: 'InferenceSource'):
        return self.get_id() != other.get_id()

    def __str__(self) -> str:
        return str(self.get_id())

    @staticmethod
    def is_inference_source_id(value: str) -> bool:
        return value == InferenceSource.Type.USER_DEFINED.value or \
               value == InferenceSource.Type.CONSTANT.value or \
               value == InferenceSource.Type.PROJECT.value or \
               value == InferenceSource.Type.APPLICATION.value or \
               value == InferenceSource.Type.PYTHON.value or \
               value == InferenceSource.Type.ARGV.value

    @staticmethod
    def get_inference_source(source_id: 'InferenceSource.Type'):
        if isinstance(source_id, str):
            if not InferenceSource.is_inference_source_id(source_id):
                raise ValueError("Unrecognised inference source id: {} | {}".format(str(source_id),
                                                                                    type(str(source_id))))

        if isinstance(source_id, InferenceSource.Type):
            source_id = source_id.value

        if source_id == InferenceSource.Type.USER_DEFINED.value:
            return InferenceSource.get_user_defined_inference_source()

        if source_id == InferenceSource.Type.CONSTANT.value:
            return InferenceSource.get_constant_inference_source()

        if source_id == InferenceSource.Type.PROJECT.value:
            return InferenceSource.get_project_inference_source()

        if source_id == InferenceSource.Type.APPLICATION.value:
            return InferenceSource.get_application_inference_source()

        if source_id == InferenceSource.Type.PYTHON.value:
            return InferenceSource.get_python_inference_source()

        if source_id == InferenceSource.Type.ARGV.value:
            return InferenceSource.get_argv_inference_source()

    @staticmethod
    def get_all_inference_source_types() -> typing.List['InferenceSource']:
        inference_sources = list()
        inference_sources.append(InferenceSource.get_user_defined_inference_source())
        inference_sources.append(InferenceSource.get_constant_inference_source())
        inference_sources.append(InferenceSource.get_project_inference_source())
        inference_sources.append(InferenceSource.get_application_inference_source())
        inference_sources.append(InferenceSource.get_python_inference_source())
        inference_sources.append(InferenceSource.get_argv_inference_source())
        return inference_sources

    @staticmethod
    def get_user_defined_inference_source() -> 'InferenceSource':
        if InferenceSource._user_defined is None:
            InferenceSource._user_defined = InferenceSource(InferenceSource.Type.USER_DEFINED, "User Defined",
                                                            InferenceSource.USER_DEFINED_DESCRIPTION)
        return InferenceSource._user_defined

    @staticmethod
    def get_constant_inference_source() -> 'InferenceSource':
        if InferenceSource._constant is None:
            InferenceSource._constant = InferenceSource(InferenceSource.Type.CONSTANT, "Constant",
                                                        InferenceSource.CONSTANT_DESCRIPTION)
        return InferenceSource._constant

    @staticmethod
    def get_project_inference_source() -> 'InferenceSource':
        if InferenceSource._project is None:
            InferenceSource._project = InferenceSource(InferenceSource.Type.PROJECT, "Project",
                                                       InferenceSource.PROJECT_DESCRIPTION)
        return InferenceSource._project

    @staticmethod
    def get_application_inference_source() -> 'InferenceSource':
        if InferenceSource._application is None:
            InferenceSource._application = InferenceSource(InferenceSource.Type.APPLICATION, "Application",
                                                           InferenceSource.APPLICATION_DESCRIPTION)
        return InferenceSource._application

    @staticmethod
    def get_python_inference_source() -> 'InferenceSource':
        if InferenceSource._python is None:
            InferenceSource._python = InferenceSource(InferenceSource.Type.PYTHON, "Python",
                                                      InferenceSource.PYTHON_DESCRIPTION)
        return InferenceSource._python

    @staticmethod
    def get_argv_inference_source() -> 'InferenceSource':
        if InferenceSource._argv is None:
            InferenceSource._argv = InferenceSource(InferenceSource.Type.ARGV, "Argv",
                                                    InferenceSource.ARGV_DESCRIPTION)
        return InferenceSource._argv
