__all__ = 'JobPresetContainer'

from PySide2 import QtCore

from . import utils
from .plugin_accessor import PluginAccessor
from .dynamic_job_preset import DynamicJobPreset

import typing

if typing.TYPE_CHECKING:
    from .submit2gm_plugin import Submit2gmPlugin
    from .job_definition import JobDefinition


class JobPresetContainer(QtCore.QObject, PluginAccessor):

    dataChanged = QtCore.Signal()
    itemAdded = QtCore.Signal(str, DynamicJobPreset)
    itemRemoved = QtCore.Signal(str, DynamicJobPreset)

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtCore.QObject'] = None):
        QtCore.QObject.__init__(self, parent if parent else plugin)
        PluginAccessor.__init__(self, plugin)

        self._items: typing.Dict[str, 'DynamicJobPreset'] = dict()

    def add(self, key: str, value: 'DynamicJobPreset') -> None:
        self._items[key] = value
        self.itemAdded.emit(key, value)

    def clear(self) -> None:
        old_items = self._items
        self._items = dict()

        for key, value in old_items:
            self.itemRemoved.emit(key, value)

        self.dataChanged.emit()

    def count(self) -> int:
        return len(self._items.values())

    def get(self, key: str, default: typing.Optional['DynamicJobPreset'] = None) -> typing.Optional['DynamicJobPreset']:
        return self._items.get(key, default=default)

    def items(self) -> typing.List[typing.Tuple[str, 'DynamicJobPreset']]:
        items = []
        for key, value in self._items.items():
            items.append((key, value))
        return items

    def keys(self) -> typing.KeysView[str]:
        return self._items.keys()

    def values(self) -> typing.ValuesView['DynamicJobPreset']:
        return  self._items.values()

    def pop(self, key: str) -> 'DynamicJobPreset':
        try:
            value = self._items.pop(key)
            self.itemRemoved.emit(key, value)
        except KeyError as e:
            raise e

        self.dataChanged.emit()
        return value

    def emplace(self, job_definition: 'JobDefinition') -> 'DynamicJobPreset':

        # get the job preset objects in the QListWidgetItems
        dynamic_job_preset_items = list(self.values())

        name_prefix = (job_definition.get_display_name() + "_Job_Preset_").replace(' ', '_')

        # create a unique id
        id = utils.create_unique_object_name(dynamic_job_preset_items,
                                             attribute_name="get_id",
                                             name_prefix=name_prefix)

        # create a unique name
        name = utils.create_unique_object_name(dynamic_job_preset_items,
                                               attribute_name="get_name",
                                               name_prefix=name_prefix)

        dynamic_job_preset = DynamicJobPreset(self.get_plugin(), name, id, job_definition)

        self.add(dynamic_job_preset.get_id(), dynamic_job_preset)

        return dynamic_job_preset
