__all__ = 'Job Processor'

import argparse
import json
import pathlib
import sys
import typing

from . import constants

if typing.TYPE_CHECKING:
    from .submit2gm_plugin import Submit2gmPlugin
    from .worker_job import WorkerJob


main_window = None
PLUGIN: typing.Optional['Submit2gmPlugin'] = None


class JobProcessor:

    SUBMISSION_POPUP_ALREADY_OPEN = 1

    def __init__(self):
        self.active_job: typing.Optional['WorkerJob'] = None

        self.parser = argparse.ArgumentParser(prog=constants.PROGRAM_NAME,
                                              description=constants.APP_DESCRIPTION)
        sub_parsers = self.parser.add_subparsers()

        # open submission popup
        parser_submission_popup = sub_parsers.add_parser('open-submission-popup', help='Submit a project')
        parser_submission_popup.add_argument('--product',
                                             type=str,
                                             dest='product',
                                             required=True,
                                             help="A GridMarkets product code.")

        parser_submission_popup.add_argument('--cps-address',
                                             type=str,
                                             dest='cps_address',
                                             required=False,
                                             help="The address of the command port server.")

        parser_submission_popup.add_argument('--auto-sign-in',
                                             type=bool,
                                             dest='auto_sign_in',
                                             default=True,
                                             required=False,
                                             help="Should {} automatically attempt to sign in via Envoy on "
                                                  "startup.".format(constants.PROGRAM_NAME))

        parser_submission_popup.add_argument('--meta-info',
                                             type=str,
                                             dest='meta_info',
                                             default="",
                                             required=False,
                                             help="Meta data to regarding the CPS plug-in.")

        parser_submission_popup.set_defaults(func=self.open_submission_popup)

        # reset
        parser_reset = sub_parsers.add_parser('reset', help='Return Submit2gm to it\'s inital state. I.e sign out the '
                                                            'user, close the UI and shutdown any command port servers.')
        parser_reset.set_defaults(func=self.reset)

        # disconnect-cps
        parser_close_cps = sub_parsers.add_parser('disconnect-cps', help='Shuts down and removes the command port '
                                                                         'server with the given address. If the CPS is '
                                                                         'active then reset Submit2gm.')
        parser_close_cps.add_argument('--cps-address',
                                      type=str,
                                      dest='cps_address',
                                      required=True,
                                      help="The address of the command port server to disconnect.")
        parser_close_cps.set_defaults(func=self.disconnect_cps)

        # reset
        parser_version = sub_parsers.add_parser('version', help='Print the submit2gm version number.')
        parser_version.set_defaults(func=self.version)

        # list job definitions
        parser_list_job_definitions = sub_parsers.add_parser('list-job-definitions', help='List all the job definition '
                                                                                          'types that can be used for '
                                                                                          'job submissions.')
        parser_list_job_definitions.set_defaults(func=self.list_job_definitions)

        # show job definition
        parser_show_job_definition = sub_parsers.add_parser('show-job-definition', help='Displays the list of '
                                                                                        'attributes required for a job '
                                                                                        'definition.')
        parser_show_job_definition.add_argument('--job-definition-id',
                                                type=str,
                                                dest='job_definition_id',
                                                default="",
                                                required=True,
                                                help="The ID of the job definition to show.")
        parser_show_job_definition.set_defaults(func=self.show_job_definition)

        # get product versions
        parser_get_product_versions = sub_parsers.add_parser('get-product-versions',
                                                             help='Displays the list of available app versions for a '
                                                                  'specified app.')

        parser_get_product_versions.add_argument('--app',
                                                 type=str,
                                                 dest='app',
                                                 required=True,
                                                 help="The app id.")

        parser_get_product_versions.set_defaults(func=self.get_product_versions)

        # get machines
        parser_get_machines = sub_parsers.add_parser('get-machines', help='Displays the list of available machines for '
                                                                          'an app based on the specified operation, '
                                                                          'machine category and plug-ins.')

        parser_get_machines.add_argument('--app',
                                         type=str,
                                         dest='app',
                                         required=True,
                                         help="The app id.")

        parser_get_machines.add_argument('--operation',
                                         type=str,
                                         default='',
                                         dest='operation',
                                         required=True,
                                         help="The operation type.")

        parser_get_machines.add_argument('--plugins',
                                         type=str,
                                         dest='plugins',
                                         default='',
                                         help="Comma separated list of plug-ins.")

        parser_get_machines.set_defaults(func=self.get_machines)

        # get user info
        parser_get_user_info = sub_parsers.add_parser('get-user-info', help='Displays all user level info.')
        parser_get_user_info.set_defaults(func=self.get_user_info)

        # submit job
        parser_submit_project = sub_parsers.add_parser('submit-project',
                                                       help='Submit a project based using project submision json data')

        parser_submit_project.add_argument(constants.SUBMIT2GM_CLI_ARGUMENTS.PROJECT_REQUEST,
                                           type=str,
                                           dest='project_request',
                                           required=True,
                                           help="An escaped JSON string containing the raw project_request data "
                                                "expected by Envoy for it's project-submit endpoint.")

        parser_submit_project.add_argument(constants.SUBMIT2GM_CLI_ARGUMENTS.USE_FILE,
                                           dest='use_file',
                                           action='store_true',
                                           required=False,
                                           help="Does the {} argument refer to a json "
                                                "file.".format(constants.SUBMIT2GM_CLI_ARGUMENTS.PROJECT_REQUEST))

        parser_submit_project.set_defaults(func=self.submit_project)

        # delete project
        parser_delete_project = sub_parsers.add_parser('delete-project',
                                                       help='Deletes a project from Envoy')

        parser_delete_project.add_argument(constants.SUBMIT2GM_CLI_ARGUMENTS.PROJECT_NAME,
                                           type=str,
                                           dest='project_name',
                                           required=True,
                                           help="The name of the project to delete as it appears within Envoy.")

        parser_delete_project.set_defaults(func=self.delete_project)

        # file dialog
        file_dialog = sub_parsers.add_parser('file-dialog',
                                             help='Open a native file / folder dialog')

        file_dialog.add_argument('--file-mode',
                                 choices=['AnyFile', 'ExistingFile', 'Directory', 'ExistingFiles'],
                                 default='AnyFile',
                                 const='all',
                                 nargs='?',
                                 dest='file_mode',
                                 help="Used to indicate what the user may select in the file dialog; i.e. what the dialog will return if the user clicks OK.")

        file_dialog.add_argument('--accept-mode',
                                 choices=['AcceptOpen', 'AcceptSave'],
                                 default='AcceptOpen',
                                 const='all',
                                 nargs='?',
                                 dest='accept_mode',
                                 help="The action mode defines whether the dialog is for opening or saving files.")

        file_dialog.add_argument('--accept-label',
                                 type=str,
                                 default='',
                                 dest='accept_label',
                                 help="Sets the dialogs accept label text.")

        file_dialog.add_argument('--reject-label',
                                 type=str,
                                 default='',
                                 dest='reject_label',
                                 help="Sets the dialogs reject label text.")

        file_dialog.add_argument('--file-name-label',
                                 type=str,
                                 default='',
                                 dest='file_name_label',
                                 help="Sets the dialogs file name label text.")

        file_dialog.set_defaults(func=self.open_file_dialog)

        """
        # submit job
        parser_submit_job = sub_parsers.add_parser('submit-job', help='Submit a job based on a job definition')

        parser_submit_job.add_argument('--project-dir',
                                       type=str,
                                       dest='project_dir',
                                       required=True,
                                       help="If --skip-upload is not specified then this is the root folder of your "
                                            "project on your local file system. If --skip-upload is specified then "
                                            "this is the name of the existing remote project as seen within Envoy "
                                            "(e.g. \"example_project\").")

        parser_submit_job.add_argument('--job-definition-id',
                                       type=str,
                                       dest='job_definition_id',
                                       default="",
                                       required=True,
                                       help="The ID of the job definition to show.")

        parser_submit_job.add_argument('--skip-upload',
                                       dest='skip_upload',
                                       action='store_true',
                                       help="Determines if file upload will be skipped for the submission.")

        parser_submit_job.add_argument('--show-ui',
                                       dest='show_ui',
                                       action='store_true',
                                       help="Determines if Submit2GM will display a user interface for the user to "
                                            "confirm their job and project settings before submission.")

        parser_submit_job.add_argument('--enable-cli-prompt',
                                       dest='enable_cli_prompt',
                                       action='store_true',
                                       help="Determines if Submit2GM will prompt the user on the cli for missing "
                                            "attributes. Has no effect when --show-ui is enabled.")

        parser_submit_job.set_defaults(func=self.submit_job)
        """

    def _sign_out_callback(self, args, unknown_args) -> None:
        from .plugin_fetcher import PluginFetcher
        global main_window
        global PLUGIN

        args.auto_sign_in = False

        # Close the active cps session (if there is one) otherwise the cps will never accept another connection
        # Only close the session and don't stop beacuse the same cps may still be needed if the user signs back in
        # using the same app session.
        try:
            PLUGIN.get_active_command_port_server().close_session()
        except ValueError as e:
            pass

        # main_window.destroyed.connect(reopen_submision_popup)
        main_window.close()
        main_window.deleteLater()
        main_window = None

        PluginFetcher.delete_cached_plugin()
        PLUGIN = None

        self.open_submission_popup(args, unknown_args)

    def _clean_up_main_window(self):
        from .plugin_fetcher import PluginFetcher
        global main_window
        global PLUGIN

        if main_window:
            main_window.close()
            main_window.deleteLater()
            main_window = None

        if PLUGIN:
            PLUGIN.shutdown_command_port_servers()
            PluginFetcher.delete_cached_plugin()
            PLUGIN = None

    def re_open_submission_popup(self, args, unknown_args) -> None:
        """
        Called if the main window and plugin already exist
        """
        global main_window
        global PLUGIN
        logger = PLUGIN.get_logging_coordinator().get_logger(__file__)
        logger.debug("re_open_submission_popup")

        # Reset the windows reset callback since the arguments might have changed
        main_window.set_reset_callback(lambda: self._sign_out_callback(args, unknown_args))

        # Get the active cps address (if one exists)
        active_cps_address = None
        try:
            active_cps_address = PLUGIN.get_active_command_port_server().get_address()
        except ValueError:
            pass

        # If the main window already exists and the user is signed in then raise the current window (even if the
        # active cps is no longer running)
        if PLUGIN.get_api_client().is_user_signed_in():
            logger.debug("user is signed in")

            # Remove all old job presets
            job_preset_container = PLUGIN.get_preferences_container().get_job_preset_container()
            job_preset_container.clear()

            # If we're reopening with a new CPS then make sure to clean up and close the old one
            if active_cps_address != args.cps_address:
                PLUGIN.shutdown_command_port_servers()
                PLUGIN.add_command_port_server(args.product, args.cps_address, meta_info=args.meta_info)

        else:
            logger.debug("user is signed out")

            # If we're reopening with a new CPS then make sure to clean up and close the old one
            if active_cps_address != args.cps_address:
                PLUGIN.shutdown_command_port_servers()
                PLUGIN.add_command_port_server(args.product, args.cps_address, meta_info=args.meta_info)

            # If the main window already exists but the user is not signed in then call the sign out call back to
            # just to reset the state of submit2gm
            self._sign_out_callback(args, unknown_args)

        # Show the window
        main_window.show()
        main_window.raise_()
        main_window.activateWindow()

    def open_submission_popup(self, args, unknown_args) -> None:
        self.active_job.mark_as_complete()

        from .plugin_fetcher import PluginFetcher
        global main_window
        global PLUGIN

        # create the plugin
        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(__file__)

        from PySide2 import QtWidgets, QtCore
        if not QtWidgets.QApplication.instance():
            app = QtWidgets.QApplication([])
        else:
            app = QtWidgets.QApplication.instance()
        app.setWindowIcon(PLUGIN.get_user_interface().get_icon())

        from submit2gm.widgets.qt_dialog_main_window import MainWindow

        if main_window:
            return self.re_open_submission_popup(args, unknown_args)
        else:
            # create the main plugin window
            def sign_out_callback():
                self._sign_out_callback(args, unknown_args)

            main_window = MainWindow(PLUGIN, sign_out_callback, parent=None)
            main_window.setWindowModality(QtCore.Qt.ApplicationModal)

        if args.product:
            PLUGIN.set_initial_product(args.product)

        if args.cps_address:
            logger.debug("adding command port server")
            PLUGIN.add_command_port_server(args.product, args.cps_address, meta_info=args.meta_info)

        if args.auto_sign_in and not PLUGIN.get_api_client().is_user_signed_in():
            main_window.get_sign_in_widget().sign_in_envoy_slot()

        #main_window.finished.connect(lambda res: PLUGIN.shutdown_command_port_servers())
        main_window.show()
        main_window.raise_()
        main_window.activateWindow()

        app.exec_()

    def reset(self, args, unknown_args) -> None:
        self._clean_up_main_window()

    def disconnect_cps(self, args, unknown_args) -> None:

        cps_address = args.cps_address

        global PLUGIN
        if PLUGIN:
            logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

            # If the cps is the active cps then we need to reset the UI so that no connection issues arise.
            try:
                active_cps = PLUGIN.get_active_command_port_server()

                if active_cps.get_address() == cps_address:
                    logger.debug("Disconnecting active cps")
                    return self.reset(args, unknown_args)
            except ValueError:
                pass

            logger.debug("Disconnecting cps")
            try:
                PLUGIN.shutdown_command_port_server(cps_address)
            except ValueError as e:
                logger.debug(str(e))
                return

    def version(self, args, unknown_args) -> None:
        from .version import VERSION
        version_str = '{} Version: '.format(constants.PROGRAM_NAME) + '.'.join(map(lambda x: str(x), VERSION))

        global PLUGIN
        if PLUGIN:
            logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)
            logger.info(version_str)
        else:
            print(version_str)

    def list_job_definitions(self, args, unknown_args):
        from submit2gm import utils
        from submit2gm.errors.api_client_error import APIClientError
        from submit2gm.inference_source import InferenceSource

        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except APIClientError as e:
                logger.info(str(e))
                return

        logger.info('\nListing job definitions:')
        api_schema = PLUGIN.get_api_client().get_api_schema()
        job_definitions = api_schema.get_job_definitions()

        table = [['Name'],['ID'],['Supports CMD submission']]
        for job_definition in job_definitions:
            table[0].append(job_definition.get_display_name())
            table[1].append(job_definition.get_definition_id())

            supports_cmd_submission = True
            argv_inference_source = InferenceSource.get_argv_inference_source()
            job_attributes = job_definition.get_attributes()

            for job_attribute in job_attributes:
                if argv_inference_source not in job_attribute.get_inference_sources():
                    supports_cmd_submission = False
                    break

            table[2].append(str(supports_cmd_submission))

        table_str = utils.format_table_str(table)
        logger.info(table_str)

    def show_job_definition(self, args, unknown_args):
        from submit2gm.errors.api_client_error import APIClientError

        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except APIClientError as e:
                logger.info(str(e))
                return

        job_definition_id = args.job_definition_id
        api_schema = PLUGIN.get_api_client().get_api_schema()

        try:
            job_definition = api_schema.get_job_definition_with_id(job_definition_id)
        except ValueError as e:
            logger.info("No job definition with ID '{}'".format(job_definition_id))
            return

        logger.info('Name: {}'.format(job_definition.get_display_name()))
        logger.info('ID: {}'.format(job_definition.get_definition_id()))

        logger.info('\nAttribute Groups:')
        attribute_groups = job_definition.get_attribute_groups()
        for attribute_group in attribute_groups:
            logger.info(attribute_group)

        logger.info('\nJob Attributes:')
        job_attributes = job_definition.get_attributes()
        for job_attribute in job_attributes:
            attribute = job_attribute.get_attribute()
            logger.info('Name: {}'.format(attribute.get_display_name()))
            logger.info('Description: {}'.format(attribute.get_description()))
            logger.info('Key: {}'.format(attribute.get_key()))
            logger.info('Type: {}'.format(attribute.get_type()))
            logger.info('Default value: {}'.format(attribute.get_default_value()))
            logger.info('Subtype Kargs: {}'.format(attribute.get_subtype_kwargs()))
            logger.info('Inference sources: {}'.format(', '.join(map(lambda x: str(x), job_attribute.get_inference_sources()))))
            logger.info('Dependencies: {}'.format(','.join(job_attribute.get_dependencies_ids())))
            logger.info('')

    def get_product_versions(self, args, unknown_args):
        from submit2gm.errors.api_client_error import APIClientError

        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except APIClientError as e:
                logger.info(str(e))
                return

        app = args.app
        products = PLUGIN.get_api_client().get_product_versions(app)

        if len(products) > 0:
            for product in products:
                logger.info("{}".format(product))
        else:
            logger.info("No versions found")

    def get_machines(self, args, unknown_args):
        from submit2gm.errors.api_client_error import APIClientError

        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except APIClientError as e:
                logger.info(str(e))
                return

        app = args.app
        operation = args.operation
        plugins = (args.plugins).split(',')

        machines = PLUGIN.get_api_client().get_machines(app, operation, plugins)

        logger.info('Found {} machine options'.format(len(machines)))

        if machines:
            max_length_name = max(map(lambda x: len(x.get_name()), machines))
            logger.info(('\n{0:{1}}  {2}').format("Name", max_length_name, "ID"))
            for machine in machines:
                logger.info(('{0:{1}}  {2}').format(machine.get_name(),
                                                    max_length_name,
                                                    machine.get_id()))

    def get_user_info(self, args, unknown_args):
        from submit2gm import utils
        from submit2gm.errors.api_client_error import APIClientError

        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except APIClientError as e:
                logger.info(str(e))
                return

        logger.info('\nUser Information:')
        serialised_user_info = PLUGIN.get_api_client().get_user_info().serialise()

        table = [['Key'], ['Value']]
        for key, value in serialised_user_info.items():
            table[0].append(key)
            table[1].append(str(value))

        table_str = utils.format_table_str(table)
        logger.info(table_str)

    def submit_project(self, args, unknown_args):
        """
        Based on the envoy python clients submit_project method.
        """

        project_request = args.project_request
        use_file = args.use_file

        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if use_file:
            try:
                with open(project_request, 'r') as request_file:
                    project_request = request_file.read()
            except FileNotFoundError as e:
                logger.error("File '{}' not found.".format(project_request))
                return

        try:
            project_request = json.loads(project_request)
        except json.JSONDecodeError as e:
            logger.error("JSON decoding error: {}".format(str(e)))
            return

        import gridmarkets
        envoy_client = gridmarkets.EnvoyClient()
        envoy_client.validate_auth()
        envoy_client.validate_credits()

        url = '{0}/project-submit'.format(envoy_client.url)

        headers = {'content-type': 'application/json'}

        try:
            resp = envoy_client.http_client.request('post', url, headers, project_request)
        except gridmarkets.errors.InsufficientCreditsError as e:
            raise e
        except Exception as e:
            raise gridmarkets.errors.APIError(str(e))
        else:
            if resp.status_code >= 200 and resp.status_code < 300:
                logger.info("Project submitted.")

            if resp.status_code == 404:
                logger.error("404: {0} not found".format(url))

            if resp.status_code == 400:
                logger.error("Invalid request: {}".format(resp.json()))

            if resp.status_code in (500, 503):
                logger.error("{0} {1}".format(resp.status_code, resp.text))

    def delete_project(self, args, unknown_args) -> None:
        from submit2gm.errors.api_client_error import APIClientError

        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except APIClientError as e:
                logger.info(str(e))
                return

        project_name: str = args.project_name

        try:
            PLUGIN.get_api_client().delete_remote_project(project_name)
            logger.info("Project '{}' deleted.".format(project_name))
        except APIClientError as e:
            logger.info(str(e))
            return
        except ValueError as e:
            logger.info(str(e))
            return

    def open_file_dialog(self, args, unknown_args):
        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        from PySide2 import QtWidgets
        from submit2gm import utils

        file_mode = args.file_mode
        if file_mode == 'AnyFile':
            file_mode = QtWidgets.QFileDialog.AnyFile
        elif file_mode == 'ExistingFile':
            file_mode = QtWidgets.QFileDialog.ExistingFile
        elif file_mode == 'Directory':
            file_mode = QtWidgets.QFileDialog.Directory
        elif file_mode == 'ExistingFiles':
            file_mode = QtWidgets.QFileDialog.ExistingFiles
        else:
            logger.error("Unknown file mode type: {}".format(file_mode))
            return

        accept_mode = args.accept_mode
        if accept_mode == 'AcceptOpen':
            accept_mode = QtWidgets.QFileDialog.AcceptMode.AcceptOpen
        elif accept_mode == 'AcceptSave':
            accept_mode = QtWidgets.QFileDialog.AcceptMode.AcceptSave
        else:
            logger.error("Unknown accept mode type: {}".format(accept_mode))
            return

        accept_label: str = args.accept_label
        if not accept_label:
            accept_label = 'Select'

        reject_label: str = args.reject_label
        if not reject_label:
            reject_label = 'Cancel'

        file_name_label: str = args.file_name_label
        if not file_name_label:
            if file_mode is QtWidgets.QFileDialog.Directory:
                file_name_label = 'Directory'
            else:
                file_name_label = 'File'

        file_dialog = QtWidgets.QFileDialog()
        file_dialog.setFileMode(file_mode)
        file_dialog.setViewMode(QtWidgets.QFileDialog.Detail)
        file_dialog.setAcceptMode(accept_mode)
        file_dialog.setLabelText(QtWidgets.QFileDialog.Accept, accept_label)
        file_dialog.setLabelText(QtWidgets.QFileDialog.Reject, reject_label)
        file_dialog.setLabelText(QtWidgets.QFileDialog.FileName, file_name_label)
        file_dialog.setDirectory(utils.get_desktop_path())

        if file_dialog.exec_():
            file_names = file_dialog.selectedFiles()

            logger.info("Selected paths start:")
            for file_name in file_names:
                logger.info(file_name)
            logger.info("Selected paths end:")
        else:
            logger.info("Selection canceled")

    def submit_job(self, args, unknown_args):
        #import gm_meta_plugin
        #from gm_meta_plugin.gridmarkets import constants as api_constants
        from submit2gm import constants, utils
        from submit2gm.errors.api_client_error import APIClientError
        from submit2gm.errors.cli_prompt_not_enabled_error import CLIPromptNotEnabledError
        from submit2gm.inference_source import InferenceSource
        from submit2gm.dynamic_job_preset import DynamicJobPreset
        from submit2gm.packed_project import PackedProject


        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except APIClientError as e:
                logger.info(str(e))
                return
            except ValueError as e:
                logger.info(str(e))
                return

        project_dir = pathlib.Path(args.project_dir)
        job_definition_id = args.job_definition_id
        skip_upload = args.skip_upload

        api_client = PLUGIN.get_api_client()
        api_schema = api_client.get_api_schema()

        try:
            job_definition = api_schema.get_job_definition_with_id(job_definition_id)
        except ValueError as e:
            logger.info("No job definition with ID '{}'".format(job_definition_id))
            return

        job_attributes = job_definition.get_attributes()
        dynamic_job_preset = DynamicJobPreset(PLUGIN, 'temp-job-preset', 'temp-job-preset', job_definition)

        show_ui = args.show_ui
        enable_cli_prompt = args.enable_cli_prompt

        if show_ui:
            raise NotImplementedError

        # Set all job preset attributes to read from sys.argv
        project_inference_source = InferenceSource.get_project_inference_source()
        argv_inference_source = InferenceSource.get_argv_inference_source()
        for job_attribute in job_attributes:
            attribute = job_attribute.get_attribute()

            job_attribute_id = job_attribute.get_id()

            job_preset_attribute = dynamic_job_preset.get_job_preset_attribute_with_job_attribute_id(job_attribute_id)

            inference_sources = job_preset_attribute.get_job_attribute().get_inference_sources()
            if project_inference_source in inference_sources:
                job_preset_attribute.set_inference_source(project_inference_source)
            else:
                try:
                    job_preset_attribute.set_inference_source(argv_inference_source)
                except ValueError:
                    logger.info("Job Definition '{}' does not support command line submissions.\n Job attribute '{}' "
                                "does not support the {} inference source.".format(job_definition_id,
                                                                                   job_attribute_id,
                                                                                   argv_inference_source.get_id()))
                    return

            """
            if not attribute.has_argv_value() and not job_preset_attribute.get_job_attribute().is_optional():
                logger.info("Missing argument '{}' for Job attribute '{}' ".format('--' + key.replace('_', '-'),
                                                                                   job_definition_id))
                return
            """

        """
        unprocessed_job_preset_attributes = set(job_preset.get_job_preset_attributes())
        processed_job_preset_attributes = set()
        while True:
            for job_preset_attribute in unprocessed_job_preset_attributes:

                # If all of the dependencies have already been processed
                if set(job_preset_attribute.get_dependencies()).issubset(processed_job_preset_attributes):

                    # And the job preset attribute is "active"
                    if job_preset_attribute.is_active():
        """
        try:
            if skip_upload:
                project = PackedProject(project_dir,
                                        pathlib.Path('/{}'.format(project_dir)),
                                        set(),
                                        dict())

                api_client.submit_to_remote_project(project, dynamic_job_preset.get_static_job_preset(project))
            else:
                raise NotImplementedError()
                """
                # get a list of all the files associated with the scene
                project_files = utils.get_files_in_directory(project_dir)

                cached_api_schema = api_client.get_cached_api_schema()
                project_definition_schema = cached_api_schema.get_project_definition_schema()
                project_instance_builder_factory = PLUGIN.get_factory_collection().get_project_instance_builder_factory()
                project_instance_builder = project_instance_builder_factory.get_project_instance_builder(project_definition_schema)

                project_instance_builder.update_states(infer_from_cmd=True)
                project_attributes = project_instance_builder.get_project_attributes(force_transition=False)

                packed_project = PackedProject(project_attributes[constants.API_KEYS.PROJECT_NAME], project_dir, project_files, {})

                api_client.submit_new_project(packed_project, dynamic_job_preset)
                """
        except CLIPromptNotEnabledError as e:
            print(e)

    def process(self, job: 'WorkerJob') -> None:

        self.active_job = job

        command_line_arguments = job.get_argv()[1:] # [1:] to strip the program's path
        try:

            # If there are no arguments then print help info
            if not len(command_line_arguments):
                command_line_arguments = ['-h']

            # Parse the args without emitting any errors
            args, unknown_args = self.parser.parse_known_args(args=command_line_arguments)

            func = None
            try:
                func = args.func
            except AttributeError:
                pass

            if func:
                func(args, unknown_args)
            else:
                args, unknown_args = self.parser.parse_known_args(args=['-h'])
                args.func(args, unknown_args)

        except SystemExit:
            exc = sys.exc_info()[1]
            #print("SystemExit:", exc)

        job.mark_as_complete()

    def clean_up(self) -> None:
        self._clean_up_main_window()
