__all__ = 'MapContainer'

from PySide2 import QtCore

import typing

from .plugin_accessor import PluginAccessor

if typing.TYPE_CHECKING:
    from .submit2gm_plugin import Submit2gmPlugin

K = typing.TypeVar('K')
V = typing.TypeVar('T')


class MapContainer(typing.Generic[K, V], QtCore.QObject, PluginAccessor):
    """
    Not actually used since PySide2 doesn't support typing generics when it comes to inheritance. See JobPresetContainer
    for a "subclass" implementation
    """

    dataChanged = QtCore.Signal()
    itemAdded = QtCore.Signal(K, V)
    itemRemoved = QtCore.Signal(K, V)

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtCore.QObject'] = None):
        QtCore.QObject.__init__(self, parent)
        PluginAccessor.__init__(self, plugin)

        self._items: typing.Dict[K, V] = dict()

    def add(self, key: K, value: V) -> None:
        self._items[key] = value
        self.itemAdded.emit(key, value)

    def clear(self) -> None:
        old_items = self._items
        self._items = dict()

        for key, value in old_items:
            self.itemRemoved.emit(key, value)

        self.dataChanged.emit()

    def get(self, key: K, default: typing.Optional[K] = None) -> typing.Optional[V]:
        return self._items.get(key, default=default)

    def items(self) -> typing.List[typing.Tuple[K, V]]:
        items = []
        for key, value in self._items:
            items.append((key, value))
        return items

    def keys(self) -> typing.KeysView[K]:
        return self._items.keys()

    def values(self) -> typing.ValuesView[V]:
        return  self._items.values()

    def pop(self, key: K) -> V:
        try:
            value = self._items.pop(key)
            self.itemRemoved.emit(key, value)
        except KeyError as e:
            raise e

        self.dataChanged.emit()
        return value
