__all__ = ['BlenderPackedSceneBuilder']

import pathlib
import time
import typing

from . import blender_cpc

from submit2gm import constants
from submit2gm.packed_project import PackedProject
from submit2gm.packed_scene_builder import PackedSceneBuilder
from submit2gm.errors.file_packing_error import FilePackingError
from submit2gm.errors.project_requires_saving_error import ProjectRequiresSavingError

if typing.TYPE_CHECKING:
    from submit2gm.dynamic_job_preset import DynamicJobPreset
    from submit2gm.widgets.node_editor.node_widget import DependencyNode


class BlenderPackedSceneBuilder(PackedSceneBuilder):

    def _build_scene_internal(self,
                              project_name: str,
                              product: str,
                              product_version: str,
                              save_project_if_required: bool,
                              attributes: typing.Dict[str, any],
                              dynamic_job_presets: typing.Optional[typing.List['DynamicJobPreset']] = None,
                              dependency_graph: typing.Optional[typing.Dict[str, 'DependencyNode']] = None) -> 'PackedProject':

        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        # Is the scene saved
        if not blender_cpc.is_saved(cps):
            raise FilePackingError("Scene file not saved.")

        # Has the scene file been modified
        if blender_cpc.is_dirty(cps):
            if save_project_if_required:
                blender_cpc.save(cps)

                # Need to wait for the save operation to actually complete
                max_wait = 10
                while blender_cpc.is_dirty(cps):
                    time.sleep(1)
                    max_wait = max_wait - 1
                    if max_wait <= 0:
                        raise FilePackingError("Unable to save blender scene.")

            else:
                def _get_current_scene() -> typing.Tuple['PackedProject',
                                                         typing.List['DynamicJobPreset'],
                                                         typing.Dict[str, 'DependencyNode']]:

                    packed_project = self.get_current_scene(project_name,
                                                            product,
                                                            product_version,
                                                            True,
                                                            attributes,
                                                            dynamic_job_presets=dynamic_job_presets,
                                                            dependency_graph=dependency_graph)

                    return packed_project, dynamic_job_presets, dependency_graph

                raise ProjectRequiresSavingError(_get_current_scene, "Scene file has been modified.")

        blend_file = pathlib.Path(blender_cpc.get_file_path(cps))
        root_dir = blend_file.parent

        project_files = blender_cpc.trace_assets(cps, blend_file)
        project_files = set(map(lambda x: pathlib.Path(x), project_files))

        # set the path attribute
        attributes[constants.API_KEYS.PATH] = (root_dir / blend_file).as_posix()

        packed_project = PackedProject(project_name,
                                       root_dir,
                                       project_files,
                                       attributes)

        return packed_project

    def get_product_type(self) -> str:
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP)

    def get_product_version(self) -> str:
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP_VERSION)
