__all__ = 'BlenderJobPresetValidator'

import pathlib
import typing

from submit2gm import constants
from submit2gm.job_preset_validator import JobPresetValidator
from . import c4d_cpc

if typing.TYPE_CHECKING:
    from submit2gm.dynamic_job_preset import DynamicJobPreset


class C4DJobPresetValidator(JobPresetValidator):

    def validate(self, job_preset: 'DynamicJobPreset') -> typing.Tuple[bool, str]:
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        is_valid = True
        validation_message = u""

        global_save = c4d_cpc.get_global_save(cps)
        is_global_multipass_enabled = c4d_cpc.is_global_multipass_enabled(cps)

        save_regular_image = c4d_cpc.should_save_regular_image(cps)
        is_multipass_save_enabled = c4d_cpc.is_multipass_save_enabled(cps)

        # Validate the output prefix
        output_prefix_attribute = job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.OUTPUT_PREFIX)

        if not global_save:
            return False, "File saving is globally disabled for your scene.\n" \
                          "Please enable 'Save' in your Render Settings otherwise your results will not be saved."

        if (not save_regular_image) and (not is_global_multipass_enabled or not is_multipass_save_enabled):
            return False, "File saving is disabled for both regular and multi-pass outputs.\n" \
                          "Please enable either your regular image output or multi-pass output otherwise your results will not be saved."

        """
        C4D still saves the image even if the regular output or global output render options are disabled in the scenes 
        render settings when -oimage is provided on the command line so there is no need to validate that these are set.
        
        If we are not explicitly passing an output_prefix value then -oimage will not be set on the command line. This 
        means we need to validate the value defined in the scene file as well as whether it is enabled.
        """
        eval_output_prefix = output_prefix_attribute.is_active() and output_prefix_attribute.is_enabled()
        if eval_output_prefix or save_regular_image:

            if eval_output_prefix:
                output_prefix: str = output_prefix_attribute.eval()
                error_message_suffix = "in the Submit2gm submission interface for '{}'.".format(output_prefix_attribute.get_job_attribute().get_attribute().get_display_name())
            else:
                output_prefix = output_prefix_attribute.eval_application_defined_value()
                error_message_suffix = "in 'Render Settings > Save > Regular Image > File...'"

            if not output_prefix:
                return False, "No output file path set for regular image.\n" \
                              "Please set a value {}".format(error_message_suffix)

        return is_valid, validation_message
