__all__ = 'C4DPackedSceneBuilder'

import os
import pathlib
import typing

from submit2gm import constants
from submit2gm.errors.file_packing_error import FilePackingError
from submit2gm.errors.project_requires_saving_error import ProjectRequiresSavingError
from submit2gm.packed_project import PackedProject
from submit2gm.packed_scene_builder import PackedSceneBuilder
from . import c4d_cpc

if typing.TYPE_CHECKING:
    from submit2gm.dynamic_job_preset import DynamicJobPreset
    from submit2gm.widgets.node_editor.node_widget import DependencyNode


class C4DPackedSceneBuilder(PackedSceneBuilder):

    def _build_scene_internal(self,
                              project_name: str,
                              save_project_if_required: bool,
                              attributes: typing.Dict[str, any],
                              dynamic_job_presets: typing.List['DynamicJobPreset'],
                              dependency_graph: typing.Dict[str, 'DependencyNode']) -> 'PackedProject':

        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        # Is the scene saved
        if not c4d_cpc.is_document_saved(cps):
            raise FilePackingError("Scene file not saved.")

        # Has the scene file been modified
        if c4d_cpc.is_document_modified(cps):
            if save_project_if_required:
                c4d_cpc.save_project(cps)
            else:
                def _get_current_scene() -> typing.Tuple['PackedProject',
                                                         typing.List['DynamicJobPreset'],
                                                         typing.Dict[str, 'DependencyNode']]:

                    packed_project = self.get_current_scene(project_name,
                                                            True,
                                                            attributes,
                                                            dynamic_job_presets,
                                                            dependency_graph)

                    return packed_project, dynamic_job_presets, dependency_graph

                raise ProjectRequiresSavingError(_get_current_scene, "Scene file has been modified.")

        document_path = pathlib.Path(c4d_cpc.get_document_file_path(cps))
        root_dir = document_path.parent

        assets = c4d_cpc.get_all_assets(cps)
        if assets is None:
            raise FilePackingError("C4D failed to trace your scene file for assets.")

        project_files = set(map(lambda x: pathlib.Path(x), assets))

        # set the path attribute
        attributes[constants.API_KEYS.PATH] = str(root_dir / c4d_cpc.get_document_name(cps))

        packed_project = PackedProject(project_name,
                                       root_dir,
                                       project_files,
                                       attributes)

        return packed_project

    def get_product_type(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP)

    def get_product_version(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP_VERSION)
