__all__ = 'MaxJobPresetValidator'

import pathlib
import typing

from submit2gm import constants
from submit2gm.job_preset_validator import JobPresetValidator

if typing.TYPE_CHECKING:
    from submit2gm.dynamic_job_preset import DynamicJobPreset


class MaxJobPresetValidator(JobPresetValidator):

    def validate(self, job_preset: 'DynamicJobPreset') -> typing.Tuple[bool, str]:
        is_valid = True
        validation_message = ""

        # Validate 3ds renderer
        renderer_attribute = job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.RENDERER)
        renderer = renderer_attribute.eval()
        if renderer and "cloud_rendering" in renderer.casefold():
            return False, "Cloud Renderer is not a supported renderer on GridMarkets, please use a production renderer."

        # Validate the output prefix
        output_prefix_attribute = job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.OUTPUT_PREFIX)

        if output_prefix_attribute.is_active() and output_prefix_attribute.is_enabled():
            output_prefix: str = output_prefix_attribute.eval()
            error_message_suffix = "in the Submit2gm submission interface."
        else:
            assert False, "output prefix must always be active for 3ds."

        if not output_prefix:
            return False, "No value set for your output path.\n" \
                          "Please enter an output path value {}".format(error_message_suffix)

        # Check that the output path is relative
        if pathlib.Path(output_prefix).is_absolute() or output_prefix.startswith('/'):
            return False, "Output path must be a relative file path.\n" \
                          "Please set a relative value {}".format(error_message_suffix)

        # Check that the output path is within the project folder
        if output_prefix.startswith('..'):
            return False, "Output path must not start with '..' otherwise results may be written outside of the project folder.\n" \
                          "Please modify the value {}".format(error_message_suffix)

        return is_valid, validation_message
