__all__ = 'MaxPackedSceneBuilder'

import os
import pathlib
import typing
import tempfile

from submit2gm import constants, utils
from submit2gm.errors.file_packing_error import FilePackingError
from submit2gm.errors.project_requires_saving_error import ProjectRequiresSavingError
from submit2gm.packed_project import PackedProject
from submit2gm.packed_scene_builder import PackedSceneBuilder

from .max_scene_exporter import MaxSceneExporter
from . import max_cpc


if typing.TYPE_CHECKING:
    from submit2gm.dynamic_job_preset import DynamicJobPreset
    from submit2gm.widgets.node_editor.node_widget import DependencyNode


class MaxPackedSceneBuilder(PackedSceneBuilder):

    def _build_scene_internal(self,
                              project_name: str,
                              save_project_if_required: bool,
                              attributes: typing.Dict[str, any],
                              dynamic_job_presets: typing.List['DynamicJobPreset'],
                              dependency_graph: typing.Dict[str, 'DependencyNode']) -> 'PackedProject':

        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        file_name = max_cpc.get_max_file_name(cps)

        # If the scene has not been saved
        if file_name == "":
            raise FilePackingError(u"Scene file not saved.")

        file_path = os.path.join(max_cpc.get_max_file_path(cps), file_name)
        pre_packed_local_root = pathlib.Path(max_cpc.get_current_project_folder(cps))

        # If the scene has been created, saved and then erased from the disk
        if os.path.exists(file_path) == 0:
            raise FilePackingError(u"Scene file '{}' does not exist on filesystem.".format(str(file_path)))

        # If the scene has been modified then save the scene
        if max_cpc.is_save_required(cps):
            if save_project_if_required:
                max_cpc.save_current_scene(cps)
            else:
                def _get_current_scene() -> typing.Tuple['PackedProject',
                                                         typing.List['DynamicJobPreset'],
                                                         typing.Dict[str, 'DependencyNode']]:

                    packed_project = self.get_current_scene(project_name,
                                                            True,
                                                            attributes,
                                                            dynamic_job_presets,
                                                            dependency_graph)

                    return packed_project, dynamic_job_presets, dependency_graph

                raise ProjectRequiresSavingError(_get_current_scene, "Scene file has been modified.")

        # get the project root dir
        # create a new temp directory
        try:
            import tempfile
            temp_dir = tempfile.mkdtemp(prefix='gm-temp-dir-')

            # Resolve the path because Max OS likes to return a symlink path instead of the real path sometimes for temp
            # files
            temp_dir_path = pathlib.Path(temp_dir).resolve()

            if os.name == 'nt':
                # Windows sometimes returns short name versions of the path so expand the path to the full name
                temp_dir_path = utils.expand_windows_short_name_path(temp_dir_path)

            root_dir = MaxSceneExporter(plugin).export(temp_dir_path)
        finally:
            # Make sure the users actual scene is loaded again once exporting is complete
            max_cpc.load_max_scene_file(cps, file_path)

        # get a list of all the files associated with the scene
        project_files = utils.get_files_in_directory(root_dir)

        # set the path attribute
        attributes[constants.API_KEYS.PATH] = str(root_dir / file_name)

        packed_project = PackedProject(project_name,
                                       root_dir,
                                       project_files,
                                       attributes,
                                       pre_packed_local_root=pre_packed_local_root)

        return packed_project

    def get_product_type(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP)

    def get_product_version(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP_VERSION)
