import typing

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import CommandPortServer


__GET_ALL_WRITE_NODES = """
def get_all_write_nodes(nodes=None):

    if nodes is None:
        nodes = get_all_nodes()

    write_nodes = filter(lambda node: node.Class() in [u'DeepWrite', u'GenerateLUT', u'Write', u'WriteGeo'], nodes)
    return write_nodes
"""


def get_version_string(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import nuke
nuke_version_string = nuke.NUKE_VERSION_STRING
""", return_variable_name='nuke_version_string')


def get_project_dir(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import nuke
project_directory = nuke.root().knob('project_directory').evaluate()
if project_directory is None:
    project_directory = ''
""", return_variable_name='project_directory')


def get_script_name(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import nuke
try:
    nuke_script_name = nuke.scriptName()
except RuntimeError as e:
    nuke_script_name = ""
""", return_variable_name='nuke_script_name')


def get_frame_range_str(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import nuke
frame_range_str = ''
root = nuke.root()
start_frame = root.firstFrame()
end_frame = root.lastFrame()
frame_step = 1
frame_range_str = "{0} {1} {2}".format(start_frame, end_frame, frame_step)
""", return_variable_name='frame_range_str')


def is_script_dirty(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import nuke
is_script_dirty = nuke.root().modified()
""", return_variable_name='is_script_dirty')


def save_script(cps: 'CommandPortServer') -> bool:
    return cps.eval_cmd("""
import nuke
successfull = nuke.scriptSave()
""", return_variable_name='successfull')


def get_write_nodes_without_create_dir_enabled(cps: 'CommandPortServer'):
    return cps.eval_cmd(f"""
import nuke

{__GET_ALL_WRITE_NODES}

write_nodes = get_all_write_nodes()
write_nodes_without_create_dir_enabled = []

for write_node in write_nodes:
    if not write_node.knob('create_directories').value():
        write_nodes_without_create_dir_enabled.append(write_node.name())

""", return_variable_name='write_nodes_without_create_dir_enabled')


def get_read_and_write_paths(cps: 'CommandPortServer', frame_ranges: typing.List[typing.Tuple[int, int, int]]):
    return cps.eval(f"""

import nuke    

READ_NODE_CLASSES = [u'AudioRead',
                     u'Axis',
                     u'Axis2',
                     u'Camera',
                     u'Camera2',
                     u'DeepRead',
                     u'OCIOFileTransform',
                     u'ParticleCache',
                     u'Precomp',
                     u'Read',
                     u'ReadGeo',
                     u'ReadGeo2',
                     u'Vectorfield' ]

PATH_KNOB_NAMES = [u'proxy', u'file', u'vfield_file']



def get_all_nodes():
    try:
        # not all nuke versions support 'recurseGroups'
        nodes = nuke.allNodes(recurseGroups=True)
    except:
        nodes = nuke.allNodes()

    return nodes


def get_all_read_nodes(nodes=None):

    if nodes is None:
        nodes = get_all_nodes()

    read_nodes = filter(lambda node: node.Class() in READ_NODE_CLASSES, nodes)
    return read_nodes


{__GET_ALL_WRITE_NODES}


def get_path_knobs_for_node(node):
    path_knobs = []

    for path_knob_name in PATH_KNOB_NAMES:
        path_knob = node.knob(path_knob_name)

        if path_knob is not None:
            path_knobs.append(path_knob)

    return path_knobs


def get_path_knobs_for_nodes(nodes):
    path_knobs = []

    for node in nodes:
        path_knobs.extend(get_path_knobs_for_node(node))

    return path_knobs


def get_indices_of_views(views):
    all_views = nuke.views()

    view_indices = []
    for view in views:
        try:
            view_indices.append(all_views.index(view) + 1)
        except ValueError as e:
            raise ValueError(u"View " + view + u" can not be found.")

    return view_indices


def evaluate_path_knobs(path_knobs, frame_ranges, views):
    # Returns a list of path strings without duplicates

    # get the index of each provided view
    view_indices = get_indices_of_views(views)

    output_context = nuke.OutputContext()
    paths = list()

    for path_knob in path_knobs:

        path_value = path_knob.value()
        evaluated_path_value = path_knob.evaluate()

        if not path_value and not evaluated_path_value:
            continue

        # Check for scripting expressions by comparing the value and the evaluated value
        if evaluated_path_value == path_value:
            if path_value and path_value not in paths:
                paths.append(path_value)
        else:
            # Evaluate expressions across every frame range
            for frame_range in frame_ranges:

                # And for every view (In case path contains '%v' or '%V')
                for view_index in view_indices:
                    current_frame = frame_range.first()
                    frame_end = frame_range.last()
                    frame_step = frame_range.stepFrame()

                    while current_frame <= frame_end:
                        # update output context
                        output_context.setFrame(current_frame)
                        output_context.setView(view_index)

                        evaluated_path_value = path_knob.getEvaluatedValue(output_context)

                        if evaluated_path_value and evaluated_path_value not in paths:
                            paths.append(evaluated_path_value)

                        current_frame = current_frame + frame_step

    return paths


def get_evaluated_read_paths(frame_ranges, views):
    read_nodes = get_all_read_nodes()
    path_knobs = get_path_knobs_for_nodes(read_nodes)
    return evaluate_path_knobs(path_knobs, frame_ranges, views)


def get_evaluated_write_paths(frame_ranges, views):
    write_nodes = get_all_write_nodes()
    path_knobs = get_path_knobs_for_nodes(write_nodes)
    return evaluate_path_knobs(path_knobs, frame_ranges, views)

# Get frame ranges
root = nuke.root()

requested_frame_ranges = {str(frame_ranges)}

#frame_start = root.firstFrame()
#frame_end = root.lastFrame()
#frame_step = 1
#frame_range = nuke.FrameRange(frame_start, frame_end, frame_step)
#frame_ranges = [frame_range]

frame_ranges = list(map(lambda x: nuke.FrameRange(x[0], x[1], x[2]), requested_frame_ranges))

# Get views
views = nuke.views()

nuke_read_paths = get_evaluated_read_paths(frame_ranges, views)
nuke_write_paths = get_evaluated_write_paths(frame_ranges, views)

""", ['nuke_read_paths', 'nuke_write_paths'])
