__all__ = 'PackedSceneBuilder'

import typing
import pathlib

from submit2gm import constants
from submit2gm.packed_project import PackedProject
from submit2gm.packed_scene_builder import PackedSceneBuilder
from submit2gm.errors.file_packing_error import FilePackingError
from submit2gm.errors.project_requires_saving_error import ProjectRequiresSavingError
from submit2gm.widgets.input_widgets.frame_range_input import FrameRangeInput
from . import nuke_cpc

if typing.TYPE_CHECKING:
    from submit2gm.dynamic_job_preset import DynamicJobPreset
    from submit2gm.widgets.node_editor.node_widget import DependencyNode


class NukePackedSceneBuilder(PackedSceneBuilder):

    def _build_scene_internal(self,
                              project_name: str,
                              save_project_if_required: bool,
                              attributes: typing.Dict[str, any],
                              dynamic_job_presets: typing.List['DynamicJobPreset'],
                              dependency_graph: typing.Dict[str, 'DependencyNode']) -> 'PackedProject':

        cps = self.get_plugin().get_active_command_port_server()

        # get the file name of the script
        file_name = pathlib.Path(nuke_cpc.get_script_name(cps))

        if not file_name:
            raise FilePackingError("Nuke script not saved. Please save and retry.")

        # If the scene has been created, saved and then erased from the disk
        if not file_name.exists():
            raise FilePackingError("Script file '" + str(file_name) + "' does not exist on filesystem.")

        # Save the script if the script has been modified
        if nuke_cpc.is_script_dirty(cps):
            if save_project_if_required:
                nuke_cpc.save_script(cps)
            else:
                def _get_current_scene() -> typing.Tuple['PackedProject',
                                                         typing.List['DynamicJobPreset'],
                                                         typing.Dict[str, 'DependencyNode']]:

                    packed_project = self.get_current_scene(project_name,
                                                            True,
                                                            attributes,
                                                            dynamic_job_presets,
                                                            dependency_graph)

                    return packed_project, dynamic_job_presets, dependency_graph

                raise ProjectRequiresSavingError(_get_current_scene, "Scene file has been modified.")

        # get the project root dir
        root_dir = pathlib.Path(nuke_cpc.get_project_dir(cps))

        read_paths = []
        for dynamic_job_preset in dynamic_job_presets:
            frame_ranges = dynamic_job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.FRAMES).eval()
            read_and_write_paths = nuke_cpc.get_read_and_write_paths(cps, FrameRangeInput.parse_frame_ranges(frame_ranges))
            read_paths.extend(read_and_write_paths['nuke_read_paths'])

        # Convert to pathlib paths
        read_paths = list(map(lambda x: pathlib.Path(x), read_paths))

        project_files = [file_name] + read_paths

        # set the path attribute
        attributes[constants.API_KEYS.PATH] = file_name.as_posix()

        # Check if create_directories is set to true for all write nodes
        write_nodes_without_create_dir_enabled = nuke_cpc.get_write_nodes_without_create_dir_enabled(cps)
        if write_nodes_without_create_dir_enabled:
            raise FilePackingError("Write node '{0}' does not have 'create directories' enabled.\n"
                                   "This is required otherwise the node may not be able to write it's results on our "
                                   "system.\n\nPlease enable 'create directories' for ""'{0}' and any other write "
                                   "nodes".format(write_nodes_without_create_dir_enabled[0]))

        packed_project = PackedProject(project_name, root_dir, set(project_files), attributes)

        return packed_project

    def get_product_type(self) -> str:
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP)

    def get_product_version(self) -> str:
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, constants.API_KEYS.APP_VERSION)
