__all__ = 'JobPresetValidator'

import typing

from . import constants
from .widgets.input_widgets.frame_range_input import FrameRangeInput
from .job_preset_validator import JobPresetValidator

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin
    from submit2gm.submit2gm_dyn.blender.blender_job_preset_validator import BlenderJobPresetValidator
    from submit2gm.submit2gm_dyn.c4d.c4d_job_preset_validator import C4DJobPresetValidator
    from submit2gm.submit2gm_dyn.max.max_job_preset_validator import MaxJobPresetValidator
    from submit2gm.submit2gm_dyn.maya.maya_job_preset_validator import MayaJobPresetValidator
    from submit2gm.submit2gm_dyn.vray.vray_job_preset_validator import VrayJobPresetValidator
    from submit2gm.dynamic_job_preset import DynamicJobPreset


class Submit2gmJobPresetValidator(JobPresetValidator):

    def __init__(self, plugin):
        JobPresetValidator.__init__(self, plugin)
        self._blender_job_preset_validator: typing.Optional['BlenderJobPresetValidator'] = None
        self._c4d_job_preset_validator: typing.Optional['C4DJobPresetValidator'] = None
        self._max_job_preset_validator: typing.Optional['MaxJobPresetValidator'] = None
        self._maya_job_preset_validator: typing.Optional['MayaJobPresetValidator'] = None
        self._vray_job_preset_validator: typing.Optional['VrayJobPresetValidator'] = None

    def get_blender_job_preset_validator(self) -> 'BlenderJobPresetValidator':
        if self._blender_job_preset_validator is None:
            from submit2gm.submit2gm_dyn.blender.blender_job_preset_validator import BlenderJobPresetValidator
            self._blender_job_preset_validator = BlenderJobPresetValidator(self.get_plugin())
        return self._blender_job_preset_validator

    def get_c4d_job_preset_validator(self) -> 'C4DJobPresetValidator':
        if self._c4d_job_preset_validator is None:
            from submit2gm.submit2gm_dyn.c4d.c4d_job_preset_validator import C4DJobPresetValidator
            self._c4d_job_preset_validator = C4DJobPresetValidator(self.get_plugin())
        return self._c4d_job_preset_validator

    def get_max_job_preset_validator(self) -> 'MaxJobPresetValidator':
        if self._max_job_preset_validator is None:
            from submit2gm.submit2gm_dyn.max.max_job_preset_validator import MaxJobPresetValidator
            self._max_job_preset_validator = MaxJobPresetValidator(self.get_plugin())
        return self._max_job_preset_validator

    def get_maya_job_preset_validator(self) -> 'MayaJobPresetValidator':
        if self._maya_job_preset_validator is None:
            from submit2gm.submit2gm_dyn.maya.maya_job_preset_validator import MayaJobPresetValidator
            self._maya_job_preset_validator = MayaJobPresetValidator(self.get_plugin())
        return self._maya_job_preset_validator

    def get_vray_job_preset_validator(self) -> 'VrayJobPresetValidator':
        if self._vray_job_preset_validator is None:
            from submit2gm.submit2gm_dyn.vray.vray_job_preset_validator import VrayJobPresetValidator
            self._vray_job_preset_validator = VrayJobPresetValidator(self.get_plugin())
        return self._vray_job_preset_validator

    def get_plugin(self) -> 'Submit2gmPlugin':
        return self._plugin

    def validate(self, job_preset: 'DynamicJobPreset') -> typing.Tuple[bool, str]:
        is_valid = True
        validation_message = u""

        is_saved_attribute = job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.IS_SAVED)
        if not is_saved_attribute.eval():
            return False, "Please save your project before submitting."

        # default comment if none passed
        job_name_attribute = job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.JOB_NAME)
        if not job_name_attribute.eval():
            return False, "No value passed for Job Name, please enter a value for your job's name."

        app_attribute = job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.APP)

        # Validate that the credits budget is greater than zero
        try:
            credits_budget_attribute = job_preset.get_job_preset_attribute_with_job_attribute_id(
                constants.API_KEYS.CREDITS_BUDGET)

            if credits_budget_attribute.is_enabled() and credits_budget_attribute.is_active():
                credits_budget = credits_budget_attribute.eval()

                if credits_budget <= 0:
                    return False, "Credits budget must be greater than zero."

        except ValueError:
            pass

        # Validate that no frame number is a float
        try:
            frame_range_attribute = job_preset.get_job_preset_attribute_with_job_attribute_id(
                constants.API_KEYS.FRAMES)

            if frame_range_attribute.is_enabled() and frame_range_attribute.is_active():
                frame_ranges_str = frame_range_attribute.eval()

                if '.' in frame_ranges_str:
                    return False, "Frame range must only contain integer values.\nI.e. no decimal points."

                frame_ranges = FrameRangeInput.parse_frame_ranges(frame_ranges_str)

                include_frame_range_number = len(frame_ranges) > 1

                for i, frame_range in enumerate(frame_ranges):
                    if frame_range[2] == 0:
                        if include_frame_range_number:
                            return False, f"Frame range number {i + 1} must have a non zero step size."
                        else:
                            return False, "Frame range must have a non zero step size."

        except ValueError:
            pass

        app = app_attribute.eval()
        if app == constants.PRODUCTS.BLENDER:
            return self.get_blender_job_preset_validator().validate(job_preset)

        if app == constants.PRODUCTS.C4D:
            return self.get_c4d_job_preset_validator().validate(job_preset)

        if app == constants.PRODUCTS.MAX:
            return self.get_max_job_preset_validator().validate(job_preset)

        if app == constants.PRODUCTS.MAYA:
            return self.get_maya_job_preset_validator().validate(job_preset)

        if app == constants.PRODUCTS.VRAY:
            return self.get_vray_job_preset_validator().validate(job_preset)

        return is_valid, validation_message
