from PySide2 import QtCore

import json
import requests
import typing

from .api_client import APIClient
from .plugin_utils import PluginUtils
from .submit2gm_attribute_source import Submit2gmAttributeSource
from . import constants

from .factory_collection import FactoryCollection
from .api_schema_factory import APISchemaFactory
from .job_preset_container import JobPresetContainer
from .widgets.log_history_container_widget import LogHistoryContainerWidget
from .preferences_container import PreferencesContainer
from .user_container import UserContainer

from .errors.command_port_server_error import CommandPortServerError

from .version import VERSION
from .plugin_version import PluginVersion
from .submit2gm_packed_scene_builder import Submit2GMPackedSceneBuilder
from .sumbit2gm_user_interface import Submit2GMUserInterface
from .submit2gm_job_preset_validator import Submit2gmJobPresetValidator
from .logging_coordinator import LoggingCoordinator

if typing.TYPE_CHECKING:
    from submit2gm.job_definition import JobDefinition


class CommandPortServer:

    def __init__(self, product: str, address: str, meta_info: typing.Optional[str] = None):
        self._product = product
        self._address = address
        self._session = requests.Session()

        if meta_info:
            self._meta_info = meta_info
        else:
            self._meta_info = ""

    def get_product(self) -> str:
        return self._product

    def get_address(self) -> str:
        return self._address

    def get_meta_info(self) -> str:
        return self._meta_info

    def is_still_running(self) -> bool:
        from urllib.parse import urlparse
        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        url = urlparse(self._address)
        netloc = url.netloc
        address, port = netloc.split(':')

        can_bind_to_socket = True

        try:
            s.bind((address, int(port)))
        except socket.error as e:
            can_bind_to_socket = False
        finally:
            try:
                s.close()
            except Exception:
                pass

        is_cps_still_running = not can_bind_to_socket
        return is_cps_still_running

    def eval(self, code: str, return_variables: typing.List[str] = None) -> typing.Dict['str', 'any']:

        if return_variables is None:
            return_variables = []

        url = '{}/eval'.format(self.get_address())
        payload = code.encode('utf8')
        headers = {
            'Content-Type': 'text/plain',
            'Return-Variables': json.dumps(return_variables)
        }

        try:
            response = self._session.post(url,
                                          data=payload,
                                          headers=headers,
                                          timeout=constants.COMMAND_PORT_SERVER_TIMEOUT)

        except requests.exceptions.ConnectTimeout as e:
            raise CommandPortServerError(
                "Connection error to GridMarkets {} plug-in. Connect timeout.".format(self._product))
        except requests.exceptions.ReadTimeout as e:
            raise CommandPortServerError(
                "Connection error to GridMarkets {} plug-in. Read timeout.".format(self._product))
        except Exception as e:
            raise CommandPortServerError(
                "Connection error to GridMarkets {} plug-in.\n{}\n{}".format(self._product, str(type(e)), str(e)))

        binary_data = response.content

        # Keep alive tokens are sent as part of the response so that the timeout can tell if the cps has crashed or if
        # it is just taking a long time
        binary_data = binary_data.strip(constants.COMMAND_PORT_SERVER_TIMEOUT_KEEP_ALIVE_TOKEN)

        try:
            utf8_data = binary_data.decode('utf8')
        except UnicodeDecodeError as e:
            raise CommandPortServerError("Unable to decode response: {}".format(str(e)))

        try:
            json_result = json.loads(utf8_data)
        except json.decoder.JSONDecodeError as e:
            # If the response is not valid json then there is a problem
            raise CommandPortServerError("Invalid response from command port server: JSON error\n" + response.text)

        if 'error' in json_result:
            exc_msg = json_result['error']

            raise CommandPortServerError("Unable to evaluate request. Code:\n{}\n\nException:{}".format(code[0:20],
                                                                                                        exc_msg))

        return json_result

    def eval_cmd(self, command: str, return_variable_name: typing.Optional[str] = None) -> any:
        if not return_variable_name:
            return_variable_name = command.split()[0]

        results = self.eval(command, [return_variable_name])
        return results[return_variable_name]

    def shutdown(self) -> None:
        url = '{}/shutdown'.format(self.get_address())
        self._session.post(url, timeout=5)

    def close_session(self) -> None:
        self._session.close()


class Submit2gmPlugin(QtCore.QObject):

    def _get_factory_collection(self):
        api_schema_factory = APISchemaFactory()

        return FactoryCollection(api_schema_factory)

    def __init__(self, parent: typing.Optional['QtCore.QObject'] = None):
        QtCore.QObject.__init__(self, parent=parent)

        version = VERSION

        self._version = PluginVersion(version[0], version[1], version[2])

        self._factory_collection = self._get_factory_collection()

        log_history_container = LogHistoryContainerWidget(self, [])
        self._logging_coordinator = LoggingCoordinator(self, log_history_container)

        user_container = UserContainer(self, [])
        self._job_preset_container = JobPresetContainer(self)
        self._preferences_container = PreferencesContainer(user_container)

        self._api_client = APIClient(self, self._logging_coordinator, self._factory_collection)
        self._user_interface = Submit2GMUserInterface(self)
        self._plugin_utils = PluginUtils(self)
        self._command_port_servers = []
        self._application_attribute_source = Submit2gmAttributeSource(self)
        self._packed_scene_builder = Submit2GMPackedSceneBuilder(self)
        self._job_preset_validator = Submit2gmJobPresetValidator(self)

        self._initial_product = None

    def get_name(self) -> str:
        return constants.PROGRAM_NAME

    def get_version(self) -> 'PluginVersion':
        return self._version

    def is_debug_build(self) -> bool:
        try:
            from .debug_info import DEBUG_INFO
            return True
        except ImportError:
            return False

    def get_debug_info(self) -> str:
        from .debug_info import DEBUG_INFO
        return DEBUG_INFO

    def get_logging_coordinator(self) -> 'LoggingCoordinator':
        return self._logging_coordinator

    def get_preferences_container(self) -> 'PreferencesContainer':
        return self._preferences_container

    def get_job_preset_container(self) -> 'JobPresetContainer':
        return self._job_preset_container

    def get_api_client(self) -> 'APIClient':
        return self._api_client

    def get_user_interface(self) -> 'Submit2GMUserInterface':
        return self._user_interface

    def add_command_port_server(self, product: str, address: str, meta_info: typing.Optional[str] = None) -> None:

        if meta_info is None:
            meta_info = ""

        cps = CommandPortServer(product, address, meta_info=meta_info)
        self._command_port_servers.append(cps)

    def get_active_command_port_server(self) -> CommandPortServer:

        if self._command_port_servers:
            return self._command_port_servers[0]

        raise ValueError("No command port servers")

    def shutdown_command_port_server(self, cps_address: str) -> CommandPortServer:

        for cps in self._command_port_servers:
            if cps.get_address() == cps_address:
                command_port_server = cps
                break
        else:
            raise ValueError("No command port server with address '{}'".format(cps_address))

        try:
            command_port_server.shutdown()
        except Exception as e:
            pass

        self._command_port_servers.remove(command_port_server)
        return command_port_server

    def shutdown_command_port_servers(self) -> None:

        for cps in self._command_port_servers:
            try:
                cps.shutdown()
            except Exception as e:
                self.get_logging_coordinator().get_logger().error(str(e))

        self._command_port_servers = []

    def get_application_attribute_source(self) -> 'Submit2gmAttributeSource':
        return self._application_attribute_source

    def get_packed_scene_builder(self) -> 'Submit2GMPackedSceneBuilder':
        return self._packed_scene_builder

    def get_factory_collection(self) -> 'FactoryCollection':
        return self._factory_collection

    def get_plugin_utils(self) -> 'PluginUtils':
        return self._plugin_utils

    def get_initial_product(self) -> typing.Optional[str]:
        return self._initial_product

    def set_initial_product(self, initial_product: str) -> None:
        self._initial_product = initial_product

    def get_compatible_job_definitions(self) -> typing.List['JobDefinition']:
        compatible_job_definitions = []

        api_schema = self.get_api_client().get_cached_api_schema()

        # if there is no active cps then don't create a default job preset
        try:
            active_cps = self.get_active_command_port_server()
            product_type = active_cps.get_product()
        except ValueError:
            product_type = self.get_initial_product()

        for job_definition in api_schema.get_job_definitions():
            if product_type in job_definition.get_supported_cps_types():
                compatible_job_definitions.append(job_definition)

        return compatible_job_definitions

    def create_default_job_presets(self) -> None:
        compatible_job_definitions = self.get_compatible_job_definitions()
        job_preset_container = self.get_job_preset_container()

        if not compatible_job_definitions:
            raise ValueError("no compatible job definitions found")

        job_definition = compatible_job_definitions[0]
        job_preset_container.emplace(job_definition)

        user_interface = self.get_user_interface()
        user_interface.get_main_window().display_signed_in_ui()

    def get_job_preset_validator(self) -> 'Submit2gmJobPresetValidator':
        return self._job_preset_validator
