import math
import os
import pathlib
import typing
from . import constants


def remove_directory(directory: pathlib.Path) -> None:
    if not directory.is_dir():
        return

    for item in directory.iterdir():
        if item.is_dir():
            remove_directory(item)
        else:
            item.unlink()

    directory.rmdir()


def get_files_in_directory(directory: pathlib.Path) -> typing.Set[pathlib.Path]:

    files = set()

    if not directory.is_dir():
        return files

    try:
        for item in directory.iterdir():
            if item.is_dir():
                files.update(get_files_in_directory(item))
            elif item.is_file():
                files.add(item)

    except PermissionError:
        pass

    return files


def get_deep_attribute(obj: object, attribute_path: str or typing.List[str]) -> any:
    if type(attribute_path) == str:
        return get_deep_attribute(obj, attribute_path.split('.')[0])

    current_object = obj
    for attribute_name in attribute_path:
        current_object = getattr(current_object, attribute_name)

    return current_object


def get_exception_with_traceback(exception: Exception) -> str:
    import traceback

    try:
        error_message = str(exception)
    except TypeError as e:
        error_message = "non string error message"

    # if the return value is an exception, add it's stack trace to the error message
    if hasattr(exception, "__traceback__"):
        stack_trace = traceback.format_tb(exception.__traceback__)
        for line in stack_trace:
            error_message = error_message + str(line)

    return error_message


def expand_windows_short_name_path(path: pathlib.Path) -> pathlib.Path:
    if os.name != 'nt':
        raise ValueError("Attempting to expand windows short path on none windows system ('{}')".format(os.name))

    import ctypes

    # This buffer size is pretty arbitrary but it should be large enough
    BUFFER_SIZE = len(str(path)) + 500
    buffer = ctypes.create_unicode_buffer(BUFFER_SIZE)
    ctypes.windll.kernel32.GetLongPathNameW(str(path), buffer, BUFFER_SIZE)

    return pathlib.Path(buffer.value)


def get_unique_id(objects: typing.List[object], attribute_name: str = 'id') -> int:
    """ Returns an integer ID that has not been used by any other object in the provided list.

    :param objects: a list of existing objects that may or may not have an id attribute
    :type objects: list
    :param attribute_name: The name of the attribute or getter method containing the id
    :type attribute_name: unicode
    :return: the new unique id
    :rtype: int
    """

    used_ids = []

    for object in objects:

        if not hasattr(object, attribute_name):
            continue

        id = getattr(object, attribute_name)
        if callable(id):
            id = id()

        if not isinstance(id, int):
            continue

        used_ids.append(id)

    # find a number which has not been used yet
    unused_id = 0

    while unused_id in used_ids:
        unused_id = unused_id + 1

    return unused_id


def create_unique_object_name(objects: typing.List[object],
                              attribute_name: str = 'name',
                              name_prefix: str = "") -> str:
    """ Creates a unique name of the form [name_prefix][x] where x is some unique positive integer.

    :param objects: the existing list of objects that this new name must not collide with. Objects in said list may or
                    may not have .name attributes or follow the provided naming scheme
    :type objects: list
    :param attribute_name: The name of the attribute or getter method which contains the objects name
    :type attribute_name: str
    :param name_prefix: a optional prefix to append before the unique number
    :type name_prefix: str
    :return: the unique name
    :rtype: str
    """

    prefix_length = len(name_prefix)

    # compile a list of used numeric integer suffixes
    used_suffixes = []
    for element in objects:

        if not hasattr(element, attribute_name):
            continue

        name = getattr(element, attribute_name)

        if callable(name):
            name = name()

        if not isinstance(name, str):
            continue

        if not name.startswith(name_prefix):
            continue

        element_suffix = name[prefix_length:]
        if element_suffix.isdigit():
            try:
                element_suffix = int(element_suffix)

                if element_suffix >= 0:
                    used_suffixes.append(element_suffix)

            except ValueError:
                pass

    # remove duplicates
    used_suffixes = list(set(used_suffixes))

    # sort
    used_suffixes.sort()

    # find a number which has not been used yet
    i = 0
    while i in used_suffixes:
        i = i+1

    return name_prefix + str(i)


def get_desktop_path() -> str:
    import os

    if os.name == 'nt':
        return os.path.join(os.path.join(os.environ['USERPROFILE']), 'Desktop')
    elif os.name == 'posix':
        return os.path.join(os.path.join(os.path.expanduser('~')), 'Desktop')
    else:
        return ""


def get_external_files_remote_path(remote_root: str, path: pathlib.Path) -> str:
    """
    Gets the external files remapped location under the external files directory.
    External files always go under the "/<project name>/gm_external_files/" directory
    """

    if not path.is_absolute():
        raise ValueError("Path '{}' must be absolute.".format(str(path)))

    if path.drive:
        # If this is a windows path with a drive then we need to remove the ':' character from the path in case the
        # image is using windows where ':' is a reserved character
        path_str = str(path)

        if path_str[1] == ':':
            path_str = path_str.replace(':', '', 1)

        path = pathlib.Path(path_str)

    return (pathlib.Path(remote_root) / constants.EXTERNAL_FILES_DIR_NAME) / path.as_posix()


def get_external_files_remote_root(remote_root: str, path: 'pathlib.Path') -> pathlib.Path:
    """
    Gets the external files remapped location under the external files directory.
    External files always go under the "/<project name>/gm_external_files/" directory
    """

    if not path.is_absolute():
        raise ValueError("Path must be absolute.")

    if path.drive:
        # If this is a windows path with a drive then we need to remove the ':' character from the path in case the
        # image is using windows where ':' is a reserved character
        path_str = path.drive

        if path_str[1] == ':':
            path_str = path_str.replace(':', '', 1)

        return (pathlib.Path(remote_root) / constants.EXTERNAL_FILES_DIR_NAME) / path_str

    return pathlib.Path(remote_root) / constants.EXTERNAL_FILES_DIR_NAME


def bytes_to_human(byte_num: int) -> str:
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

    if byte_num == 0:
        return "0 B"

    byte_num = int(byte_num)
    index = int(math.floor(math.log(byte_num, 1000)))
    power = math.pow(1000, index)
    size = round(byte_num / power, 2)
    return "{} {}".format(size, size_name[index])


def do_frame_ranges_overlap(frame_ranges: typing.List[typing.Tuple[int, int, int]]) -> bool:
    """ Detects if the the provided frame ranges overlap, respects disabled frame ranges. """

    # buffer to store frames already seen
    frames_buffer = []

    for frame_range in frame_ranges:

        start = frame_range[0]
        end = frame_range[1]
        step = frame_range[2]

        # add all frames in range to frames buffer
        i = start
        while i <= end:

            if i in frames_buffer:
                return True

            frames_buffer.append(i)
            i = i + step

    return False


def format_table_str(table: typing.List[typing.List[str]]) -> str:
    """
    Pretty prints a 2d matrix of strings in table format

    :param table: A list of table columns
    :return: str
    """

    output = ""

    columns = len(table)

    if columns:
        rows = len(table[0])

        column_widths: typing.List[int] = []

        # Calculate the maximum width of each column
        for column in table:

            max_column_width = 0

            if len(column) != rows:
                raise ValueError("Inconsistent column lengths")

            for value in column:
                str_length = len(value)

                if max_column_width < str_length:
                    max_column_width = str_length

            column_widths.append(max_column_width)

        col = 0
        row = 0

        while row < rows - 1:

            while True:
                # Add 1 to the max column width so there is some space between the columns
                col_max_width = column_widths[col] + 1

                value = table[col][row]
                cell_str = "{0:{1}}".format(value, col_max_width)
                output = output + cell_str
                col = col + 1

                if col >= columns:
                    output = output + '\n'
                    break

            row = row + 1
            col = 0

    return output


def format_invalid_request_error(e: 'any') -> str:
    e_msg = str(e)
    if hasattr(e, 'errors'):
        if 'errors' in e.errors:
            errors = e.errors['errors']
            e_msg += '\n'
            for error in errors:
                e_msg += '\n{}'.format(str(error['message']))
                if 'parameter' in error:
                    e_msg += ' ({})'.format(str(error['parameter']))

    return e_msg
