__all__ = 'DynamicJobPresetAttributeWidget'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets
import typing

from submit2gm import constants, qt_utils
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.dynamic_job_preset_attribute import DynamicJobPresetAttribute
from submit2gm.widgets.elided_label import ElidedLabel
from submit2gm.widgets.input_widgets import Constant, CheckBox, ComboBox, ListRadioInput, InputWidget
from submit2gm.widgets.resizable_stacked_widget import ResizableStackedWidget
from submit2gm.inference_source import InferenceSource

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin
    from submit2gm.dynamic_job_preset import DynamicJobPreset
    from .dynamic_job_preset_widget import DynamicJobPresetWidget
    from submit2gm.dynamic_job_preset_attribute import DynamicJobPresetAttribute


class DynamicJobPresetAttributeWidget(QtWidgets.QWidget, PluginAccessor):

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 parent_dynamic_job_preset_widget: 'DynamicJobPresetWidget',
                 dynamic_job_preset_attribute: 'DynamicJobPresetAttribute',
                 parent: typing.Optional['QtWidgets.QWidget'] = None):

        QtWidgets.QWidget.__init__(self, parent=parent)
        PluginAccessor.__init__(self, plugin)

        self._parent_dynamic_job_preset_widget = parent_dynamic_job_preset_widget
        self._dynamic_job_preset_attribute = dynamic_job_preset_attribute
        job_attribute = dynamic_job_preset_attribute.get_job_attribute()
        attribute = job_attribute.get_attribute()
        prop_id = job_attribute.get_id()
        inferences_sources = job_attribute.get_inference_sources()

        self._initialised = False

        ''' Create Widgets '''

        # display name label
        self._display_name_label = ElidedLabel(attribute.get_display_name())
        self._display_name_label.setMinimumWidth(130)
        self._display_name_label.setProperty('h5', True)

        # attribute key label
        self._attribute_key_label = ElidedLabel(dynamic_job_preset_attribute.get_key())
        self._attribute_key_label.setMinimumWidth(125)

        # various input widgets
        self._widgets = {}
        qt_utils.add_widgets_for_attribute(self.get_plugin(),
                                           job_attribute.get_attribute(),
                                           self._widgets,
                                           prop_id,
                                           dynamic_job_preset=dynamic_job_preset_attribute.get_parent_preset())

        # create the inference source widget
        self._inference_source_widget = CheckBox(self.get_plugin())
        self._inference_source_widget.setProperty(constants.STYLE_PROPS.APP_INFER_CHECKBOX, True)
        self._inference_source_widget_always_disabled = False
        self._inference_source_widget.setToolTip(qt_utils.split_tool_tip(constants.APPLICATION_LINKING_TOOL_TIP))

        # Create the is enabled widget
        self._is_enabled_widget = CheckBox(self.get_plugin())
        self._is_enabled_widget.setToolTip(qt_utils.split_tool_tip('Optional attributes can be enabled or disabled. '
                                                                   'Disabling an attribute will mean no explicit value '
                                                                   'is passed for this attribute when submitting.'))

        # create a stacked widget to hold the different input widgets
        self._stacked_widget = ResizableStackedWidget()

        # validate the inference source options ( this are temporary checks for v1)
        is_application_inferred = InferenceSource.get_application_inference_source() in inferences_sources
        is_python_inferred = InferenceSource.get_python_inference_source() in inferences_sources
        is_user_inferred = InferenceSource.get_user_defined_inference_source() in inferences_sources
        is_constant_inferred = InferenceSource.get_constant_inference_source() in inferences_sources
        is_project_inferred = InferenceSource.get_project_inference_source() in inferences_sources

        if is_constant_inferred | is_project_inferred:
            if is_application_inferred | is_user_inferred:
                assert False

        if is_application_inferred and is_python_inferred:
            assert False

        if inferences_sources[0] == InferenceSource.get_user_defined_inference_source():
            self._inference_source_widget.setCheckState(QtCore.Qt.Checked)

        # if the job preset attribute is not either app inferred or user inferred or python inferred then don't show the
        # input
        if not (is_user_inferred and (is_python_inferred or is_application_inferred)):
            self._inference_source_widget.setProperty('hidden', True)
            self._inference_source_widget.setEnabled(False)
            self._inference_source_widget_always_disabled = True

        self._inference_source_widget.input_value_changed.connect(self._handle_inference_source_widget_option_changed)
        self._is_enabled_widget.input_value_changed.connect(self._handle_is_enabled_widget_value_changed)

        if InferenceSource.get_user_defined_inference_source() in inferences_sources:
            self._user_input_widget = self._widgets[prop_id]
            self._user_input_widget.input_value_changed.connect(self._handle_user_input_value_changed)
            self._stacked_widget.addWidget(self._user_input_widget)

            # Set the user defined value to the user input value given they may not be in syn (enum app inferred values
            # for example will only be set on the widget at this point)
            dynamic_job_preset_attribute.set_user_defined_value(self._user_input_widget.get_value())

        if InferenceSource.get_constant_inference_source() in inferences_sources:
            self._constant_widget = Constant(plugin, job_attribute.get_attribute().get_default_value())
            self._constant_widget.setToolTip(qt_utils.split_tool_tip(attribute.get_description()))
            self._stacked_widget.addWidget(self._constant_widget)

        if InferenceSource.get_application_inference_source() in inferences_sources:
            # deferred until requested as it requires eval which cannot be used in the constructor
            self._app_inferred_value = None

        if InferenceSource.get_python_inference_source() in inferences_sources:
            raise NotImplementedError

            self._python_widget = Constant(plugin, self.get_python_defined_value())
            self._python_widget.setToolTip(qt_utils.split_tool_tip(attribute.get_description()))
            self._python_widget.clicked.connect(self._handle_input_widget_clicked)
            self._stacked_widget.addWidget(self._python_widget)

        if len(inferences_sources) == 1:
            self._inference_source_widget.setEnabled(False)
            self._inference_source_widget_always_disabled = True

        # --- Connect attribute value changed slot
        #self._inference_source_widget.input_value_changed.connect(
        #    lambda: dynamic_job_preset_attribute.get_parent_preset().dynamic_job_preset_attribute_value_changed.emit(job_attribute.get_id()))

        """ Create Layouts """
        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._stacked_widget)

        """ Create Connections """
        dynamic_job_preset_attribute.inference_source_changed.connect(self._handle_inference_source_changed)
        dynamic_job_preset_attribute.enabled_state_changed.connect(self._handle_enabled_state_changed)

    def get_dynamic_job_preset_attribute(self) -> 'DynamicJobPresetAttribute':
        return self._dynamic_job_preset_attribute

    def init(self) -> None:
        """ Call before adding to a layout """
        if not self._initialised:
            dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()

            job_attribute = dynamic_job_preset_attribute.get_job_attribute()

            # add the app input widget to that stacked widget
            inferences_sources = job_attribute.get_inference_sources()
            if InferenceSource.get_application_inference_source() in inferences_sources:
                self._stacked_widget.addWidget(self.get_app_inferred_widget())

            # set the starting widget
            current_widget = self.get_current_widget()

            # Check that the widget is in the stacked widget before setting to avoid warnings
            if self._stacked_widget.indexOf(current_widget) >= 0:
                self._stacked_widget.setCurrentWidget(current_widget)

            is_enabled = dynamic_job_preset_attribute.is_enabled()
            self._is_enabled_widget.set_value(is_enabled)
            self._handle_enabled_state_changed(is_enabled)

            self._initialised = True

    def get_parent_preset_widget(self) -> 'DynamicJobPresetWidget':
        return self._parent_dynamic_job_preset_widget

    def get_parent_preset(self) -> 'DynamicJobPreset':
        return self._parent_dynamic_job_preset_widget.get_dynamic_job_preset()

    def update_app_inferred_values(self) -> None:
        current_widget = self._stacked_widget.currentWidget()

        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
        is_app_inferred = dynamic_job_preset_attribute.get_inference_source() == InferenceSource.get_application_inference_source()
        is_python_inferred = dynamic_job_preset_attribute.get_inference_source() == InferenceSource.get_python_inference_source()

        # required to update combo box and radio items
        if isinstance(current_widget, ComboBox) or isinstance(current_widget, ListRadioInput):
            current_widget.update_items()

        if is_app_inferred or is_python_inferred:

            if current_widget is None:
                assert False, "current_widget is none for {}".format(dynamic_job_preset_attribute.get_id())

            old_value = current_widget.get_value()
            new_value = str(dynamic_job_preset_attribute.eval())

            if is_app_inferred:
                dynamic_job_preset_attribute.set_application_defined_value(new_value)
            else:
                dynamic_job_preset_attribute.set_python_defined_value(new_value)

            current_widget.set_value(new_value)

            if old_value != new_value:
                dynamic_job_preset_attribute.get_parent_preset().dynamic_job_preset_attribute_value_changed.emit(dynamic_job_preset_attribute.get_job_attribute().get_id())
                user_interface = self.get_plugin().get_user_interface()
                theme = user_interface.get_theme()
                original_color = QtGui.QColor(theme.FOREGROUND_COLOR)
                flash_color = QtGui.QColor(theme.BRAND_COLOR)
                qt_utils.flash_widget_color(current_widget, original_color, flash_color)

    @QtCore.Slot(int)
    def _handle_inference_source_widget_option_changed(self) -> None:
        get_dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()

        if self._inference_source_widget.isChecked():
            inference_source = InferenceSource.get_user_defined_inference_source()
        else:
            inferences_sources = get_dynamic_job_preset_attribute.get_job_attribute().get_inference_sources()
            is_application_inferred = InferenceSource.get_application_inference_source() in inferences_sources
            is_python_inferred = InferenceSource.get_python_inference_source() in inferences_sources

            if is_application_inferred:
                inference_source = InferenceSource.get_application_inference_source()
            elif is_python_inferred:
                inference_source = InferenceSource.get_python_inference_source()
            else:
                assert False

        if get_dynamic_job_preset_attribute.get_inference_source() != inference_source:
            get_dynamic_job_preset_attribute.set_inference_source(inference_source)

    @QtCore.Slot()
    def _handle_inference_source_changed(self):
        inference_source = self.get_dynamic_job_preset_attribute().get_inference_source()

        if inference_source == InferenceSource.get_application_inference_source():
            self._inference_source_widget.set_value(False)
            self._stacked_widget.setCurrentWidget(self.get_app_inferred_widget())
        elif inference_source == InferenceSource.get_project_inference_source():
            self._inference_source_widget.set_value(False)
            self._stacked_widget.setCurrentWidget(self.get_python_inferred_widget())
        else:
            self._inference_source_widget.set_value(True)
            self._stacked_widget.setCurrentWidget(self.get_user_input_widget())

    @QtCore.Slot(bool)
    def _handle_enabled_state_changed(self, enabled_state):
        self.get_display_name_label().setEnabled(enabled_state)
        self.get_display_name_label().setColor('white' if enabled_state else 'grey')

        # Enable / disable the actual dynamic job preset attribute widget
        self.setEnabled(enabled_state)

        # If the inference source widget is not set to always be disabled then set it's enabled state
        if not self._inference_source_widget_always_disabled:
            inference_source_widget = self.get_inference_source_widget()
            inference_source_widget.setEnabled(enabled_state)

    @QtCore.Slot(bool)
    def _handle_user_input_value_changed(self):
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
        dynamic_job_preset_attribute.set_user_defined_value(self._user_input_widget.get_value())

    def get_optional_parent_widget(self) -> typing.Optional['DynamicJobPresetAttributeWidget']:
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
        optional_parent_attribute = dynamic_job_preset_attribute.get_optional_parent()
        dynamic_job_preset_widget = self.get_parent_preset_widget()

        for potential_parent in dynamic_job_preset_widget.get_dynamic_job_preset_attribute_widgets():
            if potential_parent.get_dynamic_job_preset_attribute().get_id() == optional_parent_attribute.get_id():
                return potential_parent

        raise ValueError

    @QtCore.Slot()
    def _handle_is_enabled_widget_value_changed(self) -> None:
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
        checked = self._is_enabled_widget.get_value()
        dynamic_job_preset_attribute.set_enabled(checked)

        """
        is_enabled = self.is_enabled()
        self.get_display_name_label().setEnabled(is_enabled)
        self.get_display_name_label().setColor('white' if is_enabled else 'grey')
        self.setEnabled(is_enabled)

        # If the inference source widget is not set to always be disabled then set it's enabled state
        if not self._inference_source_widget_always_disabled:
            inference_source_widget = self.get_inference_source_widget()
            inference_source_widget.setEnabled(is_enabled)
        """

    def get_display_name_label(self) -> 'ElidedLabel':
        return self._display_name_label

    def get_key_label(self) -> 'ElidedLabel':
        return self._attribute_key_label

    def get_constant_widget(self) -> 'Constant':
        return self._constant_widget

    def get_user_input_widget(self) -> 'InputWidget':
        return self._user_input_widget

    def _handle_input_widget_clicked(self) -> None:
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
        inferences_sources = dynamic_job_preset_attribute.get_job_attribute().get_inference_sources()
        is_user_inferred = InferenceSource.get_user_defined_inference_source() in inferences_sources

        if is_user_inferred:
            # get the current app / python inferred value
            inferred_value = str(dynamic_job_preset_attribute.eval())

            self.get_user_input_widget().set_value(inferred_value)
            self._inference_source_widget.set_value(True)

    def get_app_inferred_widget(self) -> 'Constant':

        assert hasattr(self, '_app_inferred_value'), "JobPresetAttribute '{}' does not have an app inferred " \
                                                     "value".format(self.get_job_attribute().get_id())

        # we defer the creation of this widget as to set its value we need to use 'eval' which cannot be done in the
        # constructor
        if self._app_inferred_value is None:
            dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
            job_attribute = dynamic_job_preset_attribute.get_job_attribute()
            attribute = job_attribute.get_attribute()
            self._app_inferred_value = Constant(self.get_plugin(), u"")

            value = str(dynamic_job_preset_attribute.eval())

            self._app_inferred_value.set_value(value)
            self._app_inferred_value.setToolTip(qt_utils.split_tool_tip(attribute.get_description()))
            self._app_inferred_value.clicked.connect(self._handle_input_widget_clicked)
            self._app_inferred_value.input_value_changed.connect(
                lambda: dynamic_job_preset_attribute.get_parent_preset().dynamic_job_preset_attribute_value_changed.emit(job_attribute.get_id()))

        return self._app_inferred_value

    def get_python_inferred_widget(self) -> 'Constant':
        return self._python_widget

    def get_is_enabled_widget(self) -> 'CheckBox':
        return self._is_enabled_widget

    def get_inference_source_widget(self) -> 'CheckBox':
        return self._inference_source_widget

    def get_current_widget(self, project_inferred_value: str = "Project inferred value") -> 'InputWidget':
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
        inference_source = dynamic_job_preset_attribute.get_inference_source()

        if inference_source == InferenceSource.get_constant_inference_source():
            widget = self.get_constant_widget()

        elif inference_source == InferenceSource.get_application_inference_source():
            widget = self.get_app_inferred_widget()

        elif inference_source == InferenceSource.get_python_inference_source():
            widget = self.get_python_inferred_widget()

        elif inference_source == InferenceSource.get_user_defined_inference_source():
            widget = self.get_user_input_widget()

        elif inference_source == InferenceSource.get_project_inference_source():
            widget = Constant(self.get_plugin(), project_inferred_value)
            widget.setToolTip(qt_utils.split_tool_tip(dynamic_job_preset_attribute.get_job_attribute().get_attribute().get_description()))
        else:
            raise ValueError("Unsupported inference source type: " + inference_source.get_display_name())

        return widget

    """
    def get_user_defined_value(self) -> any:
        return self.get_user_input_widget().get_value()

    def set_user_defined_value(self, value: any) -> None:
        gm_meta_plugin.JobPresetAttribute.set_user_defined_value(value)
        self.get_user_input_widget().set_value(value)
    

    def eval_user_defined_value(self, fail_on_eval_lock: bool = True) -> any:
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()
        if self._is_eval_locked:
            if fail_on_eval_lock:
                raise ValueError("Attempted to evaluate job preset attribute ('{}') value while evaluation lock is "
                                 "enabled".format(dynamic_job_preset_attribute.get_id()))
            return dynamic_job_preset_attribute.get_user_defined_value()

        user_defined_value = self.get_user_input_widget().get_value()
        dynamic_job_preset_attribute.set_user_defined_value(user_defined_value)
        return dynamic_job_preset_attribute.get_user_defined_value()

    def set_python_defined_value(self, value: any) -> None:
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()

        try:
            self.get_python_inferred_widget().set_value(value)
        except AttributeError as e:
            # The widget may not exist yet since set value is called by the super class
            pass

        dynamic_job_preset_attribute.set_python_defined_value(value)

    def get_value(self):
        inference_source = self.get_inference_source()

        if inference_source == gm_meta_plugin.InferenceSource.get_constant_inference_source():
            value = self.get_job_attribute().get_attribute().get_default_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_application_inference_source():
            value = self.get_app_inferred_widget().get_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
            value = self.get_user_input_widget().get_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_project_inference_source():
            raise NotImplementedError
        else:
            raise ValueError("Unsupported inference source type: " + inference_source.get_display_name())

        return value

    def set_value(self, value):
        self.get_user_input_widget().set_value(value)
    """

    def get_inference_source(self) -> 'InferenceSource':
        dynamic_job_preset_attribute = self.get_dynamic_job_preset_attribute()

        """
        index = self.get_inference_source_widget().currentIndex()
        inference_source = self.get_job_attribute().get_inference_sources()[index]
        return inference_source
        """
        inference_source_widget = self.get_inference_source_widget()
        inferences_sources = dynamic_job_preset_attribute.get_job_attribute().get_inference_sources()

        if InferenceSource.get_user_defined_inference_source() in inferences_sources and \
                (InferenceSource.get_application_inference_source() in inferences_sources or
                 InferenceSource.get_python_inference_source() in inferences_sources):

            if inference_source_widget.isChecked():
                return InferenceSource.get_user_defined_inference_source()
            else:
                is_application_inferred = InferenceSource.get_application_inference_source() in inferences_sources
                is_python_inferred = InferenceSource.get_python_inference_source() in inferences_sources

                if is_application_inferred:
                    return InferenceSource.get_application_inference_source()
                elif is_python_inferred:
                    return InferenceSource.get_python_inference_source()
                else:
                    assert False

        return inferences_sources[0]
