__all__ = 'DynamicJobPresetTabWidget'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets
import typing

from submit2gm import constants
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.widgets.dynamic_job_preset_widget import DynamicJobPresetWidget

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin
    from submit2gm.dynamic_job_preset import DynamicJobPreset


class DynamicJobPresetTabWidget(QtWidgets.QWidget, PluginAccessor):

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 dynamic_job_preset: 'DynamicJobPreset',
                 parent: typing.Optional['QtWidgets.QWidget'] = None):

        QtWidgets.QWidget.__init__(self, parent=parent)
        PluginAccessor.__init__(self, plugin)

        self._dynamic_job_preset = dynamic_job_preset

        """ Create Widgets """
        spinner_movie = QtGui.QMovie(str(constants.SPINNER_WHITE_2D2D2D_GIF_PATH), parent=self)
        spinner_movie.setCacheMode(QtGui.QMovie.CacheAll)
        spinner_movie.start()

        self._loading_label = QtWidgets.QLabel('Loading job settings')
        self._loading_spinner = QtWidgets.QLabel('')
        self._loading_spinner.setMovie(spinner_movie)

        self._dynamic_job_preset_widget = None

        self.scroll_area = QtWidgets.QScrollArea(parent=self)
        self.scroll_area.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.scroll_area.hide()

        """ Create Layouts """
        self._loading_layout = QtWidgets.QHBoxLayout()
        self._loading_layout.setContentsMargins(0, 0, 0, 0)
        self._loading_layout.addStretch()
        self._loading_layout.addWidget(self._loading_label)
        self._loading_layout.addWidget(self._loading_spinner)
        self._loading_layout.addStretch()

        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self.scroll_area)
        self._layout.addLayout(self._loading_layout)

        # Finally trigger a worker thread to create the dynamic job preset widget
        self._create_dynamic_job_preset_widget()

    def ready(self) -> bool:
        """
        Tests whether the dynamic job preset widget for this tab has finished being created.
        """
        return not (self._dynamic_job_preset_widget is None)

    def get_dynamic_job_preset(self) -> 'DynamicJobPreset':
        return self._dynamic_job_preset

    def get_dynamic_job_preset_widget(self) -> 'DynamicJobPresetWidget':
        return self._dynamic_job_preset_widget

    def _create_dynamic_job_preset_widget(self) -> None:
        plugin = self.get_plugin()
        user_interface = plugin.get_user_interface()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)
        logger.debug(u"Creating DynamicJobPresetWidget Preset...")

        worker_thread = user_interface.get_worker_thread(u"Creating job preset...")

        ################################################################################################################

        target = DynamicJobPresetWidget

        args = (
            self.get_plugin(),
            self.get_dynamic_job_preset(),
        )

        kwargs = dict()

        worker_thread.process(target, args, kwargs, self._handle_dynamic_job_preset_widget_created)

    def _handle_dynamic_job_preset_widget_created(self, result: any) -> bool:
        if isinstance(result, DynamicJobPresetWidget):
            """
            This is very hacky but it works.

            In Qt it is technically an error to create a QWidget in any thread other than the Main GUI thread. This 
            means we can not use the job preset just created because it was created in a different thread. However
            because all the long running parts of creating a DynamicJobPresetWidget get cached (machine options, etc) it 
            is very quick to re-create the same DynamicJobPresetWidget here in the Gui thread and use the new version 
            instead. 
            """

            # Delete the job preset created in another thread Otherwise
            # 'QBasicTimer::stop: Failed. Possibly trying to stop from a different thread' warnings occur when the
            # application quits
            result.deleteLater()

            self._dynamic_job_preset_widget = DynamicJobPresetWidget(self.get_plugin(), self.get_dynamic_job_preset())

            # Update the submission view
            self._layout.removeItem(self._loading_layout)
            self._loading_label.hide()
            self._loading_spinner.hide()

            self.scroll_area.setWidget(self._dynamic_job_preset_widget)
            self.scroll_area.show()

            return True
        return False
