__all__ = 'ElidedLabel'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets
import typing


class ElidedLabel(QtWidgets.QFrame):
    elisionChanged = QtCore.Signal(bool)

    def __init__(self, text: str, parent: typing.Optional[QtWidgets.QWidget] = None):
        QtWidgets.QFrame.__init__(self, parent=parent)
        self.content = text
        self.elided = False
        self._color = QtGui.QColor("white")
        self._alignment = QtCore.Qt.AlignVCenter

    def getColor(self) -> 'QtGui.QColor':
        return self._color

    def setColor(self, color: 'QtGui.QColor' or str) -> None:
        if isinstance(color, str):
            self._color = QtGui.QColor(color)
        else:
            self._color = color

    def setText(self, newText: str) -> None:
        self.content = newText
        self.update()

    def text(self) -> str:
        return self.content

    def alignment(self) -> QtCore.Qt.Alignment:
        return self._alignment

    def setAlignment(self, alignment: QtCore.Qt.Alignment) -> None:
        self._alignment = alignment

    def isElided(self) -> bool:
        return self.elided

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        QtWidgets.QFrame.paintEvent(self, event)

        painter = QtGui.QPainter(self)
        painter.setPen(self._color)
        font_metrics = painter.fontMetrics()
        rect = self.rect()

        did_elide = False
        textLayout = QtGui.QTextLayout(self.content, painter.font())
        textLayout.beginLayout()

        while True:
            line = textLayout.createLine()

            if not line.isValid():
                break

            line.setLineWidth(self.width())
            elided_line = font_metrics.elidedText(self.content, QtCore.Qt.ElideRight, self.width())
            painter.drawText(rect, int(self._alignment), elided_line)
            line = textLayout.createLine()
            did_elide = line.isValid()

        textLayout.endLayout()

        if did_elide != self.elided:
            self.elided = did_elide
            self.elisionChanged.emit(did_elide)