__all__ = 'JobEventHandlersInput'

from PySide2 import QtCore
from PySide2 import QtWidgets
import typing

from submit2gm import constants
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.widgets.qhline import QHLine
from submit2gm.widgets.qt_dialog_statement_creator import StatementCreatorWindow
from submit2gm.widgets.input_widgets.input_widget import InputWidget
from submit2gm.widgets.input_widgets.combo_box import ComboBox

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class JobEventHandlerInput(QtWidgets.QGroupBox, PluginAccessor):

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtWidgets.QWidget'] = None):
        QtWidgets.QGroupBox.__init__(self, parent=parent)
        PluginAccessor.__init__(self, plugin)

        ''' Create Widgets '''
        self.setProperty("background-level-2", True)

        event_label = QtWidgets.QLabel("Trigger on:")
        event_label.setProperty('h5', True)
        event_label.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        event_combo_box = ComboBox(plugin)
        event_combo_box.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        event_combo_box.addItem("State Change Event", "job_state_changed_event")
        event_combo_box.setItemData(0, "Trigger this event handler whenever the jobs's state changes. An example state change could be your job changing from a running state to a stopped state.", QtCore.Qt.ToolTipRole)

        event_combo_box.addItem("Credits Billed Change Event", "job_credits_billed_changed_event")
        event_combo_box.setItemData(1, "Trigger this event handler whenever the jobs's total billed credits changes.", QtCore.Qt.ToolTipRole)

        event_combo_box.addItem("Run Time Change Event", "job_run_time_changed_event")
        event_combo_box.setItemData(2, "Trigger this event handler whenever the jobs's total run time changes. A jobs total run time is the sum time of all it's tasks.", QtCore.Qt.ToolTipRole)

        separator = QHLine()

        add_statement_push_btn = QtWidgets.QPushButton("Add statement +")
        add_statement_push_btn.setFlat(True)
        add_statement_push_btn.setProperty(constants.STYLE_PROPS.BORDERED, True)
        add_statement_push_btn.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        ''' Create Layouts '''
        event_header_layout = QtWidgets.QHBoxLayout()
        event_header_layout.setSpacing(10)
        event_header_layout.setContentsMargins(0, 0, 0, 0)
        event_header_layout.addWidget(event_label)
        event_header_layout.addWidget(event_combo_box)
        event_header_layout.addStretch()

        self._top_level_statements_layout = QtWidgets.QVBoxLayout()
        self._top_level_statements_layout.setSpacing(5)
        self._top_level_statements_layout.setContentsMargins(0, 0, 0, 0)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(event_header_layout)
        layout.addWidget(separator)
        layout.addLayout(self._top_level_statements_layout)
        layout.addWidget(add_statement_push_btn)

        ''' Create Connections '''
        add_statement_push_btn.clicked.connect(self._handle_add_statement)

    def _handle_add_statement(self) -> None:
        # open preflight window
        statement_creator_window = StatementCreatorWindow(self.get_plugin(), parent=self)
        statement_creator_window.exec_()

        if statement_creator_window.cancel_statement_creation():
            return

        statement_widget_type = statement_creator_window.chosen_statement_type()
        statement_widget = statement_widget_type(self.get_plugin())

        self._top_level_statements_layout.addWidget(statement_widget)
        statement_widget.set_parent_layout(self._top_level_statements_layout)

        # show the main window once the preflight window has closed
        #user_interface.get_main_window().show()


class JobEventHandlersInput(QtWidgets.QWidget, InputWidget, PluginAccessor):

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtWidgets.QWidget'] = None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        PluginAccessor.__init__(self, plugin)

        ''' Create Widgets '''
        self._push_btn = QtWidgets.QPushButton("Add event handler +")
        self._push_btn.setFlat(True)
        self._push_btn.setProperty(constants.STYLE_PROPS.BORDERED, True)

        ''' Create Layouts '''
        self._job_event_handlers_layout = QtWidgets.QVBoxLayout()
        self._job_event_handlers_layout.setSpacing(10)
        self._job_event_handlers_layout.setContentsMargins(0, 0, 0, 0)

        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setSpacing(10)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addLayout(self._job_event_handlers_layout)
        self._layout.addWidget(self._push_btn)

        ''' Create Connections '''
        self._push_btn.clicked.connect(self._add_event_handler)

    def _add_event_handler(self) -> None:
        self._job_event_handlers_layout.addWidget(JobEventHandlerInput(self.get_plugin()))

    def get_value(self) -> str:
        return "test value"

    def set_value(self, value: str) -> None:
        pass
