__all__ = 'MachineInstancesInput'

from PySide2 import QtWidgets
import typing

from submit2gm import constants, qt_utils
from submit2gm.widgets.input_widgets.input_widget import InputWidget
from submit2gm.widgets.input_widgets.spin_box import SpinBox
from submit2gm.widgets.input_widgets.check_box import CheckBox
from submit2gm.widgets.input_widgets.constant import Constant

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class MachineInstancesInput(QtWidgets.QWidget, InputWidget):

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtWidgets.QWidget'] = None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        InputWidget.__init__(self, plugin)

        """ Create Widgets """
        self.default_count_widget = QtWidgets.QWidget()
        self.label = Constant(plugin, u"Use Portal Settings")
        self.label.setToolTip(qt_utils.split_tool_tip(u"Do not add a machine instance limit. Instead use the default machine count as specified" 
                              " under your profile settings in the render manager portal."))

        self.custom_count_widget = QtWidgets.QWidget()
        self.spin_box = SpinBox(plugin)
        self.spin_box.setMinimum(1)
        self.spin_box.setMaximum(constants.MAX_INT)
        self.spin_box.setToolTip(qt_utils.split_tool_tip(u"The maximum number of machines to use for this job."))

        self.stacked_widget = QtWidgets.QStackedWidget()
        self.default_count_index = self.stacked_widget.addWidget(self.default_count_widget)
        self.custom_count_index = self.stacked_widget.addWidget(self.custom_count_widget)

        self.toggle_check_box = CheckBox(plugin)
        self.toggle_check_box.set_value(True)

        """ Create Layout """
        self.custom_count_layout = QtWidgets.QHBoxLayout()
        self.custom_count_layout.setContentsMargins(0,0,0,0)
        self.custom_count_layout.addWidget(self.spin_box)
        self.custom_count_widget.setLayout(self.custom_count_layout)

        self.default_count_layout = QtWidgets.QHBoxLayout()
        self.default_count_layout.setContentsMargins(0,0,0,0)
        self.default_count_layout.addWidget(self.label)
        self.default_count_widget.setLayout(self.default_count_layout)

        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(0,0,0,0)
        self.layout.addWidget(self.stacked_widget)
        self.layout.addWidget(self.toggle_check_box)

        """ Create Connections """
        self.spin_box.input_value_changed.connect(self._catch_value_change)
        self.toggle_check_box.input_value_changed.connect(self._catch_value_change)
        self.toggle_check_box.input_value_changed.connect(self._toggle_check_box)

    def _toggle_check_box(self) -> None:
        is_checked = self.toggle_check_box.get_value()

        if is_checked:
            self.stacked_widget.setCurrentIndex(self.default_count_index)
        else:
            self.stacked_widget.setCurrentIndex(self.custom_count_index)

    def set_maximum(self, max: int) -> None:
        self.spin_box.setMaximum(max)

    def set_minimum(self, min: int) -> None:
        self.spin_box.setMinimum(min)

    def get_value(self) -> int:
        if self.toggle_check_box.isChecked():
            return 0
        else:
            return self.spin_box.get_value()

    def set_value(self, value: int) -> None:
        self.spin_box.set_value(value)
