__all__ = 'PathInput'

import pathlib
from PySide2 import QtWidgets
import typing

from submit2gm import ui_constants, utils
from submit2gm.widgets.input_widgets.input_widget import InputWidget
from submit2gm.widgets.input_widgets.line_edit import LineEdit

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class PathInput(QtWidgets.QWidget, InputWidget):

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 file_mode: typing.Optional['QtWidgets.QFileDialog.FileMode'] = None,
                 parent: typing.Optional['QtWidgets.QWidget'] = None):

        QtWidgets.QWidget.__init__(self, parent=parent)
        InputWidget.__init__(self, plugin)

        if file_mode is None:
            file_mode = QtWidgets.QFileDialog.AnyFile

        self._file_mode = file_mode

        self._line_edit_widget = LineEdit(plugin)
        self._line_edit_widget.editingFinished.connect(self._catch_value_change)

        self._file_selection_btn = QtWidgets.QPushButton()
        self._file_selection_btn.setIcon(ui_constants.DIRECTORY_ICON_WHITE)
        self._file_selection_btn.setFlat(True)
        self._file_selection_btn.clicked.connect(self._file_selector)

        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0,0,0,0)
        self.layout.addWidget(self._line_edit_widget)
        self.layout.addWidget(self._file_selection_btn)

    def _get_directory(self) -> str:
        if self.get_value() and pathlib.Path(self.get_value()).exists():
            return self.get_value()

        return utils.get_desktop_path()

    def _file_selector(self) -> None:
        file_dialog = QtWidgets.QFileDialog()
        file_dialog.setFileMode(self._file_mode)
        file_dialog.setViewMode(QtWidgets.QFileDialog.Detail)
        file_dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptOpen)
        file_dialog.setLabelText(QtWidgets.QFileDialog.Accept, "Select")
        file_dialog.setLabelText(QtWidgets.QFileDialog.Reject, "Cancel")
        file_dialog.setLabelText(QtWidgets.QFileDialog.FileName, "File Name")
        file_dialog.setDirectory(self._get_directory())

        if file_dialog.exec_():
            file_names = file_dialog.selectedFiles()

            # only one file can be selected anyway
            selected_path = file_names[0]

            self.set_value(selected_path)
            self.input_value_changed.emit()

    def set_tool_tip(self, text: str) -> None:
        self._line_edit_widget.setToolTip(text)

    def set_placeholder_text(self, text: str) -> None:
        self._line_edit_widget.setPlaceholderText(text)

    def get_value(self) -> str:
        return self._line_edit_widget.text()

    def set_value(self, value: str) -> None:
        self._line_edit_widget.setText(value)
