__all__ = 'ProjectInput'

from PySide2 import QtCore, QtWidgets
import typing

from submit2gm import constants, ui_constants
from submit2gm.widgets.right_aligned_qmenu import RightAlignedQMenu
from submit2gm.widgets.input_widgets.input_widget import InputWidget
from submit2gm.widgets.input_widgets.line_edit import LineEdit

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class ProjectInput(QtWidgets.QWidget, InputWidget):

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 parent: typing.Optional['QtWidgets.QWidget'] = None):

        QtWidgets.QWidget.__init__(self, parent=parent)
        InputWidget.__init__(self, plugin)

        self._line_edit_widget = LineEdit(plugin)
        self._line_edit_widget.editingFinished.connect(self._catch_value_change)

        self._project_selection_btn = QtWidgets.QPushButton()
        self._project_selection_btn.setIcon(ui_constants.PROJECT_SELECT_ICON_WHITE)
        self._project_selection_btn.setFlat(True)
        self._project_selection_btn.setToolTip('Select existing project')
        self._project_selection_btn.setFixedSize(20, 20)
        self._project_selection_btn.setProperty(constants.STYLE_PROPS.BORDERED, True)

        self._project_selection_btn.clicked.connect(self._project_selector)

        self._v_layout = QtWidgets.QVBoxLayout()
        self._v_layout.setSpacing(0)
        self._v_layout.setContentsMargins(0,0,0,0)
        self._v_layout.setAlignment(QtCore.Qt.AlignVCenter)
        self._v_layout.addSpacing(3)
        self._v_layout.addWidget(self._project_selection_btn)

        self.layout = QtWidgets.QHBoxLayout(self)

        self.layout.setSpacing(5)
        self.layout.setContentsMargins(0,0,0,0)
        self.layout.addWidget(self._line_edit_widget)
        self.layout.addLayout(self._v_layout)

    def _about_to_show_project_selection_menu(self):
        self._project_selection_menu.clear()

        projects = self.get_plugin().get_api_client().get_remote_projects()

        for project in projects:
            self._project_selection_menu.addAction(project.get_name())

    def _project_selector(self) -> None:
        current_value = self._line_edit_widget.get_value()
        projects = self.get_plugin().get_api_client().get_remote_projects()

        project_names = list(map(lambda x: x.get_name(), projects))

        try:
            current_index = project_names.index(current_value)
        except ValueError:
            current_index = 0

        if project_names:
            help_test = 'Existing projects:'
        else:
            help_test = 'No existing projects found, please enter a new project name'

        selected_project_name, ok = QtWidgets.QInputDialog.getItem(self,
                                                                   'Select an existing project',
                                                                   help_test,
                                                                   project_names,
                                                                   current_index,
                                                                   False,
                                                                   flags=QtCore.Qt.Window)

        if selected_project_name and ok:
            self.set_value(selected_project_name)

    def set_tool_tip(self, text: str) -> None:
        self._line_edit_widget.setToolTip(text)

    def setCompleter(self, completer: 'QtWidgets.QCompleter') -> None:
        self._line_edit_widget.setCompleter(completer)

    def setText(self, text: str) -> None:
        self._line_edit_widget.setText(text)

    def setPlaceholderText(self, text: str) -> None:
        self.set_placeholder_text(text)

    def set_placeholder_text(self, text: str) -> None:
        self._line_edit_widget.setPlaceholderText(text)

    def get_value(self) -> str:
        return self._line_edit_widget.text()

    def set_value(self, value: str) -> None:
        self._line_edit_widget.setText(value)
