__all__ = 'DynamicJobPresetNode'

from PySide2 import QtCore
from PySide2 import QtWidgets

from submit2gm import constants
from submit2gm.widgets.node_editor.node import Node

import typing

if typing.TYPE_CHECKING:
    from submit2gm.dynamic_job_preset import DynamicJobPreset


class DynamicJobPresetNode(Node):

    def __init__(self, dynamic_job_preset: 'DynamicJobPreset'):
        Node.__init__(self)

        self._dynamic_job_preset = dynamic_job_preset

        self.title = self._dynamic_job_preset.get_job_preset_attribute_by_key(constants.API_KEYS.JOB_NAME).eval()
        self.type_text = self._dynamic_job_preset.get_job_definition().get_definition_id()

        self.add_port(name="input", is_output=False)
        self.add_port(name="output", is_output=True)

        dynamic_job_preset.job_name_changed.connect(self._handle_job_name_changed)
        dynamic_job_preset.selected.connect(self._handle_dynamic_job_preset_selected)
        self._handle_job_name_changed()

    def itemChange(self, change: 'QtWidgets.QGraphicsItem.GraphicsItemChange', value: any):
        if change == QtWidgets.QGraphicsItem.ItemSelectedChange and self.scene():
            if value:
                # Emit via a QTimer so that a recursion loop does not occur
                QtCore.QTimer.singleShot(1, lambda: self._dynamic_job_preset.selected.emit())

        return QtWidgets.QGraphicsItem.itemChange(self, change, value)

    @QtCore.Slot()
    def _handle_dynamic_job_preset_selected(self):
        if not self.isSelected():
            scene = self.scene()
            if scene:
                scene.clearSelection()
                self.setSelected(True)

    @QtCore.Slot()
    def _handle_job_name_changed(self) -> None:
        new_job_name = self._dynamic_job_preset.get_job_preset_attribute_with_job_attribute_id(
            constants.API_KEYS.JOB_NAME).get_value()

        self.title = new_job_name
        self.build()
        self.update()

    def get_dynamic_job_preset(self) -> 'DynamicJobPreset':
        return self._dynamic_job_preset

    def get_inputs(self) -> typing.List['DynamicJobPresetNode']:
        inputs = []

        for port in self._ports:
            if not port.is_output():
                connected_input_nodes = port.get_connected_nodes()
                for connected_input_node in connected_input_nodes:
                    if isinstance(connected_input_node, DynamicJobPresetNode):
                        inputs.append(connected_input_node)

        return inputs

    def get_outputs(self) -> typing.List['DynamicJobPresetNode']:
        outputs = []

        for port in self._ports:
            if port.is_output():
                connected_input_nodes = port.get_connected_nodes()
                for connected_input_node in connected_input_nodes:
                    if isinstance(connected_input_node, DynamicJobPresetNode):
                        outputs.append(connected_input_node)

        return outputs

    def delete(self):
        self._dynamic_job_preset.get_plugin().get_job_preset_container().pop(self._dynamic_job_preset.get_id())