from PySide2 import QtWidgets, QtGui, QtCore

from submit2gm.widgets.node_editor.port import Port

import typing


class Node(QtWidgets.QGraphicsPathItem):

    def __init__(self):
        super(Node, self).__init__()

        self.setFlag(QtWidgets.QGraphicsPathItem.ItemIsMovable)
        self.setFlag(QtWidgets.QGraphicsPathItem.ItemIsSelectable)

        self._title_text = "Title"
        self._type_text = "base"

        self._width = 30  # The Width of the node
        self._height = 30  # the height of the node
        self._ports = []  # A list of ports

        self.node_color = QtGui.QColor(20, 20, 20, 200)

        self.title_path = QtGui.QPainterPath()  # The path for the title
        self.type_path = QtGui.QPainterPath()  # The path for the type
        self.misc_path = QtGui.QPainterPath()  # a bunch of other stuff

        self.horizontal_margin = 30  # horizontal margin
        self.vertical_margin = 20  # vertical margin

        self.min_width = 200

    @property
    def title(self):
        return self._title_text

    @title.setter
    def title(self, title):
        self._title_text = title

    @property
    def type_text(self):
        return self._type_text

    @type_text.setter
    def type_text(self, type_text):
        self._type_text = type_text

    def paint(self, painter, option=None, widget=None):
        if self.isSelected():
            painter.setPen(QtGui.QPen(QtGui.QColor(241, 175, 0), 2))
            painter.setBrush(self.node_color)
        else:
            painter.setPen(self.node_color.lighter())
            painter.setBrush(self.node_color)

        painter.drawPath(self.path())
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtCore.Qt.white)

        painter.drawPath(self.title_path)
        painter.drawPath(self.type_path)
        painter.drawPath(self.misc_path)

    def add_port(self, name, is_output=False, flags=0, ptr=None):
        port = Port(self, self.scene())
        port.set_is_output(is_output)
        port.set_name(name)
        port.set_node(node=self)
        port.set_port_flags(flags)
        port.set_ptr(ptr)

        self._ports.append(port)

    def get_input_ports(self) -> typing.List['Port']:
        input_ports = []
        for port in self._ports:
            if not port.is_output():
                input_ports.append(port)
        return input_ports

    def get_output_ports(self) -> typing.List['Port']:
        output_ports = []
        for port in self._ports:
            if port.is_output():
                output_ports.append(port)
        return output_ports

    def remove_input_port(self):
        i = len(self._ports) - 1
        while i >= 0:
            port = self._ports[i]

            if not port.is_output():
                port.clear_connection()
                self._ports.remove(port)
                self.scene().removeItem(port)
                break
            i = i - 1

    def remove_output_port(self):
        i = len(self._ports) - 1
        while i >= 0:
            port = self._ports[i]

            if port.is_output():
                port.clear_connection()
                self._ports.remove(port)
                self.scene().removeItem(port)
                break
            i = i - 1

    def cycles_check(self, seen_nodes: typing.Optional[typing.List['Node']] = None) -> None:
        """
        Is this node dependent on the provided node.
        """

        if seen_nodes is None:
            seen_nodes = [self]
        else:
            seen_nodes.append(self)

        input_ports = self.get_input_ports()
        for input_port in input_ports:
            if input_port.is_connected():
                connection = input_port.connection

                start_node = connection.start_port.node()
                end_node = connection.end_port.node()
                input_node = end_node if connection.end_port.is_output() else start_node

                if input_node in seen_nodes:
                    raise ValueError()

                # Check every input node for cycles
                input_node.cycles_check(seen_nodes=seen_nodes)

    def build(self):
        """ Build the node
        """

        self.title_path = QtGui.QPainterPath()  # reset
        self.type_path = QtGui.QPainterPath()  # The path for the type
        self.misc_path = QtGui.QPainterPath()  # a bunch of other stuff

        total_width = 0
        total_height = 0
        path = QtGui.QPainterPath()  # The main path

        # The fonts what will be used
        title_font = QtGui.QFont("Open Sans", pointSize=16)
        title_type_font = QtGui.QFont("Open Sans", pointSize=8)
        port_font = QtGui.QFont("Open Sans")

        # Get the dimensions of the title and type
        title_dim = {
            "w": QtGui.QFontMetrics(title_font).width(self._title_text),
            "h": QtGui.QFontMetrics(title_font).height(),
        }

        title_type_dim = {
            "w": QtGui.QFontMetrics(title_type_font).width("(" + self._type_text + ")"),
            "h": QtGui.QFontMetrics(title_type_font).height(),
        }

        # Get the max width
        for dim in [title_dim["w"], title_type_dim["w"]]:
            if dim > total_width:
                total_width = dim

        # Add both the title and type height together for the total height
        for dim in [title_dim["h"], title_type_dim["h"]]:
            total_height += dim

        total_input_height = 0
        total_output_height = 0

        # Add the height for each of the ports
        for port in self._ports:
            port_dim = {
                "w": QtGui.QFontMetrics(port_font).width(port.name()),
                "h": QtGui.QFontMetrics(port_font).height(),
            }

            if port_dim["w"] > total_width:
                total_width = port_dim["w"]

            if port.is_output():
                total_output_height += port_dim["h"]
            else:
                total_input_height += port_dim["h"]

        total_height += max(total_input_height, total_output_height)

        # Add the margin to the total_width
        total_width += self.horizontal_margin
        total_height += self.vertical_margin

        if total_width < self.min_width:
            total_width = self.min_width

        # Draw the background rectangle
        path.addRoundedRect(
            QtCore.QRectF(-total_width / 2, -total_height / 2, total_width, total_height), 5, 5
        )

        # Draw the title
        self.title_path.addText(
            -title_dim["w"] / 2,
            (-total_height / 2) + title_dim["h"],
            title_font,
            self._title_text,
        )

        # Draw the type
        self.type_path.addText(
            -title_type_dim["w"] / 2,
            (-total_height / 2) + title_dim["h"] + title_type_dim["h"],
            title_type_font,
            "(" + self._type_text + ")",
        )

        y = (-total_height / 2) + title_dim["h"] + title_type_dim["h"] + port_dim["h"]

        y_input = y
        y_output = y

        for port in self._ports:
            if port.is_output():
                port.setPos(total_width / 2 - 10, y_output)
                y_output += port_dim["h"]
            else:
                port.setPos(-total_width / 2 + 10, y_input)
                y_input += port_dim["h"]

        self.setPath(path)

        self._width = total_width
        self._height = total_height

    def select_connections(self, value):
        for port in self._ports:
            if port.connection:
                port.connection._do_highlight = value
                port.connection.update_path()
            # for connection in port.connections():
            #     connection._do_highlight = value
            #     connection.update_path()

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu(self)
        pos = event.pos()

        # actions
        delete_node = QtWidgets.QAction("Delete Node")
        edit_node = QtWidgets.QAction("Edit Node")
        menu.addAction(delete_node)

        action = menu.exec_(self.mapToGlobal(pos))

    def delete(self):
        self.remove()

    def remove(self):
        """Delete the connection.
        Remove any found connections ports by calling :any:`Port.remove_connection`.  After connections
        have been removed set the stored :any:`Port` to None. Lastly call :any:`QGraphicsScene.removeItem`
        on the scene to remove this widget.
        """

        to_delete = []

        for port in self._ports:
            if port.connection:
                to_delete.append(port.connection)

        for connection in to_delete:
            connection.delete()

        self.scene().removeItem(self)
