# from PySide2.QtWidgets import QWidget, QVBoxLayout, QGraphicsScene

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

from submit2gm import constants
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.widgets.node_editor.view import View
from submit2gm.widgets.node_editor.node import Node
from submit2gm.widgets.node_editor.dynamic_job_preset_node import DynamicJobPresetNode
from submit2gm.widgets.node_editor.node_editor import NodeEditor

import typing

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class NodeScene(QtWidgets.QGraphicsScene):
    def dragEnterEvent(self, e):
        e.acceptProposedAction()

    def dropEvent(self, e):
        # find item at these coordinates
        item = self.itemAt(e.scenePos())
        if item.setAcceptDrops == True:
            # pass on event to item at the coordinates
            try:
                item.dropEvent(e)
            except RuntimeError:
                pass  # This will suppress a Runtime Error generated when dropping into a widget with no ProxyWidget

    def dragMoveEvent(self, e):
        e.acceptProposedAction()


class DependencyNode():

    def __init__(self, inputs: typing.List[str], outputs: typing.List[str]):
        self._inputs = inputs
        self._outputs = outputs

    def get_inputs(self) -> typing.List[str]:
        return self._inputs

    def get_outputs(self) -> typing.List[str]:
        return self._outputs


class NodeWidget(QtWidgets.QWidget, PluginAccessor):
    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtWidgets.QWidget'] = None):
        QtWidgets.QWidget.__init__(self, parent)
        PluginAccessor.__init__(self, plugin)

        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        self.node_editor = NodeEditor(self)
        self.scene = NodeScene()
        self.scene.setSceneRect(0, 0, 9999, 9999)
        self.view = View(self)
        self.view.setScene(self.scene)
        self.node_editor.install(self.scene)

        self._zoom_in = QtWidgets.QAction('Zoom in')
        self._zoom_out = QtWidgets.QAction('Zoom out')
        self._reset_view = QtWidgets.QAction('Reset View')

        self._view_menu = QtWidgets.QMenu('View', parent=self)
        self._view_menu.addAction(self._zoom_in)
        self._view_menu.addAction(self._zoom_out)
        self._view_menu.addAction(self._reset_view)

        self._view_menu_btn = QtWidgets.QPushButton('View')
        self._view_menu_btn.setProperty(constants.STYLE_PROPS.TAB_BTN, True)
        self._view_menu_btn.setMenu(self._view_menu)

        self._add_job_menu = QtWidgets.QMenu('Add job', parent=self)

        self._add_job = QtWidgets.QPushButton('Add job')
        self._add_job.setProperty(constants.STYLE_PROPS.TAB_BTN, True)
        self._add_job.setMenu(self._add_job_menu)

        self._menu_layout = QtWidgets.QHBoxLayout()
        self._menu_layout.setContentsMargins(0, 0, 0, 0)
        self._menu_layout.addWidget(self._view_menu_btn)
        self._menu_layout.addWidget(self._add_job)
        self._menu_layout.addStretch()

        main_layout.addLayout(self._menu_layout)
        main_layout.addWidget(self.view)

        self._add_job_menu.aboutToShow.connect(self._handle_add_job_menu_about_to_show)
        self._reset_view.triggered.connect(self.view.home_viewport)
        self._zoom_in.triggered.connect(self.view.zoom_in)
        self._zoom_out.triggered.connect(self.view.zoom_out)

        #self.view.request_node.connect(self.create_node)

    def _handle_add_job_menu_about_to_show(self) -> None:
        plugin = self.get_plugin()
        job_preset_container = plugin.get_job_preset_container()
        compatible_job_definitions = plugin.get_compatible_job_definitions()

        self._add_job_menu.clear()

        for compatible_job_definition in compatible_job_definitions:
            menu_action = QtWidgets.QAction(compatible_job_definition.get_display_name(), parent=self._add_job_menu)
            menu_action.triggered.connect(lambda: job_preset_container.emplace(compatible_job_definition))
            self._add_job_menu.addAction(menu_action)

    def add_node(self, node: 'Node') -> None:
        self.scene.addItem(node)

        scene_rect = self.view.sceneRect()
        x = scene_rect.x() + (scene_rect.width() / 2)
        y = scene_rect.y() + (scene_rect.height() / 2)
        pos = QtCore.QPointF(x, y)

        node.setPos(pos)

    def remove_node(self, node: 'Node') -> None:
        node.remove()

    def get_nodes(self) -> typing.List['Node']:
        items = self.scene.items()

        nodes = []
        for item in items:
            if isinstance(item, Node):
                nodes.append(item)

        return nodes

    def get_dynamic_job_preset_nodes(self) -> typing.List['DynamicJobPresetNode']:
        dynamic_job_preset_nodes = []
        for node in self.get_nodes():
            if isinstance(node, DynamicJobPresetNode):
                dynamic_job_preset_nodes.append(node)

        return dynamic_job_preset_nodes

    def get_dependency_graph(self) -> typing.Dict[str, 'DependencyNode']:
        dynamic_job_preset_nodes = self.get_dynamic_job_preset_nodes()

        dependency_graph = dict()

        for dynamic_job_preset_node in dynamic_job_preset_nodes:
            dynamic_job_preset = dynamic_job_preset_node.get_dynamic_job_preset()
            inputs = dynamic_job_preset_node.get_inputs()
            outputs = dynamic_job_preset_node.get_outputs()

            input_ids = list(map(lambda x: x.get_dynamic_job_preset().get_id(), inputs))
            output_ids = list(map(lambda x: x.get_dynamic_job_preset().get_id(), outputs))

            dependency_graph[dynamic_job_preset.get_id()] = DependencyNode(input_ids, output_ids)

        return dependency_graph
