__all__ = 'PixmapLabel'

from PySide2 import QtWidgets, QtGui, QtCore
import typing

from .clickable_qlabel import ClickableQLabel


class PixmapLabel(ClickableQLabel):

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None):
        ClickableQLabel.__init__(self, parent=parent)

        self._text = None
        self._pixmap = None

        # scale content by default
        self.setScaledContents(True)

        # QImage to cache the pixmap()
        # to avoid constructing a new QImage on every scale operation
        self.cachedImage = None

        # used to cache the last scaled pixmap
        # to avoid calling scale again when the size is still at the same
        self.scaledPixmap = None

        # key for the currently cached QImage and QPixmap
        # used to make sure the label was not set to another QPixmap
        self.cacheKey = None

    def text(self) -> str:
        return self._text

    def setText(self, text: str):
        self._text = text

    def pixmap(self) -> 'QtGui.QPixmap':
        return self._pixmap

    def setPixmap(self, pixmap: 'QtGui.QPixmap'):
        self._pixmap = pixmap

    def paintEvent(self, event: 'QtGui.QPaintEvent'):

        # if self is assigned a pixmap
        if self.pixmap():
            style = self.style()
            painter = QtGui.QPainter(self)

            self.drawFrame(painter)
            cr = self.contentsRect()
            cr.adjust(self.margin(), self.margin(), -self.margin(), -self.margin())

            align = QtWidgets.QStyle.visualAlignment(self.layoutDirection(), self.alignment())
            if self.hasScaledContents():  # if scaling is enabled
                scaled_size = cr.size() * self.devicePixelRatioF()

                # if scaledPixmap is invalid
                if self.scaledPixmap is None or self.scaledPixmap.size() != scaled_size or \
                        self.pixmap().cacheKey() != self.cacheKey:

                    # if cachedImage is also invalid
                    if self.pixmap().cacheKey() != self.cacheKey:

                        # reconstruct cachedImage
                        self.cachedImage = self.pixmap().toImage()

                    scaled_image = self.cachedImage.scaled(scaled_size,
                                                           QtCore.Qt.KeepAspectRatioByExpanding,
                                                           QtCore.Qt.SmoothTransformation)

                    self.scaledPixmap = QtGui.QPixmap.fromImage(scaled_image)
                    self.scaledPixmap.setDevicePixelRatio(self.devicePixelRatioF())

                pix = self.scaledPixmap
            else:  # no scaling, Just use pixmap()
                pix = self.pixmap()

            opt = QtWidgets.QStyleOption()
            opt.initFrom(self)

            if not self.isEnabled():
                pix = style.generatedIconPixmap(QtGui.QIcon.Disabled, pix, opt)

            style.drawItemPixmap(painter, cr, align, pix)
            style.drawItemText(painter,
                               cr,
                               QtWidgets.QStyle.visualAlignment(self.layoutDirection(),
                                                                QtCore.Qt.AlignVCenter | QtCore.Qt.AlignLeft),
                               self.palette(), True, self.text())

        else:  # otherwise (if the label is not assigned to a pixmap)
            # call base paintEvent
            QtWidgets.QLabel.paintEvent(self, event)
