__all__ = "ExceptionPopup"

import io
import pathlib
from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets
import json
import typing
import sys

from submit2gm import constants, qt_utils, ui_constants, utils
from submit2gm.plugin_accessor import PluginAccessor
from .input_widgets.check_box import CheckBox

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class EquivalentRequestPopup(QtWidgets.QDialog, PluginAccessor):
    """
    This popup is used when an unexpected exception is encountered and the user must be informed that something went
    very wrong.
    """

    INITIAL_WINDOW_WIDTH = 900
    INITIAL_WINDOW_HEIGHT = 500

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 request: typing.Dict,
                 parent: typing.Optional['QtWidgets.QWidget'] = None):


        window_title = "Equivalent Command Line Request"

        QtWidgets.QDialog.__init__(self, parent)
        PluginAccessor.__init__(self, plugin)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.setWindowModality(QtCore.Qt.ApplicationModal)

        # apply style sheet
        plugin.get_user_interface().set_style_sheet(self)

        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Set the window title to the plugin name
        self.setWindowTitle(window_title)

        # resize the window to it's initial size
        self.resize(EquivalentRequestPopup.INITIAL_WINDOW_WIDTH, EquivalentRequestPopup.INITIAL_WINDOW_HEIGHT)

        # set fixed dimensions
        self.setFixedWidth(EquivalentRequestPopup.INITIAL_WINDOW_WIDTH)
        self.setFixedHeight(EquivalentRequestPopup.INITIAL_WINDOW_HEIGHT)

        self.setWindowIcon(plugin.get_user_interface().get_icon())

        """ Create Widgets """
        self._header = QtWidgets.QLabel(window_title)
        self._header.setProperty(constants.STYLE_PROPS.H2, True)

        self._extra_message = QtWidgets.QLabel("\"Command Line Request\" is the submit2gm command line command to submit "
                                               "with the parameters you have selected.\n\"HTTP JSON Request\" is for "
                                               "use with Envoy's /project-submit HTTP endpoint if you need to submit "
                                               "programmatically.")

        # Get the python executable used to start Submit2gm
        python_exe_path = pathlib.Path(sys.executable).absolute().as_posix()

        # Use non pythonw exe version as pythonw will suppress info logs
        python_exe_path = python_exe_path.replace('pythonw', 'python')

        submit2gm_path = pathlib.Path(sys.argv[0]).absolute()

        command_line_request_str = f"""### Submit Request via CLI & JSON File 

To submit your request via a CLI command that takes a json request file as an argument use the following command. To 
generate a json request file switch to the HTTP Request tab and save the displayed json to a file.

```"{python_exe_path}" "{submit2gm_path.as_posix()}" {constants.SUBMIT2GM_CLI_ARGUMENTS.SUBMIT_PROJECT} {constants.SUBMIT2GM_CLI_ARGUMENTS.USE_FILE} {constants.SUBMIT2GM_CLI_ARGUMENTS.PROJECT_REQUEST} "path/to/request/file.json"```
<br />
<br />
---
### Submit Request directly via CLI

Make sure the command does not exceed your CLIs maximum character limit

```"{python_exe_path}" "{submit2gm_path.as_posix()}" {constants.SUBMIT2GM_CLI_ARGUMENTS.SUBMIT_PROJECT} {constants.SUBMIT2GM_CLI_ARGUMENTS.PROJECT_REQUEST} {json.dumps(json.dumps(request))}```
"""

        self._cmd_command = QtWidgets.QTextEdit()
        self._cmd_command.setReadOnly(True)
        self._cmd_command.setMarkdown(command_line_request_str)
        self._cmd_command.setWordWrapMode(QtGui.QTextOption.WordWrap)
        self._cmd_command.setProperty(constants.STYLE_PROPS.BACKGROUND_LEVEL_2, True)

        self._json_command = QtWidgets.QPlainTextEdit()
        self._json_command.setReadOnly(True)
        self._json_command.setPlainText(json.dumps(request, indent=3))
        self._json_command.setWordWrapMode(QtGui.QTextOption.WordWrap)
        self._json_command.setProperty(constants.STYLE_PROPS.BACKGROUND_LEVEL_2, True)

        self._tab_bar = QtWidgets.QTabWidget()
        self._tab_bar.addTab(self._cmd_command, ui_constants.DIRECTORY_ICON_WHITE, "Command Line Request")
        self._tab_bar.addTab(self._json_command, ui_constants.EYE_ICON_WHITE, "HTTP JSON Request")
        self._tab_bar.tabBar().setDocumentMode(True)
        self._tab_bar.tabBar().setExpanding(True)

        self._enable_word_wrap = CheckBox(self.get_plugin())
        self._enable_word_wrap.setText("Line wrapping")
        self._enable_word_wrap.set_value(self._cmd_command.wordWrapMode() == QtGui.QTextOption.WordWrap)
        self._enable_word_wrap.toggled.connect(self._handle__line_wrapping_toggled)

        self._save_to_file_btn = QtWidgets.QPushButton(ui_constants.NEW_FILE_ICON_BLACK, "Save to file")
        self._save_to_file_btn.setAutoDefault(False)
        self._save_to_file_btn.clicked.connect(self._handle_save_to_file)

        self._copy_to_clipboard_btn = QtWidgets.QPushButton(ui_constants.COPY_ICON_BLACK, "Copy to clipboard")
        self._copy_to_clipboard_btn.setAutoDefault(False)
        self._copy_to_clipboard_btn.clicked.connect(self._handle_copy_to_clipboard)

        self._close_btn = QtWidgets.QPushButton("Close")
        self._close_btn.clicked.connect(self.close)
        self._close_btn.setMaximumWidth(120)
        self._close_btn.setProperty(constants.STYLE_PROPS.BRAND, True)
        self._close_btn.setAutoDefault(True)

        self._col_1_widget = QtWidgets.QWidget()
        self._col_2_widget = QtWidgets.QWidget()

        """ Create Layouts """
        self._column_1 = QtWidgets.QVBoxLayout()
        self._column_1.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self._column_1.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)
        self._col_1_widget.setLayout(self._column_1)
        self._col_1_widget.setMaximumWidth(64)

        self._controls_layout = QtWidgets.QHBoxLayout()
        self._controls_layout.setAlignment(QtCore.Qt.AlignRight)
        self._controls_layout.addWidget(self._enable_word_wrap)
        self._controls_layout.addStretch()
        self._controls_layout.addWidget(self._save_to_file_btn)
        self._controls_layout.addWidget(self._copy_to_clipboard_btn)
        self._controls_layout.addWidget(self._close_btn)

        self._column_2 = QtWidgets.QVBoxLayout()
        self._column_2.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)
        self._column_2.addWidget(self._header)
        self._column_2.addWidget(self._extra_message)
        self._column_2.addSpacing(10)
        self._column_2.addWidget(self._tab_bar)
        self._column_2.addSpacing(10)
        self._column_2.addLayout(self._controls_layout)
        self._col_2_widget.setLayout(self._column_2)

        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.addWidget(self._col_1_widget)
        self._layout.addWidget(self._col_2_widget)

    @QtCore.Slot(bool)
    def _handle__line_wrapping_toggled(self, checked: bool) -> None:
        check_value = QtGui.QTextOption.WordWrap if checked else QtGui.QTextOption.NoWrap
        self._cmd_command.setWordWrapMode(check_value)
        self._json_command.setWordWrapMode(check_value)

    @QtCore.Slot()
    def _handle_save_to_file(self) -> None:
        logger = self.get_plugin().get_logging_coordinator().get_logger('EquivalentRequestPopup._handle_save_to_file')

        file_dialog = QtWidgets.QFileDialog()
        file_dialog.setFileMode(QtWidgets.QFileDialog.AnyFile)
        file_dialog.setViewMode(QtWidgets.QFileDialog.Detail)
        file_dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
        file_dialog.setNameFilter("*")
        file_dialog.setDefaultSuffix('.txt' if self._tab_bar.currentWidget() == self._cmd_command else '.json')
        file_dialog.setLabelText(QtWidgets.QFileDialog.Accept, "Save")
        file_dialog.setLabelText(QtWidgets.QFileDialog.Reject, "Cancel")
        file_dialog.setLabelText(QtWidgets.QFileDialog.FileName, "File Name")
        file_dialog.setDirectory(utils.get_desktop_path())

        if file_dialog.exec_():
            file_names = file_dialog.selectedFiles()

            # only one file can be selected anyway
            save_file = file_names[0]

            try:
                with io.open(save_file, 'w', encoding='utf-8') as f:
                    f.write(self._tab_bar.currentWidget().toPlainText())
                    f.close()

                QtWidgets.QMessageBox.information(self,
                                                  "GridMarkets",
                                                  "File saved successfully",
                                                  QtWidgets.QMessageBox.Ok,
                                                  defaultButton=QtWidgets.QMessageBox.Ok)

            except Exception as e:
                qt_utils.log_error(self, logger, str(e))

    @QtCore.Slot()
    def _handle_copy_to_clipboard(self) -> None:
        self.get_plugin().get_plugin_utils().copy_to_clipboard(self._tab_bar.currentWidget().toPlainText())