__all__ = 'MainWindow'

from PySide2 import QtCore
from PySide2 import QtWidgets
from PySide2 import QtGui

import pathlib
import os
import time
import typing

from submit2gm import constants, qt_utils
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.sumbit2gm_user_interface import Submit2GMUserInterface
from submit2gm.widgets.main_ui import MainUI
from submit2gm.widgets.sign_in_widget import SignInWidget

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class MainWindow(QtWidgets.QDialog, PluginAccessor):
    SHADOW_BLUR_RADIUS = 30

    def __init__(self,
                 plugin: 'Submit2gmPlugin',
                 reset_callback: typing.Callable[[], None],
                 parent: typing.Optional['QtWidgets.QWidget'] = None):

        QtWidgets.QDialog.__init__(self, parent)
        PluginAccessor.__init__(self, plugin)

        self._reset_callback = reset_callback

        # add the MainWindow to the plugins user interface class
        plugin.get_user_interface().set_main_window(self)

        # load custom fonts
        Submit2GMUserInterface.load_fonts()

        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Make the window appear on top to begin with. This is because new windows on Windows OS do not get brought
        # to the front if another process has focus
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)

        # Set the window title to the plugin name
        self.setWindowTitle(plugin.get_name())

        self.setWindowIcon(plugin.get_user_interface().get_icon())

        self._closed = False

        # The main window has two main states:
        # 1. The user has yet to sign in
        #    - Display a simple sign-in layout
        #
        # 2. The user is signed in
        #    - Display the main plugin user interface
        #
        # The layout for (2) can not be generated until the user is signed in so the creation of that layout is deferred
        # until after the user has signed-in.

        # the container widget is the widget sub widgets (1) & (2) will be parented to. This is done so that
        self._container_widget = QtWidgets.QStackedWidget()
        self._container_widget.setProperty(constants.STYLE_PROPS.MAIN_CONTAINER_WIDGET, True)

        # Override the style object and apply it to the container, this way any styles applied by the parent do not
        # affect our widgets.
        self._style_override = QtWidgets.QStyleFactory().create('windows')
        self._container_widget.setStyle(self._style_override)

        self._sign_in_widget = None
        self._main_ui_widget = None

        self._sign_in_widget_index = None
        self._main_ui_widget_index = None

        self._shadow_widget = QtWidgets.QWidget(self)
        self._shadow_widget.setAutoFillBackground(True)
        self._shadow_widget.setGeometry(MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS,
                                        SignInWidget.WINDOW_WIDTH - MainWindow.SHADOW_BLUR_RADIUS * 2,
                                        SignInWidget.WINDOW_HEIGHT - MainWindow.SHADOW_BLUR_RADIUS * 2)

        self._shadow_effect = QtWidgets.QGraphicsDropShadowEffect(self)
        self._shadow_effect.setColor(QtGui.QColor('#444444'))
        self._shadow_effect.setBlurRadius(MainWindow.SHADOW_BLUR_RADIUS)
        self._shadow_effect.setOffset(0, 0)
        self._shadow_widget.setGraphicsEffect(self._shadow_effect)

        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.addWidget(self._container_widget)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)

        self.display_signed_out_ui()

        # Watch ipc folder for ipc signals
        logger = self.get_plugin().get_logging_coordinator().get_logger('ipc')
        logger.info('Watching: %s' % constants.FILE_SYSTEM_WATCHER_PATH.as_posix())
        self._fs_watcher = QtCore.QFileSystemWatcher([constants.FILE_SYSTEM_WATCHER_PATH.as_posix()])
        self._fs_watcher.directoryChanged.connect(self.directory_changed)

        # Clear any 'ipc-signal-' files that exist in the file system watcher path
        # There shouldn't be any 'ipc-signal-' files as each file should be deleted after it is detected but it is
        # conceivable that a file could be left hanging if something crashed unexpectedly so best to clean up.
        # Note that this could incorrectly delete a valid 'ipc-signal-' file for another running instance of Submit2gm
        # but that is very unlikely.
        for path in constants.FILE_SYSTEM_WATCHER_PATH.iterdir():
            if path.is_file() and path.name.startswith('.ipc-signal-'):
                path.unlink(missing_ok=True)


    def directory_changed(self, path):
        pid = os.getpid()
        raise_window_signal = pathlib.Path(constants.FILE_SYSTEM_WATCHER_PATH) / ('.ipc-signal-raise-window-' + str(pid))

        if raise_window_signal.exists():
            # Remove the ipc signal raise window file
            raise_window_signal.unlink(missing_ok=True)

            # Do nothing if the user is not signed in and the active cps is not alive
            if not self.get_plugin().get_api_client().is_user_signed_in():
                try:
                    cps = self.get_plugin().get_active_command_port_server()
                    if not cps.is_still_running():
                        return
                except ValueError:
                    return

            self.show()
            self.raise_()
            self.activateWindow()

            # Reload scene settings
            self.get_main_ui_widget().get_main_ui_switcher().get_submission_view_widget().reload_scene_settings()

    def is_closed(self) -> bool:
        return self._closed

    def get_sign_in_widget(self) -> 'SignInWidget':

        if not self._sign_in_widget:
            self._sign_in_widget = SignInWidget(self.get_plugin())

        return self._sign_in_widget

    def get_main_ui_widget(self) -> 'MainUI':

        if not self._main_ui_widget:
            self._main_ui_widget = MainUI(self.get_plugin())

        return self._main_ui_widget

    def get_sign_in_widget_index(self) -> int:

        if self._sign_in_widget_index is None:
            self._sign_in_widget_index = self._container_widget.addWidget(self.get_sign_in_widget())

        return self._sign_in_widget_index

    def get_main_ui_widget_index(self) -> int:

        if self._main_ui_widget_index is None:
            self._main_ui_widget_index = self._container_widget.addWidget(self.get_main_ui_widget())

        return self._main_ui_widget_index

    def display_signed_in_ui(self) -> None:
        main_ui_widget_index = self.get_main_ui_widget_index()

        self._shadow_widget.hide()
        self._layout.setContentsMargins(0, 0, 0, 0)

        # set min and max values (fixed) for both width and height
        self.setMinimumWidth(MainUI.MIN_WINDOW_WIDTH)
        self.setMinimumHeight(MainUI.MIN_WINDOW_HEIGHT)

        self.setMaximumWidth(MainUI.MAX_WINDOW_WIDTH)
        self.setMaximumHeight(MainUI.MAX_WINDOW_HEIGHT)

        # perform the resize
        resize_width = MainUI.INITIAL_WINDOW_WIDTH
        screen_width = QtWidgets.QApplication.instance().desktop().screenGeometry().width()
        if resize_width > screen_width:
            resize_width = screen_width

        resize_height = MainUI.INITIAL_WINDOW_HEIGHT
        screen_height = QtWidgets.QApplication.instance().desktop().screenGeometry().height()
        if resize_height > screen_height:
            resize_height = screen_height

        settings = QtCore.QSettings()
        window_size = settings.value("windowSize", QtCore.QSize(resize_width, resize_height))

        self.resize(window_size)

        qt_utils.center_window(self)

        # load style sheet
        # self.get_plugin().get_user_interface().set_style_sheet(main_window)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.FramelessWindowHint)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowMaximizeButtonHint)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowMinimizeButtonHint)

        # Remove the stay on top hint after sign in. We know the user has focused this window at this point so it
        # can be removed.
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowStaysOnTopHint)
        self.show()

        self.get_plugin().get_user_interface().set_style_sheet(self._container_widget)
        self._container_widget.setCurrentIndex(main_ui_widget_index)

    def display_signed_out_ui(self) -> None:

        self._layout.setContentsMargins(MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS)

        self._shadow_widget.show()

        # set min and max values (fixed) for both width and height
        self.setFixedWidth(SignInWidget.WINDOW_WIDTH)
        self.setFixedHeight(SignInWidget.WINDOW_HEIGHT)

        # perform the resize
        self.resize(SignInWidget.WINDOW_WIDTH, SignInWidget.WINDOW_HEIGHT)

        qt_utils.center_window(self)

        # load style sheet
        # self.get_plugin().get_user_interface().set_style_sheet(main_window, qt_constants.WhiteTheme)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.FramelessWindowHint)

        self.get_plugin().get_user_interface().set_style_sheet(self._container_widget, constants.WhiteTheme)
        self._container_widget.setCurrentIndex(self.get_sign_in_widget_index())

    def resizeEvent(self, event: 'QtGui.QResizeEvent') -> None:
        if self.get_plugin().get_api_client().is_user_signed_in() and \
                self._container_widget.currentIndex() == self.get_main_ui_widget_index():
            settings = QtCore.QSettings()
            settings.setValue("windowSize", event.size())

        QtWidgets.QWidget.resizeEvent(self, event)

    def set_reset_callback(self, reset_callback: typing.Callable[[], None]):
        self._reset_callback = reset_callback

    def reset(self) -> None:
        self._reset_callback()

    def close_window(self) -> None:
        self._closed = True

        while self.get_main_ui_widget().get_header().is_polling_cps():
            time.sleep(0.3)

        self.close()
        Submit2GMUserInterface.unload_fonts()
