__all__ = ['PreflightWindow']

from PySide2 import QtCore
from PySide2 import QtWidgets
import pathlib
import typing

from submit2gm import constants, qt_utils, ui_constants, utils
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.packed_project import PackedProject
from .qt_model_file_tree import FileTreeModel
from .qt_model_watch_pattern_table import WatchPatternTable
from .input_widgets.path_input import PathInput
from .notification import Notification

if typing.TYPE_CHECKING:
    from PySide2 import QtGui
    from submit2gm.submit2gm_plugin import Submit2gmPlugin
    from submit2gm.dynamic_job_preset import DynamicJobPreset
    from submit2gm.widgets.node_editor.node_widget import DependencyNode


class PreflightWindow(QtWidgets.QDialog, PluginAccessor):

    INITIAL_WINDOW_WIDTH = 1200
    INITIAL_WINDOW_HEIGHT = 700

    MINIMUM_WINDOW_WIDTH = 600
    MINIMUM_WINDOW_HEIGHT = 600

    UPLOAD_AND_SUBMIT = "Upload and Submit"
    UPLOAD = "Upload"
    SUBMIT = "Submit"

    UPLOAD_AND_SUBMIT_TOOL_TIP = "Tell Envoy to upload the above selected files and submit your job once all files " \
                                 "have been uploaded."
    UPLOAD_TOOL_TIP = "Tell Envoy to upload the above selected files."
    SUBMIT_TOOL_TIP = "Tell Envoy to submit your job without uploading any new files."

    def __init__(self, plugin: 'Submit2gmPlugin',
                 packed_project: 'PackedProject',
                 submission_name: str,
                 submission_comment: str,
                 dynamic_job_presets: typing.List['DynamicJobPreset'],
                 dependency_graph: typing.Dict[str, 'DependencyNode'],
                 parent: typing.Optional['QtWidgets.QWidget'] = None):

        QtWidgets.QDialog.__init__(self, parent)
        PluginAccessor.__init__(self, plugin)

        self._packed_project = packed_project
        self._submission_name = submission_name
        self._submission_comment = submission_comment
        self._dynamic_job_presets = dynamic_job_presets
        self._dependency_graph = dependency_graph

        self.setWindowModality(QtCore.Qt.ApplicationModal)

        # apply style sheet
        plugin.get_user_interface().set_style_sheet(self)
        self.setProperty(constants.STYLE_PROPS.BACKGROUND_LEVEL_1, True)

        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Set the window title to the plugin name
        self.setWindowTitle("GridMarkets Preflight")

        # resize the window to it's initial size
        self.resize(PreflightWindow.INITIAL_WINDOW_WIDTH, PreflightWindow.INITIAL_WINDOW_HEIGHT)

        # set minimum dimensions
        self.setMinimumWidth(PreflightWindow.MINIMUM_WINDOW_WIDTH)
        self.setMinimumHeight(PreflightWindow.MINIMUM_WINDOW_HEIGHT)

        self.setWindowIcon(plugin.get_user_interface().get_icon())

        self._cancel_submission = True

        """ Create Widgets """
        self._header_layout_widget = QtWidgets.QWidget()
        self._header_layout_widget.setProperty(constants.STYLE_PROPS.BACKGROUND_LEVEL_0, True)

        self._title = QtWidgets.QLabel(u"Preflight")
        self._title.setProperty(constants.STYLE_PROPS.H3, True)

        self._upload_files_disclaimer = Notification("Please select all files you wish to upload. Upload progress will "
                                                     "show within Envoy upon submission.")

        self._watch_patterns_disclaimer = Notification("Please verify the below watch patterns. Watch patterns "
                                                       "determine which files are returned and where they are "
                                                       "downloaded to.")


        self._watch_pattern_model = WatchPatternTable(packed_project, dynamic_job_presets)

        self._tree_view = QtWidgets.QTreeView()
        self._tree_view.setProperty(constants.STYLE_PROPS.BACKGROUND_LEVEL_2, True)

        self._upload_file_stats_label = QtWidgets.QLabel('')

        self.file_tree_model = FileTreeModel(self.get_plugin(), packed_project, self._tree_view)
        self._tree_view.setModel(self.file_tree_model)

        self.file_tree_model.set_default_expanded_state(self._tree_view)
        self._tree_view.header().setStretchLastSection(False)
        self._tree_view.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self._tree_view.header().swapSections(FileTreeModel.PATH_INDEX, FileTreeModel.DISPLAY_HIDDEN_INDEX)
        self._tree_view.setAlternatingRowColors(True)
        self._tree_view.resizeColumnToContents(FileTreeModel.DISPLAY_HIDDEN_INDEX)

        self._tree_view.expanded.connect(self.file_tree_model.handle_expanded_index)
        self.file_tree_model.collapse.connect(self._tree_view.collapse)
        self.file_tree_model.expand.connect(self._tree_view.expand)
        self.file_tree_model.upload_files_changed.connect(self._handle_upload_files_changed)
        self._handle_upload_files_changed()

        self._watch_patterns_view = QtWidgets.QTableView()
        self._watch_patterns_view.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self._watch_patterns_view.setProperty(constants.STYLE_PROPS.BACKGROUND_LEVEL_2, True)
        self._watch_patterns_view.setModel(self._watch_pattern_model)
        self._watch_patterns_view.verticalHeader().hide()
        self._watch_patterns_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

        self._add_watch_pattern_btn = QtWidgets.QPushButton('add')
        self._add_watch_pattern_btn.setFlat(True)
        self._add_watch_pattern_btn.setToolTip(qt_utils.split_tool_tip('Add a new watch pattern.'))
        self._add_watch_pattern_btn.clicked.connect(self._add_watch_pattern)
        self._add_watch_pattern_btn.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        self._remove_watch_pattern_btn = QtWidgets.QPushButton('remove')
        self._remove_watch_pattern_btn.setFlat(True)
        self._remove_watch_pattern_btn.setToolTip(qt_utils.split_tool_tip('Remove the selected watch patterns.'))
        self._remove_watch_pattern_btn.clicked.connect(self._remove_watch_pattern)
        self._remove_watch_pattern_btn.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self._remove_watch_pattern_btn.setEnabled(False)

        watch_patterns_header = self._watch_patterns_view.horizontalHeader()

        #watch_patterns_header.setDefaultSectionSize()
        #watch_patterns_header.setStretchLastSection(False)

        #watch_patterns_header.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
        #watch_patterns_header.resizeSections(QtWidgets.QHeaderView.Stretch)

        watch_patterns_header.resizeSection(0, (PreflightWindow.INITIAL_WINDOW_WIDTH / 2) - 75)
        watch_patterns_header.resizeSection(1, (PreflightWindow.INITIAL_WINDOW_WIDTH / 2) - 75)
        watch_patterns_header.setStretchLastSection(True)


        #watch_patterns_header.setSectionResizeMode(0, QtWidgets.QHeaderView.Interactive)
        #watch_patterns_header.setSectionResizeMode(1, QtWidgets.QHeaderView.Interactive)
        #watch_patterns_header.setSectionResizeMode(2, QtWidgets.QHeaderView.Fixed)
        #watch_patterns_header.resizeSection(2, 80)

        #watch_patterns_header.resizeSections(QtWidgets.QHeaderView.Stretch)
        self._watch_patterns_view.setAlternatingRowColors(True)
        self._watch_patterns_view.selectionModel().selectionChanged.connect(self._watch_pattern_selection_changed)
        #self._watch_patterns_view.setTextElideMode(QtCore.Qt.ElideMiddle)

        watch_patterns_view = self._watch_patterns_view
        self._size_hint = QtCore.QSize(20, 29)
        size_hint = self._size_hint

        class subclassedQStyledItemDelegate(QtWidgets.QStyledItemDelegate):
            def initStyleOption(self, option, index):
                option.textElideMode = QtCore.Qt.ElideNone
                #option.backgroundBrush = QtGui.QBrush(QtGui.QColor(255, 150, 150))
                QtWidgets.QStyledItemDelegate.initStyleOption(self, option, index)


            def paint(self, painter, option, index):
                if not index.isValid():
                    return

                opt = option

                self.initStyleOption(opt, index)
                padding = 3

                textElideMode = opt.textElideMode
                text = opt.text

                opt.textElideMode = QtCore.Qt.ElideNone
                opt.text = ""

                QtWidgets.QApplication.style().drawControl(QtWidgets.QStyle.CE_ItemViewItem, opt, painter)

                opt.textElideMode = textElideMode
                opt.text = text

                painter.save()
                painter.setClipRect(opt.rect)
                opt.rect = opt.rect.adjusted(padding, padding, -padding, -padding)
                if not watch_patterns_view.isPersistentEditorOpen(index):
                    painter.drawText(opt.rect, QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter,
                                     opt.fontMetrics.elidedText(opt.text, QtCore.Qt.ElideMiddle, opt.rect.width()))
                painter.restore()

            def setEditorData(self, editorWidget, modelIndex):
                column = modelIndex.column()

                if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
                    editorWidget.set_value(modelIndex.data())
                else:
                    editorWidget.setText(modelIndex.data())

            def setModelData(self, editorWidget, model, modelIndex):
                column = modelIndex.column()

                if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
                    model.setData(modelIndex, editorWidget.get_value())
                else:
                    QtWidgets.QStyledItemDelegate.setModelData(self, editorWidget, model, modelIndex)

            def updateEditorGeometry(self, editorWidget, option, modelIndex):
                rect = option.rect
                rect.setWidth(watch_patterns_header.sectionSize(modelIndex.column()))
                editorWidget.setMaximumSize(rect.width(), rect.height())
                editorWidget.setMinimumSize(rect.width(), rect.height())
                editorWidget.setGeometry(rect)
                editorWidget.setFixedSize(rect.width(), rect.height())
                editorWidget.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
                editorWidget.updateGeometry()
                editorWidget.sizeHint = size_hint

            def createEditor(self, parent, option, modelIndex):
                column = modelIndex.column()

                if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
                    editorWidget = PathInput(plugin, file_mode= QtWidgets.QFileDialog.Directory, parent=parent)
                    editorWidget.input_value_changed.connect(lambda: self.setModelData(editorWidget,
                                                                                       modelIndex.model(),
                                                                                       modelIndex))
                else:
                    editorWidget = QtWidgets.QLineEdit(parent)

                return editorWidget

            def destroyEditor(self, editorWidget, modelIndex):
                QtWidgets.QStyledItemDelegate.destroyEditor(self, editorWidget, modelIndex)

        self._styled_item_delegate = subclassedQStyledItemDelegate()
        self._watch_patterns_view.setItemDelegate(self._styled_item_delegate)

        self._controls_layout_widget = QtWidgets.QWidget()

        self._cancel_btn = QtWidgets.QPushButton(u"Cancel")

        self._primary_action_1 = QtWidgets.QAction(PreflightWindow.UPLOAD_AND_SUBMIT, self)
        self._primary_action_1.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_AND_SUBMIT_TOOL_TIP))
        self._primary_action_2 = QtWidgets.QAction(PreflightWindow.UPLOAD, self)
        self._primary_action_2.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_TOOL_TIP))
        self._primary_action_3 = QtWidgets.QAction(PreflightWindow.SUBMIT, self)
        self._primary_action_3.setToolTip(qt_utils.split_tool_tip(PreflightWindow.SUBMIT_TOOL_TIP))

        self._primary_actions_menu = QtWidgets.QMenu()
        self._primary_actions_menu.setToolTipsVisible(True)
        self._primary_actions_menu.addAction(self._primary_action_1)
        self._primary_actions_menu.addAction(self._primary_action_2)

        # only show the submit action if the project already exists
        if packed_project.get_name() in plugin.get_api_client().get_root_directories():
            self._primary_actions_menu.addAction(self._primary_action_3)

        plugin.get_user_interface().set_style_sheet(self._primary_actions_menu)

        self._primary_btn = QtWidgets.QToolButton()
        self._primary_btn.setProperty(constants.STYLE_PROPS.BRAND, True)
        self._primary_btn.setMinimumWidth(175)
        self._primary_btn.setMenu(self._primary_actions_menu)
        self._primary_btn.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        self.primary_action_1_slot()

        self._add_additional_files_action = QtWidgets.QAction(ui_constants.FILE_ICON_WHITE, "Add Additional Files")
        self._get_equivalent_cmd_request_action = QtWidgets.QAction(ui_constants.CONSOLE_ICON_WHITE, "Get Equivalent Command Line Request")

        self._hamburger_menu = QtWidgets.QMenu(parent=self)
        self._hamburger_menu.setLayoutDirection(QtCore.Qt.LeftToRight)
        self._hamburger_menu.addAction(self._add_additional_files_action)
        self._hamburger_menu.addSeparator()
        self._hamburger_menu.addAction(self._get_equivalent_cmd_request_action)

        self._hamburger_menu_btn = QtWidgets.QPushButton(ui_constants.MENU_ICON_WHITE, "")
        self._hamburger_menu_btn.setMenu(self._hamburger_menu)
        self._hamburger_menu_btn.setAutoDefault(False)
        self._hamburger_menu_btn.setFlat(True)
        self._hamburger_menu_btn.setLayoutDirection(QtCore.Qt.RightToLeft)

        self._upload_files_widget = QtWidgets.QWidget()
        self._watch_patterns_widget = QtWidgets.QWidget()
        self._watch_patterns_widget.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)

        self._tab_bar = QtWidgets.QTabWidget()
        tab1_index = self._tab_bar.addTab(self._upload_files_widget, ui_constants.DIRECTORY_ICON_WHITE, "Upload Files")
        tab2_index = self._tab_bar.addTab(self._watch_patterns_widget, ui_constants.EYE_ICON_WHITE, "Watch Patterns")
        self._tab_bar.tabBar().setDocumentMode(True)
        self._tab_bar.tabBar().setExpanding(True)

        self._tab_bar.setTabToolTip(tab1_index, "Select the files you would like to upload.")
        self._tab_bar.setTabToolTip(tab2_index, "Set watch patterns to choose which files get returned.")

        self._generating_command_line_request_dialog: 'QtWidgets.QMessageBox' = QtWidgets.QMessageBox()
        self._generating_command_line_request_dialog.setWindowModality(QtCore.Qt.ApplicationModal)
        self._generating_command_line_request_dialog.setText('Generating equivalent command line request.')
        self._generating_command_line_request_dialog.setStandardButtons(QtWidgets.QMessageBox.NoButton)
        self.get_plugin().get_user_interface().set_style_sheet(self._generating_command_line_request_dialog)

        # Only open the _generating_command_line_request_dialog dialog if it takes a while to generate the command. This
        # timer is used to delay the triggering of show.
        self._generating_command_line_request_dialog_open_timer = QtCore.QTimer()
        self._generating_command_line_request_dialog_open_timer.setInterval(200)
        self._generating_command_line_request_dialog_open_timer.setSingleShot(True)

        """ Create Connections """
        self._add_additional_files_action.triggered.connect(self._add_additional_files)
        self._get_equivalent_cmd_request_action.triggered.connect(self._get_equivalent_cmd_request)

        self._cancel_btn.clicked.connect(self.cancel)
        self._primary_btn.clicked.connect(self.submit)
        self._primary_actions_menu.aboutToShow.connect(self._about_to_show_primary_actions_menu)

        self._primary_action_1.triggered.connect(self.primary_action_1_slot)
        self._primary_action_2.triggered.connect(self.primary_action_2_slot)
        self._primary_action_3.triggered.connect(self.primary_action_3_slot)

        self._generating_command_line_request_dialog_open_timer.timeout.connect(self._generating_command_line_request_dialog.show)

        """ Create Layouts """
        header_layout = QtWidgets.QHBoxLayout()
        header_layout.addWidget(self._title)
        header_layout.addStretch()
        header_layout.addWidget(self._hamburger_menu_btn)
        self._header_layout_widget.setLayout(header_layout)

        body_layout = QtWidgets.QVBoxLayout()
        body_layout.addWidget(self._upload_files_disclaimer)
        body_layout.addWidget(self._tree_view)
        body_layout.setContentsMargins(0,10,0,0)
        self._upload_files_widget.setLayout(body_layout)

        self._watch_pattern_controls_layout = QtWidgets.QHBoxLayout()
        self._watch_pattern_controls_layout.addWidget(self._add_watch_pattern_btn)
        self._watch_pattern_controls_layout.addWidget(self._remove_watch_pattern_btn)
        self._watch_pattern_controls_layout.addStretch()

        watch_pattern_layout = QtWidgets.QVBoxLayout()
        watch_pattern_layout.setContentsMargins(0,10,0,0)
        watch_pattern_layout.addWidget(self._watch_patterns_disclaimer)
        watch_pattern_layout.addWidget(self._watch_patterns_view)
        watch_pattern_layout.addLayout(self._watch_pattern_controls_layout)

        self._watch_patterns_widget.setLayout(watch_pattern_layout)

        controls_layout = QtWidgets.QHBoxLayout()
        controls_layout.setContentsMargins(0,0,0,0)
        controls_layout.setAlignment(QtCore.Qt.AlignRight)
        controls_layout.addWidget(self._upload_file_stats_label)
        controls_layout.addStretch()
        controls_layout.addWidget(self._primary_btn)
        controls_layout.addWidget(self._cancel_btn)
        self._controls_layout_widget.setLayout(controls_layout)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self._header_layout_widget)
        layout.addWidget(self._tab_bar)
        layout.addWidget(self._controls_layout_widget)

    def _handle_upload_files_changed(self):
        upload_count = self.file_tree_model.get_upload_count()
        upload_total_size = self.file_tree_model.get_upload_total_size()

        upload_count_str = '{} files marked for upload  |  '.format(upload_count)
        if upload_count == 1:
            upload_count_str = '1 file marked for upload  |  '

        upload_count_str = upload_count_str + utils.bytes_to_human(upload_total_size)

        self._upload_file_stats_label.setText(upload_count_str)

    def _add_additional_files(self):
        file_dialog = QtWidgets.QFileDialog(parent=self)
        file_dialog.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
        if file_dialog.exec_():
            urls = file_dialog.selectedUrls()
            for url in urls:
                self.file_tree_model.insert_file_into_model(pathlib.Path(url.toLocalFile()))

    def _get_equivalent_cmd_request(self):
        api_client = self.get_plugin().get_api_client()

        files_to_upload = set(self.get_files_to_upload())
        packed_project = PackedProject(self._packed_project.get_name(),
                                       self._packed_project.get_root_dir(),
                                       files_to_upload,
                                       self._packed_project.get_attributes())

        watch_patterns = self._watch_pattern_model.get_watch_patterns()
        packed_project.set_watch_patterns(watch_patterns)

        # Convert the dynamic job preset to a static job preset
        static_job_presets = []
        for dynamic_job_preset in self._dynamic_job_presets:
            static_job_preset = dynamic_job_preset.get_static_job_preset(packed_project)
            static_job_presets.append(static_job_preset)

        self._generating_command_line_request_dialog_open_timer.start()

        worker_thread = self.get_plugin().get_user_interface().get_worker_thread("Generating command line request...")

        ################################################################################################################

        target = api_client.get_epc_project

        args = (
            packed_project,
            self._submission_name,
            self._submission_comment
        )

        kwargs = {
            "static_job_presets": static_job_presets,
            "dependency_graph": self._dependency_graph
        }

        worker_thread.process(target, args, kwargs, self._handle_get_epc_project_result)

    def _handle_get_epc_project_result(self, result: any) -> bool:
        import gridmarkets

        self._generating_command_line_request_dialog_open_timer.stop()
        self._generating_command_line_request_dialog.done(0)

        if isinstance(result, gridmarkets.Project):
            import json
            import sys
            from .qt_dialog_equivalent_request_popup import EquivalentRequestPopup

            popup = EquivalentRequestPopup(self.get_plugin(), result.serialize)
            popup.show()
            popup.raise_()
            popup.activateWindow()

            return True

        return False

    def _add_watch_pattern(self) -> None:
        self._watch_pattern_model.add_row()

    def _remove_watch_pattern(self) -> None:
        rows = set()
        for index in self._watch_patterns_view.selectedIndexes():
            rows.add(index.row())

        for row in sorted(rows, reverse=True):
            self._watch_pattern_model.removeRow(row)

    def _watch_pattern_selection_changed(self,
                                         selected: typing.List['WatchPattern'],
                                         deselected: typing.List['WatchPattern']):
        self._remove_watch_pattern_btn.setEnabled(len(selected) > 0)

    def primary_action_1_slot(self) -> None:
        self._primary_btn.setText(PreflightWindow.UPLOAD_AND_SUBMIT)
        self._primary_btn.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_AND_SUBMIT_TOOL_TIP))

    def primary_action_2_slot(self) -> None:
        self._primary_btn.setText(PreflightWindow.UPLOAD)
        self._primary_btn.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_TOOL_TIP))

    def primary_action_3_slot(self) -> None:
        self._primary_btn.setText(PreflightWindow.SUBMIT)
        self._primary_btn.setToolTip(qt_utils.split_tool_tip(PreflightWindow.SUBMIT_TOOL_TIP))

    def cancel(self) -> None:
        self._cancel_submission = True
        self.hide()

    def submit(self) -> None:
        watch_patterns = self._watch_pattern_model.get_watch_patterns()
        self._packed_project.set_watch_patterns(watch_patterns)
        self._cancel_submission = False
        self.hide()

    def _about_to_show_primary_actions_menu(self) -> None:
        self._primary_actions_menu.setFixedWidth(self._primary_btn.width())

    def cancel_submission(self) -> bool:
        return self._cancel_submission

    def get_next_action(self) -> str:
        return self._primary_btn.text()

    def get_files_to_upload(self) -> typing.List[pathlib.Path]:
        return self.file_tree_model.get_files_to_upload()

    def get_forbidden_files_to_upload(self) -> typing.List[pathlib.Path]:
        return self.file_tree_model.get_forbidden_files_to_upload()