__all__ = ['FileTreeModel']

import pathlib
from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets
import operator
import os
import time
import typing
import weakref

from submit2gm import constants, ui_constants, utils
from submit2gm.worker_thread import WorkerThread
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.exception_wrapper import ExceptionWrapper

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin
    from submit2gm.packed_project import PackedProject


class PathlibTreeNode:
    def __init__(self, path: pathlib.Path, is_loaded: bool, is_file: bool):
        self._path = path
        self._posix_path = path.as_posix() # Cache the posix path for faster sorting
        self._is_loaded = is_loaded
        self._is_file = is_file
        self.children: typing.List['PathlibTreeNode'] = []

    def get_path(self) -> pathlib.Path:
        return self._path

    def get_posix_path(self) -> str:
        return self._posix_path

    def is_loaded(self) -> bool:
        return self._is_loaded

    def is_file(self) -> bool:
        return self._is_file

    def is_directory(self) -> bool:
        return not self.is_file()

    def add_child(self, new_child: 'PathlibTreeNode') -> None:
        # Find alphabetical insert position
        insert_pos = 0

        for child in self.children:
            if child.get_posix_path() < new_child.get_posix_path():
                insert_pos = insert_pos + 1
            else:
                break

        self.children.insert(insert_pos, new_child)

    def get_children(self) -> typing.List['PathlibTreeNode']:
        return self.children


class TreeNode(QtCore.QObject):
    """ A Simple class to represent a node in a tree for use with tree models. """

    # use str instead of int as int maps to c++ 4 byte int which will crash the program if we pass too large a value
    upload_changed = QtCore.Signal(bool, str)

    def __init__(self, model: 'FileTreeModel', path: pathlib.Path, size: int, display_hidden: bool, upload: bool, parent_node: typing.Optional['TreeNode'] = None):
        QtCore.QObject.__init__(self)

        self._model = model

        if path.is_file():
            self.upload_changed.connect(self._model._handle_file_upload_changed)
            self.destroyed.connect(self.file_destroyed_handler)

        self._column_path = path
        self._column_size = size
        self._column_display_hidden = display_hidden
        self._column_upload = False
        self.set_upload(upload)

        forbidden_paths = list(map(lambda x: pathlib.Path(x), constants.FORBIDDEN_LINUX_PATHS))

        # A TreeNode is forbidden if any of the paths parents are a forbidden path
        paths = [path]
        paths.extend(path.parents)
        for _path in paths:
            if _path in forbidden_paths:
                self._forbidden = True
                break
        else:
            self._forbidden = False

        self._parent_node = parent_node
        self._child_nodes = []

        # Directories will be lazy loaded if display hidden is not enabled
        self._loaded = display_hidden or not self._column_path.is_dir()

    def file_destroyed_handler(self) -> None:
        if self._column_upload:
            self.upload_changed.emit(False, str(self._column_size))

    def append_child(self, child: 'TreeNode'):
        self._child_nodes.append(child)
        #self._child_nodes.sort(key=operator.attrgetter('_column_path'))

    def insert_child(self, index: int, child: 'TreeNode') -> None:
        self._child_nodes.insert(index, child)

    def pop_child(self, index: int) -> 'TreeNode':
        return self._child_nodes.pop(index)

    def child(self, row: int) -> typing.Optional['TreeNode']:
        if row < 0 or row >= len(self._child_nodes):
            return None
        return self._child_nodes[row]

    def child_count(self) -> int:
        return len(self._child_nodes)

    def children(self) -> typing.List['TreeNode']:
        return self._child_nodes

    def sort_children(self, order: 'QtCore.Qt.SortOrder' = QtCore.Qt.AscendingOrder):
        self._child_nodes.sort(key=operator.attrgetter('_column_path'), reverse=order == QtCore.Qt.DescendingOrder)
        for child in self._child_nodes:
            child.sort_children(order)

    def set_loaded(self, loaded: bool) -> None:
        self._loaded = loaded

    def loaded(self) -> bool:
        return self._loaded

    def get_descendants(self) -> typing.List['TreeNode']:
        descendants = []
        remaining_nodes = []
        remaining_nodes.extend(self.children())

        while remaining_nodes:
            node = remaining_nodes.pop()
            descendants.append(node)
            remaining_nodes.extend(node.children())

        return descendants

    def column_count(self) -> int:
        return len(FileTreeModel.FileTreeModelHeaders)

    def is_forbidden(self):
        return self._forbidden

    def get_path(self) -> pathlib.Path:
        return self._column_path

    def set_path(self, path: pathlib.Path) -> None:
        self._column_path = path

    def get_size(self) -> int:
        return self._column_size

    def set_size(self, size: int) -> None:
        self._column_size = size

    def display_hidden(self) -> bool:
        return self._column_display_hidden

    def set_display_hidden(self, display_hidden: bool) -> None:
        self._column_display_hidden = display_hidden

    def upload(self) -> bool:
        return self._column_upload

    def set_upload(self, upload: bool) -> None:
        if self._column_upload != upload:
            self._column_upload = upload
            self.upload_changed.emit(upload, str(self._column_size))

    def contains_upload(self) -> bool:
        if self._column_upload:
            return True
        else:
            for child in self.children():
                if child.contains_upload():
                    return True

    def contains_auto_detected_file(self) -> bool:
        path_str = self.get_path().as_posix()

        for auto_detected_path in self._model.get_auto_detected_paths():
            if auto_detected_path.as_posix().startswith(path_str):
                return True

        return False

    def contains_unloaded_dir(self) -> bool:
        if not self.loaded():
            return True

        for child in self._child_nodes:
            if child.contains_unloaded_dir():
                return True

    def data(self, column: int) -> typing.Optional[any]:
        if column < 0 or column >= self.column_count():
            return None

        if column == FileTreeModel.PATH_INDEX:
            return self.get_path()
        elif column == FileTreeModel.SIZE_INDEX:
            return self.get_size()
        elif column == FileTreeModel.DISPLAY_HIDDEN_INDEX:
            return self.display_hidden()
        elif column == FileTreeModel.UPLOAD_INDEX:
            return self.upload()

        return None

    def set_data(self, column: int, value: any) -> bool:
        if column < 0 or column >= self.column_count():
            return False

        if column == FileTreeModel.PATH_INDEX:
            self.set_path(value)
        elif column == FileTreeModel.SIZE_INDEX:
            self.set_size(value)

        elif column == FileTreeModel.DISPLAY_HIDDEN_INDEX:
            self.set_display_hidden(value)

        elif column == FileTreeModel.UPLOAD_INDEX:
            self.set_upload(value)

            # if upload has been checked then set all the upload flag for all children to True and iterate through
            # parents to update their upload flags accordingly
            if value:

                # Update children's upload flags
                children = self.children()[:]

                # For each child
                while children:
                    # remove it from the children list
                    child = children.pop(0)

                    # set it's upload flag to True
                    child.set_upload(True)

                    # add all its children to the children list so that we can recurse through the tree
                    children.extend(child._child_nodes)

                # Update parent's upload flags
                child_node = self
                while child_node.parent_item():
                    parent_node = child_node.parent_item()

                    # if there is a sibling with its upload flag set to false
                    if not parent_node.loaded() or any(filter(lambda x: not x.upload(), child_node.siblings())):
                        # then don't update the parent and break the recursion
                        break
                    else:
                        # if all the siblings have their upload flag set to True then set the paretns upload flag to
                        # True and recurse up the tree
                        parent_node.set_upload(True)
                        child_node = parent_node

                return True

            # if the upload flag has been un-checked then unset the upload flag for all children and all parents
            # Update children's upload flags
            children = self.children()[:]

            # For each child
            while children:
                # remove it from the children list
                child = children.pop(0)

                # set it's upload flag to False
                child.set_upload(False)

                # add all its children to the children list so that we can recurse through the tree
                children.extend(child._child_nodes)

            # Update parent's upload flags
            child_node = self
            while child_node.parent_item():
                parent_node = child_node.parent_item()

                if not parent_node.upload():
                    break

                parent_node.set_upload(False)
                child_node = parent_node

        return True

    def row(self) -> int:
        if self._parent_node and self in self._parent_node._child_nodes:
            return self._parent_node._child_nodes.index(self)
        return 0

    def parent_item(self) -> typing.Optional['TreeNode']:
        return self._parent_node

    def siblings(self) -> typing.List['TreeNode']:
        if self._parent_node:
            # get all the parents children
            siblings = self._parent_node._child_nodes[:]

            # get index of node in question
            row = siblings.index(self)

            # remove self from the parents children to get all siblings
            siblings.pop(row)

            return siblings

        return []

    def parents(self) -> typing.List['TreeNode']:
        parents = []

        node = self
        while node.parent_item():
            parents.append(node.parent_item())
            node = node.parent_item()

        return parents

    def get_pathlib_tree_node(self) -> 'PathlibTreeNode':
        pathlib_tree_node = PathlibTreeNode(self.get_path(), self.loaded(), self.get_path().is_file())

        for child in self.children():
            pathlib_tree_node.add_child(child.get_pathlib_tree_node())

        return pathlib_tree_node


class FileTreeModelHeader:
    def __init__(self, display_text: str, tool_tip: str, icon: typing.Optional['QtGui.QIcon'] = None):
        self._display_text = display_text
        self._tool_tip = tool_tip
        self._icon = icon

    def get_display_text(self) -> str:
        return self._display_text

    def get_tool_tip(self) -> str:
        return self._tool_tip

    def get_icon(self) -> typing.Optional['QtGui.QIcon']:
        return self._icon


class FileTreeModel(QtCore.QAbstractItemModel, PluginAccessor):
    PATH_HEADER = FileTreeModelHeader("Local Path", "The local path of the file or directory on your file system.")
    PATH_INDEX = 0

    DISPLAY_HIDDEN_HEADER = FileTreeModelHeader("",
                                                "Select to display all files in this directory, not just the files selected for upload.",
                                                icon=ui_constants.DIRECTORY_HIDDEN_ICON_WHITE)
    DISPLAY_HIDDEN_INDEX = 1

    SIZE_HEADER = FileTreeModelHeader("Size", "The size of the file.")
    SIZE_INDEX = 2

    UPLOAD_HEADER = FileTreeModelHeader("Upload", "Select to upload the file to the remote folder.")
    UPLOAD_INDEX = 3

    FileTreeModelHeaders = [PATH_HEADER, DISPLAY_HIDDEN_HEADER, SIZE_HEADER, UPLOAD_HEADER]

    DIRECTORY_LOADING_INDICATOR_LOADED_CODE = 0
    DIRECTORY_LOADING_INDICATOR_LOADING_CODE = 1
    DIRECTORY_LOADING_INDICATOR_ERROR_CODE = 2

    collapse = QtCore.Signal(QtCore.QModelIndex)
    expand = QtCore.Signal(QtCore.QModelIndex)
    update_loading_indicator = QtCore.Signal()
    upload_files_changed = QtCore.Signal()

    def __init__(self, plugin: 'Submit2gmPlugin', packed_project: 'PackedProject', parent: typing.Optional[QtCore.QObject] = None):
        QtCore.QAbstractItemModel.__init__(self, parent=parent)
        PluginAccessor.__init__(self, plugin)

        # Use a WeakValueDictionary to prevent seg faults as per
        # https://www.riverbankcomputing.com/pipermail/pyqt/2017-June/039280.html
        # This is because Python can not track objects passed to QAbstractItemModel.createIndex which results in
        # QModelIndex.internalPointer() either SEGFAULTing the program or returning a random object if the object has
        # been garbage collected.
        self.objects = weakref.WeakValueDictionary()

        self._directory_loader_worker_thread: 'WorkerThread' = WorkerThread(self.get_plugin())
        self._loading_indicator: 'QtWidgets.QMessageBox' = QtWidgets.QMessageBox()
        self._loading_indicator.setWindowModality(QtCore.Qt.ApplicationModal)
        self._loading_indicator.setStandardButtons(QtWidgets.QMessageBox.Cancel)
        self._loading_indicator.setDefaultButton(QtWidgets.QMessageBox.Cancel)
        self._loading_indicator.setInformativeText("This may take a few minutes if the directory contains many files.")
        self.get_plugin().get_user_interface().set_style_sheet(self._loading_indicator)
        self._directory_loader_files_counter = 0
        self._directory_loader_start_time = 0
        self._directory_loader_last_update_time = 0
        self._directory_loader_pathlib_tree: typing.Optional['utils.PathlibTreeNode'] = None

        # Create the root node
        self.root_node = TreeNode(self, pathlib.Path(''), 0, False, False)

        # Get the project files
        project_files = packed_project.get_files()

        self._upload_count = 0
        self._upload_total_size = 0

        # Cache the files that were auto detected
        self._auto_detected_paths = packed_project.get_files()

        for project_file in list(project_files):
            self.insert_file_into_model(project_file)

        # On Windows make sure all drives are lists
        if os.name == 'nt':
            drives: typing.List[QtCore.QFileInfo] = QtCore.QDir.drives()
            for drive in drives:
                drive_path = pathlib.Path(drive.absoluteFilePath())
                for drive_node in self.root_node.children():
                    if drive_node.get_path() == drive_path:
                        break
                else:
                    drive_node = TreeNode(self, drive_path, 0, False, False, parent_node=self.root_node)
                    self.root_node.append_child(drive_node)

        self.root_node.sort_children()

        self.update_loading_indicator.connect(self.handle_update_loading_indicator, type=QtCore.Qt.QueuedConnection)

    @QtCore.Slot(str)
    def handle_update_loading_indicator(self):
        self._directory_loader_files_counter = self._directory_loader_files_counter + 1

        current_time = time.time()

        if not self._loading_indicator.isVisible() and \
                self._loading_indicator.result() == FileTreeModel.DIRECTORY_LOADING_INDICATOR_LOADING_CODE and \
                current_time - self._directory_loader_start_time > 0.5:
            self._loading_indicator.show()

        if current_time - self._directory_loader_last_update_time > 0.1:
            self._loading_indicator.setText("Loaded {} files.".format(str(self._directory_loader_files_counter)))
            self._directory_loader_last_update_time = time.time()

    def reset_loaded_files(self) -> None:
        self._directory_loader_files_counter = 0
        self._directory_loader_start_time = time.time()
        self._directory_loader_last_update_time = self._directory_loader_start_time
        self._directory_loader_pathlib_tree = None
        self._loading_indicator.setText("Loading files (0 files processed).")
        self._loading_indicator.setResult(FileTreeModel.DIRECTORY_LOADING_INDICATOR_LOADING_CODE)

    def headerData(self,
                   section: int,
                   orientation: 'QtCore.Qt.Orientation',
                   role: 'QtCore.Qt.ItemDataRole' = QtCore.Qt.DisplayRole):

        if orientation == QtCore.Qt.Horizontal:

            header = FileTreeModel.FileTreeModelHeaders[section]

            if role == QtCore.Qt.DisplayRole:
                return header.get_display_text()

            if role == QtCore.Qt.DecorationRole:
                return header.get_icon()

            if role == QtCore.Qt.ToolTipRole or role == QtCore.Qt.WhatsThisRole:
                return header.get_tool_tip()

        return None

    def data(self, index: 'QtCore.QModelIndex', role: 'QtCore.Qt.ItemDataRole' = QtCore.Qt.DisplayRole) -> any:
        if not index.isValid():
            return None

        item = self.retrieve_object(index)

        if item is None:
            return None

        column = index.column()

        path: pathlib.Path = item.data(FileTreeModel.PATH_INDEX)
        size: typing.Optional[int] = item.data(FileTreeModel.SIZE_INDEX)
        display_hidden: bool = item.data(FileTreeModel.DISPLAY_HIDDEN_INDEX)
        upload: bool = item.data(FileTreeModel.UPLOAD_INDEX)

        if column == FileTreeModel.PATH_INDEX:
            if role == QtCore.Qt.DisplayRole:

                # Do not show slash on windows drives
                if os.name == 'nt' and len(path.parts) == 1:
                    return path.parts[-1].replace('\\', '')

                return path.parts[-1]

            if role == QtCore.Qt.ToolTipRole or role == QtCore.Qt.WhatsThisRole:
                return path.as_posix() + '(loaded: {})'.format(item.loaded())

            if role == QtCore.Qt.DecorationRole:
                if path.is_dir():
                    if item.upload():
                        return ui_constants.DIRECTORY_ICON_WHITE
                    elif item.contains_upload():
                        return ui_constants.DIRECTORY_OUTLINE_ICON_WHITE
                    return ui_constants.DIRECTORY_OUTLINE_ICON_OFF_WHITE

                if path.is_file():
                    if item.upload():
                        return ui_constants.FILE_ICON_WHITE
                    return ui_constants.FILE_OUTLINE_ICON_OFF_WHITE

        elif column == FileTreeModel.SIZE_INDEX:
            if role == QtCore.Qt.DisplayRole:
                if path.is_file():
                    return utils.bytes_to_human(size)

        elif column == FileTreeModel.DISPLAY_HIDDEN_INDEX:
            if role == QtCore.Qt.CheckStateRole:

                # Nodes that are not directories cannot toggle their 'display hidden' status
                if not item.get_path().is_dir():
                    return None

                return QtCore.Qt.Checked if display_hidden else QtCore.Qt.Unchecked

        elif column == FileTreeModel.UPLOAD_INDEX:
            if role == QtCore.Qt.CheckStateRole:
                return QtCore.Qt.Checked if upload else QtCore.Qt.Unchecked

        if column != FileTreeModel.UPLOAD_INDEX and role == QtCore.Qt.ForegroundRole:

            # Display auto-detected files using a green foreground colour
            if item.get_path().is_file() and item.get_path() in self._auto_detected_paths:
                if item.upload():
                    return QtGui.QBrush(QtGui.QColor(0,255,0))
                else:
                    return QtGui.QBrush(QtGui.QColor(75,150,75))

            if upload:
                return QtGui.QBrush(QtGui.QColor(255, 255 ,255))
            else:
                return QtGui.QBrush(QtGui.QColor(150, 150, 150))

        if role == QtCore.Qt.BackgroundRole:
            if item.is_forbidden():
                return QtGui.QBrush(QtGui.QColor(255,0,0), QtCore.Qt.BDiagPattern)

        return None

    @QtCore.Slot('QtCore.QModelIndex')
    def handle_expanded_index(self, index: 'QtCore.QModelIndex') -> None:
        """
        Custom slot that must be connected to the tree views `expanded` signal to enable lazy loading.
        """
        if not index.isValid():
            return

        item = self.retrieve_object(index)

        if item is None:
            return

        if not item.loaded() and not item.contains_upload():
            self.setData(index.siblingAtColumn(FileTreeModel.DISPLAY_HIDDEN_INDEX), None, QtCore.Qt.CheckStateRole)

    def setData(self, index: 'QtCore.QModelIndex', value: any, role: int) -> bool:

        if not index.isValid():
            return False

        item = self.retrieve_object(index)

        if item is None:
            return False

        column = index.column()

        # toggle the display hidden checkbox
        if column == FileTreeModel.DISPLAY_HIDDEN_INDEX:
            if role == QtCore.Qt.CheckStateRole:
                # get the checked state for the current item
                is_checked = item.data(FileTreeModel.DISPLAY_HIDDEN_INDEX)

                loaded = item.loaded()

                # toggle the check state for the current the item
                item.set_data(FileTreeModel.DISPLAY_HIDDEN_INDEX, not is_checked)

                # Get the children after the item has been loaded
                if is_checked and loaded:
                    # Get the children before loading so we know what has changed
                    old_children = item.children()

                    new_children = list(filter(lambda x: x.contains_auto_detected_file() or x.contains_upload(), old_children))

                    current_range_start = None
                    remove_ranges = []

                    for i, old_child in enumerate(old_children):

                        if old_child in new_children:
                            old_child_index = i
                            old_model_index = self.createIndex(old_child_index, 0, old_child)
                            new_child_index = new_children.index(old_child)
                            new_model_index = self.createIndex(new_child_index, 0, old_child)
                            self.changePersistentIndex(old_model_index, new_model_index)

                        if current_range_start is None:
                            # Get the start of the insert range
                            if old_child not in new_children:
                                current_range_start = i
                        else:
                            # Find the end of the current insert range
                            if old_child in new_children:
                                remove_ranges.append((current_range_start, i - 1))
                                current_range_start = None

                    if current_range_start is not None:
                        remove_ranges.append((current_range_start, len(old_children) - 1))

                    total_rows_removed = 0
                    for remove_range in remove_ranges:
                        self.beginRemoveRows(index, remove_range[0], remove_range[1])

                        for i in range(remove_range[0], remove_range[1] + 1):
                            item.pop_child(remove_range[0] - total_rows_removed)

                        total_rows_removed = total_rows_removed + ((remove_range[1] + 1) - remove_range[0])

                        self.endRemoveRows()

                    item.set_loaded(False)

                    # Set the upload flag for all parents to false
                    parent_node = item
                    while parent_node:
                        parent_node.set_upload(False)
                        parent_node = parent_node.parent_item()

                    # collapse this item if it doesn't contain any files for upload or autodetected files
                    if not item.contains_upload() and not item.contains_auto_detected_file():
                        self.collapse.emit(index.siblingAtColumn(0))

                elif not is_checked and not loaded:
                    new_child_paths = list(item.get_path().iterdir())
                    self._unhide_directory(new_child_paths, item, index, expand=True)

                # emitting dataChanged for items with different parents is undefined in qt tree models so just update
                # the whole data structure
                self.dataChanged.emit(QtCore.QModelIndex(), QtCore.QModelIndex())
                self.layoutChanged.emit()

        elif column == FileTreeModel.UPLOAD_INDEX:
            if role == QtCore.Qt.CheckStateRole:
                # get the checked state for the current item
                is_checked = item.data(FileTreeModel.UPLOAD_INDEX)

                #if not item.loaded() and item.get_path().is_dir():
                if not is_checked and item.get_path().is_dir():
                    self.load_directory(index, item, upload=True)
                else:
                    # toggle the check state for the current the item
                    item.set_data(FileTreeModel.UPLOAD_INDEX, not is_checked)

                    # emitting dataChanged for items with different parents is undefined in qt tree models so just
                    # update the whole data structure
                    self.dataChanged.emit(QtCore.QModelIndex(), QtCore.QModelIndex())

        return QtCore.QAbstractItemModel.setData(self, index, value, role)

    def flags(self, index: 'QtCore.QModelIndex') -> 'QtCore.Qt.ItemFlags':
        if not index.isValid():
            return QtCore.Qt.NoItemFlags

        item = self.retrieve_object(index)
        column = index.column()

        if item.is_forbidden() and (column == FileTreeModel.DISPLAY_HIDDEN_INDEX or column == FileTreeModel.UPLOAD_INDEX):
            return QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsSelectable

        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsSelectable

    def index(self,
              row: int,
              column: int,
              parent: typing.Optional['QtCore.Qt.QModelIndex'] = None) -> 'QtCore.Qt.QModelIndex':

        if parent is None:
            parent = QtCore.QModelIndex()

        # validate the index exists
        if not self.hasIndex(row, column, parent):
            return QtCore.QModelIndex()

        if parent.isValid():
            parent_item = self.retrieve_object(parent)
        else:
            parent_item = None

        if parent_item is None:
            parent_item = self.root_node

        child_item = parent_item.child(row)
        if child_item:
            return self.createIndex(row, column, child_item)
        return QtCore.QModelIndex()

    def parent(self, index: 'QtCore.Qt.QModelIndex') -> 'QtCore.Qt.QModelIndex':
        if not index.isValid():
            return QtCore.QModelIndex()

        child_item = self.retrieve_object(index)

        if child_item is None:
            return QtCore.QModelIndex()

        if child_item == self.root_node:
            return QtCore.QModelIndex()

        parent_item = child_item.parent_item()

        if parent_item == self.root_node:
            return QtCore.QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def createIndex(self, row: int, col: int, obj: any) -> 'QtCore.Qt.QModelIndex':
        if isinstance(obj, int):  # for those who prefer to store ids
            return QtCore.QAbstractItemModel.createIndex(self, row, col, obj)

        # Store the objects id instead of the object itself to avoid garbage collection issues
        self.objects[id(obj)] = obj
        return QtCore.QAbstractItemModel.createIndex(self, row, col, id(obj))

    def retrieve_object(self, index: 'QtCore.Qt.QModelIndex') -> typing.Optional['TreeNode']:
        # this will return None if the object has already been collected
        return self.objects.get(index.internalId())

    def rowCount(self, parent: typing.Optional['QtCore.Qt.QModelIndex'] = QtCore.QModelIndex()) -> int:
        if parent.column() > 0:
            return 0

        if parent.isValid():
            parent_item = self.retrieve_object(parent)
        else:
            parent_item = None

        if parent_item is None:
            parent_item = self.root_node

        child_count = parent_item.child_count()

        # Since we are lazy loading folders we need to check that hasChildren is not true if there are actually no
        # children
        if not child_count and self.hasChildren(parent):
            return 1

        return child_count

    def columnCount(self, parent: typing.Optional['QtCore.Qt.QModelIndex'] = QtCore.QModelIndex()) -> int:

        if parent.isValid():
            parent_item = self.retrieve_object(parent)
            if parent_item:
                return parent_item.column_count()

        return self.root_node.column_count()

    def hasChildren(self, index: typing.Optional['QtCore.Qt.QModelIndex'] = QtCore.QModelIndex()):
        if index.column() > 0:
            return 0

        if index.isValid():
            item = self.retrieve_object(index)
        else:
            item = None

        if item is None:
            item = self.root_node

        if not item.loaded():
            try:
                return any(item.get_path().iterdir())
            except PermissionError:
                pass

        return item.child_count() > 0

    def _directory_loading_file_callback(self,
                                         item: 'TreeNode',
                                         index: 'QtCore.QModelIndex',
                                         upload: bool) -> typing.Tuple['TreeNode', 'QtCore.QModelIndex', bool, 'utils.PathlibTreeNode']:
        """
        The target for the _directory_loader_worker_thread.
        """
        return item, index, upload, self._get_pathlib_tree(item.get_pathlib_tree_node())

    def _get_pathlib_tree(self, directory: 'PathlibTreeNode') -> 'PathlibTreeNode':

        if self._loading_indicator.result() == QtWidgets.QMessageBox.Cancel:
            raise StopIteration

        # If the directory is loaded then build the tree using the cached children of the TreeNode instead of
        # re-checking the file system
        if directory.is_loaded():
            for child in directory.get_children():
                if child.is_directory():
                    self._get_pathlib_tree(child)
        else:
            try:
                for item in directory.get_path().iterdir():

                    if self._loading_indicator.result() == QtWidgets.QMessageBox.Cancel:
                        raise StopIteration

                    if item.is_dir():
                        directory.add_child(self._get_pathlib_tree(PathlibTreeNode(item, False, False)))
                    elif item.is_file():
                        self.update_loading_indicator.emit()
                        directory.add_child(PathlibTreeNode(item, True, True))

            except PermissionError:
                pass

        return directory

    def load_directory(self, index: 'QtCore.QModelIndex', item: 'TreeNode', upload: bool = False):
        self._directory_loader_worker_thread = WorkerThread(self.get_plugin())
        self.reset_loaded_files()

        # Create a worker thread to load the files in the directory
        target = self._directory_loading_file_callback

        args = (
            item,
            index,
            upload
        )

        kwargs = {}

        self._directory_loader_worker_thread.process(target, args, kwargs, self._load_directory_handler)

    def _load_directory_handler(self, result: any) -> bool:
        """
        Handles the result from the directory _directory_loader_worker_thread
        """

        if isinstance(result, ExceptionWrapper):
            self._loading_indicator.done(FileTreeModel.DIRECTORY_LOADING_INDICATOR_ERROR_CODE)

            exception = result.get_exception()

            if isinstance(exception, StopIteration):
                return True

            return False

        item = result[0]
        index = result[1]
        upload = result[2]
        self._directory_loader_pathlib_tree = result[3]

        # Set the display hidden to true as we have now loaded every path under the directory
        item.set_loaded(True)
        item.set_data(FileTreeModel.DISPLAY_HIDDEN_INDEX, True)

        if self._directory_loader_pathlib_tree:
            # Update the QMessageBox text
            self._loading_indicator.setText("Adding files to files list...")

            # insert_pathlib_tree will block the UI so process events now to make sure text is updated
            QtCore.QCoreApplication.processEvents()

            self.insert_pathlib_tree(self._directory_loader_pathlib_tree, item, index, upload=upload)

        self._loading_indicator.done(FileTreeModel.DIRECTORY_LOADING_INDICATOR_LOADED_CODE)

        # emitting dataChanged for items with different parents is undefined in qt tree models so just update
        # the whole data structure
        self.dataChanged.emit(QtCore.QModelIndex(), QtCore.QModelIndex())
        self.layoutChanged.emit()

        return True

    """
    Given a file path this method creates TreeNodes for every part of the file path (should the TreeNode not already 
    exist). It does not add every file / directory along the way, only the parts required to add this file. 
    
    Should only be called during the model constructor.
    """
    def insert_file_into_model(self, file_path: pathlib.Path) -> None:
        logger = self.get_plugin().get_logging_coordinator().get_logger(__file__)

        if not file_path.is_file():
            logger.warning("File path \"{}\" is not a file.".format(file_path.as_posix()))
            return

        current_parent_node = self.root_node
        parts = file_path.parts

        last_part_index = len(parts) - 1

        # Iterate through the path parts for the new file and create any TreeNodes which do not already exist
        for i, part in enumerate(parts):

            # Get the pathlib.Path of this part of the path
            path_part = pathlib.Path(parts[0])
            for _part in parts:
                path_part /= pathlib.Path(_part)
                if part == _part:
                    break

            # Check to see if the current parent node already contains a child with this path
            for child_node in current_parent_node.children():

                # If it does then set the parent node to the child and move on
                if child_node.get_path() == path_part:
                    current_parent_node = child_node
                    break

            else:
                # If the path part is not already a Tree node attached to the parent then create it
                new_child_node = TreeNode(self,
                                          path_part,
                                          file_path.stat().st_size if i == last_part_index else 0,
                                          False,
                                          i == last_part_index,
                                          parent_node=current_parent_node)

                insert_position = 0
                insert_index = self.createIndex(insert_position, 0, current_parent_node)
                self.beginInsertRows(insert_index, insert_position, 0)
                current_parent_node.insert_child(insert_position, new_child_node)
                self.endInsertRows()

                current_parent_node = new_child_node

                # Expand the file items parents
                if last_part_index == i:
                    node_to_expand = new_child_node.parent_item()

                    while node_to_expand:
                        self.expand.emit(self.createIndex(node_to_expand.row(), 0, node_to_expand))
                        node_to_expand = node_to_expand.parent_item()

    def _unhide_directory(self,
                          new_child_paths: typing.List[pathlib.Path],
                          directory: 'TreeNode',
                          index: 'QtCore.QModelIndex',
                          upload: bool = False,
                          expand: bool = False):

        old_dir_children = directory.children()
        old_dir_child_paths = list(map(lambda x: x.get_path(), old_dir_children))

        current_range_start = None
        insert_ranges = []

        for i, new_child in enumerate(new_child_paths):
            try:
                # If the new child already exists in the old child array then update it's persistent index
                old_child_index = old_dir_child_paths.index(new_child)
                old_child = old_dir_children[old_child_index]

                if upload:
                    old_child.set_upload(True)

                old_model_index = self.createIndex(old_child_index, 0, old_child)
                new_model_index = self.createIndex(i, 0, old_child)
                self.changePersistentIndex(old_model_index, new_model_index)

                # Find the end of the current insert range
                if current_range_start is not None:
                    insert_ranges.append((current_range_start, i - 1))
                    current_range_start = None

            except ValueError:
                # If the new child is not already in the old child array then set the start of the insert range if not
                # already set.
                if current_range_start is None:
                    current_range_start = i

        if current_range_start is not None:
            insert_ranges.append((current_range_start, len(new_child_paths) - 1))

        for insert_range in insert_ranges:
            self.beginInsertRows(index, insert_range[0], insert_range[1])

            for i in range(insert_range[0], insert_range[1] + 1):
                new_child_path = new_child_paths[i]
                new_tree_node = TreeNode(self,
                                         new_child_path,
                                         new_child_path.stat().st_size if new_child_path.is_file() else 0,
                                         display_hidden=upload,
                                         upload=upload,
                                         parent_node=directory)
                directory.insert_child(i, new_tree_node)

        directory.set_loaded(True)

        if upload:
            directory.set_upload(True)
            directory.set_display_hidden(True)

            # Check all parent nodes to see if their upload flag also needs setting
            parent_node = directory.parent_item()
            while parent_node:
                # Check that the parent is set to display hidden items
                if not parent_node.display_hidden():
                    break

                # Check that all children are marked for upload
                for child in parent_node.children():
                    if not child.upload():
                        should_break = True
                        break
                else:
                    should_break = False

                if should_break:
                    break

                parent_node.set_upload(True)
                parent_node = parent_node.parent_item()

        else:
            # If the directory now contains files not set for upload then set all parents upload flags to False
            if directory.contains_unloaded_dir() or list(filter(lambda x: not x.upload(), directory.get_descendants())):
                parent_node = directory
                while parent_node:
                    parent_node.set_upload(False)
                    parent_node = parent_node.parent_item()
            else:
                directory.set_upload(True)

        # Expand the item
        if expand:
            self.expand.emit(index.siblingAtColumn(0))

    def insert_pathlib_tree(self,
                            pathlib_tree: 'PathlibTreeNode',
                            directory: 'TreeNode',
                            index: 'QtCore.QModelIndex',
                            upload: bool = False) -> None:
        """
        Given a PathlibTreeNode 'pathlib_tree' insert all items which exist in the tree but which are missing in the
        model. Directory is the corresponding model TreeNode for the PathlibTreeNode, index is it's QModelIndex in the
        model and upload is a flag which determines if these new files should be marked for upload.
        """

        # Unhide the directory
        self._unhide_directory(list(map(lambda x: x.get_path(), pathlib_tree.get_children())), directory, index, upload=upload, expand=False)

        # Then recurse over all child directories in the pathlib tree
        for child in pathlib_tree.get_children():
            if child.is_directory():

                # Get the corresponding tree node
                for i, tree_node_child in enumerate(directory.children()):
                    if tree_node_child.get_path().as_posix() == child.get_posix_path():
                        self.insert_pathlib_tree(child, tree_node_child,self.createIndex(i, 0, tree_node_child), upload=upload)

    def _handle_file_upload_changed(self, upload: bool, file_size: str):
        file_size = int(file_size)

        if upload:
            self._upload_count = self._upload_count + 1
            self._upload_total_size = self._upload_total_size + file_size
        else:
            self._upload_count = self._upload_count - 1
            self._upload_total_size = self._upload_total_size - file_size

        self.upload_files_changed.emit()

    def get_upload_count(self) -> int:
        return self._upload_count

    def get_upload_total_size(self) -> int:
        return self._upload_total_size

    def get_auto_detected_paths(self) -> typing.Set[pathlib.Path]:
        return self._auto_detected_paths

    def get_forbidden_files_to_upload(self) -> typing.List[pathlib.Path]:
        forbidden_files_to_upload = []

        descendants = self.root_node.get_descendants()

        while descendants:
            descendant = descendants.pop(0)

            descendant_path = descendant.data(FileTreeModel.PATH_INDEX)
            descendant_upload_flag = descendant.data(FileTreeModel.UPLOAD_INDEX)

            if descendant_path.is_file() and descendant_upload_flag and descendant.is_forbidden():
                forbidden_files_to_upload.append(descendant_path)

        return forbidden_files_to_upload

    def get_files_to_upload(self, include_forbidden: bool = False) -> typing.List[pathlib.Path]:
        files_to_upload = []

        descendants = self.root_node.get_descendants()

        while descendants:
            descendant = descendants.pop(0)

            descendant_path = descendant.data(FileTreeModel.PATH_INDEX)
            descendant_upload_flag = descendant.data(FileTreeModel.UPLOAD_INDEX)

            if descendant_path.is_file() and descendant_upload_flag:

                # ignore forbidden files if required
                if not include_forbidden and descendant.is_forbidden():
                    continue

                files_to_upload.append(descendant_path)

        return files_to_upload

    def set_default_expanded_state(self, tree_view: 'QtWidgets.QTreeView') -> None:
        tree_view.collapseAll()

        root_nodes = self.root_node.get_descendants()

        indexes_to_expand = []

        def add_node_to_expanded_list(node):
            index = self.createIndex(node.row(), 0, node)
            indexes_to_expand.append(index)

        for root_node in root_nodes:

            if root_node.upload():
                # append the node to the expanded nodes list
                add_node_to_expanded_list(root_node)

                # expand all parents as well
                for parent in root_node.parents():
                    if parent:
                        add_node_to_expanded_list(parent)

        for index in indexes_to_expand:
            tree_view.expand(index)
