__all__ = "WatchPatternTable"

from PySide2 import QtCore
import typing

from submit2gm import constants, utils
from submit2gm.watch_pattern import WatchPattern

if typing.TYPE_CHECKING:
    from submit2gm.packed_project import PackedProject
    from submit2gm.dynamic_job_preset import DynamicJobPreset


class WatchPatternModelHeader:
    def __init__(self, display_text: str, tool_tip: str):
        self._display_text = display_text
        self._tool_tip = tool_tip

    def get_display_text(self) -> str:
        return self._display_text

    def get_tool_tip(self) -> str:
        return self._tool_tip


class WatchPatternTable(QtCore.QAbstractTableModel):

    PATTERN_HEADER = WatchPatternModelHeader("Watch Pattern",
                                             "The regex pattern that will be used to match files created by your job.")
    PATTERN_INDEX = 0

    DOWNLOAD_PATH_HEADER = WatchPatternModelHeader("Download Path",
                                                   "The local path that the matched files should be downloaded to.")
    DOWNLOAD_PATH_INDEX = 1

    WatchPatternModelHeaders = [PATTERN_HEADER, DOWNLOAD_PATH_HEADER]

    def __init__(self,
                 packed_project: 'PackedProject',
                 dynamic_job_presets: typing.List['DynamicJobPreset'],
                 parent: typing.Optional['QtCore.QObject'] = None):

        QtCore.QAbstractTableModel.__init__(self, parent=parent)

        self._packed_project = packed_project
        self._dynamic_job_presets = dynamic_job_presets

        # Serialise and parse watch patterns so that we do not modify the original list
        self._watch_patterns = WatchPattern.parse_watch_patterns(WatchPattern.serialise_watch_patterns(packed_project.get_watch_patterns()))

    def headerData(self,
                   section: int,
                   orientation: 'QtCore.Qt.Orientation',
                   role: 'QtCore.Qt.ItemDataRole' = QtCore.Qt.DisplayRole) -> str:

        if orientation == QtCore.Qt.Horizontal:
            header = WatchPatternTable.WatchPatternModelHeaders[section]

            if role == QtCore.Qt.DisplayRole:
                return header.get_display_text()

            if role == QtCore.Qt.ToolTipRole or role == QtCore.Qt.WhatsThisRole:
                return header.get_tool_tip()

        return QtCore.QAbstractTableModel.headerData(self, section, orientation, role=role)

    def rowCount(self, parent: 'QtCore.Qt.QModelIndex' = QtCore.QModelIndex()) -> int:

        #if parent.isValid():
        #    return parent.internalPointer().column_count()

        return len(self._watch_patterns)

    def columnCount(self, parent: 'QtCore.Qt.QModelIndex' = QtCore.QModelIndex()) -> int:

        #if parent.isValid():
        #    return parent.internalPointer().column_count()

        return len(WatchPatternTable.WatchPatternModelHeaders)

    def get_watch_patterns(self) -> typing.List['WatchPattern']:
        return self._watch_patterns

    def data(self, index: 'QtCore.QModelIndex', role: int = QtCore.Qt.DisplayRole) -> any:
        if role == QtCore.Qt.DisplayRole:

            row = index.row()
            column = index.column()


            if column < 2:
                watch_file = self._watch_patterns[row]

                if column == 0:
                    return watch_file.get_remote_watch_pattern()

                if column == 1:
                    return watch_file.get_local_download_path()

    def add_row(self) -> None:
        self._watch_patterns.append(WatchPattern("/{}/.+".format(self._packed_project.get_name()),
                                                 utils.get_desktop_path().replace('\\', '/')))
        self.layoutChanged.emit()

    def removeRows(self, position: int, rows: int = 1, index: 'QtCore.QModelIndex' = QtCore.QModelIndex()) -> bool:
        self.beginRemoveRows(QtCore.QModelIndex(), position, position + rows - 1)

        i = 0
        while i < rows:
            del self._watch_patterns[position]
            i = i + 1

        self.endRemoveRows()

        return True

    def setData(self, modelIndex: 'QtCore.QModelIndex', value: any, role: int = QtCore.Qt.EditRole) -> bool:
        if not modelIndex.isValid():
            return False

        row = modelIndex.row()
        column = modelIndex.column()

        watch_pattern = self._watch_patterns[row]
        if column == WatchPatternTable.PATTERN_INDEX:
            watch_pattern.set_remote_watch_pattern(value)

        if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
            watch_pattern.set_local_download_path(value.replace('\\', '/'))

        return True

    def flags(self, index: 'QtCore.QModelIndex') -> 'QtCore.Qt.ItemFlags':
        if not index.isValid():
            return QtCore.Qt.NoItemFlags

        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable
