__all__ = 'SignInWidget'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets
import typing

from submit2gm import constants, qt_utils, ui_constants
from submit2gm.exception_wrapper import ExceptionWrapper
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.qt_none import QtNone
from submit2gm.user import User
from submit2gm.worker_thread import WorkerThread
from submit2gm.errors.invalid_email_error import InvalidEmailError
from submit2gm.errors.not_signed_in_error import NotSignedInError
from submit2gm.errors.invalid_user_error import InvalidUserError
from submit2gm.errors.api_client_error import APIClientError
from submit2gm.errors.command_port_server_error import CommandPortServerError

from .input_widgets.line_edit import LineEdit
from .validated_widget import ValidatedWidget

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin
    from submit2gm.logger import Logger


class SignInWidget(QtWidgets.QWidget, PluginAccessor):
    WINDOW_WIDTH = 600
    WINDOW_HEIGHT = 350
    WINDOW_MARGINS = (0, 0, 0, 0)

    MARGIN_LEFT = 20
    MARGIN_TOP = 15
    MARGIN_RIGHT = 20
    MARGIN_BOTTOM = 20

    CONTENT_WIDTH = WINDOW_WIDTH - (MARGIN_LEFT + MARGIN_RIGHT)

    SIGN_IN_TEXT = u"Sign-in"
    SIGNING_IN_TEXT = u"Signing in..."

    SIGN_IN_ENVOY_TEXT = u"Sign-in with Envoy"
    SIGNING_IN_ENVOY_TEXT = u"Signing in with Envoy..."

    SIGN_IN_USING_ENVOY = u"Sign-in using Envoy"
    SIGN_IN_USING_CREDENTIALS = u"Sign-in using your Email and Access Key"

    CLOSE_BUTTON_WIDTH = 20
    PLUGIN_TITLE_IMAGE_WIDTH = 200

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtWidgets.QWidget'] = None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        PluginAccessor.__init__(self, plugin)

        # used for storing the position of the mouse for custom move events
        self._current_mouse_position = None
        self._worker_thread = None
        self.setProperty(constants.STYLE_PROPS.MAIN_CONTAINER_WIDGET, True)

        """ Create Widgets """
        # title image
        self.title_image = QtWidgets.QLabel()
        self.title_image.setPixmap(ui_constants.GRIDMARKETS_IMAGE.scaledToWidth(150, QtCore.Qt.SmoothTransformation))
        self.title_image.setAlignment(QtCore.Qt.AlignCenter)

        # close window button
        self.close_window_btn = QtWidgets.QPushButton(u"X")
        self.close_window_btn.setFlat(True)
        self.close_window_btn.setFixedWidth(SignInWidget.CLOSE_BUTTON_WIDTH)
        self.close_window_btn.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.close_window_btn.clicked.connect(self.close_window_slot)

        # the loading bar shown when the user is signing in
        self.loading_bar_movie = QtGui.QMovie(constants.LOADING_BAR_GIF, parent=self)
        self.loading_bar_movie.setScaledSize(QtCore.QSize(75, 20))
        self.loading_bar_movie.setCacheMode(QtGui.QMovie.CacheAll)
        self.loading_bar = QtWidgets.QLabel("")
        self.loading_bar.setMovie(self.loading_bar_movie)

        ''' Standard Sign in screen widgets'''
        self.standard_sign_in_widget = QtWidgets.QWidget()

        # email input
        self.auth_email_input = ValidatedWidget(LineEdit(plugin))
        self.auth_email_input.get_widget().setPlaceholderText(u"GridMarkets Email")

        # access key input
        self.auth_access_key_input = ValidatedWidget(LineEdit(plugin))
        self.auth_access_key_input.get_widget().setPlaceholderText(u"GridMarkets Access Key")
        self.auth_access_key_input.get_widget().setEchoMode(QtWidgets.QLineEdit.Password)

        # sign in button
        self.sign_in_btn = ValidatedWidget(QtWidgets.QPushButton(SignInWidget.SIGN_IN_TEXT))
        self.sign_in_btn.align_warning_message_center(True)
        self.sign_in_btn.get_widget().setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.sign_in_btn.get_widget().clicked.connect(self.sign_in_slot)

        # lost your password button
        self.lost_password_link = QtWidgets.QPushButton(u"Lost your password?")
        self.lost_password_link.setFlat(True)
        self.lost_password_link.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.lost_password_link.clicked.connect(
            lambda: qt_utils.open_url(constants.FORGOT_PASSWORD_URL))

        ''' Envoy sign in screen widgets'''
        self.envoy_sign_in_widget = QtWidgets.QWidget()
        self.envoy_sign_in_widget.setFixedWidth(220)
        self.sign_in_envoy_btn = ValidatedWidget(QtWidgets.QPushButton(self.SIGN_IN_ENVOY_TEXT))
        self.sign_in_envoy_btn.align_warning_message_center(True)
        self.sign_in_envoy_btn.get_widget().setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.sign_in_envoy_btn.get_widget().clicked.connect(self.sign_in_envoy_slot)
        self.sign_in_envoy_btn.get_widget().setFocus()
        self.sign_in_envoy_btn.get_widget().setIcon(ui_constants.GM_ICON_WHITE)

        # register
        self.register_label = QtWidgets.QLabel(u"Don't have Envoy installed?")
        self.register_link = QtWidgets.QPushButton(u"See here")
        self.register_link.setFlat(True)
        self.register_link.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.register_link.clicked.connect(lambda: qt_utils.open_url(constants.ENVOY_URL))
        self.register_link.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        """ Create Layouts """
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(self.MARGIN_LEFT, self.MARGIN_TOP, self.MARGIN_RIGHT, self.MARGIN_BOTTOM)

        self.loading_bar_layout = QtWidgets.QHBoxLayout()
        self.loading_bar_layout.setAlignment(QtCore.Qt.AlignHCenter)
        self.loading_bar_layout.addWidget(self.loading_bar)

        self.envoy_sign_in_layout = QtWidgets.QHBoxLayout()
        self.envoy_sign_in_layout.setContentsMargins(0, 0, 0, 0)
        self.envoy_sign_in_layout.setAlignment(QtCore.Qt.AlignVCenter)
        # self.envoy_sign_in_layout.addStretch()
        self.envoy_sign_in_layout.addWidget(self.sign_in_envoy_btn)
        # self.envoy_sign_in_layout.addStretch()
        self.envoy_sign_in_widget.setLayout(self.envoy_sign_in_layout)

        self.sign_in_container_layout = QtWidgets.QVBoxLayout()
        self.sign_in_container_layout.setAlignment(QtCore.Qt.AlignVCenter)
        self.sign_in_container_layout.setAlignment(QtCore.Qt.AlignHCenter)
        self.sign_in_container_layout.addSpacing(40)
        self.sign_in_container_layout.addWidget(self.envoy_sign_in_widget)

        self.register_layout = QtWidgets.QHBoxLayout()
        self.register_layout.setAlignment(QtCore.Qt.AlignHCenter)
        self.register_layout.setSpacing(0)
        self.register_layout.addWidget(self.register_label)
        self.register_layout.addWidget(self.register_link)

        self.top_corner_layout = QtWidgets.QVBoxLayout()
        self.top_corner_layout.setContentsMargins(0, 0, 0, 0)
        self.top_corner_layout.addWidget(self.close_window_btn)
        self.top_corner_layout.addSpacing(20)

        self.title_layout = QtWidgets.QHBoxLayout()
        self.title_layout.setAlignment(QtCore.Qt.AlignTop)
        self.title_layout.addSpacing(20)
        self.title_layout.addWidget(self.title_image)
        self.title_layout.addLayout(self.top_corner_layout)

        # self.layout.addWidget(self.title_image)
        self.layout.addLayout(self.title_layout)
        self.layout.addLayout(self.sign_in_container_layout)
        self.layout.addLayout(self.loading_bar_layout)
        self.layout.addLayout(self.register_layout)

        """ Create Connections """
        user_interface = plugin.get_user_interface()
        user_interface.operation_started.connect(self.show_status_message)
        user_interface.operation_finished.connect(self.hide_status_message)

    def get_title_image(self) -> 'QtGui.QPixmap':
        return ui_constants.GRIDMARKETS_IMAGE.scaledToWidth(self.layout.width(), QtCore.Qt.SmoothTransformation)

    def _reset_input_validity(self) -> None:
        """ Resets all the error messages from the sign in window """

        user_interface = self.get_plugin().get_user_interface()
        user_interface.set_auth_email_validity_flag(True)
        user_interface.set_auth_access_key_validity_flag(True)
        user_interface.set_user_validity_flag(True)
        user_interface.set_auth_email_validity_message("")
        user_interface.set_auth_access_key_validity_message("")
        user_interface.set_user_validity_message("")

    def sign_in_slot(self) -> None:
        """ Sign in using the email and access key provided by the user """

        # get the email
        auth_email = self.auth_email_input.get_widget().text()

        # create a user instance with the above credentials
        user = User(auth_email)

        self._worker_thread = WorkerThread(self.get_plugin())
        self._worker_thread.started.connect(self.show_status_message)
        self._worker_thread.finished.connect(self.hide_status_message)

        # set all inputs to valid while signing in
        self._reset_input_validity()

        ################################################################################################################

        target = self.get_plugin().get_api_client().sign_in

        args = (
            user,
        )

        kwargs = {

        }

        self._worker_thread.process(target, args, kwargs, self.sign_in_result_handler)

    def sign_in_envoy_slot(self) -> None:
        user_interface = self.get_plugin().get_user_interface()
        self._worker_thread = user_interface.get_worker_thread(SignInWidget.SIGNING_IN_ENVOY_TEXT)

        # disconnect finished signal so that the operation finished signal is never emitted (so that the sign in
        # button doesn't flicker as it enables and instantly disables again due to the 'create job preset' operation)
        self._worker_thread.finished.disconnect()

        # set all inputs to valid while signing in
        self._reset_input_validity()

        ################################################################################################################

        target = self.get_plugin().get_api_client().sign_in_with_envoy

        args = set()

        kwargs = {

        }

        self._worker_thread.process(target, args, kwargs, self.sign_in_result_handler)

    def sign_in_result_handler(self, result: any) -> bool:
        """ Handel the result of the sign in operation """

        plugin = self.get_plugin()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        user_interface = plugin.get_user_interface()
        main_window = user_interface.get_main_window()

        # None is returned when the sign in operation was successful
        if isinstance(result, QtNone):

            user_info = plugin.get_api_client().get_user_info()
            plugin_versions = user_info.get_plugin_versions()

            version = list(plugin.get_version().to_string().split('.'))

            # Check that submit2gm is in the versions list
            if not ('submit2gm' in plugin_versions):
                self._log_error(logger, "Your account is incorrectly setup, please contact support.")

            min_version = plugin_versions['submit2gm']['minimum'].split('.')
            latest_version_label = plugin_versions['submit2gm']['latest']
            latest_version = latest_version_label.split('.')

            # Prevent sign in if not using a supported Submit2gm version
            envoy_update_msg = "Please restart Envoy to trigger an automatic update."
            if version < min_version:
                self._log_error(logger, "You are using a deprecated version of Submit2gm ({})\n\n{}".format(
                    plugin.get_version().to_string(), envoy_update_msg))
            else:

                # Warn if not using the latest Submit2gm version
                if version < latest_version:
                    self._log_error(logger, "A newer version of Submit2gm is available ({} -> {}).\n\n{}".format(
                        plugin.get_version().to_string(), latest_version_label, envoy_update_msg))

                try:
                    plugin.create_default_job_presets()
                    return True

                except Exception as e:

                    err_msg = "Exception during sign in: " + str(e)
                    traceback = e.__traceback__
                    while traceback:
                        err_msg = err_msg + ("\n    {}: {}".format(traceback.tb_frame.f_code.co_filename,
                                                                   traceback.tb_lineno))
                        traceback = traceback.tb_next
                    logger.error(err_msg)
                    print(err_msg)

            # Cleanup and display the signed out ui again
            plugin.get_api_client().sign_out()
            user_interface.hide_running_operation_message()
            main_window.display_signed_out_ui()
            main_window.show()
            return True

        # If the response was not QtNOne then sign in failed
        user_interface.hide_running_operation_message()

        # handle expected exceptions
        if isinstance(result, ExceptionWrapper):
            exception = result.get_exception()

            plugin.get_api_client().sign_out()
            main_window.display_signed_out_ui()
            main_window.show()

            if isinstance(exception, InvalidEmailError):
                self._log_error(logger, "Unable to Sign in\n" + exception.user_message)
                return True

            elif isinstance(exception, NotSignedInError):
                self._log_error(logger, exception.user_message)
                return True

            elif isinstance(exception, InvalidUserError):
                self._log_error(logger, "Unable to Sign in\n" + exception.user_message)
                return True

            elif isinstance(exception, APIClientError):
                self._log_error(logger, "Unable to Sign in\n" + exception.user_message)
                return True

            if isinstance(exception, CommandPortServerError):
                self._log_error(logger, "Unable to Sign in\n" + exception.user_message)
                return True

        return False

    def _log_error(self, logger: 'Logger', message: str) -> None:
        qt_utils.log_error(self, logger, message)

    def show_status_message(self) -> None:
        # play the loading gif
        self.loading_bar_movie.start()

        # disable the email input
        self.auth_email_input.get_widget().setEnabled(False)

        # disable the access key input
        self.auth_access_key_input.get_widget().setEnabled(False)

        # disable the sign in buttons
        self.sign_in_envoy_btn.get_widget().setEnabled(False)

        # update the text in the sign in buttons to reflect what is going on
        message = self.get_plugin().get_user_interface().get_running_operation_message()
        self.sign_in_envoy_btn.get_widget().setText(message)

    def hide_status_message(self) -> None:
        # stop and rewind the loading gif
        self.loading_bar_movie.stop()
        self.loading_bar_movie.jumpToFrame(0)

        # enable the email input
        self.auth_email_input.get_widget().setEnabled(True)

        # enable the access key input
        self.auth_access_key_input.get_widget().setEnabled(True)

        # enable the sign in buttons
        self.sign_in_envoy_btn.get_widget().setEnabled(True)

        # reset the sign in button's text labels
        self.sign_in_envoy_btn.get_widget().setText(SignInWidget.SIGN_IN_ENVOY_TEXT)

    def close_window_slot(self) -> None:
        """ Called when the user presses the close window button 'X' """
        main_window = self.get_plugin().get_user_interface().get_main_window()
        main_window.close()

    def mousePressEvent(self, event: 'QtGui.QMouseEvent'):
        """ Cache the position of the mouse when first pressed """
        self._current_mouse_position = event.globalPos()

    def mouseMoveEvent(self, event: 'QtGui.QMouseEvent') -> None:
        """ Move the window with the user's mouse when the user clicks and drags on any part of this widget """

        if self._current_mouse_position is not None:
            main_window = self.get_plugin().get_user_interface().get_main_window()

            # calculate the delta between the current position of the mouse and the mouses previous position
            delta = QtCore.QPoint(event.globalPos() - self._current_mouse_position)

            # move the window to be under the mouse
            main_window.move(main_window.x() + delta.x(), main_window.y() + delta.y())

            # cache the new position of the mouse
            self._current_mouse_position = event.globalPos()
