__all__ = 'SubmissionViewWidget'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets
import time
import typing

#import gm_meta_plugin
#from gm_meta_plugin.gridmarkets import constants as api_constants
#from gm_meta_plugin.errors import *


from .elided_label import ElidedLabel
from .qhline import QHLine
from .resizable_stacked_widget import ResizableStackedWidget
from .right_aligned_qmenu import RightAlignedQMenu
from .qt_dialog_preflight_window import PreflightWindow

from submit2gm import constants, qt_utils, ui_constants
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.exception_wrapper import ExceptionWrapper
from submit2gm.inference_source import InferenceSource
from submit2gm.qt_none import QtNone
from submit2gm.packed_project import PackedProject
from submit2gm.errors.application_attribute_not_found import ApplicationAttributeNotFound
from submit2gm.job_preset import JobPreset
from submit2gm.watch_pattern import WatchPattern
from submit2gm.errors.application_attribute_not_found import ApplicationAttributeNotFound
from submit2gm.errors.command_port_server_error import CommandPortServerError
from submit2gm.errors.project_requires_saving_error import ProjectRequiresSavingError
from submit2gm.errors.file_packing_error import FilePackingError
from submit2gm.errors.not_signed_in_error import NotSignedInError
from submit2gm.errors.credits_error import CreditsError
from submit2gm.errors.api_client_error import APIClientError
from submit2gm.errors.unsupported_application_error import UnsupportedApplicationError
from submit2gm.widgets import input_widgets
from submit2gm.widgets.dynamic_job_preset_tab_widget import DynamicJobPresetTabWidget
from submit2gm.widgets.node_editor.node_widget import NodeWidget
from submit2gm.widgets.node_editor.dynamic_job_preset_node import DynamicJobPresetNode


if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import Submit2gmPlugin
    from submit2gm.dynamic_job_preset import DynamicJobPreset
    from submit2gm.widgets.node_editor.node_widget import DependencyNode


class SubmissionViewWidget(QtWidgets.QWidget, PluginAccessor):

    resized = QtCore.Signal()

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional[QtWidgets.QWidget] = None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        PluginAccessor.__init__(self, plugin)

        def display_job_preset_container():
            main_window = self.get_plugin().get_user_interface().get_main_window()
            main_window.get_main_ui_widget().get_main_ui_switcher().display_job_preset_container()

        """ Create Widgets """
        self.no_job_presets_label = ElidedLabel("No jobs to view.\nPlease create a new job node in the dependency view..")
        self.no_job_presets_label.setAlignment(QtCore.Qt.AlignCenter)
        self.no_job_preset_widget = QtWidgets.QWidget()
        self.no_job_preset_widget.setProperty(constants.STYLE_PROPS.BACKGROUND_LEVEL_1, True)

        self.group_box_2 = QtWidgets.QGroupBox()

        # If connected to a cps then set the project and submission name
        try:
            application_inference_source = plugin.get_application_attribute_source()
            project_name = application_inference_source.get_attribute_value(None, None, constants.API_KEYS.PROJECT_NAME)
            submission_name = application_inference_source.get_attribute_value(None, None, constants.API_KEYS.SUBMISSION_NAME)
        except (ValueError, ApplicationAttributeNotFound):
            project_name = ''
            submission_name = ''

        self.project_input_label = QtWidgets.QLabel('Project Name')
        self.project_input_label.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.project_input_label.setMinimumWidth(140)
        self.project_input_label.setProperty('h5', True)

        self.project_input_widget = input_widgets.ProjectInput(plugin)
        #self.project_input_widget.setMaximumWidth(600)
        self.project_input_widget.set_placeholder_text('Project Name')
        self.project_input_widget.setToolTip(qt_utils.split_tool_tip(constants.PROJECT_NAME_DESCRIPTION))
        self.project_input_widget.hide()

        self._on_remote_projects_changed_handler()
        plugin.get_api_client().onRemoteProjectsChanged.connect(self._on_remote_projects_changed_handler)

        self.project_infer_widget = input_widgets.Constant(plugin, project_name)
        self.project_infer_widget.setToolTip(qt_utils.split_tool_tip(constants.PROJECT_NAME_DESCRIPTION))
        self._project_input_override_widget = input_widgets.CheckBox(self.get_plugin())
        self._project_input_override_widget.setProperty(constants.STYLE_PROPS.APP_INFER_CHECKBOX, True)
        self._project_input_override_widget.setToolTip(qt_utils.split_tool_tip(constants.APPLICATION_LINKING_TOOL_TIP))

        self.submission_name_input_label = QtWidgets.QLabel('Submission Name')
        self.submission_name_input_label.setMinimumWidth(140)
        self.submission_name_input_label.setProperty('h5', True)
        self.submission_name_input_label.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.submission_name_input_widget = input_widgets.LineEdit(plugin)
        #self.submission_name_input_widget.setMaximumWidth(600)
        self.submission_name_input_widget.setPlaceholderText('Submission Name')
        self.submission_name_input_widget.setToolTip(qt_utils.split_tool_tip(constants.SUBMISSION_NAME_DESCRIPTION))
        self.submission_name_input_widget.hide()
        self.submission_name_infer_widget = input_widgets.Constant(plugin, submission_name)
        self.submission_name_infer_widget.setToolTip(qt_utils.split_tool_tip(constants.SUBMISSION_NAME_DESCRIPTION))

        self._submission_name_override_widget = input_widgets.CheckBox(self.get_plugin())
        self._submission_name_override_widget.setProperty(constants.STYLE_PROPS.APP_INFER_CHECKBOX, True)
        self._submission_name_override_widget.setToolTip(qt_utils.split_tool_tip(constants.APPLICATION_LINKING_TOOL_TIP))

        self.submission_comment_input_label = QtWidgets.QLabel('Submission Comment')
        self.submission_comment_input_label.setMinimumWidth(140)
        self.submission_comment_input_label.setProperty('h5', True)
        self.submission_comment_input_label.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.submission_comment_input_widget = input_widgets.LineEdit(plugin)
        self.submission_comment_input_widget.setPlaceholderText('Submission Comment')
        self.submission_comment_input_widget.setToolTip(qt_utils.split_tool_tip(constants.SUBMISSION_COMMENT_DESCRIPTION))

        self._spacer_item = QtWidgets.QSpacerItem(8, 8)

        self.reload_scene_settings_btn = QtWidgets.QPushButton(u"Reload Scene Settings")
        self.reload_scene_settings_btn.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.reload_scene_settings_btn.clicked.connect(self.reload_scene_settings)
        self.reload_movie = QtGui.QMovie(str(constants.SPINNER_WHITE_666666_GIF_PATH))

        self.tab_widget = QtWidgets.QTabWidget()
        self._node_widget = NodeWidget(self.get_plugin())
        self._node_view_btn = QtWidgets.QPushButton(ui_constants.DEPENDENCY_VIEW_ICON_BRAND, '')
        self._node_view_btn.setProperty(constants.STYLE_PROPS.TAB_BTN, True)

        self._toggle_node_view_action = QtWidgets.QAction('Open dependency view')
        self._node_view_menu = RightAlignedQMenu('', parent=self._node_view_btn)
        self._node_view_menu.addAction(self._toggle_node_view_action)
        self._node_view_btn.setMenu(self._node_view_menu)

        self.tab_widget.setCornerWidget(self._node_view_btn)

        self._stacked_widget = ResizableStackedWidget()
        self._stacked_widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._stacked_widget.addWidget(self.tab_widget)
        self._stacked_widget.addWidget(self.no_job_preset_widget)

        # Handle dynamic job presets that already exist
        for key, dynamic_job_preset in self.get_plugin().get_job_preset_container().items():
            self._on_dynamic_job_preset_added(key, dynamic_job_preset)

        self._splitter_old_pos = [700, 0]
        self._splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        self._splitter.setOrientation(QtCore.Qt.Horizontal)
        self._splitter.addWidget(self._stacked_widget)
        self._splitter.addWidget(self._node_widget)
        self._splitter.setSizes(self._splitter_old_pos)
        self._splitter.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        self.submit_btn = QtWidgets.QPushButton(u"Submit")
        self.submit_btn.setProperty(constants.STYLE_PROPS.BRAND, True)
        self.submit_btn.setMinimumWidth(300)
        self.submit_btn.clicked.connect(self._submit)

        user_interface = plugin.get_user_interface()
        user_interface.operation_started.connect(lambda: self.reload_scene_settings_btn.setDisabled(True))
        user_interface.operation_finished.connect(lambda: self.reload_scene_settings_btn.setEnabled(True))
        user_interface.operation_started.connect(lambda: self.submit_btn.setDisabled(True))
        user_interface.operation_finished.connect(lambda: self.submit_btn.setEnabled(True))

        """ Create Layouts """
        self.job_preset_warning_layout = QtWidgets.QHBoxLayout()
        self.job_preset_warning_layout.setContentsMargins(0, 0, 0, 0)
        self.job_preset_warning_layout.addWidget(self.no_job_presets_label)
        self.no_job_preset_widget.setLayout(self.job_preset_warning_layout)

        vboxlayout = QtWidgets.QVBoxLayout()
        vboxlayout.setContentsMargins(8, 4, 0, 0)
        vboxlayout.addWidget(self._project_input_override_widget)

        self.project_input_layout = QtWidgets.QHBoxLayout()
        self.project_input_layout.setContentsMargins(0, 0, 0, 0)
        self.project_input_layout.setAlignment(QtCore.Qt.AlignRight)
        self.project_input_layout.addWidget(self.project_input_label)
        self.project_input_layout.addWidget(self.project_input_widget)
        self.project_input_layout.addWidget(self.project_infer_widget)
        self.project_input_layout.addLayout(vboxlayout)

        vboxlayout = QtWidgets.QVBoxLayout()
        vboxlayout.setContentsMargins(8, 4, 0, 0)
        vboxlayout.addWidget(self._submission_name_override_widget)

        self.submission_name_layout = QtWidgets.QHBoxLayout()
        self.submission_name_layout.setContentsMargins(0, 0, 0, 0)
        self.submission_name_layout.setAlignment(QtCore.Qt.AlignRight)
        self.submission_name_layout.addWidget(self.submission_name_input_label)
        self.submission_name_layout.addWidget(self.submission_name_input_widget)
        self.submission_name_layout.addWidget(self.submission_name_infer_widget)
        self.submission_name_layout.addLayout(vboxlayout)

        vboxlayout = QtWidgets.QVBoxLayout()
        vboxlayout.setContentsMargins(8, 4, 0, 0)

        self.submission_comment_layout = QtWidgets.QHBoxLayout()
        self.submission_comment_layout.setContentsMargins(0, 0, 31, 0)
        self.submission_comment_layout.setAlignment(QtCore.Qt.AlignRight)
        self.submission_comment_layout.addWidget(self.submission_comment_input_label)
        self.submission_comment_layout.addWidget(self.submission_comment_input_widget)
        self.submission_comment_layout.addLayout(vboxlayout)

        self._old_width = None
        self.resized.connect(self.update_submission_grid_layout_columns)

        self.submission_grid_layout = QtWidgets.QGridLayout()
        #self.submission_grid_layout.setSpacing(0)
        self.update_submission_grid_layout_columns()

        self.group_box_2.setLayout(self.submission_grid_layout)

        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        # hide project / job preset selection for v1
        # self.layout.addWidget(self.group_box)

        self.layout.addWidget(self._splitter)
        self.layout.addSpacing(8)
        self.layout.addWidget(self.group_box_2)

        """ Create connections """
        self._project_input_override_widget.input_value_changed.connect(self._project_input_override_widget_value_changed)
        self._submission_name_override_widget.input_value_changed.connect(self._submission_name_input_override_widget_value_changed)

        self.project_infer_widget.clicked.connect(self._project_override_input_clicked)
        self.submission_name_infer_widget.clicked.connect(self._submission_name_override_input_clicked)

        self.tab_widget.tabBarClicked.connect(self._handle_tab_bar_clicked)
        #self._node_view_btn.clicked.connect(self._handle_node_btn)
        self._toggle_node_view_action.triggered.connect(self._toggle_node_view)
        self._splitter.splitterMoved.connect(self._handle_splitter_moved)

        job_preset_container = plugin.get_job_preset_container()
        job_preset_container.itemAdded.connect(self._on_dynamic_job_preset_added)
        job_preset_container.itemRemoved.connect(self._on_dynamic_job_preset_removed)

        self.on_active_cps_changed()

    @QtCore.Slot()
    def on_active_cps_changed(self):
        try:
            self.get_plugin().get_active_command_port_server()
        except ValueError:
            self._project_input_override_widget.set_value(True)
            self._project_input_override_widget.hide()
            self._submission_name_override_widget.set_value(True)
            self._submission_name_override_widget.hide()
            self.submission_comment_layout.setContentsMargins(0, 0, 0, 0)

    @QtCore.Slot(int)
    def _handle_tab_bar_clicked(self, index: int) -> None:
        if index >= 0:
            dynamic_job_preset_tab_widget = self.tab_widget.widget(index)
            if isinstance(dynamic_job_preset_tab_widget, DynamicJobPresetTabWidget):
                dynamic_job_preset_tab_widget.get_dynamic_job_preset().selected.emit()

    def _handle_splitter_moved(self) -> None:
        was_open = self._splitter_old_pos[1] > 0
        is_open = self._splitter.sizes()[1] > 0
        if was_open and not is_open:
            self._toggle_node_view_action.setText('Open dependency view')
        elif not was_open and is_open:
            self._toggle_node_view_action.setText('Close dependency view')

        self._splitter_old_pos = self._splitter.sizes()

    def _toggle_node_view(self) -> None:
        if self._splitter.sizes()[1] > 0:
            self._splitter.setSizes([700, 0])
        else:
            self._splitter.setSizes([350, 350])
        self._handle_splitter_moved()

    def _on_dynamic_job_preset_added(self, key: str, dynamic_job_preset: 'DynamicJobPreset') -> None:
        self._node_widget.add_node(DynamicJobPresetNode(dynamic_job_preset))
        dynamic_job_preset_tab_widget = DynamicJobPresetTabWidget(self.get_plugin(), dynamic_job_preset)

        def _handle_dynamic_job_preset_selected() -> None:
            index = self.tab_widget.indexOf(dynamic_job_preset_tab_widget)
            if index >= 0:
                self.tab_widget.setCurrentWidget(dynamic_job_preset_tab_widget)

        def _handle_job_name_changed():
            index = self.tab_widget.indexOf(dynamic_job_preset_tab_widget)
            if index >= 0:
                new_job_name = dynamic_job_preset.get_job_preset_attribute_with_job_attribute_id(constants.API_KEYS.JOB_NAME).get_value()
                self.tab_widget.setTabText(index, new_job_name)

        self.tab_widget.addTab(dynamic_job_preset_tab_widget, '')
        _handle_job_name_changed() # set the tab name

        dynamic_job_preset.selected.connect(_handle_dynamic_job_preset_selected)
        dynamic_job_preset.job_name_changed.connect(_handle_job_name_changed)

        if self._stacked_widget.currentWidget() != self.tab_widget:
            self._stacked_widget.setCurrentWidget(self.tab_widget)

    def _on_dynamic_job_preset_removed(self, key: str, dynamic_job_preset: 'DynamicJobPreset') -> None:
        nodes = self._node_widget.get_nodes()

        nodes_to_remove = []
        for node in nodes:
            if isinstance(node, DynamicJobPresetNode):
                if node.get_dynamic_job_preset() is dynamic_job_preset:
                    nodes_to_remove.append(node)

        while nodes_to_remove:
            node = nodes_to_remove.pop()
            self._node_widget.remove_node(node)

        for i in range(self.tab_widget.count() - 1, -1, -1):
            widget = self.tab_widget.widget(i)
            if isinstance(widget, DynamicJobPresetTabWidget):
                if widget.get_dynamic_job_preset() is dynamic_job_preset:
                    self.tab_widget.removeTab(i)

        job_preset_count = self.get_plugin().get_job_preset_container().count()
        if job_preset_count <= 0 and self._stacked_widget.currentWidget() != self.no_job_preset_widget:
            self._stacked_widget.setCurrentWidget(self.no_job_preset_widget)

    def _on_remote_projects_changed_handler(self):
        plugin = self.get_plugin()
        word_list = plugin.get_api_client().get_root_directories()
        completer = QtWidgets.QCompleter(word_list, self.project_input_widget)
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        plugin.get_user_interface().set_style_sheet(completer.popup())
        self.project_input_widget.setCompleter(completer)

    def get_project_name(self) -> str:
        if self._project_input_override_widget.isChecked():
            return self.project_input_widget.get_value()
        return self.project_infer_widget.get_value()

    def get_submission_name(self) -> str:
        if self._submission_name_override_widget.isChecked():
            return self.submission_name_input_widget.get_value()
        return self.submission_name_infer_widget.get_value()

    def get_submission_comment(self) -> str:
        return self.submission_comment_input_widget.get_value()

    def update_project_name(self) -> None:
        if not self._project_input_override_widget.isChecked():
            plugin = self.get_plugin()
            application_inference_source = plugin.get_application_attribute_source()
            user_interface = plugin.get_user_interface()

            product = plugin.get_active_command_port_server().get_product()

            old_value = self.project_infer_widget.get_value()
            new_value = application_inference_source.get_attribute_value(product, None, constants.API_KEYS.PROJECT_NAME)

            if new_value != old_value:
                theme = user_interface.get_theme()
                original_color = QtGui.QColor(theme.FOREGROUND_COLOR)
                flash_color = QtGui.QColor(theme.BRAND_COLOR)
                qt_utils.flash_widget_color(self.project_infer_widget, original_color, flash_color)
                self.project_infer_widget.set_value(new_value)

    def update_submission_name(self) -> None:
        if not self._submission_name_override_widget.isChecked():
            plugin = self.get_plugin()
            application_inference_source = plugin.get_application_attribute_source()
            user_interface = plugin.get_user_interface()

            product = plugin.get_active_command_port_server().get_product()

            old_value = self.submission_name_infer_widget.get_value()
            new_value = application_inference_source.get_attribute_value(product, None, constants.API_KEYS.SUBMISSION_NAME)

            if new_value != old_value:
                theme = user_interface.get_theme()
                original_color = QtGui.QColor(theme.FOREGROUND_COLOR)
                flash_color = QtGui.QColor(theme.BRAND_COLOR)
                qt_utils.flash_widget_color(self.submission_name_infer_widget, original_color, flash_color)
                self.submission_name_infer_widget.set_value(new_value)

    def _project_input_override_widget_value_changed(self):
        if self._project_input_override_widget.isChecked():
            self.project_infer_widget.hide()
            self.project_input_widget.show()
        else:
            self.project_input_widget.hide()
            self.project_infer_widget.show()
            self.update_project_name()

    def _submission_name_input_override_widget_value_changed(self):
        if self._submission_name_override_widget.isChecked():
            self.submission_name_infer_widget.hide()
            self.submission_name_input_widget.show()
        else:
            self.submission_name_input_widget.hide()
            self.submission_name_infer_widget.show()
            self.update_submission_name()

    def _project_override_input_clicked(self) -> None:
        self.project_input_widget.set_value(self.project_infer_widget.get_value())
        self._project_input_override_widget.set_value(True)

    def _submission_name_override_input_clicked(self) -> None:
        self.submission_name_input_widget.set_value(self.submission_name_infer_widget.get_value())
        self._submission_name_override_widget.set_value(True)

    def resizeEvent(self, event):
        self.resized.emit()
        return QtWidgets.QWidget.resizeEvent(self, event)

    @QtCore.Slot()
    def update_submission_grid_layout_columns(self):
        break_point = 1200

        if self.width() < break_point and (self._old_width is None or self._old_width >= break_point) :
            self.submission_grid_layout.removeItem(self.project_input_layout)
            self.submission_grid_layout.removeItem(self.submission_name_layout)
            self.submission_grid_layout.removeItem(self.submission_comment_layout)
            self.submission_grid_layout.removeItem(self._spacer_item)
            self.submission_grid_layout.removeWidget(self.reload_scene_settings_btn)
            self.submission_grid_layout.removeWidget(self.submit_btn)

            self.submission_grid_layout.addLayout(self.project_input_layout, 0, 0, 1, 2, alignment = QtCore.Qt.AlignLeft)
            self.submission_grid_layout.addLayout(self.submission_name_layout, 1, 0, 1, 2, alignment = QtCore.Qt.AlignLeft)
            self.submission_grid_layout.addLayout(self.submission_comment_layout, 2, 0, 1, 2, alignment = QtCore.Qt.AlignLeft)
            self.submission_grid_layout.addItem(self._spacer_item, 3, 0, 1, 2)
            self.submission_grid_layout.addWidget(self.reload_scene_settings_btn, 4, 0)
            self.submission_grid_layout.addWidget(self.submit_btn, 4, 1)

        elif self.width() >= break_point and (self._old_width is None or self._old_width < break_point):
            self.submission_grid_layout.removeItem(self.project_input_layout)
            self.submission_grid_layout.removeItem(self.submission_name_layout)
            self.submission_grid_layout.removeItem(self.submission_comment_layout)
            self.submission_grid_layout.removeItem(self._spacer_item)
            self.submission_grid_layout.removeWidget(self.reload_scene_settings_btn)
            self.submission_grid_layout.removeWidget(self.submit_btn)

            self.submission_grid_layout.addLayout(self.project_input_layout, 0, 0, alignment = QtCore.Qt.AlignLeft)
            self.submission_grid_layout.addLayout(self.submission_name_layout, 1, 0, alignment = QtCore.Qt.AlignLeft)
            self.submission_grid_layout.addLayout(self.submission_comment_layout, 2, 0, alignment = QtCore.Qt.AlignLeft)
            self.submission_grid_layout.addItem(self._spacer_item, 0, 1, 3, 1)
            self.submission_grid_layout.addWidget(self.reload_scene_settings_btn, 1, 2)
            self.submission_grid_layout.addWidget(self.submit_btn, 2, 2)

        self._old_width = self.width()

    def reload_scene_settings(self):

        def set_icon(frame):
            pixmap = self.reload_movie.currentPixmap()

            icon = QtGui.QIcon()
            icon.addPixmap(pixmap, QtGui.QIcon.Normal)
            icon.addPixmap(pixmap, QtGui.QIcon.Disabled)

            self.reload_scene_settings_btn.setIcon(icon)

        # start reload spinner
        self.reload_movie.start()
        self.reload_movie.frameChanged[int].connect(set_icon)

        worker_thread = self.get_plugin().get_user_interface().get_worker_thread(u"Reloading scene settings...")

        ################################################################################################################

        def target(dynamic_job_presets: typing.List['DynamicJobPreset']) -> bool:
            time_start = time.time()

            # Attempt to infer the value of all application and user defined job preset attributes.
            # This is so any api requests made by the application_attribute_source can be cached before redrawing the
            # job preset
            for dynamic_job_preset in dynamic_job_presets:
                for dynamic_job_preset_attribute in dynamic_job_preset.get_job_preset_attributes():

                    inference_source = dynamic_job_preset_attribute.get_inference_source()
                    is_app_defined = inference_source == InferenceSource.get_application_inference_source()
                    is_user_defined = inference_source == InferenceSource.get_user_defined_inference_source()

                    if is_app_defined or is_user_defined:
                        key = dynamic_job_preset_attribute.get_key()

                        # enum items
                        try:
                            self.get_plugin().get_application_attribute_source().get_enum_items(None, None, key)
                        except ApplicationAttributeNotFound as e:
                            pass

                        # attribute values
                        try:
                            self.get_plugin().get_application_attribute_source().get_attribute_value(None, None, key)
                        except ApplicationAttributeNotFound as e:
                            pass

            time_end = time.time()

            # get elapsed time in seconds
            total_time = time_end - time_start

            # if elapsed time is less than RELOAD_SCENE_SETTINGS_TIMER_INTERVAL then sleep for the difference. This is
            # to give the user some visual feedback that something has happened
            if total_time < constants.RELOAD_SCENE_SETTINGS_SLEEP_INTERVAL:
                time.sleep(constants.RELOAD_SCENE_SETTINGS_SLEEP_INTERVAL - total_time)

            return True

        args = (
            self.get_plugin().get_job_preset_container().values(),
        )

        kwargs = {}

        worker_thread.process(target, args, kwargs, self._reload_scene_settings_handler)

    def _reload_scene_settings_handler(self, result: any) -> bool:
        plugin = self.get_plugin()
        user_interface = plugin.get_user_interface()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        # stop reload spinner
        self.reload_movie.frameChanged[int].disconnect()
        self.reload_scene_settings_btn.setIcon(QtGui.QIcon())

        if isinstance(result, bool) and result == True:
            self.redraw_job_presets()
            self.update_project_name()
            self.update_submission_name()
            return True

        if isinstance(result, ExceptionWrapper):
            exception = result.get_exception()

            if isinstance(exception, CommandPortServerError):
                self._log_error(logger, "Unable to reload scene settings - " + exception.user_message)
                user_interface.get_main_window().close()
                return True

        return False

    def get_selected_project_option(self):
        remote_project = self.project_option_selector.currentData()

        if remote_project is None:
            remote_project = constants.CURRENT_SCENE_PROJECT_OPTION_INDEX

        return remote_project

    def redraw_job_presets(self):
        i = 0
        while i < self.tab_widget.count():
            widget = self.tab_widget.widget(i)

            if isinstance(widget, DynamicJobPresetTabWidget) and widget.ready():
                dynamic_job_preset_widget = widget.get_dynamic_job_preset_widget()
                dynamic_job_preset_widget.update_app_inferred_values()

            i = i + 1

    def _submit(self):
        user_interface = self.get_plugin().get_user_interface()
        logging_coordinator = self.get_plugin().get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        # Prevent submit if already running an operation in another thread
        logger.debug(u"Checking for running operation...")
        if user_interface.is_running_operation():
            QtWidgets.QMessageBox.warning(self,
                                          u"Operation already running",
                                          u"The plugin is already busy running another operation, wait for the first "
                                          "operation to finish and then try again.",
                                          QtWidgets.QMessageBox.Ok,
                                          defaultButton=QtWidgets.QMessageBox.Ok)
            return

        dynamic_job_presets = list(self.get_plugin().get_job_preset_container().values())
        dependency_graph = self._node_widget.get_dependency_graph()

        # if no job preset is selected ask the user if they want to create one
        if not dynamic_job_presets:
            QtWidgets.QMessageBox.warning(self,
                                          "No jobs found",
                                          "You must have at least one job node in the dependency editor to "
                                          "submit.\n\nPlease create a job node.",
                                          QtWidgets.QMessageBox.Ok,
                                          defaultButton=QtWidgets.QMessageBox.Ok)
            return

        # validate the job preset
        logger.info(u"Validating job preset...")

        try:
            for dynamic_job_preset in dynamic_job_presets:
                is_valid, validation_message = self.get_plugin().get_job_preset_validator().validate(dynamic_job_preset)

                if not is_valid:
                    QtWidgets.QMessageBox.warning(self,
                                                  u"Invalid Job Preset",
                                                  validation_message,
                                                  QtWidgets.QMessageBox.Ok,
                                                  defaultButton=QtWidgets.QMessageBox.Ok)
                    return

            self._submit_to_new_project(dynamic_job_presets, dependency_graph)

        except Exception as e:
            logger.error(str(e))
            QtWidgets.QMessageBox.warning(self,
                                          u"Submit2gm Encountered An Error",
                                          str(e),
                                          QtWidgets.QMessageBox.Ok,
                                          defaultButton=QtWidgets.QMessageBox.Ok)

    def _execute_submit(self,
                        action,
                        packed_project,
                        submission_name: str,
                        submission_comment: str,
                        dynamic_job_presets: typing.List['DynamicJobPreset'],
                        dependency_graph: typing.Dict[str, 'DependencyNode']):

        api_client = self.get_plugin().get_api_client()

        if action == PreflightWindow.UPLOAD_AND_SUBMIT:
            return api_client.submit_new_project(packed_project,
                                                 submission_name,
                                                 submission_comment,
                                                 dynamic_job_presets,
                                                 dependency_graph,
                                                 delete_local_files_after_upload=False)

        if action == PreflightWindow.UPLOAD:
            return api_client.upload_project(packed_project,
                                             submission_name,
                                             submission_comment,
                                             delete_local_files_after_upload=False)

        if action == PreflightWindow.SUBMIT:
            static_job_presets = []

            for dynamic_job_preset in dynamic_job_presets:
                static_job_preset = dynamic_job_preset.get_static_job_preset(packed_project)
                static_job_presets.append(static_job_preset)

            return api_client.submit_to_remote_project(packed_project,
                                                       submission_name,
                                                       submission_comment,
                                                       static_job_presets,
                                                       dependency_graph)

    def _execute_pack_new_project(self,
                                  project_name,
                                  save_project_if_required,
                                  attributes,
                                  dynamic_job_presets: typing.List['DynamicJobPreset'],
                                  dependency_graph: typing.Dict[str, 'DependencyNode']) -> typing.Tuple['PackedProject', typing.List['DynamicJobPreset'], typing.Dict[str, 'DependencyNode']]:

        packed_scene_builder = self.get_plugin().get_packed_scene_builder()

        packed_project = packed_scene_builder.get_current_scene(project_name,
                                                                save_project_if_required,
                                                                attributes,
                                                                dynamic_job_presets,
                                                                dependency_graph)

        return packed_project, dynamic_job_presets, dependency_graph

    def _submit_to_new_project(self,
                               dynamic_job_presets: typing.List['DynamicJobPreset'],
                               dependency_graph: typing.Dict[str, 'DependencyNode']):

        plugin = self.get_plugin()
        user_interface = plugin.get_user_interface()
        api_client = plugin.get_api_client()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)
        logger.debug("Submitting to new project...")

        # Validate the project name input
        project_name = self.get_project_name()
        if not project_name:
            QtWidgets.QMessageBox.warning(self,
                                          u"Invalid Project Name",
                                          u"No project name provided",
                                          QtWidgets.QMessageBox.Ok,
                                          defaultButton=QtWidgets.QMessageBox.Ok)
            return

        # Validate the project name input
        submission_name = self.get_submission_name()
        if not submission_name:
            QtWidgets.QMessageBox.warning(self,
                                          u"Invalid Submission Name",
                                          u"No submission name provided",
                                          QtWidgets.QMessageBox.Ok,
                                          defaultButton=QtWidgets.QMessageBox.Ok)
            return

        # if project already exists warn the user
        logger.debug("Getting root directories...")
        root_directories = api_client.get_root_directories(ignore_cache=True)
        if project_name in root_directories:
            msb_box = QtWidgets.QMessageBox.warning(self,
                                                    u"Project Already Exists",
                                                    u"A project already exists with that name.\nWould you like to "
                                                    "submit anyway? Any project files with the same name will be "
                                                    "overwritten.",
                                                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                    defaultButton=QtWidgets.QMessageBox.Yes)

            if msb_box == QtWidgets.QMessageBox.Yes:
                pass
            else:
                logger.debug("Submission canceled by user (Project already exists)")
                return

        logger.info(u"Packing new project...")

        worker_thread = user_interface.get_worker_thread(u"Packing Scene...")

        ################################################################################################################

        target = self._execute_pack_new_project

        args = (
            project_name,
            False,
            {},
            dynamic_job_presets,
            dependency_graph
        )

        kwargs = {}

        worker_thread.process(target, args, kwargs, self._handle_packed_project_result)

    def _submit_to_existing_project(self, remote_project, static_job_preset):
        plugin = self.get_plugin()
        user_interface = plugin.get_user_interface()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        logger.info(u"Submitting Job...")

        worker_thread = user_interface.get_worker_thread(u"Submitting job...")

        ################################################################################################################

        target = plugin.get_api_client().submit_to_remote_project

        args = (
            remote_project,
            static_job_preset
        )

        kwargs = {}

        worker_thread.process(target, args, kwargs, self._handle_result_slot)

    def _handle_packed_project_result(self, result):
        plugin = self.get_plugin()
        user_interface = plugin.get_user_interface()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        if isinstance(result, tuple) and \
                len(result) == 3 and \
                isinstance(result[0], PackedProject) and \
                isinstance(result[1], list) and \
                isinstance(result[2], dict):

            packed_project = result[0]
            dynamic_job_presets: typing.List['DynamicJobPreset'] = result[1]
            dependency_graph: typing.Dict[str, 'DependencyNode'] = result[2]

            # hide the main window while the preflight window is open
            main_window = user_interface.get_main_window()
            main_window.hide()

            def show_main_window():
                main_window.show()
                main_window.raise_()
                main_window.activateWindow()

            # Set the watch patterns
            default_watch_pattern = WatchPattern('/{}/.+'.format(packed_project.get_name()), '')
            packed_project.set_watch_patterns([default_watch_pattern])

            submission_name = self.get_submission_name()
            submission_comment = self.get_submission_comment()

            # open preflight window
            preflight_window = PreflightWindow(self.get_plugin(),
                                               packed_project,
                                               submission_name,
                                               submission_comment,
                                               dynamic_job_presets,
                                               dependency_graph)

            preflight_window.exec_()

            # if the user has pressed cancel or closed the window then stop the submission
            if preflight_window.cancel_submission():
                QtWidgets.QMessageBox.information(self,
                                                  u"",
                                                  u"Submission cancelled.",
                                                  QtWidgets.QMessageBox.Ok,
                                                  defaultButton=QtWidgets.QMessageBox.Ok)

                show_main_window()
                return True

            # otherwise update the packed projects files to only contain the files selected by the user for upload
            files_to_upload = set(preflight_window.get_files_to_upload())
            next_action = preflight_window.get_next_action()

            if next_action == PreflightWindow.UPLOAD_AND_SUBMIT or next_action == PreflightWindow.UPLOAD:
                # Warn the user if they have files selected for upload that exist under a forbidden path
                forbidden_upload_files = preflight_window.get_forbidden_files_to_upload()
                if forbidden_upload_files:
                    emsg = "You have {} file selected for upload that exists under a forbidden directory.<br/><br/>"
                    if len(forbidden_upload_files) > 1:
                        emsg = "You have {} files selected for upload that exist under forbidden directories.<br/><br/>"

                    emsg += "See <a href=\"http://www.google.com\">here</a> for documentation on the supported " \
                            "directories on GirdMarkets.<br/><br/>Select 'Ok' to continue with the submission (files " \
                            "under forbidden directories will not be uploaded) or cancel to abort the submission."

                    emsg = emsg.format(len(forbidden_upload_files))

                    message_box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Icon.Warning,
                                                        "Warning",
                                                        emsg,
                                                        QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
                                                        parent=self)
                    message_box.setWindowModality(QtCore.Qt.ApplicationModal)
                    message_box.setTextFormat(QtCore.Qt.RichText)
                    result = message_box.exec_()

                    if result == QtWidgets.QMessageBox.Cancel:
                        show_main_window()
                        return True

                # check the user has selected at least one file to upload
                if not files_to_upload:
                    QtWidgets.QMessageBox.warning(self,
                                                  u"Warning",
                                                  u"No Files selected for upload, please select at least one file before uploading.",
                                                  QtWidgets.QMessageBox.Ok,
                                                  defaultButton=QtWidgets.QMessageBox.Ok)
                    show_main_window()
                    return True

            packed_project.set_files(files_to_upload)

            if next_action == PreflightWindow.UPLOAD_AND_SUBMIT:
                logger.info(u"Uploading new project and submitting job...")
                result_handler = self._handle_submit_and_upload_slot

            elif next_action == PreflightWindow.UPLOAD:
                logger.info(u"Uploading new project...")
                result_handler = self._handle_upload_project_slot

            elif next_action == PreflightWindow.SUBMIT:
                logger.info(u"Submitting job...")
                result_handler = self._handle_submit_slot

            else:
                assert False, "Unknown preflight action '{}'".format(next_action)

            # show the main window once the preflight window has closed
            show_main_window()

            worker_thread = user_interface.get_worker_thread(u"Submitting...")

            ################################################################################################################

            target = self._execute_submit

            args = (
                next_action,
                packed_project,
                submission_name,
                submission_comment,
                dynamic_job_presets,
                dependency_graph
            )

            kwargs = {}

            worker_thread.process(target, args, kwargs, result_handler)
            return True

        if isinstance(result, ExceptionWrapper):
            exception = result.get_exception()

            if isinstance(exception, ProjectRequiresSavingError):
                logger.exception(exception.user_message)

                btn = QtWidgets.QMessageBox.question(self,
                                               "Project requires saving",
                                               "{}\n\nPress Save to save the project and submit, otherwise press cancel".format(exception.user_message),
                                               QtWidgets.QMessageBox.Save | QtWidgets.QMessageBox.Cancel,
                                               defaultButton=QtWidgets.QMessageBox.Save)

                if btn == QtWidgets.QMessageBox.Save:
                    worker_thread = user_interface.get_worker_thread(u"Packing Scene...")
                    target = exception.get_current_scene_cb()
                    args = tuple()
                    kwargs = {}
                    worker_thread.process(target, args, kwargs, self._handle_packed_project_result)

                return True

            if isinstance(exception, FilePackingError):
                self._log_error(logger, exception.user_message)
                return True

        return False

    def _log_error(self, logger, message):
        qt_utils.log_error(self, logger, message)

    def _handle_upload_project_slot(self, result):
        plugin = self.get_plugin()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        if isinstance(result, ExceptionWrapper):
            exception = result.get_exception()

            if isinstance(exception, NotSignedInError):
                msg = u"Could not upload your current scene. You are not signed in."
                self._log_error(logger, msg)
                return True

            if isinstance(exception, CreditsError):
                msg = u"Could not submit your project. You have insufficient credits."
                self._log_error(logger, msg)
                return True

            if isinstance(exception, APIClientError):
                self._log_error(logger, exception.user_message)
                return True

            if isinstance(exception, UnsupportedApplicationError):
                self._log_error(logger, exception.user_message)
                return True

        # a remote project is returned when submitting a new project or uploading a project
        if isinstance(result, QtNone):
            QtWidgets.QMessageBox.information(self,
                                              u"",
                                              u"Project upload started.\nCheck Envoy to see your project's upload status.",
                                              QtWidgets.QMessageBox.Ok,
                                              defaultButton=QtWidgets.QMessageBox.Ok)
            plugin.get_api_client().get_remote_projects(ignore_cache=True)
            return True

        return False

    def _handle_submit_slot(self, result):
        plugin = self.get_plugin()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        if isinstance(result, ExceptionWrapper):
            exception = result.get_exception()

            if isinstance(exception, NotSignedInError):
                msg = u"Could not upload your current scene. You are not signed in."
                self._log_error(logger, msg)
                return True

            if isinstance(exception, CreditsError):
                msg = u"Could not submit your project. You have insufficient credits."
                self._log_error(logger, msg)
                return True

            if isinstance(exception, APIClientError):
                self._log_error(logger, exception.user_message)
                return True

            if isinstance(exception, UnsupportedApplicationError):
                self._log_error(logger, exception.user_message)
                return True

        if isinstance(result, QtNone):  # None is returned when submitting to an existing project
            QtWidgets.QMessageBox.information(self,
                                              u"",
                                              u"Your job has been submitted.",
                                              QtWidgets.QMessageBox.Ok,
                                              defaultButton=QtWidgets.QMessageBox.Ok)
            return True

        return False

    def _handle_submit_and_upload_slot(self, result):
        plugin = self.get_plugin()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        if isinstance(result, ExceptionWrapper):
            exception = result.get_exception()

            if isinstance(exception, NotSignedInError):
                msg = u"Could not upload your current scene. You are not signed in."
                self._log_error(logger, msg)
                return True

            if isinstance(exception, CreditsError):
                msg = u"Could not submit your project. You have insufficient credits."
                self._log_error(logger, msg)
                return True

            if isinstance(exception, APIClientError):
                self._log_error(logger, exception.user_message)
                return True

            if isinstance(exception, UnsupportedApplicationError):
                self._log_error(logger, exception.user_message)
                return True

        # a remote project is returned when submitting a new project or uploading a project
        if isinstance(result, QtNone):
            QtWidgets.QMessageBox.information(self,
                                              u"",
                                              u"Project upload started.\nYour job will start when your files have "
                                              "finished uploading.\nCheck Envoy to see your project's upload status.",
                                              QtWidgets.QMessageBox.Ok,
                                              defaultButton=QtWidgets.QMessageBox.Ok)
            plugin.get_api_client().get_remote_projects(ignore_cache=True)
            return True

        return False
