__all__ = 'WorkerThread'


from PySide2 import QtCore
import sys
import typing

from submit2gm.qt_none import QtNone
from submit2gm.plugin_accessor import PluginAccessor
from submit2gm.exception_wrapper import ExceptionWrapper
from submit2gm.widgets.qt_dialog_exception_popup import ExceptionPopup

if typing.TYPE_CHECKING:
    from PySide2 import QtWidgets
    from submit2gm.submit2gm_plugin import Submit2gmPlugin


class WorkerThread(QtCore.QThread, PluginAccessor):
    """
    Used for processing long running methods in a separate QThread.
    """

    # Signals
    accepted = QtCore.Signal()
    unhandled = QtCore.Signal()
    output = QtCore.Signal(object)

    NO_ARGUMENTS_STRING = u"NA"
    INDENT_STRING = u"   "

    def __init__(self, plugin: 'Submit2gmPlugin', parent: typing.Optional['QtWidgets.QWidget'] = None):
        super(WorkerThread, self).__init__(parent)
        PluginAccessor.__init__(self, plugin)

        self._target = None
        self._args = ()
        self._kwargs = {}
        self._result_handler_callback = None
        self.output.connect(self._handle_result_slot)

    def process(self,
                target: typing.Callable,
                args: typing.Tuple,
                kwargs: typing.Dict[str, any],
                result_handler_callback: typing.Callable[[any], bool]) -> None:

        self._target = target
        self._args = args if args else ()
        self._kwargs = kwargs if kwargs else {}
        self._result_handler_callback = result_handler_callback

        # add logging lock
        logging_coordinator = self.get_plugin().get_logging_coordinator()

        self.start()

    def run(self) -> None:
        try:
            result = self._target(*self._args, **self._kwargs)

        except Exception as e:
            exc_type, exc_value, exc_tb = sys.exc_info()

            # wrap exceptions in an exception wrapper class so that the traceback is retained
            result = ExceptionWrapper(e, exc_type, exc_value, exc_tb)

        # Wrap None objects to prevent PySide2 crashing
        if result is None:
            result = QtNone()

        self.output.emit(result)

    def _get_extra_exception_info(self, window_title: str, exception_message: str):

        line_break = u"\n" + u'-'*80 + u"\n\n"

        sys_info = self.get_plugin().get_plugin_utils().get_sys_info()

        # convert the arguments into a pretty formatted string
        args_string = self.INDENT_STRING + self.NO_ARGUMENTS_STRING
        if self._args:
            args_string = u'\n'.join(list(map(lambda x: self.INDENT_STRING + str(x), self._args)))

        # convert the key word arguments into a pretty formatted string
        kwargs = []
        for key, value in self._kwargs.items():
            kwargs.append(self.INDENT_STRING + str(key) + u": " + str(value))

        kwargs_string = self.INDENT_STRING + self.NO_ARGUMENTS_STRING
        if kwargs:
            kwargs_string = u'\n'.join(kwargs)

        # get the name of the result handler callback method
        handler_name = u"   " + str(self._result_handler_callback.__name__)

        exception_message = (u"\n" + window_title + u"\n" +
                             line_break +
                             sys_info + u"\n" +
                             line_break +
                             u"Arguments:\n" +
                             args_string +
                             u"\n\nKey Word Arguments:\n" +
                             kwargs_string +
                             u"\n\nHandler:\n " +
                             handler_name +
                             u"\n\n" + exception_message)

        return exception_message

    def _handle_result_slot(self, result: any) -> None:

        # remove logging lock and flush output
        logging_coordinator = self.get_plugin().get_logging_coordinator()

        try:
            if self._result_handler_callback(result):
                self.accepted.emit()
            else:
                # if the result is not handled by the result handler then the result must be of an unexpected type
                if isinstance(result, ExceptionWrapper):
                    # Unexpected exceptions indicate some insufficiency in the the code base. All exceptions should be
                    # handled so we should indicate to the user that something went very wrong and to contact support.

                    window_title = u"Unexpected Exception"
                    exception_message = result.to_string()
                else:
                    # If the worker has returned a object that is of an unexpected type and NOT an exception then
                    # something is very wrong.

                    window_title = u"Unexpected Return Type"
                    exception_message = u"Result: " + str(result) + u"\nResult Type: " + str(type(result))

                exception_message = self._get_extra_exception_info(window_title, exception_message)
                popup = ExceptionPopup(self.get_plugin(), exception_message, window_title=window_title)
                popup.show()
                popup.raise_()
                popup.activateWindow()

                self.unhandled.emit()

        except Exception as e:
            # The result handler should never return an exception then but if it does then inform the user that
            # something has gone wrong.

            exc_type, exc_value, exc_tb = sys.exc_info()
            ew = ExceptionWrapper(e, exc_type, exc_value, exc_tb)

            window_title = u"Exception in Handler"
            exception_message = self._get_extra_exception_info(window_title, ew.to_string())
            popup = ExceptionPopup(self.get_plugin(), exception_message)
            popup.show()
            popup.raise_()
            popup.activateWindow()

            self.unhandled.emit()
