__all__ = 'APISchema'

import typing

from gm_meta_plugin.plugin_accessor import PluginAccessor
from gm_meta_plugin.errors.rejected_transition_input_error import RejectedTransitionInputError

if typing.TYPE_CHECKING:
    from gm_meta_plugin.job_definition import JobDefinition
    from gm_meta_plugin.project_definition_schema import ProjectDefinitionSchema
    from gm_meta_plugin.project_attribute import ProjectAttribute
    from gm_meta_plugin.attribute_types import EnumAttributeType


class APISchema(PluginAccessor):

    def __init__(self, plugin, job_definitions: typing.List['JobDefinition'], project_definition_schema):
        PluginAccessor.__init__(self, plugin)
        self._job_definitions = job_definitions
        self._project_definition_schema = project_definition_schema

    def get_job_definitions(self) -> typing.List['JobDefinition']:
        return self._job_definitions

    def get_job_definition_with_id(self, id) -> 'JobDefinition':

        for job_definition in self.get_job_definitions():
            if job_definition.get_definition_id() == id:
                return job_definition

        raise ValueError("No job definition with id '" + id + "'")

    def get_project_definition_schema(self) -> 'ProjectDefinitionSchema':
        return self._project_definition_schema

    def get_supported_applications(self, job_definition: 'JobDefinition') -> typing.List[str]:
        """
        Returns a list of all the application ids that the job definition works with
        """

        from gm_meta_plugin.gridmarkets import constants as api_constants

        project_attribute = self.get_project_definition_schema().get_project_attribute_with_id(api_constants.PROJECT_ATTRIBUTE_IDS.PRODUCT)
        attribute: 'EnumAttributeType' = project_attribute.get_attribute()

        def _recursive(p: 'ProjectAttribute'):
            if job_definition in p.get_compatible_job_definitions():
                return True

            for transition in p.get_transitions():
                if _recursive(transition.get_project_attribute()):
                    return True

        supported_applications = []
        for item in attribute.get_items():
            key = item.get_key()

            try:
                child_attribute = project_attribute.transition(key)
                if _recursive(child_attribute):
                    supported_applications.append(key)

            except RejectedTransitionInputError as e:
                pass

        return supported_applications




