__all__ = 'AttributeType', 'Attribute'

import enum
from abc import abstractmethod
import sys
import typing

from gm_meta_plugin.errors import CLIPromptNotEnabledError

if typing.TYPE_CHECKING:
    from gm_meta_plugin.plugin import Plugin
    from gm_meta_plugin.job_preset import JobPreset


class AttributeType(enum.Enum):
    STRING = "STRING"
    ENUM = "ENUM"
    NULL = "NULL"
    BOOLEAN = "BOOLEAN"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    LIST = "LIST"


class Attribute:

    def __init__(self,
                 key,
                 display_name,
                 description,
                 attribute_type,
                 subtype_kwargs):

        self._key = key
        self._display_name = display_name
        self._description = description
        self._attribute_type = attribute_type
        self._subtype_kwargs = subtype_kwargs

    def get_key(self):
        return self._key

    def get_display_name(self):
        return self._display_name

    def get_description(self):
        return self._description

    def get_type(self):
        return self._attribute_type

    def get_subtype_kwargs(self):
        return self._subtype_kwargs

    def get_argv_value(self, plugin: 'Plugin', key: str = None, job_preset: 'JobPreset' = None):
        attribute_type = self.get_type()
        assert attribute_type != AttributeType.NULL, 'Unable to get ARGV values for NULL type attributes'

        if key is None:
            arg_key = '--' + self.get_key()
        else:
            arg_key = '--' + key

        arg_key_2 = arg_key.replace('_', '-')

        # Try to get the index of the argument key in sys.argv
        index = -1
        try:
            index = sys.argv.index(arg_key)
        except ValueError:
            try:
                # job preset keys use '_' where as argparser arguments use '-'
                index = sys.argv.index(arg_key_2)
            except ValueError:
                pass

        # If the index was found then return it's value
        if index >=0:

            try:
                value = sys.argv[index + 1]
            except IndexError as e:
                raise CLIPromptNotEnabledError("CLI argument '{}' missing value.\nPlease pass a value after the CLI argument.".format(arg_key_2))

            if self.get_type() == AttributeType.BOOLEAN:
                value = bool(value)
            elif self.get_type() == AttributeType.INTEGER:
                value = int(value)
            elif self.get_type() == AttributeType.FLOAT:
                value = float(value)

            return value

        if '--enable-cli-prompt' not in sys.argv:
            raise CLIPromptNotEnabledError("CLI argument '{}' not passed.\nEither the argument on the CLI or pass '--enable-cli-prompt' to enable dynamic prompting for missing values.".format(arg_key_2))

        from gm_meta_plugin.gridmarkets import constants as api_constants
        from gm_meta_plugin.attribute_types import StringAttributeType, StringSubtype, \
            EnumAttributeType, EnumSubtype, BooleanAttributeType, IntegerAttributeType, IntegerSubtype, \
            FloatAttributeType

        from gm_meta_plugin.errors import InvalidAttributeError
        import PyInquirer
        import re

        PROMPT_STYLE = PyInquirer.style_from_dict({
            PyInquirer.Token.Separator: '#cc5454',
            PyInquirer.Token.QuestionMark: '#673ab7 bold',
            PyInquirer.Token.Selected: '#cc5454',  # default
            PyInquirer.Token.Pointer: '#673ab7 bold',
            PyInquirer.Token.Instruction: '',  # default
            PyInquirer.Token.Answer: '#f44336 bold',
            PyInquirer.Token.Question: '',
        })

        display_name = self.get_display_name()
        description = self.get_description()
        default = self.get_default_value()
        subtype_kwargs = self.get_subtype_kwargs()
        message = "No value specified for '{}' ('{}'):\n  Description: {}\n\n" \
                  "  Please enter a value:".format(display_name,
                                                   '--' + self.get_key().replace('_', '-'),
                                                   description)

        if attribute_type == AttributeType.STRING:
            # noinspection PyTypeChecker
            string_attribute: StringAttributeType = self

            subtype = string_attribute.get_subtype()
            placeholder = string_attribute.get_placeholder()

            def string_validator(value: str):
                try:
                    string_attribute.validate_value(value)
                except InvalidAttributeError as e:
                    return str(e)

                return True

            answers = PyInquirer.prompt([
                {
                    'type': 'input',
                    'name': self.get_key(),
                    'default': default,
                    'message': message,
                    'validate': string_validator
                }
            ], style=PROMPT_STYLE)

            return answers[self.get_key()]

        if attribute_type == AttributeType.ENUM:
            # noinspection PyTypeChecker
            enum_attribute: EnumAttributeType = self

            subtype = enum_attribute.get_subtype()
            message = "No value specified for '{}' ('{}'):\n  Description: {}\n\n" \
                      "  Please select a value:".format(display_name,
                                                        '--' + self.get_key().replace('_', '-'),
                                                        description)
            choices = []

            if subtype == EnumSubtype.PRODUCT_VERSIONS.value:
                product = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.PRODUCT_VERSIONS.PRODUCT)

                match = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.PRODUCT_VERSIONS.MATCH)
                match = p = re.compile(match) if match is not None else None

                versions = plugin.get_api_client().get_product_versions(product)

                if match is not None:
                    versions = [s for s in versions if p.match(s)]

                choices = [{'name': item, 'value': item} for item in list(versions)]

                # if the product matches the current cps
                try:
                    if plugin.get_active_command_port_server().get_product() == product:
                        # then we can get current version of the application and use it as the default item
                        application_attribute_source = plugin.get_application_attribute_source()
                        items, default_item_index = application_attribute_source.get_enum_items(product, None,
                                                                                                self.get_key())
                        default_version = items[default_item_index]

                        # attempt to set the default item to the version returned by the app attribute source.
                        # If the default version does not match any version in the versions list then this does nothing.
                        default = default_version.get_key()
                except ValueError as e:
                    pass

            elif subtype == EnumSubtype.MACHINE_TYPE.value:

                app = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.MACHINE_TYPE.PRODUCT)
                operation = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.MACHINE_TYPE.OPERATION)
                if operation is None:
                    if job_preset:
                        operation = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.OPERATION,
                                                                               filter_inactive_attributes=True).get_value()

                use_gpu_default = subtype_kwargs.get(api_constants.SUBTYPE_KEYS.ENUM.MACHINE_TYPE.USE_GPU_DEFAULT) == "True"
                machine_options = plugin.get_api_client().get_machines(app, operation)

                for i, machine_option in enumerate(machine_options):
                    choices.append({'name': machine_option.get_name(), 'value': machine_option.get_id()})

                    # give the default machine id 0 so it starts selected
                    if machine_option.is_default() and machine_option.is_gpu_machine() == use_gpu_default:
                        default = machine_option.get_id()

            elif subtype == EnumSubtype.APP_INFERRED.value:
                try:
                    application_attribute_source = plugin.get_application_attribute_source()
                    res = application_attribute_source.get_enum_items(None, None, self.get_key())
                    try:
                        items, default_item_index = res

                        for item in items:
                            choices.append({'name': item.get_display_name(), 'value': item.get_key()})

                        default = default_item_index[default_item_index].get_key()

                    except TypeError as e:
                        print("res:", res)
                        raise e
                except Exception:

                    answers = PyInquirer.prompt([
                        {
                            'type': 'input',
                            'name': self.get_key(),
                            'message': message,
                            'default': str(default) if default else ''
                        }
                    ], style=PROMPT_STYLE)

                    return answers[self.get_key()]

            else:
                enum_items = enum_attribute.get_items()
                for enum_item in enum_items:
                    choices.append({'name': enum_item.get_display_name(), 'value': enum_item.get_key()})

            if len(choices) > 1:
                answers = PyInquirer.prompt([
                    {
                        'type': 'list',
                        'name': self.get_key(),
                        'message': message,
                        'choices': choices,
                        'default': default
                    }
                ], style=PROMPT_STYLE)
            else:
                answers = {self.get_key(): choices[0]['value']}

            return answers[self.get_key()]

        if attribute_type == AttributeType.BOOLEAN:

            message = "No value specified for '{}' ('{}'):\n  Description: {}\n\n" \
                      "  Please select a value:".format(display_name,
                                                        '--' + self.get_key().replace('_', '-'),
                                                        description)

            answers = PyInquirer.prompt([
                {
                    'type': 'confirm',
                    'name': self.get_key(),
                    'message': message,
                    'default': default
                }
            ], style=PROMPT_STYLE)

            return answers[self.get_key()]

        if attribute_type == AttributeType.INTEGER:
            # noinspection PyTypeChecker
            integer_attribute: IntegerAttributeType = self

            if integer_attribute.get_subtype() == IntegerSubtype.INSTANCES.value:
                user_info = plugin.get_api_client().get_user_info()

                job_throughput_limit = user_info.get_job_throughout_limit()
                max_throughput = user_info.get_max_throughout()

                def validate_machines(value):
                    try:
                        value = int(value)
                        if value < 0:
                            return False
                        elif value > max_throughput:
                            return False
                        else:
                            return True

                    except ValueError:
                        return False

                answers = PyInquirer.prompt([
                    {
                        'type': 'input',
                        'name': self.get_key(),
                        'message': message,
                        'default': '0',
                        'validate': validate_machines
                    }
                ], style=PROMPT_STYLE)

                return int(answers[self.get_key()])

            else:
                def int_validator(value: str):
                    try:
                        int(value)
                    except ValueError:
                        return False
                    return True

                answers = PyInquirer.prompt([
                    {
                        'type': 'input',
                        'name': self.get_key(),
                        'default': str(default),
                        'message': message,
                        'validate': int_validator
                    }
                ], style=PROMPT_STYLE)

                return int(answers[self.get_key()])

        if attribute_type == AttributeType.FLOAT:
            # noinspection PyTypeChecker
            float_attribute: FloatAttributeType = self

            def float_validator(value: str):
                try:
                    float(value)
                except ValueError:
                    return False
                return True

            answers = PyInquirer.prompt([
                {
                    'type': 'input',
                    'name': self.get_key(),
                    'default': str(default),
                    'message': message,
                    'validate': float_validator
                }
            ], style=PROMPT_STYLE)

            return float(answers[self.get_key()])

        raise NotImplementedError(self.get_key(), self.get_type())

    @abstractmethod
    def get_default_value(self):
        raise NotImplementedError

    @abstractmethod
    def validate_value(self, value):
        raise NotImplementedError
