__all__ = 'DynamicJobPreset'

from .job_preset import JobPreset
from .dynamic_job_preset_attribute import DynamicJobPresetAttribute
from .static_job_preset import StaticJobPreset
from .static_job_preset import StaticJobPresetAttribute

import typing

if typing.TYPE_CHECKING:
    from .job_definition import JobDefinition
    from .plugin import Plugin
    from .remote_project import RemoteProject


class DynamicJobPreset(JobPreset):

    def __init__(self,
                 plugin: 'Plugin',
                 name: str,
                 id: str,
                 job_definition: 'JobDefinition',
                 job_preset_attributes: typing.List['DynamicJobPresetAttribute'] = None,
                 is_locked: bool = False):

        JobPreset.__init__(self, plugin, name, id, job_definition)

        if job_preset_attributes is None:
            job_preset_attributes = list(map(lambda job_attribute: DynamicJobPresetAttribute(plugin, self, job_attribute),
                                             job_definition.get_attributes()))

        self._job_preset_attributes = job_preset_attributes
        self._is_locked = is_locked

    def get_job_preset_attributes(self) -> typing.List['DynamicJobPresetAttribute']:
        return self._job_preset_attributes

    def get_job_preset_attribute_with_job_attribute_id(self, id: str) -> 'DynamicJobPresetAttribute':
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_job_attribute().get_id() == id:
                return job_preset_attribute

        raise ValueError(u"No JobPresetAttribute with JobAttribute id: '{}'".format(str(id)))

    def get_job_preset_attribute_by_key(self,
                                        key: str,
                                        filter_inactive_attributes: bool = True,
                                        allow_multiple: bool = False) -> 'DynamicJobPresetAttribute':

        result = None

        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_key() == key:

                # Filter the inactive attributes (do after key comparison as this is a potentially expensive operation)
                if filter_inactive_attributes and not job_preset_attribute.is_active():
                    continue

                if allow_multiple:
                    return job_preset_attribute

                if result is not None:
                    raise ValueError("Multiple JobPresetAttributes with the key: " + str(key))

                result = job_preset_attribute

        if result:
            return result

        raise ValueError(u"No JobPresetAttribute with the key: " + str(key))

    def get_job_preset_attribute_by_id(self, id) -> 'DynamicJobPresetAttribute':
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_id() == id:
                return job_preset_attribute

        raise ValueError(u"No JobPresetAttribute with the id: " + str(id))

    def get_job_preset_attributes_in_group(self, attribute_group) -> typing.List['DynamicJobPresetAttribute']:
        job_preset_attributes = []
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_job_attribute().get_attribute_group() == attribute_group:
                job_preset_attributes.append(job_preset_attribute)
        return job_preset_attributes

    def get_active_job_preset_attributes(self) -> typing.List['DynamicJobPresetAttribute']:
        job_preset_attributes = []
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.is_active():
                job_preset_attributes.append(job_preset_attribute)

        return job_preset_attributes

    def is_locked(self) -> bool:
        return self._is_locked

    def set_locked_state(self, is_locked: bool) -> None:
        self._is_locked = is_locked

    def toggle_locked_state(self) -> None:
        self.set_locked_state(not self._is_locked)

    def get_static_job_preset(self, remote_project: 'RemoteProject') -> 'StaticJobPreset':
        # Create the static job preset without any attributes (they are added later)
        static_job_preset = StaticJobPreset(self.get_plugin(),
                                            self.get_name(),
                                            self.get_id(),
                                            self.get_job_definition(),
                                            [])

        # Create the static job preset attributes
        static_job_preset_attributes = []
        dynamic_job_preset_attributes = self.get_job_preset_attributes()[:]
        processed_job_preset_attributes = set()

        eval_lock_states = dict()
        for dynamic_job_preset_attribute in self.get_job_preset_attributes():
            eval_lock_states[dynamic_job_preset_attribute.get_id()] = dynamic_job_preset_attribute.is_eval_locked()

        while True:

            # Get the first job preset attribute for which all of it's dependencies have been processed
            dynamic_job_preset_attribute: typing.Optional['DynamicJobPresetAttribute'] = None
            for jpa in dynamic_job_preset_attributes:
                if set(jpa.get_dependencies()).issubset(processed_job_preset_attributes):
                    dynamic_job_preset_attribute = jpa
                    break

            # If there are no more dynamic job preset attributes left to process that have all of their dependencies met
            # then break
            if dynamic_job_preset_attribute is None:
                break

            dynamic_job_preset_attributes.remove(dynamic_job_preset_attribute)
            processed_job_preset_attributes.add(dynamic_job_preset_attribute)

            # In-active job preset attributes should be skipped as their preconditions have not been met
            if not dynamic_job_preset_attribute.is_active():
                continue

            if not dynamic_job_preset_attribute.is_enabled():
                continue

            value = dynamic_job_preset_attribute.eval(source_project=remote_project, fail_on_eval_lock=False)

            # Eval lock after getting the value so that other job preset attributes which depend on this job preset
            # attribute cannot re-evaluate it. This is required to prevent argv defined job preset attributes from
            # prompting the user multiple times for the same attribute however it improves performance with respect to
            # the other inference sources as it reduces the number of times the attribute is evaluated overall.
            dynamic_job_preset_attribute.set_eval_locked(True)

            static_job_preset_attributes.append(StaticJobPresetAttribute(
                self.get_plugin(),
                static_job_preset,
                dynamic_job_preset_attribute.get_job_attribute(),
                True,
                value,
                dynamic_job_preset_attribute.should_skip()
            ))

        # Add a static job preset attribute without a value for all the dynamic job preset attributes that were not
        # already created
        for dynamic_job_preset_attribute in self.get_job_preset_attributes():
            for static_job_preset_attribute in static_job_preset_attributes:
                if dynamic_job_preset_attribute.get_id() == static_job_preset_attribute.get_id():
                    break
            else:
                static_job_preset_attributes.append(StaticJobPresetAttribute(
                    self.get_plugin(),
                    static_job_preset,
                    dynamic_job_preset_attribute.get_job_attribute(),
                    False,
                    None,
                    dynamic_job_preset_attribute.should_skip()
                ))

        # Finally restore the eval locks to their previous state
        for dynamic_job_preset_attribute in self.get_job_preset_attributes():
            dynamic_job_preset_attribute.set_eval_locked(eval_lock_states[dynamic_job_preset_attribute.get_id()])

        # noinspection PyProtectedMember
        static_job_preset._set_job_preset_attributes(static_job_preset_attributes)
        return static_job_preset
