__all__ = 'EnvoyProject', 'EnvoyFile', 'GridMarketsAPIClient'

import pathlib

import json
import platform
import PySide2
import sys
import typing

from gm_meta_plugin.gridmarkets.remote_project import convert_packed_project
from gm_meta_plugin.gridmarkets.xml_api_schema_parser import XMLAPISchemaParser
from . import constants as api_constants
from gm_meta_plugin.api_client import APIClient as MetaAPIClient
from gm_meta_plugin.cached_value import CachedValue
from gm_meta_plugin.attribute_types import EnumItem
from gm_meta_plugin.user_info import UserInfo
from gm_meta_plugin.machine_option import MachineOption
from gm_meta_plugin.product import Product
from gm_meta_plugin.watch_pattern import WatchPattern
from gm_meta_plugin.inference_source import InferenceSource
from gm_meta_plugin.utils import get_files_in_directory, get_exception_with_traceback, get_external_files_remote_root
from gm_meta_plugin.errors import *
from .. import attribute_types

import gridmarkets
from gridmarkets.errors import AuthenticationError, APIError, InvalidRequestError, InsufficientCreditsError

if typing.TYPE_CHECKING:
    from gm_meta_plugin.factory_collection import FactoryCollection
    from gm_meta_plugin.gridmarkets.api_schema import APISchema
    from gm_meta_plugin.dynamic_job_preset import DynamicJobPreset
    from gm_meta_plugin.static_job_preset import StaticJobPreset
    from gm_meta_plugin.static_job_preset_attribute import StaticJobPresetAttribute
    from gm_meta_plugin.project import Project
    from gm_meta_plugin.packed_project import PackedProject
    from gm_meta_plugin.remote_project import RemoteProject


class EnvoyProject:
    def __init__(self, name, deleting, num_files, total_size, auto_downloading,
                 watched_list):
        self._name = name
        self._deleting = deleting
        self._num_files = num_files
        self._total_size = total_size
        self._auto_downloading = auto_downloading
        self._watched_list = watched_list

    def get_name(self):
        return self._name

    def is_deleting(self):
        return self._deleting

    def get_file_count(self):
        return self._num_files

    def get_total_size(self):
        return self._total_size

    def is_auto_downloading(self):
        return self._auto_downloading

    def get_watched_list(self):
        return self._watched_list


class EnvoyFile:
    def __init__(self, check_sum, last_modified, name, key, size, fmt_name):
        self._check_sum = check_sum
        self._last_modified = last_modified
        self._name = name
        self._key = key
        self._size = size
        self._fmt_name = fmt_name

    def get_check_sum(self):
        return self._check_sum

    def get_last_modified(self):
        return self._last_modified

    def get_name(self):
        return self._name

    def get_key(self):
        return self._key

    def get_size(self):
        return self._size

    def get_fmt_name(self):
        return self._fmt_name


class GridMarketsAPIClient(MetaAPIClient):
    http_api_endpoint = u"https://api.gridmarkets.com:8003/api/render/1.0/"

    def __init__(self, plugin, logging_coordinator, factory_collection: 'FactoryCollection'):
        MetaAPIClient.__init__(self, plugin)
        self._factory_collection = factory_collection
        self._signed_in = False
        self._envoy_client = None
        self._log = logging_coordinator.get_logger(u"GridMarketsAPIClient")

        # used for caching projects
        self._api_schema_cache = CachedValue['APISchema']()
        self._remote_projects_cache = CachedValue([])
        self._project_files_dictionary_cache = {}
        self._product_resolver = CachedValue(None)
        self._products_cache = CachedValue([])
        self._cached_machine_options = CachedValue({})
        self._cached_user_info = CachedValue['UserInfo']()

    def clear_all_caches(self):
        self._log.info(u"Clearing cached API data")
        self._remote_projects_cache.reset_and_clear_cache()
        self._project_files_dictionary_cache.clear()
        self._product_resolver.reset_and_clear_cache()
        self._products_cache.reset_and_clear_cache()
        self._cached_machine_options.reset_and_clear_cache()
        self._cached_user_info.reset_and_clear_cache()

    def sign_in(self, user, skip_validation=False):
        self._log.info(u"Signing in as " + user.get_auth_email())

        MetaAPIClient.sign_in(self, user, skip_validation=skip_validation)

        self._envoy_client = gridmarkets.EnvoyClient()

        # refresh caches
        self.clear_all_caches()
        self.get_root_directories(ignore_cache=True)
        self.get_api_schema(ignore_cache=True)
        self.get_user_info(ignore_cache=True)
        self._signed_in = True

    def sign_out(self):
        if self.is_user_signed_in():
            self._log.info(u"Signing out user " + self.get_signed_in_user().get_auth_email())

        MetaAPIClient.sign_out(self)
        self._signed_in = False

    def is_user_signed_in(self):
        return self._signed_in

    def connected(self):
        raise NotImplementedError

    def get_signed_in_user(self):
        return MetaAPIClient.get_signed_in_user(self)

    def get_cached_api_schema(self):
        return self._api_schema_cache.get_value()

    def get_api_schema(self, factory_collection: 'FactoryCollection' = None, ignore_cache: bool = False) -> 'APISchema':
        self._log.info(u"Getting api schema. ignore_cache=" + str(ignore_cache))

        if factory_collection is None:
            factory_collection = self._factory_collection

        if ignore_cache or not self._api_schema_cache.is_cached():
            self._log.info(u"Fetching api schema from XML source.")

            try:
                api_schema = XMLAPISchemaParser.parse(self.get_plugin(), factory_collection)
                self._api_schema_cache.set_value(api_schema)
            except Exception as e:
                msg = u"Could not fetch api schema." + get_exception_with_traceback(e)
                self._log.error(msg, exc_info=True)
                raise APIClientError(msg)

        self._log.info(u"Returning api schema")
        return self._api_schema_cache.get_value()

    @staticmethod
    def validate_credentials(auth_email):

        # check an email address has been entered
        has_email = isinstance(auth_email, str) and len(auth_email) > 0

        if has_email:
            try:
                client = gridmarkets.EnvoyClient()
                client.validate_auth()
            except AuthenticationError as e:
                raise InvalidUserError(message=str(e))
            except APIError as e:
                e_msg = str(e)

                # pretty error message for common errors
                if e_msg.startswith(u"Unexpected error communicating with Envoy.Please start Envoy"):
                    e_msg = u"Unable to connect to Envoy."

                raise APIClientError(message=e_msg)

        elif not has_email:
            message = u"No " + api_constants.COMPANY_NAME + u" email address provided."
            raise InvalidEmailError(message=message)

        return True

    def upload_project(self,
                       packed_project,
                       upload_root_dir,
                       delete_local_files_after_upload=False):

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to upload a project.")

        self._log.info(u"Uploading packed project")
        gm_project = self._get_gm_project(packed_project)

        if upload_root_dir:
            # add all files to packed project files list
            files = get_files_in_directory(packed_project.get_root_dir())
            packed_project.add_files(gm_project.local_root, gm_project.remote_root, files)

            for file in files:
                self._log.info(u"Marking \"" + file + u"\" for upload.")

            gm_project.add_files(gm_project.local_root, gm_project.remote_root, *files)
        else:
            # Add project files to gm_project
            self._add_files_to_upload(gm_project, packed_project)

        try:
            self._generate_summary_contents(gm_project)
            self._envoy_client.upload_project_files(gm_project)
        except InsufficientCreditsError as e:
            raise CreditsError(str(e))
        except APIError as e:
            raise APIClientError(str(e))

        self._log.info(u"Uploaded project")

        return convert_packed_project(packed_project, self.get_plugin())

    def _get_gm_job(self, static_job_preset: 'StaticJobPreset', remote_project):
        self._log.info(u"Calculating Job parameters...")

        job = gridmarkets.Job(**static_job_preset.get_job_settings())

        # Handle credits budget attributes
        if (static_job_preset.contains_job_attribute_with_id(api_constants.API_KEYS.CREDITS_BUDGET) and
                static_job_preset.contains_job_attribute_with_id(api_constants.API_KEYS.CREDITS_BUDGET_ACTION)):

            credits_budget_attribute = static_job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.CREDITS_BUDGET)
            credits_budget_action_attribute = static_job_preset.get_job_preset_attribute_by_key(
                api_constants.API_KEYS.CREDITS_BUDGET_ACTION)

            if credits_budget_attribute.has_value() and credits_budget_action_attribute.has_value():
                credits_budget = credits_budget_attribute.get_value()
                credits_budget_action = credits_budget_action_attribute.get_value()

                if credits_budget_action == api_constants.JOB_ACTIONS.CREDITS_BUDGET_ACTION_ALERT:
                    credits_budget_action = gridmarkets.Job.CREDITS_BUDGET_ACTION_ALERT
                elif credits_budget_action == api_constants.JOB_ACTIONS.CREDITS_BUDGET_ACTION_STOP:
                    credits_budget_action = gridmarkets.Job.CREDITS_BUDGET_ACTION_STOP
                elif credits_budget_action == api_constants.JOB_ACTIONS.CREDITS_BUDGET_ACTION_SUSPEND:
                    credits_budget_action = gridmarkets.Job.CREDITS_BUDGET_ACTION_SUSPEND
                else:
                    raise Exception("{} is not a known job action key".format(credits_budget_action))

                job.set_credits_budget(credits_budget, credits_budget_action)

        # Check the app is supported
        if job.app not in self.get_product_app_types():
            msg = u"App \"" + job.app + u"\" is not supported for your account. Please contact " + \
                  api_constants.COMPANY_NAME + u" support for details."
            self._log.error(msg)
            raise UnsupportedApplicationError(msg)

        # Check the version is supported
        if job.app_version not in self.get_product_versions(job.app):
            msg = u"Version \"" + job.app_version + u"\" for app_type \"" + job.app + \
                  u"\" is not supported for your account. Please contact " + api_constants.COMPANY_NAME + \
                  u" support for details."

            self._log.error(msg)
            raise UnsupportedApplicationError(msg)

        return job

    def _get_submission_name(self, static_job_preset: 'StaticJobPreset') -> typing.Optional[str]:
        try:
            attribute = static_job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.SUBMISSION_NAME)
            submission_name = attribute.get_value()
        except ValueError:
            submission_name = None

        return submission_name

    def _get_submission_comment(self, static_job_preset: 'StaticJobPreset') -> typing.Optional[str]:
        try:
            attribute = static_job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.SUBMISSION_COMMENT)
            submission_comment = attribute.get_value()
        except ValueError:
            submission_comment = None

        return submission_comment

    def _set_watch_patterns(self, project: 'Project', gm_project):

        watch_patterns = project.get_watch_patterns()

        # The python envoy clients skip_auto_download attribute determines if any watch files are set for the request
        # If we have no watch files then do not attempt to download anything
        gm_project.skip_auto_download = len(watch_patterns) <= 0

        self._log.info(u"Setting watch patterns...")
        for watch_pattern in watch_patterns:
            self._log.info(u"{0} : {1}".format(watch_pattern.get_remote_watch_pattern(),
                                               watch_pattern.get_local_download_path()))

            gm_project.add_watch_files(gridmarkets.WatchFile(watch_pattern.get_remote_watch_pattern(),
                                                             watch_pattern.get_local_download_path()))

        gm_project.remote_output_folder_name = project.get_name()
        gm_project.remote_output_folder = project.get_name()

    def _add_files_to_upload(self, gm_project, packed_project):
        for project_file in packed_project.get_files():
            if packed_project.get_root_dir() not in project_file.parents:
                local_root = project_file.anchor
                remote_root = get_external_files_remote_root(gm_project.remote_root, project_file).as_posix()
            else:
                local_root = gm_project.local_root
                remote_root = gm_project.remote_root

            unicode_project_file = str(project_file)

            self._log.info(u"Marking \"" + unicode_project_file + u"\" for upload.")
            self._log.debug(u"Local root: " + local_root)
            self._log.debug(u"Remote root: " + remote_root)
            gm_project.add_files(local_root, remote_root, *[unicode_project_file])

    def _get_gm_project(self, project: 'Project', static_job_preset: 'StaticJobPreset' = None):
        project_name = project.get_name()
        project_dir = str(project.get_root_dir())

        # Set the submission name
        submission_name = self._get_submission_name(static_job_preset) if static_job_preset else None

        # Set the submission comment
        submission_comment = self._get_submission_comment(static_job_preset) if static_job_preset else None

        plugin = self.get_plugin()

        sys_info = {
            'os': platform.platform(),
            'envoy_python_client_version': gridmarkets.version.VERSION,
            'submit2gm_version': plugin.get_version().to_string(),
            'pyside2_version': PySide2.__version__,
        }

        try:
            cps = plugin.get_active_command_port_server()
            sys_info['cps_info'] = {
                'address': cps.get_address(),
                'product': cps.get_product(),
                'meta_info': cps.get_meta_info()
            }
        except ValueError:
            pass

        return gridmarkets.Project(project_dir,
                                   project_name,
                                   submission_name=submission_name,
                                   submission_comment=submission_comment,
                                   sys_info=sys_info)

    def _generate_summary_contents(self, gm_project):
        summary = """
PROJECT NAME:\t\t""" + gm_project.name + """ 
SUBMISSION NAME:\t""" + str(gm_project.submission_name) + """

PYTHON ENVOY CLIENT VERSION:\t""" + gridmarkets.version.VERSION + """

SYS INFO:
""" + self.get_plugin().get_plugin_utils().get_sys_info() + """
"""

        self._log.debug(json.dumps(gm_project.serialize, indent=3))

        gm_project.params['summary'] = summary

    def submit_new_project(self, packed_project: 'PackedProject', job_preset: 'DynamicJobPreset', delete_local_files_after_upload=False):

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to Submit a job.")

        self._log.info(u"Preparing to submit current scene as project \"" + packed_project.get_name() +
                       u"\" with job \"" + job_preset.get_name() + u"\".")

        # Convert the packed project to a remote project
        remote_project = convert_packed_project(packed_project, self.get_plugin())

        # Convert the dynamic job preset to a static job preset
        static_job_preset = job_preset.get_static_job_preset(remote_project)

        client = self._envoy_client
        gm_project = self._get_gm_project(packed_project, static_job_preset=static_job_preset)
        self._add_files_to_upload(gm_project, packed_project)

        job = self._get_gm_job(static_job_preset, remote_project)

        # add job to project
        gm_project.add_jobs(job)

        # set the watch files
        self._set_watch_patterns(remote_project, gm_project)

        # submit the job to the remote project
        self._log.info(u"Submitting current scene with job \"" + static_job_preset.get_name() + u"\" to project \"" +
                       remote_project.get_name() + u"\".")

        try:
            self._generate_summary_contents(gm_project)
            client.submit_project(gm_project, skip_upload=False)

        except InvalidRequestError as e:
            e_msg = "Submission Failed.\n\nInvalid Request Error:\n{}".format(str(e))

            if 'errors' in e.errors:
                errors = e.errors['errors']
                e_msg += '\n'

                for error in errors:
                    e_msg += '\n{}'.format(str(error['message']))

            raise APIClientError(e_msg)
        except InsufficientCreditsError as e:
            raise CreditsError(str(e))
        except APIError as e:
            raise APIClientError(str(e))

        self._log.info(u"Job submitted - See Render Manager for details")

        return remote_project

    def submit_to_remote_project(self, remote_project: 'RemoteProject', static_job_preset: 'StaticJobPreset'):

        if not self.is_user_signed_in():
            raise NotSignedInError(u"Must be signed-in to Submit a job.")

        """
        # check the project exists
        if len(remote_project.get_files(force_update=True)) <= 0:
            msg = u"Can not submit job to remote project \"" + remote_project.get_name() + \
                  u"\", project does not exist."
            self._log.warning(msg)
            raise APIClientError(msg)
        

        # check the project has finished uploading
        if remote_project.get_remaining_bytes_to_upload(force_update=True) != 0:
            msg = u"Can not submit job to remote project \"" + remote_project.get_name() + \
                  u"\", project has not finished uploading."
            self._log.warning(msg)
            raise APIClientError(msg)
        """

        self._log.info(u"Preparing job " + str(static_job_preset.get_name()) + u" for submission to remote project " +
                       str(remote_project.get_name()))

        client = self._envoy_client
        gm_project = self._get_gm_project(remote_project, static_job_preset=static_job_preset)

        job = self._get_gm_job(static_job_preset, remote_project)

        # add job to project
        gm_project.add_jobs(job)

        # set the watch files
        self._set_watch_patterns(remote_project, gm_project)

        # submit the job to the remote project
        self._log.info(u"Submitting job " + str(static_job_preset.get_name()) + u" to existing remote project " +
                       str(remote_project.get_name()))

        try:
            self._generate_summary_contents(gm_project)
            client.submit_project(gm_project, skip_upload=True)

        except InvalidRequestError as e:
            e_msg = "Submission Failed.\n\nInvalid Request Error:\n{}".format(str(e))

            if 'errors' in e.errors:
                errors = e.errors['errors']
                e_msg += '\n'

                for error in errors:
                    e_msg += '\n{}'.format(str(error['message']))

            raise APIClientError(e_msg)
        except InsufficientCreditsError as e:
            raise CreditsError(str(e))
        except APIError as e:
            raise APIClientError(str(e))

        self._log.info(u"Job submitted - See Render Manager for details")

    def update_remote_project_status(self, remote_project):

        self._log.info(u"Fetching project status for project \"" + remote_project.get_name() + u"\"...")

        http_client = self._envoy_client.http_client
        url = self._envoy_client.url + u"/project-status/" + remote_project.get_name()

        try:
            resp = http_client.request('get', url)
        except Exception as e:
            msg = u"Could not fetch project status for project \"" + remote_project.get_name() + u"\". " + \
                  get_exception_with_traceback(e)
            self._log.error(msg)
            raise APIClientError(msg)
        else:
            if resp.status_code == 200:
                self._log.info(u"Fetched project status for project \"" + remote_project.get_name() + u"\".")

                json = resp.json()

                details = json.get(u'Details')

                total_bytes = 0
                total_bytes_done = 0
                files = set()

                if len(details.items()) > 0:
                    for detail_key, detail in details.items():
                        total_bytes_for_detail = detail.get(u'BytesTotal', 1)
                        bytes_done_for_detail = detail.get(u'BytesDone', 0)

                        total_bytes += total_bytes_for_detail
                        total_bytes_done += bytes_done_for_detail
                        files.add(pathlib.Path(detail_key))
                else:
                    # It is possible that the project has already been uploaded and Envoy is not bothering to show the
                    # project status for it any more.  Check to see if there are any files uploaded for this project.
                    uploaded_files = self.get_remote_project_files(remote_project.get_name(), ignore_cache=True)
                    if len(uploaded_files) >= 0:
                        for file in uploaded_files:
                            total_bytes += file.get_size()
                            total_bytes_done += file.get_size()
                            files.add(pathlib.Path(file.get_key()))

                remote_project.set_size(total_bytes)
                remote_project.set_total_bytes_done_uploading(total_bytes_done)
                remote_project.set_file_list(files)

            if resp.status_code == 404:
                msg = u"404: {0} not found".format(url)
                self._log.error(msg)
                raise APIClientError(msg)

    def get_cached_root_directories(self):
        return list(map(lambda x: x.get_name(), self._remote_projects_cache.get_value()))

    def get_root_directories(self, ignore_cache=False):
        self._log.info(u"Getting root directories. ignore_cache=" + str(ignore_cache))

        if ignore_cache or not self._remote_projects_cache.is_cached():
            self.get_remote_projects(ignore_cache=True)

        return self.get_cached_root_directories()

    def get_remote_projects(self, ignore_cache=False):
        self._log.info(u"Getting Envoy Projects. ignore_cache=" + str(ignore_cache))

        if ignore_cache or not self._remote_projects_cache.is_cached():
            url = self._envoy_client.url + u'/projects'
            self._log.info(u"Fetching existing remote project directories from \"" + url + u"\"")

            try:
                import requests
                r = requests.get(url)
                json = r.json()
                projects = json.get(u"projects", [])

                envoy_projects = []

                for project in projects:
                    envoy_projects.append(EnvoyProject(project.get(u'name', u''),
                                                       project.get(u'deleting', False),
                                                       project.get(u'num_files', 0),
                                                       project.get(u'total_size', 0),
                                                       project.get(u'auto_downloading', False),
                                                       project.get(u'watched_list', [])))

                projects = sorted(envoy_projects, key=lambda s: s.get_name().lower())

            except Exception as e:
                self._log.error(u"Could not make request to envoy API. " + str(e))
                self.sign_out()
                projects = []

            self._remote_projects_cache.set_value(projects)

        remote_projects = self._remote_projects_cache.get_value()
        self._log.info(u"Returning " + str(len(remote_projects)) + u" Envoy Project(s).")
        return remote_projects

    def get_product_versions(self, product, ignore_cache=False):
        if ignore_cache or not self._product_resolver.is_cached():
            # get product resolver
            resolver = self._envoy_client.get_product_resolver()

            # cache all products
            self._product_resolver.set_value(resolver.get_all_types())

        # sorted by latest version to oldest version
        return sorted(self._product_resolver.get_value().get(product, {}).get(u"versions", []),
                      key=lambda s: s.lower(), reverse=True)

    def get_closest_matching_product_version(self, product, product_version, ignore_cache=False):
        product_versions = self.get_product_versions(product, ignore_cache=ignore_cache)

        if not len(product_versions):
            return None

        # first check to see if the provided version exactly matches a version option
        for version_option in product_versions:
            if version_option == product_version:
                return version_option

        # if not then we are down to heuristics
        import difflib
        return difflib.get_close_matches(product_version, product_versions)[0]

    def get_remote_project_files(self, project_name, ignore_cache=False):
        """ Returns a list of Envoy File objects, if project does not exist it returns an empty list. """

        if ignore_cache or project_name not in self._project_files_dictionary_cache:
            import requests

            try:
                r = requests.get(self._envoy_client.url + u'/project/' + project_name + u'/Files')
            except Exception as e:
                raise APIClientError(str(e))

            json = r.json()
            all_files = json.get(u'project_files', {}).get(u'all_files', [])

            envoy_files = []
            for file in all_files:
                envoy_files.append(EnvoyFile(
                    file.get(u'CheckSum', u''),
                    file.get(u'LastModified', u''),
                    file.get(u'Name', u''),
                    file.get(u'Key', u''),
                    file.get(u'Size', 0),
                    file.get(u'FmtName', u''),
                ))

            self._project_files_dictionary_cache[project_name] = envoy_files

        return self._project_files_dictionary_cache[project_name]

    def clear_project_files_cache(self):
        self._project_files_dictionary_cache.clear()

    def get_products(self, ignore_cache=False):
        self._log.info(u"Getting products list...")

        if not self.is_user_signed_in():
            msg = u"Must be signed-in to get products list."
            self._log.error(msg)
            raise NotSignedInError(msg)

        if ignore_cache or not self._products_cache.is_cached():
            url = u'{0}/products'.format(self._envoy_client.url)

            self._log.info(u"Fetching products list from " + url)

            try:
                resp = self._envoy_client.http_client.request('get', url)
            except Exception as e:
                msg = u"Could not fetch products list: " + get_exception_with_traceback(e)
                self._log.error(msg)
                raise APIClientError(msg)
            else:
                if resp.status_code == 200:
                    content = resp.json()

                    def parse_product(json):
                        return Product(json.get(u'id', -1),
                                       json.get(u'name', u''),
                                       json.get(u'app_type', u''),
                                       json.get(u'version', u''))

                    self._log.info(u"Caching fetched products...")
                    self._products_cache.set_value(list(map(lambda x: parse_product(x), content)))

                if resp.status_code == 404:
                    msg = u"404: {0} not found".format(url)
                    self._log.error(msg)
                    raise APIClientError(msg)

        cached_products = self._products_cache.get_value()
        self._log.info(u"Returning " + str(len(cached_products)) + u" cached products...")
        return cached_products

    def get_products_with_app_type(self, app_type, ignore_cache=False):
        products = self.get_products(ignore_cache=ignore_cache)
        return list(filter(lambda x: x.get_app_type() == app_type, products))

    def get_product_app_types(self, ignore_cache=False):
        return sorted(list(set(map(lambda x: x.get_app_type(), self.get_products(ignore_cache=ignore_cache)))),
                      key=lambda s: s.lower(), reverse=True)

    def get_user_info(self, ignore_cache: bool = False) -> UserInfo:
        self._log.info(u"Getting User info. ignore_cache=" + str(ignore_cache))

        if ignore_cache or not self._cached_user_info.is_cached():
            self._log.info(u"Fetching user info")

            try:
                user_info_raw = self._envoy_client.get_user_info()

                business_verticals = user_info_raw.get('business_verticals', [])
                created = user_info_raw.get('created', '')
                credits_available = user_info_raw.get('credits_available', 0)
                credits_consumed = user_info_raw.get('credits_consumed', 0)
                email_address = user_info_raw.get('email_address', '')
                job_throughput_limit = user_info_raw.get('job_throughput_limit', 30)
                max_throughput = user_info_raw.get('max_throughput', 9999)
                name = user_info_raw.get('name', '')
                plugin_versions = user_info_raw.get('plugin_versions', {})
                service_plan = user_info_raw.get('service_plan', '')
                username = user_info_raw.get('username', '')

                user_info = UserInfo(business_verticals,
                                     created,
                                     credits_available,
                                     credits_consumed,
                                     email_address,
                                     job_throughput_limit,
                                     max_throughput,
                                     name,
                                     plugin_versions,
                                     service_plan,
                                     username)

                self._log.info(u"Caching user info...")
                self._cached_user_info.set_value(user_info)

            except Exception as e:
                msg = u"Could not fetch user info: " + get_exception_with_traceback(e)
                self._log.error(msg)
                raise APIClientError(msg)

        self._log.info(u"Returning User info.")
        return self._cached_user_info.get_value()

    def get_machines(self,
                     app,
                     operation,
                     plugins=None,
                     ignore_cache=False):

        cached_machine_options = self._cached_machine_options.get_value()

        if plugins is None:
            plugins = []

        _app_cache_id = app + ''.join(plugins)

        if _app_cache_id in cached_machine_options:
            app_machine_options = cached_machine_options[_app_cache_id]
            if operation in app_machine_options:
                operation_machine_options = app_machine_options[operation]

                if not ignore_cache:
                    return operation_machine_options

        try:
            self._log.info(u"Fetching machine types for product \"" + app + u"\" (operation=\"" + operation +
                           u"\" plugins=[" + ','.join(plugins) + u"])")

            machine_options = self._envoy_client.get_machines(app,
                                                              operation,
                                                              gridmarkets.EnvoyClient.MACHINE_CATEGORY_ALL,
                                                              *plugins)

        except APIError as e:
            msg = str(e)
            self._log.error(msg)
            raise APIClientError(message=msg)

        def parse_machine_option(machine_option):
            return MachineOption(machine_option[u'id'], machine_option[u'name'], machine_option[u'is_default'], machine_option[u'gpu'])

        combined_machine_options = list(map(lambda x: parse_machine_option(x), machine_options))

        # It is possible no machine options were returned, in that case we can still let users choose between their
        # default machine options (cpu vs gpu)
        if len(combined_machine_options) <= 0:
            combined_machine_options.append(MachineOption(u"default_machine_cpu",
                                                          u"Default CPU Machine (If Supported)",
                                                          u"Use your default cpu machine (As defined on your profile page in the portal)",
                                                          False))
            combined_machine_options.append(MachineOption(u"default_machine_gpu",
                                                          u"Default GPU Machine (If Supported)",
                                                          u"Use your default gpu machine (As defined on your profile page in the portal)",
                                                          True))

        app_machine_options = cached_machine_options.get(_app_cache_id, {})
        app_machine_options[operation] = combined_machine_options
        cached_machine_options[_app_cache_id] = app_machine_options

        self._cached_machine_options.set_value(cached_machine_options)

        return combined_machine_options

    def get_machines_options(self,
                     app: str,
                     operation: str,
                     plugins: typing.Optional[typing.List[str]] = None,
                     ignore_cache: bool = False,
                     use_gpu_default: bool = False,
                     filter_cpu: bool = False,
                     filter_gpu: bool = False):

            items = []

            machines = self.get_machines(app, operation, plugins=plugins)

            index = 0
            default = 0
            default_machines = []
            for machine in machines:

                if filter_cpu and not machine.is_gpu_machine():
                    continue

                if filter_gpu and machine.is_gpu_machine():
                    continue

                items.append(EnumItem(machine.get_id(), machine.get_name(), u""))

                if machine.is_default():
                    default_machines.append((machine, index))

                index = index + 1

            # if at least one machine was marked as default
            if default_machines:

                # use the first default machine who's is_gpu_machine() getter matches the use_gpu_default key word argument
                for default_machine in default_machines:
                    if default_machine[0].is_gpu_machine() == use_gpu_default:
                        default = default_machine[1]
                        break
                else:  # if no mach is found then use the first default machine in the list
                    default = default_machines[0][1]

            return items, default

    def _is_gpu_enabled_machine(self, machine_id):

        # get all cached machine types
        cached_machine_options_dict = self._cached_machine_options.get_value()

        cached_machine_options = set()
        for app, operations in cached_machine_options_dict.items():
            for operation, machine_options in operations.items():
                cached_machine_options.update(machine_options)

        for machine_option in cached_machine_options:
            if machine_option.get_id() == machine_id:
                return machine_option.is_gpu_machine()

        raise ValueError(u"Could not find cached machine option with id '" + str(machine_id) + u"'")
