__all__ = 'RemoteProject', 'convert_packed_project'

from gm_meta_plugin.attribute import AttributeType
from gm_meta_plugin.attribute_types import StringSubtype
from gm_meta_plugin.remote_project import RemoteProject as MetaRemoteProject
from gm_meta_plugin.project import Project
from gm_meta_plugin.cached_value import CachedValue
from gm_meta_plugin.utils import get_external_files_remote_path

import pathlib
import typing

if typing.TYPE_CHECKING:
    from gm_meta_plugin.packed_project import PackedProject


class RemoteProject(MetaRemoteProject):

    def __init__(self,
                 name,
                 root_dir,
                 files,
                 attributes,
                 api_client):

        Project.__init__(self,
                         name,
                         root_dir,
                         files,
                         attributes)

        self._api_client = api_client
        self._cached_size = CachedValue(0)
        self._cached_total_uploaded_bytes_done_uploading = CachedValue(0)
        self.update_project_status()

    def get_api_client(self):
        return self._api_client

    def get_files(self, force_update=False):
        if force_update:
            self.get_api_client().update_remote_project_status(self)

        return Project.get_files(self)

    def set_file_list(self, files):
        self._files = files

    def update_project_status(self):
        self.get_api_client().update_remote_project_status(self)

    def set_name(self, name):
        raise NotImplementedError

    def exists(self):
        raise NotImplementedError

    def delete(self):
        raise NotImplementedError

    def get_size(self, force_update=False):

        if force_update or not self._cached_size.is_cached():
            self.get_api_client().update_remote_project_status(self)

        return self._cached_size.get_value()

    def set_size(self, size):
        self._cached_size.set_value(size)

    def get_total_bytes_done_uploading(self, force_update = False):

        if force_update or not self._cached_total_uploaded_bytes_done_uploading.is_cached():
            self.get_api_client().update_remote_project_status(self)

        return self._cached_total_uploaded_bytes_done_uploading.get_value()

    def set_total_bytes_done_uploading(self, bytes):
        self._cached_total_uploaded_bytes_done_uploading.set_value(bytes)

    def get_remaining_bytes_to_upload(self, force_update = False):

        if force_update or \
                not self._cached_size.is_cached() or \
                not self._cached_total_uploaded_bytes_done_uploading.is_cached():

            self.get_api_client().update_remote_project_status(self)

        return self.get_size() - self.get_total_bytes_done_uploading()


def convert_packed_project(packed_project: 'PackedProject', plugin):
    root_dir = pathlib.Path(packed_project.get_name())

    def convert_local_file_to_remote_file(file):

        # External files
        if packed_project.get_root_dir() not in file.parents:
            return get_external_files_remote_path(str(pathlib.Path('/') / root_dir), file)

        return pathlib.Path('/') / root_dir / packed_project.get_relative_file_path(file)

    # convert files from absolute local paths to absolute remote paths
    files = set(map(convert_local_file_to_remote_file, packed_project.get_files()))

    # do the same to all path type project attributes
    attributes = {}
    attributes.update(packed_project.get_attributes())

    project_attribute = plugin.get_api_client().get_cached_api_schema().get_project_definition_schema().get_root_project_attribute()
    attribute = project_attribute.get_attribute()

    while attribute.get_type() != AttributeType.NULL:
        value = packed_project.get_attribute(attribute.get_key())

        if attribute.get_type() == AttributeType.STRING and attribute.get_subtype() == StringSubtype.PATH.value:
            # update the file path attribute
            value = convert_local_file_to_remote_file(pathlib.Path(value)).as_posix()
            attributes[attribute.get_key()] = value

        project_attribute = project_attribute.transition(value)
        attribute = project_attribute.get_attribute()

    remote_project = RemoteProject(packed_project.get_name(),
                                   root_dir,
                                   files,
                                   attributes,
                                   plugin.get_api_client())

    remote_project.set_watch_patterns(packed_project.get_watch_patterns().copy())
    return remote_project