__all__ = 'XMLAPISchemaParser'

from . import constants
from .project_definition_schema import ProjectDefinitionSchema

from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.job_definition import JobDefinition
from gm_meta_plugin.attribute import AttributeType
from gm_meta_plugin.attribute_types import *
from gm_meta_plugin.job_attribute import JobAttribute
from gm_meta_plugin.inference_source import InferenceSource
from gm_meta_plugin.transition import Transition

import xml.etree.ElementTree as ET
import pathlib
import typing

if typing.TYPE_CHECKING:
    from gm_meta_plugin.api_schema import APISchema


def get_all_sub_elements(element, tag_name, expect_at_least_one = False):
    elements = element.findall(tag_name)

    if expect_at_least_one and len(elements) < 1:
        raise ValueError(u"The <" + element.tag + u"> element must contain at least one <" +
                         str(tag_name) + u"> child element")

    return elements


def get_sub_element(element,
                    tag_name,
                    expect_unique_tag = True):

    elements = element.findall(tag_name)

    if len(elements) == 0:
        raise ValueError(u"The <" + element.tag + u"> element must contain exactly one <" + tag_name +
                         u"> child element")

    if expect_unique_tag and len(elements) != 1:
        raise ValueError(u"The <" + element.tag + u"> element must contain exactly one <" + tag_name +
                         u"> child element")

    return elements[0]


def get_attribute(element, attribute_key):

    if attribute_key not in element.attrib:
        raise ValueError(u"All <" + element.tag + u"> tags must contain an '" + attribute_key +
                         u"' attribute")

    return element.attrib[attribute_key]


def get_all_sub_elements_text(element, tag_name, expect_at_least_one = True):
    elements = get_all_sub_elements(element, tag_name, expect_at_least_one)
    return list(map(lambda x: x.text, elements))


def get_text(element, tag_name, expect_unique_tag = True):
    element = get_sub_element(element, tag_name, expect_unique_tag)

    if element.text is None:
        return u""

    return element.text


def get_text_optional(element, tag_name) -> typing.Optional[str]:
    elements = element.findall(tag_name)

    if len(elements) <= 0:
        return None

    if len(elements) > 1:
        raise ValueError(u"More than one optional '" + tag_name + u"' tag.")

    if elements[0].text is None:
        return None

    return elements[0].text


def get_optional_number(element, tag_name):
    text = get_text_optional(element, tag_name)

    if text is None:
        return None

    return int(text)


def to_bool(value):
    if value == u"True":
        return True
    elif value == u"False":
        return False
    else:
        raise ValueError(u"Could not convert value: " + value +
                         u" to python bool. Expects either 'True' or 'False'.")


SCHEMA_DEFINITION_FILE = pathlib.Path(__file__).parent / u"APISchema.xml"

TAG_API_SCHEMA = u"APISchema"
TAG_JOB_DEFINITIONS = u"JobDefinitions"
TAG_JOB_DEFINITION = u"JobDefinition"
TAG_ATTRIBUTE_GROUPS = u"AttributeGroups"
TAG_ATTRIBUTE_GROUP = u"AttributeGroup"
ID_ATTRIBUTE = u"id"
TAG_DISPLAY_NAME = u"DisplayName"
TAG_DESCRIPTION = u"Description"
TAG_TYPE = u"Type"
TAG_PLACEHOLDER = u"Placeholder"
TAG_SUBTYPE = u"Subtype"
TAG_ITEMS = u"Items"
TAG_KEY = u"Key"
TAG_VALUE = u"Value"
TAG_ATTRIBUTE = u"Attribute"
TAG_JOB_ATTRIBUTE = u"JobAttribute"
TAG_INFERENCE_SOURCES = u"InferenceSources"
TAG_INFERENCE_SOURCE = u"InferenceSource"
TAG_ATTRIBUTE_KEY_ATTRIBUTE = u"key"
TAG_PROJECT_ATTRIBUTES = u"ProjectAttributes"
TAG_PROJECT_ATTRIBUTE = u"ProjectAttribute"
TAG_DEFAULT_VALUE = u"DefaultValue"
TAG_APP_INFER_DEFAULT_VALUE = u"AppInferDefaultValue"
TAG_TRANSITIONS = u"Transitions"
TAG_TRANSITION = u"Transition"
TAG_PROJECT_ATTRIBUTE_ID = u"ProjectAttributeId"
ATTRIBUTE_ID = u"id"
TAG_TRANSITION_FORMULA = u"TransitionFormula"
TAG_REJECTION_MESSAGE = u"RejectionMessage"
TAG_COMPATIBLE_JOB_DEFINITIONS = u"CompatibleJobDefinitions"
TAG_COMPATIBLE_JOB_DEFINITION = u"CompatibleJobDefinition"
TAG_JOB_DEFINITION_ID = u"JobDefinitionId"
TAG_MAX_LENGTH = u"MaxLength"
TAG_MIN_LENGTH = u"MinLength"
TAG_SUBTYPE_KWARGS = u"SubtypeKwargs"
TAG_SUBTYPE_KWARG = u"SubtypeKwarg"
TAG_SKIP = u"Skip"
TAG_OPTIONAL = u"Optional"
TAG_PRECONDITION_EVALUATOR = "PreconditionEvaluator"
TAG_DEPENDENCIES = "Dependencies"
TAG_DEPENDENCY = "Dependency"
TAG_LIST_TYPE = "ListType"
TAG_ITEMS_INFERENCE_SOURCE = "ItemsInferenceSource"
TAG_INHERIT = "Inherit"
TAG_PARENT = "Parent"


class XMLAPISchemaParser:
    """ Reads the xml api schema file and builds a APISchema class instance """

    @staticmethod
    def _parse_enum_items(enum_items_element):
        # instantiate the output list
        enum_items = []

        # parse each enum item
        for enum_item_element in enum_items_element:
            enum_item_key = get_text(enum_item_element, TAG_KEY)
            enum_item_display_name = get_text(enum_item_element, TAG_DISPLAY_NAME)
            enum_item_description = get_text(enum_item_element, TAG_DESCRIPTION)
            enum_items.append(EnumItem(enum_item_key, enum_item_display_name, enum_item_description))

        return enum_items

    @staticmethod
    def _parse_subtype_kwargs(attribute_element):
        subtype_kwargs_attributes = get_all_sub_elements(attribute_element, TAG_SUBTYPE_KWARGS)

        if len(subtype_kwargs_attributes) > 1:
            raise ValueError("The <" + attribute_element.tag +
                             "> element must contain at most one <" + TAG_SUBTYPE_KWARGS + "> child element")

        subtype_kwargs = {}

        if len(subtype_kwargs_attributes):
            subtype_kwargs_attribute = subtype_kwargs_attributes[0]
            subtype_kwarg_attributes = get_all_sub_elements(subtype_kwargs_attribute, TAG_SUBTYPE_KWARG)
            for subtype_kwarg_attribute in subtype_kwarg_attributes:
                key = get_text(subtype_kwarg_attribute, TAG_KEY)
                value = get_text(subtype_kwarg_attribute, TAG_VALUE)
                subtype_kwargs[key] = value

        return subtype_kwargs

    @staticmethod
    def _parse_attribute(attribute_element):
        attribute_key = get_text(attribute_element, TAG_KEY, expect_unique_tag=True)
        attribute_display_name = get_text(attribute_element, TAG_DISPLAY_NAME, expect_unique_tag=True)
        attribute_description = get_text(attribute_element, TAG_DESCRIPTION, expect_unique_tag=True)
        attribute_type = get_text(attribute_element, TAG_TYPE, expect_unique_tag=True)
        attribute_default_value = get_text_optional(attribute_element, TAG_DEFAULT_VALUE)
        attribute_subtype = get_text_optional(attribute_element, TAG_SUBTYPE)
        attribute_placeholder = get_text_optional(attribute_element, TAG_PLACEHOLDER)
        attribute_subtype_kwargs = XMLAPISchemaParser._parse_subtype_kwargs(attribute_element)

        if attribute_type == AttributeType.STRING.value:

            max_length = get_optional_number(attribute_element, TAG_MAX_LENGTH)
            min_length = get_optional_number(attribute_element, TAG_MIN_LENGTH)

            return StringAttributeType(attribute_key,
                                       attribute_display_name,
                                       attribute_description,
                                       attribute_subtype_kwargs,
                                       default_value=attribute_default_value,
                                       subtype=attribute_subtype,
                                       placeholder=attribute_placeholder,
                                       max_length=max_length,
                                       min_length=min_length)

        elif attribute_type == AttributeType.ENUM.value:

            if attribute_subtype == constants.ENUM_SUBTYPE_PRODUCT_VERSIONS:
                enum_items = []
            elif attribute_subtype == EnumSubtype.MACHINE_TYPE.value:
                enum_items = []
            elif attribute_subtype == EnumSubtype.APP_INFERRED.value:
                enum_items = []
            else:
                enum_items_element = get_sub_element(attribute_element, TAG_ITEMS)
                enum_items = XMLAPISchemaParser._parse_enum_items(enum_items_element)

            return EnumAttributeType(attribute_key,
                                     attribute_display_name,
                                     attribute_description,
                                     attribute_subtype_kwargs,
                                     enum_items,
                                     default_value=attribute_default_value,
                                     subtype=attribute_subtype)

        elif attribute_type == AttributeType.NULL.value:
            return NullAttributeType(attribute_key,
                                     attribute_display_name,
                                     attribute_description,
                                     attribute_subtype_kwargs)

        elif attribute_type == AttributeType.BOOLEAN.value:
            return BooleanAttributeType(attribute_key,
                                        attribute_display_name,
                                        attribute_description,
                                        attribute_subtype_kwargs,
                                        default_value=to_bool(attribute_default_value))

        elif attribute_type == AttributeType.INTEGER.value:
            if attribute_default_value is not None:
                attribute_default_value = int(attribute_default_value)

            return IntegerAttributeType(attribute_key,
                                        attribute_display_name,
                                        attribute_description,
                                        attribute_subtype_kwargs,
                                        default_value=attribute_default_value,
                                        subtype=attribute_subtype)
        elif attribute_type == AttributeType.FLOAT.value:

            if attribute_default_value is not None:
                attribute_default_value = float(attribute_default_value)

            return FloatAttributeType(attribute_key,
                                        attribute_display_name,
                                        attribute_description,
                                        attribute_subtype_kwargs,
                                        default_value=attribute_default_value)

        elif attribute_type == AttributeType.LIST.value:

            if attribute_default_value is not None:
                assert False, "Default values are currently not supported for list types."

            list_type = get_text(attribute_element, TAG_LIST_TYPE, expect_unique_tag=True)

            if list_type not in set(item.value for item in AttributeType):
                assert False, "List type '{}' unrecognised".format(list_type)

            items_inference_source = InferenceSource.get_inference_source(get_text(attribute_element,
                                                                                   TAG_ITEMS_INFERENCE_SOURCE,
                                                                                   expect_unique_tag=True))

            if items_inference_source == InferenceSource.get_user_defined_inference_source():
                items_element = get_sub_element(attribute_element, TAG_ITEMS)
                enum_items = XMLAPISchemaParser._parse_enum_items(items_element)
            else:
                enum_items = []

            return ListAttributeType(attribute_key,
                                     attribute_display_name,
                                     attribute_description,
                                     attribute_subtype_kwargs,
                                     items_inference_source,
                                     enum_items,
                                     list_type,
                                     default_value=attribute_default_value,
                                     subtype=attribute_subtype)
        else:
            raise ValueError("Unrecognised attribute type.")

    @staticmethod
    def _parse_job_inference_sources(inference_sources_element):

        attribute_inference_sources = get_all_sub_elements_text(inference_sources_element, TAG_INFERENCE_SOURCE)

        return list(map(lambda x: InferenceSource.get_inference_source(x), attribute_inference_sources))

    @staticmethod
    def _parse_job_attribute(job_attribute_element) -> 'JobAttribute':
        attribute_element = get_sub_element(job_attribute_element, TAG_ATTRIBUTE, expect_unique_tag=True)
        attribute = XMLAPISchemaParser._parse_attribute(attribute_element)

        try:
            job_attribute_id = get_attribute(job_attribute_element, ATTRIBUTE_ID)
        except ValueError:
            job_attribute_id = attribute.get_key()

        attribute_group = get_text_optional(job_attribute_element, TAG_ATTRIBUTE_GROUP)
        inference_sources_element = get_sub_element(job_attribute_element, TAG_INFERENCE_SOURCES)
        attribute_inference_sources = XMLAPISchemaParser._parse_job_inference_sources(inference_sources_element)
        precondition_evaluator_str = get_text_optional(job_attribute_element, TAG_PRECONDITION_EVALUATOR)


        try:
            dependencies_element = get_sub_element(job_attribute_element, TAG_DEPENDENCIES, expect_unique_tag=True)
            dependencies = get_all_sub_elements_text(dependencies_element, TAG_DEPENDENCY)
        except ValueError:
            dependencies = []

        def default_precondition_evaluator_func(job_preset_attribute: 'JobPresetAttribute',
                                                job_preset: 'JobPreset') -> bool:
            return True

        precondition_evaluator_func = default_precondition_evaluator_func
        if precondition_evaluator_str:

            scope = {}
            exec(precondition_evaluator_str, scope)

            if 'precondition_evaluator' not in scope:
                raise ValueError('"precondition_evaluator" not in locals()')

            precondition_evaluator = scope['precondition_evaluator']

            # noinspection PyUnresolvedReferences
            if not callable(precondition_evaluator):
                raise ValueError('"precondition_evaluator" is not callable')

            # noinspection PyUnresolvedReferences
            precondition_evaluator_func = precondition_evaluator

        attribute_skip = get_text_optional(job_attribute_element, TAG_SKIP)
        if attribute_skip is None:
            attribute_skip = 'False'
        attribute_skip = to_bool(attribute_skip)

        if attribute.get_key() in api_constants.ALWAYS_SKIPPED_API_KEYS and not attribute_skip:
            raise ValueError("Attribute '{}' must be marked as skip for job attribute '{}'.".format(
                attribute.get_key(), job_attribute_id))

        try:
            attribute_optional_elements = get_all_sub_elements(job_attribute_element, TAG_OPTIONAL, expect_at_least_one=True)

            if len(attribute_optional_elements) > 1:
                raise ValueError("To many <Optional> elements for attribute '{}'.".format(job_attribute_id))

            attribute_optional_element = attribute_optional_elements[0]
            attribute_optional_value = to_bool(get_text(attribute_optional_element, TAG_VALUE, expect_unique_tag=True))

            if attribute_optional_element.findall(TAG_DEFAULT_VALUE):
                attribute_optional_default_enabled_state = to_bool(get_text(attribute_optional_element, TAG_DEFAULT_VALUE, expect_unique_tag=True))
                attribute_optional_parent = None
            elif attribute_optional_element.findall(TAG_PARENT):
                attribute_optional_default_enabled_state = None
                attribute_optional_parent = get_text(attribute_optional_element, TAG_PARENT, expect_unique_tag=True)
            else:
                raise ValueError()

        except ValueError:
            attribute_optional_value = False
            attribute_optional_default_enabled_state = True
            attribute_optional_parent = None

        return JobAttribute(job_attribute_id,
                            attribute,
                            attribute_group,
                            attribute_inference_sources,
                            attribute_skip,
                            attribute_optional_value,
                            precondition_evaluator_func,
                            dependencies,
                            default_enabled_state=attribute_optional_default_enabled_state,
                            optional_parent=attribute_optional_parent)

    @staticmethod
    def _parse_job_definition(job_definition_element) -> 'JobDefinition':

        if ID_ATTRIBUTE not in job_definition_element.attrib:
            raise ValueError("All <" + TAG_JOB_DEFINITION + "> tags must contain an 'id' attribute")

        job_definition_id = get_attribute(job_definition_element, ID_ATTRIBUTE)
        job_definition_display_name = get_text(job_definition_element, TAG_DISPLAY_NAME)

        job_definition_attribute_groups_elements = get_all_sub_elements(job_definition_element, TAG_ATTRIBUTE_GROUPS)
        if job_definition_attribute_groups_elements:
            job_definition_attribute_groups = get_all_sub_elements_text(job_definition_attribute_groups_elements[0], TAG_ATTRIBUTE_GROUP)
        else:
            job_definition_attribute_groups = []

        attribute_elements = get_all_sub_elements(job_definition_element, TAG_JOB_ATTRIBUTE, expect_at_least_one=True)

        attributes: typing.List['JobAttribute'] = []
        for attribute_element in attribute_elements:
            job_attribute = XMLAPISchemaParser._parse_job_attribute(attribute_element)

            # Check the new job attribute doesn't have the same id as an existing job attribute
            for attribute in attributes:
                if attribute.get_id() == job_attribute.get_id():
                    raise ValueError("Two job attributes can not have the same id ('{}')".format(attribute.get_id()))

            attributes.append(job_attribute)

        job_definition_inherited_job_definition_ids = get_all_sub_elements_text(job_definition_element,
                                                                                TAG_INHERIT,
                                                                                expect_at_least_one=False)

        job_definition = JobDefinition(job_definition_id,
                                       job_definition_display_name,
                                       job_definition_attribute_groups,
                                       attributes,
                                       job_definition_inherited_job_definition_ids)

        return job_definition

    @staticmethod
    def _parse_job_definitions(api_schema_element, factory_collection) -> typing.List['JobDefinition']:

        # instantiate the output list
        job_definitions = dict()

        # parse job definitions
        job_definitions_element = get_sub_element(api_schema_element, TAG_JOB_DEFINITIONS)
        job_definition_elements = get_all_sub_elements(job_definitions_element, TAG_JOB_DEFINITION)

        for job_definition_element in job_definition_elements:
            job_definition = XMLAPISchemaParser._parse_job_definition(job_definition_element)
            job_definitions[job_definition.get_definition_id()] = job_definition

        for job_definition_id, job_definition in job_definitions.items():
            # Get the job definitions to inherit
            inherited_job_definitions = list(map(lambda x: job_definitions[x],
                                                 job_definition.get_inherited_job_definition_ids()))

            job_definition._set_inherited_job_definitions(inherited_job_definitions)

        return list(job_definitions.values())

    @staticmethod
    def _parse_transition(transition_element, project_attributes):

        transition_project_attribute_id = get_text(transition_element, TAG_PROJECT_ATTRIBUTE_ID)

        for project_attribute in project_attributes:
            if project_attribute.get_id() == transition_project_attribute_id:
                transition_project_attribute = project_attribute
                break
        else:
            raise ValueError("No ProjectAttribute with id '" + transition_project_attribute_id + "' found.")

        transition_formula = get_text(transition_element, TAG_TRANSITION_FORMULA)
        rejection_message = get_text_optional(transition_element, TAG_REJECTION_MESSAGE)

        return Transition(transition_formula, transition_project_attribute, rejection_message=rejection_message)

    @staticmethod
    def _parse_transitions(transitions_element, project_attributes):

        # instantiate the output list
        transitions = []

        transition_elements = get_all_sub_elements(transitions_element, TAG_TRANSITION)
        for transition_element in transition_elements:
            transitions.append(XMLAPISchemaParser._parse_transition(transition_element, project_attributes))

        return transitions

    @staticmethod
    def _parse_compatible_job_definitions(compatible_job_definitions_element,
                                          job_definitions):
        # instantiate the output list
        compatible_job_definitions = []

        job_definition_ids = get_all_sub_elements_text(compatible_job_definitions_element,
                                                       TAG_COMPATIBLE_JOB_DEFINITION,
                                                       False)

        for job_definition_id in job_definition_ids:
            for job_definition in job_definitions:
                if job_definition_id == job_definition.get_definition_id():
                    compatible_job_definitions.append(job_definition)

        return compatible_job_definitions

    @staticmethod
    def _parse_project_attribute(project_attribute_element,
                                 project_attributes,
                                 job_definitions,
                                 factory_collection):

        project_attribute_id = get_attribute(project_attribute_element, ATTRIBUTE_ID)

        attribute_element = get_sub_element(project_attribute_element, TAG_ATTRIBUTE, expect_unique_tag=True)
        attribute = XMLAPISchemaParser._parse_attribute(attribute_element)

        # parse transitions
        transitions_element = get_sub_element(project_attribute_element, TAG_TRANSITIONS)
        project_attribute_transitions = XMLAPISchemaParser._parse_transitions(transitions_element, project_attributes)

        # parse compatible job definitions
        compatible_job_definitions_element = get_sub_element(project_attribute_element,
                                                             TAG_COMPATIBLE_JOB_DEFINITIONS)
        compatible_job_definitions = XMLAPISchemaParser._parse_compatible_job_definitions(
            compatible_job_definitions_element, job_definitions)

        factory = factory_collection.get_project_attribute_factory()

        return factory.get_project_attribute(project_attribute_id,
                                             attribute,
                                             project_attribute_transitions,
                                             compatible_job_definitions)

    @staticmethod
    def _parse_project_attributes(api_schema_element,
                                  job_definitions,
                                  factory_collection):

        # instantiate the output list
        project_attributes = []

        # parse project attributes
        project_attributes_element = get_sub_element(api_schema_element, TAG_PROJECT_ATTRIBUTES)
        project_attribute_elements = get_all_sub_elements(project_attributes_element, TAG_PROJECT_ATTRIBUTE)

        for project_attribute_element in project_attribute_elements:
            project_attributes.append(XMLAPISchemaParser._parse_project_attribute(project_attribute_element,
                                                                                  project_attributes,
                                                                                  job_definitions,
                                                                                  factory_collection))

        return project_attributes

    @staticmethod
    def parse(plugin, factory_collection) -> 'APISchema':

        # parse the xml file using xml.etree.ElementTree
        tree = ET.parse(SCHEMA_DEFINITION_FILE)
        api_schema_element = tree.getroot()

        if api_schema_element.tag != TAG_API_SCHEMA:
            raise ValueError(u"API schema file must have a <" + TAG_API_SCHEMA + u"> tag as the root")

        job_definitions = XMLAPISchemaParser._parse_job_definitions(api_schema_element, factory_collection)
        project_attributes = XMLAPISchemaParser._parse_project_attributes(api_schema_element, job_definitions,
                                                                          factory_collection)

        project_definition_schema = ProjectDefinitionSchema(plugin, project_attributes)

        return factory_collection.get_api_schema_factory().get_api_schema(plugin,
                                                                          job_definitions,
                                                                          project_definition_schema)
