__all__ = ['JobAttribute']

import typing

if typing.TYPE_CHECKING:
    from gm_meta_plugin.attribute import Attribute
    from gm_meta_plugin.job_definition import JobDefinition
    from gm_meta_plugin.job_preset import JobPreset
    from gm_meta_plugin.job_preset_attribute import JobPresetAttribute


class JobAttribute:

    def __init__(self,
                 job_attribute_id: str,
                 attribute: 'Attribute',
                 attribute_group: str,
                 inference_sources,
                 skip: bool,
                 optional: bool,
                 precondition_evaluator: typing.Callable[['JobPresetAttribute', 'JobPreset'], bool],
                 dependencies: typing.List['str'],

                 default_enabled_state: bool = None,
                 optional_parent: str = None):

        if inference_sources is None or len(inference_sources) < 1:
            raise ValueError("JobAttribute has no inference sources")

        self._id = job_attribute_id
        self._job_definition = None
        self._attribute = attribute
        self._attribute_group = attribute_group
        self._inference_sources = inference_sources
        self._skip = skip

        if default_enabled_state is None and optional_parent is None:
            raise ValueError("Neither a default enabled state or optional parent have been provided.")

        if default_enabled_state is not None and optional_parent is not None:
            raise ValueError("Both default enabled state and optional parent have been provided. Only one should be passed")

        self._optional = optional
        self._default_enabled_state = default_enabled_state
        self._optional_parent = optional_parent

        self._precondition_evaluator = precondition_evaluator
        self._dependencies = dependencies

    def get_id(self) -> str:
        return self._id

    def get_job_definition(self) -> 'JobDefinition':
        return self._job_definition

    def _set_job_definition(self, job_definition: 'JobDefinition') -> None:
        """
        Internal setter so that the JobAttribute's parent JobDefinition can set. The JobDefinition and
        JobAttribute classes depend on each other which is why this extra setter is necessary.
        """
        self._job_definition = job_definition

    def get_attribute(self) -> 'Attribute':
        return self._attribute

    def get_attribute_group(self):
        return self._attribute_group

    def get_inference_sources(self):
        return self._inference_sources

    def set_inference_sources(self, inference_sources):
        self._inference_sources = inference_sources

    def get_default_inference_source(self):
        try:
            default = self.get_inference_sources()[0]
        except IndexError as e:
            raise e

        return self.get_inference_sources()[0]

    def should_skip(self) -> bool:
        return self._skip

    def is_optional(self) -> bool:
        return self._optional

    def default_enabled_state(self) -> bool:
        if self._optional_parent is not None:
            return self.get_job_definition().get_attribute_with_id(self._optional_parent).default_enabled_state()
        return self._default_enabled_state

    def get_optional_parent(self) -> typing.Optional[str]:
        return self._optional_parent

    def get_precondition_evaluator(self) -> typing.Callable[['JobPresetAttribute', 'JobPreset'], bool]:
        """
        A precondition evaluator is something that is required to evaluate to true in order for the job preset attribute
        to be considered 'active'.
        """
        return self._precondition_evaluator

    def get_dependencies_ids(self) -> typing.List['str']:
        return self._dependencies

    def get_dependencies(self) -> typing.List['JobAttribute']:
        dependencies = []
        job_definition = self.get_job_definition()

        for dependency_id in self.get_dependencies_ids():
            dependencies.append(job_definition.get_attribute_with_id(dependency_id))

        return dependencies
