__all__ = 'JobDefinition'

import typing

if typing.TYPE_CHECKING:
    from gm_meta_plugin.job_attribute import JobAttribute


class JobDefinition:

    def __init__(self, id: str,
                 display_name: str,
                 attribute_groups: typing.List['str'],
                 attributes: typing.List['JobAttribute'],
                 inherited_job_definition_ids: typing.List['str']):

        self._id = id
        self._attribute_groups = attribute_groups
        self._display_name = display_name
        self._attributes = attributes
        self._inherited_job_definition_ids = inherited_job_definition_ids
        self._inherited_job_definitions = None  # Used to cache the job definition references

        for attribute in attributes:
            attribute._set_job_definition(self)

    def get_definition_id(self) -> str:
        return self._id

    def get_display_name(self) -> str:
        return self._display_name

    def get_attribute_groups(self)  -> typing.List['str']:
        attribute_groups = self._attribute_groups.copy()

        for inherited_job_definition in self.get_inherited_job_definitions():
            for inherited_job_attribute_group in reversed(inherited_job_definition.get_attribute_groups()):
                if inherited_job_attribute_group not in attribute_groups:
                    attribute_groups = [inherited_job_attribute_group] + attribute_groups

        return attribute_groups

    def get_attributes(self) -> typing.List['JobAttribute']:
        job_attributes = self._attributes.copy()

        for inherited_job_definition in self.get_inherited_job_definitions():
            inherited_job_attributes = []
            for inherited_job_attribute in inherited_job_definition.get_attributes():

                # Inherit job attributes with ids that are not already part of this job definition
                if not any(y.get_id() == inherited_job_attribute.get_id() for y in job_attributes):
                    inherited_job_attributes.append(inherited_job_attribute)

            job_attributes = inherited_job_attributes + job_attributes

        return job_attributes

    def get_inherited_job_definition_ids(self) -> typing.List[str]:
        return self._inherited_job_definition_ids

    def get_inherited_job_definitions(self) -> typing.List['JobDefinition']:
        if self._inherited_job_definitions is None:
            assert False, "Attempting to get the inherited job definitions before they have been set."

        return self._inherited_job_definitions

    def _set_inherited_job_definitions(self, inherited_job_definitions: typing.List['JobDefinition']):
        """
        Called from the xml schema parser
        """
        self._inherited_job_definitions = inherited_job_definitions

        # Check for circular dependencies
        for attribute in self._attributes:
            dependencies = attribute.get_dependencies()[:]
            seen_dependencies = []

            while dependencies:
                dependency = dependencies.pop(0)

                if dependency == attribute:
                    raise ValueError("Circular dependency chain found '{}' -> '{}'".format(attribute.get_id(), dependency.get_id()))

                if dependency in seen_dependencies:
                    continue

                seen_dependencies.append(dependency)
                for dep in dependency.get_dependencies():
                    dependencies.append(dep)

        # Validate that the job attributes belong to an attribute group that exists
        attribute_groups = self.get_attribute_groups()
        for attribute in self.get_attributes():
            attribute_group = attribute.get_attribute_group()
            if (not attribute_group is None) and attribute_group not in attribute_groups:
                raise ValueError("Job Attribute '{}' belongs to a job attribute group that does not "
                                 "exist ('{}')".format(attribute.get_id(), attribute_group))


    def get_attribute_with_id(self, id: str) -> 'JobAttribute':
        for job_attribute in self.get_attributes():
            if job_attribute.get_id() == id:
                return job_attribute
        else:
            raise ValueError("Could not find attribute with id '" + id + "'")

    def get_attributes_with_key(self, key: str) -> typing.List['JobAttribute']:
        job_attributes = []
        for job_attribute in self.get_attributes():
            if job_attribute.get_attribute().get_key() == key:
                job_attributes.append(job_attribute)
        return job_attributes

    def get_attributes_in_group(self, attribute_group) -> typing.List['JobAttribute']:
        attributes = []
        for attribute in self.get_attributes():
            if attribute.get_attribute_group() == attribute_group:
                attributes.append(attribute)
        return attributes
