__all__ = 'StaticJobPreset'

from .job_preset import JobPreset
from .static_job_preset_attribute import StaticJobPresetAttribute
from .gridmarkets import constants as api_constants

import typing

if typing.TYPE_CHECKING:
    from .job_definition import JobDefinition
    from .plugin import Plugin


class StaticJobPreset(JobPreset):

    def __init__(self,
                 plugin: 'Plugin',
                 name: str,
                 id: str,
                 job_definition: 'JobDefinition',
                 job_preset_attributes: typing.List['StaticJobPresetAttribute']):

        JobPreset.__init__(self, plugin, name, id, job_definition)
        self._job_preset_attributes = job_preset_attributes

    def get_job_preset_attributes(self) -> typing.List['StaticJobPresetAttribute']:
        return self._job_preset_attributes

    def _set_job_preset_attributes(self, job_preset_attributes: typing.List['StaticJobPresetAttribute']) -> None:
        self._job_preset_attributes = job_preset_attributes

    def get_job_preset_attribute_with_job_attribute_id(self, id: str) -> 'StaticJobPresetAttribute':
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_job_attribute().get_id() == id:
                return job_preset_attribute

        raise ValueError(u"No JobPresetAttribute with JobAttribute id: '{}'".format(str(id)))

    def get_job_preset_attribute_by_key(self,
                                        key: str,
                                        allow_multiple: bool = False) -> 'StaticJobPresetAttribute':

        result = None

        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_key() == key:

                if allow_multiple:
                    return job_preset_attribute

                if result is not None:
                    raise ValueError("Multiple JobPresetAttributes with the key: " + str(key))

                result = job_preset_attribute

        if result:
            return result

        raise ValueError(u"No JobPresetAttribute with the key: " + str(key))

    def get_job_preset_attribute_by_id(self, id) -> 'StaticJobPresetAttribute':
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_id() == id:
                return job_preset_attribute

        raise ValueError(u"No JobPresetAttribute with the id: " + str(id))

    def get_job_preset_attributes_in_group(self, attribute_group) -> typing.List['StaticJobPresetAttribute']:
        job_preset_attributes = []
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.get_job_attribute().get_attribute_group() == attribute_group:
                job_preset_attributes.append(job_preset_attribute)
        return job_preset_attributes

    def get_job_settings(self) -> typing.Dict['str', typing.Any]:
        """
        Stores the value of every attribute with a value in a dictionary indexed by each attributes key.
        """

        job_settings = {}

        for static_job_preset_attribute in self.get_job_preset_attributes():

            # Skip attributes which are marked as skip
            if static_job_preset_attribute.should_skip():
                continue

            # Skip attributes which have no value
            if not static_job_preset_attribute.has_value():
                continue

            key = static_job_preset_attribute.get_key()
            value = static_job_preset_attribute.get_value()

            job_settings[key] = value

            # The python envoy client treats machine type and gpu as two different attributes so we must infer
            # the gpu attribute from the machine type here.
            # todo this should probably be handled by the new dependent job preset attribute logic
            if key == api_constants.API_KEYS.MACHINE_TYPE:
                # we need to find the machine type with this id and work out if it is a gpu machine or not
                machine_id = value

                if machine_id:
                    is_gpu = self.get_plugin().get_api_client()._is_gpu_enabled_machine(machine_id)
                else:
                    subtype_kwargs = static_job_preset_attribute.get_job_attribute().get_attribute().get_subtype_kwargs()
                    is_gpu = subtype_kwargs[api_constants.SUBTYPE_KEYS.ENUM.MACHINE_TYPE.USE_GPU_DEFAULT]

                job_settings[api_constants.API_KEYS.GPU] = is_gpu

        return job_settings
