__all__ = 'MainWindow'

from PySide2 import QtCore
from PySide2 import QtWidgets
from PySide2 import QtGui

import time
import typing

from gm_qt_plugin import qt_constants, qt_utils
from gm_qt_plugin.widgets import MainUI, SignInWidget
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin.qt_user_interface import QtUserInterface


class MainWindow(QtWidgets.QDialog, QtPluginAccessor):
    
    SHADOW_BLUR_RADIUS = 30

    def __init__(self, plugin, reset_callback, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        QtPluginAccessor.__init__(self, plugin)

        self._reset_callback = reset_callback

        # add the MainWindow to the plugins user interface class
        plugin.get_user_interface().set_main_window(self)

        # load custom fonts
        QtUserInterface.load_fonts()

        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Make the window appear on top to begin with. This is because new windows on Windows OS do not get brought
        # to the front if another process has focus
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)

        # Set the window title to the plugin name
        self.setWindowTitle(plugin.get_name())

        self.setWindowIcon(plugin.get_user_interface().get_icon())

        self._closed = False

        # The main window has two main states:
        # 1. The user has yet to sign in
        #    - Display a simple sign-in layout
        #
        # 2. The user is signed in
        #    - Display the main plugin user interface
        #
        # The layout for (2) can not be generated until the user is signed in so the creation of that layout is deferred
        # until after the user has signed-in.

        # the container widget is the widget sub widgets (1) & (2) will be parented to. This is done so that
        self._container_widget = QtWidgets.QStackedWidget()
        self._container_widget.setProperty(qt_constants.STYLE_PROPS.MAIN_CONTAINER_WIDGET, True)

        # Override the style object and apply it to the container, this way any styles applied by the parent do not
        # affect our widgets.
        self._style_override = QtWidgets.QStyleFactory().create('windows')
        self._container_widget.setStyle(self._style_override)

        self._sign_in_widget = None
        self._main_ui_widget = None

        self._sign_in_widget_index = None
        self._main_ui_widget_index = None

        self._shadow_widget = QtWidgets.QWidget(self)
        self._shadow_widget.setAutoFillBackground(True)
        self._shadow_widget.setGeometry(MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS,
                                        SignInWidget.WINDOW_WIDTH - MainWindow.SHADOW_BLUR_RADIUS * 2,
                                        SignInWidget.WINDOW_HEIGHT - MainWindow.SHADOW_BLUR_RADIUS * 2)
        self._shadow_widget.setStyleSheet("background-color:red;")

        self._shadow_effect = QtWidgets.QGraphicsDropShadowEffect(self)
        self._shadow_effect.setColor(QtGui.QColor('#444444'))
        self._shadow_effect.setBlurRadius(MainWindow.SHADOW_BLUR_RADIUS)
        self._shadow_effect.setOffset(0, 0)
        self._shadow_widget.setGraphicsEffect(self._shadow_effect)

        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.addWidget(self._container_widget)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)

        self.display_signed_out_ui()

    def is_closed(self):
        return self._closed

    def get_sign_in_widget(self):

        if not self._sign_in_widget:
            self._sign_in_widget = SignInWidget(self.get_plugin())

        return self._sign_in_widget

    def get_main_ui_widget(self):

        if not self._main_ui_widget:
            self._main_ui_widget = MainUI(self.get_plugin())

        return self._main_ui_widget

    def get_sign_in_widget_index(self):

        if self._sign_in_widget_index is None:
            self._sign_in_widget_index = self._container_widget.addWidget(self.get_sign_in_widget())

        return self._sign_in_widget_index

    def get_main_ui_widget_index(self):

        if self._main_ui_widget_index is None:
            self._main_ui_widget_index = self._container_widget.addWidget(self.get_main_ui_widget())

        return self._main_ui_widget_index

    def display_signed_in_ui(self):
        main_ui_widget_index = self.get_main_ui_widget_index()

        self._shadow_widget.hide()
        self._layout.setContentsMargins(0,0,0,0)

        # set min and max values (fixed) for both width and height
        self.setMinimumWidth(MainUI.MIN_WINDOW_WIDTH)
        self.setMinimumHeight(MainUI.MIN_WINDOW_HEIGHT)

        self.setMaximumWidth(MainUI.MAX_WINDOW_WIDTH)
        self.setMaximumHeight(MainUI.MAX_WINDOW_HEIGHT)

        # perform the resize
        resize_width = MainUI.INITIAL_WINDOW_WIDTH
        screen_width = QtWidgets.QApplication.instance().desktop().screenGeometry().width()
        if resize_width > screen_width:
            resize_width = screen_width

        resize_height = MainUI.INITIAL_WINDOW_HEIGHT
        screen_height = QtWidgets.QApplication.instance().desktop().screenGeometry().height()
        if resize_height > screen_height:
            resize_height = screen_height

        settings = QtCore.QSettings()
        window_size = settings.value("windowSize", QtCore.QSize(resize_width, resize_height))

        self.resize(window_size)

        qt_utils.center_window(self)

        # load style sheet
        # self.get_plugin().get_user_interface().set_style_sheet(main_window)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.FramelessWindowHint)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowMaximizeButtonHint)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowMinimizeButtonHint)

        # Remove the stay on top hint after sign in. We know the user has focused this window at this point so it
        # can be removed.
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowStaysOnTopHint)
        self.show()

        self.get_plugin().get_user_interface().set_style_sheet(self._container_widget)
        self._container_widget.setCurrentIndex(main_ui_widget_index)

    def display_signed_out_ui(self):

        self._layout.setContentsMargins(MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS,
                                        MainWindow.SHADOW_BLUR_RADIUS)

        self._shadow_widget.show()

        # set min and max values (fixed) for both width and height
        self.setFixedWidth(SignInWidget.WINDOW_WIDTH)
        self.setFixedHeight(SignInWidget.WINDOW_HEIGHT)

        # perform the resize
        self.resize(SignInWidget.WINDOW_WIDTH, SignInWidget.WINDOW_HEIGHT)

        qt_utils.center_window(self)

        # load style sheet
        # self.get_plugin().get_user_interface().set_style_sheet(main_window, qt_constants.WhiteTheme)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.FramelessWindowHint)

        self.get_plugin().get_user_interface().set_style_sheet(self._container_widget, qt_constants.WhiteTheme)
        self._container_widget.setCurrentIndex(self.get_sign_in_widget_index())

    def resizeEvent(self, event):
        if self.get_plugin().get_api_client().is_user_signed_in() and \
                self._container_widget.currentIndex() == self.get_main_ui_widget_index():
            settings = QtCore.QSettings()
            settings.setValue("windowSize", event.size())

        QtWidgets.QWidget.resizeEvent(self, event)

    def set_reset_callback(self, reset_callback: typing.Callable):
        self._reset_callback = reset_callback

    def reset(self):
        self._reset_callback()

    def close_window(self):
        self._closed = True

        while self.get_main_ui_widget().get_header().is_polling_cps():
            time.sleep(0.3)

        self.close()
        QtUserInterface.unload_fonts()
