__all__ = 'PreflightWindow'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.watch_pattern import WatchPattern
from gm_meta_plugin.utils import get_desktop_path

from gm_qt_plugin import qt_constants, qt_utils
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin.qt_model_file_tree import FileTreeModel
from gm_qt_plugin.qt_model_watch_pattern_table import WatchPatternTable
from gm_qt_plugin.widgets.input_widgets.path_input import PathInput
from gm_qt_plugin.widgets.notification import Notification


class PreflightWindow(QtWidgets.QDialog, QtPluginAccessor):

    INITIAL_WINDOW_WIDTH = 1200
    INITIAL_WINDOW_HEIGHT = 700

    MINIMUM_WINDOW_WIDTH = 600
    MINIMUM_WINDOW_HEIGHT = 600

    UPLOAD_AND_SUBMIT = u"Upload and Submit"
    UPLOAD = u"Upload"
    SUBMIT = u"Submit"

    UPLOAD_AND_SUBMIT_TOOL_TIP = u"Tell Envoy to upload the above selected files and submit your job once all files have been uploaded."
    UPLOAD_TOOL_TIP = u"Tell Envoy to upload the above selected files."
    SUBMIT_TOOL_TIP = u"Tell Envoy to submit your job without uploading any new files."

    def __init__(self, plugin, packed_project, job_preset, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        QtPluginAccessor.__init__(self, plugin)

        self._packed_project = packed_project
        self._job_preset = job_preset

        self.setWindowModality(QtCore.Qt.ApplicationModal)

        # apply style sheet
        plugin.get_user_interface().set_style_sheet(self)
        self.setProperty(qt_constants.STYLE_PROPS.BACKGROUND_LEVEL_1, True)

        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Set the window title to the plugin name
        self.setWindowTitle(u"GridMarkets Preflight")

        # resize the window to it's inital size
        self.resize(PreflightWindow.INITIAL_WINDOW_WIDTH, PreflightWindow.INITIAL_WINDOW_HEIGHT)

        # set minimum dimensions
        self.setMinimumWidth(PreflightWindow.MINIMUM_WINDOW_WIDTH)
        self.setMinimumHeight(PreflightWindow.MINIMUM_WINDOW_HEIGHT)

        self.setWindowIcon(plugin.get_user_interface().get_icon())

        self._cancel_submission = True

        """ Create Widgets """
        self._header_layout_widget = QtWidgets.QWidget()
        self._header_layout_widget.setProperty(qt_constants.STYLE_PROPS.BACKGROUND_LEVEL_0, True)

        self._title = QtWidgets.QLabel(u"Preflight")
        self._title.setProperty(qt_constants.STYLE_PROPS.H3, True)
        self._upload_files_disclaimer = Notification(u"Please select all files you wish to upload. Upload progress will show within Envoy upon submission.")
        self._watch_patterns_disclaimer = Notification(u"Please verify the below watch patterns. Watch patterns determine which files are returned and where they are downloaded to.")

        self.file_tree_model = FileTreeModel(packed_project)
        self._watch_pattern_model = WatchPatternTable(packed_project, job_preset)

        self._tree_view = QtWidgets.QTreeView()
        self._tree_view.setProperty(qt_constants.STYLE_PROPS.BACKGROUND_LEVEL_2, True)
        self._tree_view.setModel(self.file_tree_model)
        self.file_tree_model.set_default_expanded_state(self._tree_view)
        self._tree_view.header().setStretchLastSection(False)
        self._tree_view.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self._tree_view.setAlternatingRowColors(True)

        self._watch_patterns_view = QtWidgets.QTableView()
        self._watch_patterns_view.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self._watch_patterns_view.setProperty(qt_constants.STYLE_PROPS.BACKGROUND_LEVEL_2, True)
        self._watch_patterns_view.setModel(self._watch_pattern_model)
        self._watch_patterns_view.verticalHeader().hide()
        self._watch_patterns_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        #self._watch_patterns_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

        self._add_watch_pattern_btn = QtWidgets.QPushButton('add')
        self._add_watch_pattern_btn.setFlat(True)
        self._add_watch_pattern_btn.setToolTip(qt_utils.split_tool_tip('Add a new watch pattern.'))
        self._add_watch_pattern_btn.clicked.connect(self._add_watch_pattern)
        self._add_watch_pattern_btn.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        self._remove_watch_pattern_btn = QtWidgets.QPushButton('remove')
        self._remove_watch_pattern_btn.setFlat(True)
        self._remove_watch_pattern_btn.setToolTip(qt_utils.split_tool_tip('Remove the selected watch patterns.'))
        self._remove_watch_pattern_btn.clicked.connect(self._remove_watch_pattern)
        self._remove_watch_pattern_btn.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self._remove_watch_pattern_btn.setEnabled(False)

        watch_patterns_header = self._watch_patterns_view.horizontalHeader()

        #watch_patterns_header.setDefaultSectionSize()
        #watch_patterns_header.setStretchLastSection(False)

        #watch_patterns_header.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
        #watch_patterns_header.resizeSections(QtWidgets.QHeaderView.Stretch)

        watch_patterns_header.resizeSection(0, (PreflightWindow.INITIAL_WINDOW_WIDTH / 2) - 75)
        watch_patterns_header.resizeSection(1, (PreflightWindow.INITIAL_WINDOW_WIDTH / 2) - 75)
        watch_patterns_header.setStretchLastSection(True)


        #watch_patterns_header.setSectionResizeMode(0, QtWidgets.QHeaderView.Interactive)
        #watch_patterns_header.setSectionResizeMode(1, QtWidgets.QHeaderView.Interactive)
        #watch_patterns_header.setSectionResizeMode(2, QtWidgets.QHeaderView.Fixed)
        #watch_patterns_header.resizeSection(2, 80)

        #watch_patterns_header.resizeSections(QtWidgets.QHeaderView.Stretch)
        self._watch_patterns_view.setAlternatingRowColors(True)
        self._watch_patterns_view.selectionModel().selectionChanged.connect(self._watch_pattern_selection_changed)
        #self._watch_patterns_view.setTextElideMode(QtCore.Qt.ElideMiddle)

        watch_patterns_view = self._watch_patterns_view
        self._size_hint = QtCore.QSize(20, 29)
        size_hint = self._size_hint

        class subclassedQStyledItemDelegate(QtWidgets.QStyledItemDelegate):
            def initStyleOption(self, option, index):
                option.textElideMode = QtCore.Qt.ElideNone
                #option.backgroundBrush = QtGui.QBrush(QtGui.QColor(255, 150, 150))
                QtWidgets.QStyledItemDelegate.initStyleOption(self, option, index)


            def paint(self, painter, option, index):
                if not index.isValid():
                    return

                opt = option

                self.initStyleOption(opt, index)
                padding = 3

                textElideMode = opt.textElideMode
                text = opt.text

                opt.textElideMode = QtCore.Qt.ElideNone
                opt.text = ""

                QtWidgets.QApplication.style().drawControl(QtWidgets.QStyle.CE_ItemViewItem, opt, painter)

                opt.textElideMode = textElideMode
                opt.text = text

                painter.save()
                painter.setClipRect(opt.rect)
                opt.rect = opt.rect.adjusted(padding, padding, -padding, -padding)
                if not watch_patterns_view.isPersistentEditorOpen(index):
                    painter.drawText(opt.rect, QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter,
                                     opt.fontMetrics.elidedText(opt.text, QtCore.Qt.ElideMiddle, opt.rect.width()))
                painter.restore()

            def setEditorData(self, editorWidget, modelIndex):
                column = modelIndex.column()

                if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
                    editorWidget.set_value(modelIndex.data())
                else:
                    editorWidget.setText(modelIndex.data())

            def setModelData(self, editorWidget, model, modelIndex):
                column = modelIndex.column()

                if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
                    model.setData(modelIndex, editorWidget.get_value())
                else:
                    QtWidgets.QStyledItemDelegate.setModelData(self, editorWidget, model, modelIndex)

            def updateEditorGeometry(self, editorWidget, option, modelIndex):
                rect = option.rect
                rect.setWidth(watch_patterns_header.sectionSize(modelIndex.column()))
                editorWidget.setMaximumSize(rect.width(), rect.height())
                editorWidget.setMinimumSize(rect.width(), rect.height())
                editorWidget.setGeometry(rect)
                editorWidget.setFixedSize(rect.width(), rect.height())
                editorWidget.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
                editorWidget.updateGeometry()
                editorWidget.sizeHint = size_hint

            def createEditor(self, parent, option, modelIndex):
                column = modelIndex.column()

                if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
                    editorWidget = PathInput(plugin, file_mode= QtWidgets.QFileDialog.Directory, parent=parent)
                    editorWidget.input_value_changed.connect(lambda: self.setModelData(editorWidget, modelIndex.model(), modelIndex))
                else:
                    editorWidget = QtWidgets.QLineEdit(parent)

                return editorWidget

            def destroyEditor(self, editorWidget, modelIndex):
                QtWidgets.QStyledItemDelegate.destroyEditor(self, editorWidget, modelIndex)

        self._styled_item_delegate = subclassedQStyledItemDelegate()
        self._watch_patterns_view.setItemDelegate(self._styled_item_delegate)

        self._controls_layout_widget = QtWidgets.QWidget()

        self._cancel_btn = QtWidgets.QPushButton(u"Cancel")

        self._primary_action_1 = QtWidgets.QAction(PreflightWindow.UPLOAD_AND_SUBMIT, self)
        self._primary_action_1.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_AND_SUBMIT_TOOL_TIP))
        self._primary_action_2 = QtWidgets.QAction(PreflightWindow.UPLOAD, self)
        self._primary_action_2.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_TOOL_TIP))
        self._primary_action_3 = QtWidgets.QAction(PreflightWindow.SUBMIT, self)
        self._primary_action_3.setToolTip(qt_utils.split_tool_tip(PreflightWindow.SUBMIT_TOOL_TIP))

        self._primary_actions_menu = QtWidgets.QMenu()
        self._primary_actions_menu.setToolTipsVisible(True)
        self._primary_actions_menu.addAction(self._primary_action_1)
        self._primary_actions_menu.addAction(self._primary_action_2)

        # only show the submit action if the project already exists
        if packed_project.get_name() in plugin.get_api_client().get_root_directories():
            self._primary_actions_menu.addAction(self._primary_action_3)

        plugin.get_user_interface().set_style_sheet(self._primary_actions_menu)

        self._primary_btn = QtWidgets.QToolButton()
        self._primary_btn.setProperty(qt_constants.STYLE_PROPS.BRAND, True)
        self._primary_btn.setMinimumWidth(175)
        self._primary_btn.setMenu(self._primary_actions_menu)
        self._primary_btn.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        self.primary_action_1_slot()

        self._upload_files_widget = QtWidgets.QWidget()
        self._watch_patterns_widget = QtWidgets.QWidget()
        self._watch_patterns_widget.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)

        self._tab_bar = QtWidgets.QTabWidget()
        tab1_index = self._tab_bar.addTab(self._upload_files_widget, qt_constants.DIRECTORY_ICON_WHITE, "Upload Files")
        tab2_index = self._tab_bar.addTab(self._watch_patterns_widget, qt_constants.EYE_ICON_WHITE, "Watch Patterns")
        self._tab_bar.tabBar().setDocumentMode(True)
        self._tab_bar.tabBar().setExpanding(True)

        self._tab_bar.setTabToolTip(tab1_index, "Select the files you would like to upload.")
        self._tab_bar.setTabToolTip(tab2_index, "Set watch patterns to choose which files get returned.")

        """ Create Connections """
        self._cancel_btn.clicked.connect(self.cancel)
        self._primary_btn.clicked.connect(self.submit)
        self._primary_actions_menu.aboutToShow.connect(self._about_to_show_primary_actions_menu)

        self._primary_action_1.triggered.connect(self.primary_action_1_slot)
        self._primary_action_2.triggered.connect(self.primary_action_2_slot)
        self._primary_action_3.triggered.connect(self.primary_action_3_slot)

        """ Create Layouts """
        header_layout = QtWidgets.QVBoxLayout()
        header_layout.addWidget(self._title)
        self._header_layout_widget.setLayout(header_layout)

        body_layout = QtWidgets.QVBoxLayout()
        body_layout.addWidget(self._upload_files_disclaimer)
        body_layout.addWidget(self._tree_view)
        body_layout.setContentsMargins(0,10,0,0)
        self._upload_files_widget.setLayout(body_layout)

        self._watch_pattern_controls_layout = QtWidgets.QHBoxLayout()
        self._watch_pattern_controls_layout.addWidget(self._add_watch_pattern_btn)
        self._watch_pattern_controls_layout.addWidget(self._remove_watch_pattern_btn)
        self._watch_pattern_controls_layout.addStretch()

        watch_pattern_layout = QtWidgets.QVBoxLayout()
        watch_pattern_layout.setContentsMargins(0,10,0,0)
        watch_pattern_layout.addWidget(self._watch_patterns_disclaimer)
        watch_pattern_layout.addWidget(self._watch_patterns_view)
        watch_pattern_layout.addLayout(self._watch_pattern_controls_layout)

        self._watch_patterns_widget.setLayout(watch_pattern_layout)

        controls_layout = QtWidgets.QHBoxLayout()
        controls_layout.setContentsMargins(0,0,0,0)
        controls_layout.setAlignment(QtCore.Qt.AlignRight)
        controls_layout.addWidget(self._primary_btn)
        controls_layout.addWidget(self._cancel_btn)
        self._controls_layout_widget.setLayout(controls_layout)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self._header_layout_widget)
        layout.addWidget(self._tab_bar)
        layout.addWidget(self._controls_layout_widget)

    def _add_watch_pattern(self):
        self._watch_pattern_model.add_row()

    def _remove_watch_pattern(self):
        rows = set()
        for index in self._watch_patterns_view.selectedIndexes():
            rows.add(index.row())

        for row in sorted(rows, reverse=True):
            self._watch_pattern_model.removeRow(row)

    def _watch_pattern_selection_changed(self, selected, deselected):
        self._remove_watch_pattern_btn.setEnabled(len(selected) > 0)

    def primary_action_1_slot(self):
        self._primary_btn.setText(PreflightWindow.UPLOAD_AND_SUBMIT)
        self._primary_btn.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_AND_SUBMIT_TOOL_TIP))

    def primary_action_2_slot(self):
        self._primary_btn.setText(PreflightWindow.UPLOAD)
        self._primary_btn.setToolTip(qt_utils.split_tool_tip(PreflightWindow.UPLOAD_TOOL_TIP))

    def primary_action_3_slot(self):
        self._primary_btn.setText(PreflightWindow.SUBMIT)
        self._primary_btn.setToolTip(qt_utils.split_tool_tip(PreflightWindow.SUBMIT_TOOL_TIP))

    def cancel(self):
        self._cancel_submission = True
        self.hide()

    def submit(self):
        watch_patterns = self._watch_pattern_model.get_watch_patterns()
        self._packed_project.set_watch_patterns(watch_patterns)
        self._cancel_submission = False
        self.hide()

    def _about_to_show_primary_actions_menu(self):
        self._primary_actions_menu.setFixedWidth(self._primary_btn.width())

    def cancel_submission(self):
        return self._cancel_submission

    def get_next_action(self):
        return self._primary_btn.text()

    def get_files_to_upload(self):
        return self.file_tree_model.get_files_to_upload()
