__all__ = 'StatementCreatorWindow'

from PySide2 import QtCore
from PySide2 import QtWidgets
from PySide2 import QtGui

from gm_qt_plugin import qt_constants, qt_utils
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin.widgets.input_widgets.spin_box import SpinBox
from gm_qt_plugin.widgets.input_widgets.double_spin_box import DoubleSpinBox


class ExpressionWidget(QtWidgets.QGroupBox, QtPluginAccessor):

    __expression_widgets = []

    def __init__(self, plugin, remove_callback, parent=None):
        QtWidgets.QGroupBox.__init__(self, parent)
        QtPluginAccessor.__init__(self, plugin)

        self.setAlignment(QtCore.Qt.AlignVCenter)
        self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        self._remove_callback = remove_callback
        self._parent_layout = None

        self._test_action = QtWidgets.QAction(u"Remove", self)
        self._test_action.triggered.connect(self.remove_expression)

        self._controls_menu = QtWidgets.QMenu()
        self._controls_menu.setToolTipsVisible(True)
        self._controls_menu.addAction(self._test_action)
        plugin.get_user_interface().set_style_sheet(self._controls_menu)

        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION, True)
        self.setMouseTracking(True)
        self.setAttribute(QtCore.Qt.WA_NoMousePropagation, True)
        self.setAttribute(QtCore.Qt.WA_Hover, True)

        ExpressionWidget.__expression_widgets.append(self)

    def mousePressEvent(self, event):
        self._controls_menu.exec_(QtGui.QCursor.pos())
        event.setAccepted(True)

    def mouseMoveEvent(self, event):
        # Un hover all the others
        for expression_widget in ExpressionWidget.__expression_widgets:
            expression_widget.setProperty('hover', False)
            expression_widget.style().unpolish(expression_widget)
            expression_widget.style().polish(expression_widget)

        self.setProperty('hover', True)
        self.style().unpolish(self)
        self.style().polish(self)
        event.setAccepted(True)

    def leaveEvent(self, event):
        self.setProperty('hover', False)
        self.style().unpolish(self)
        self.style().polish(self)
        event.setAccepted(True)

    def remove_expression(self):
        self._remove_callback()
        ExpressionWidget.__expression_widgets.remove(self)
        self.deleteLater()


class TrueExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION_TRUE, True)

        """ Create Widgets """
        self._label = QtWidgets.QLabel("True")
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._label)
        self._layout.addStretch()
        self.setLayout(self._layout)


class FalseExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION_FALSE, True)

        """ Create Widgets """
        self._label = QtWidgets.QLabel("False")
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._label)
        self._layout.addStretch()
        self.setLayout(self._layout)


class IntegerExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION_INTEGER, True)

        """ Create Widgets """
        self._spin_box = SpinBox(self.get_plugin())
        self._spin_box.set_value(0)
        self._spin_box.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)
        self._spin_box.setMaximum(qt_constants.MAX_INT)
        self._spin_box.setMinimum(qt_constants.MIN_INT)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._spin_box)
        self._layout.addStretch()
        self.setLayout(self._layout)


class DecimalExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION_INTEGER, True)

        """ Create Widgets """
        self._spin_box = DoubleSpinBox(plugin)
        self._spin_box.set_value(0)
        self._spin_box.setStepType(QtWidgets.QAbstractSpinBox.AdaptiveDecimalStepType)
        self._spin_box.setSingleStep(0.1)
        self._spin_box.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)
        self._spin_box.setMaximum(qt_constants.MAX_INT)
        self._spin_box.setMinimum(qt_constants.MIN_INT)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._spin_box)
        self._layout.addStretch()
        self.setLayout(self._layout)


class DateTimeExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION_INTEGER, True)

        """ Create Widgets """
        self._date_time_edit = QtWidgets.QDateTimeEdit()
        self._date_time_edit.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._date_time_edit)
        self._layout.addStretch()
        self.setLayout(self._layout)


class DateExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION_INTEGER, True)

        """ Create Widgets """
        self._date_edit = QtWidgets.QDateEdit()
        self._date_edit.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._date_edit)
        self._layout.addStretch()
        self.setLayout(self._layout)


class TimeExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.EXPRESSION_INTEGER, True)

        """ Create Widgets """
        self._time_edit = QtWidgets.QTimeEdit()
        self._time_edit.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._time_edit)
        self._layout.addStretch()
        self.setLayout(self._layout)


class NowExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.RUNTIME_EXPRESSION, True)

        """ Create Widgets """
        self._label = QtWidgets.QLabel("Now()")
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._label)
        self._layout.addStretch()
        self.setLayout(self._layout)


class TodayExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)
        self.setProperty(qt_constants.STYLE_PROPS.RUNTIME_EXPRESSION, True)

        """ Create Widgets """
        self._label = QtWidgets.QLabel("Today()")
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._label)
        self._layout.addStretch()
        self.setLayout(self._layout)


class UnaryExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, name, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)

        """ Create Widgets """
        self._label = QtWidgets.QLabel(name)
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        self._right_expression_widget = None

        self._add_r_expression_btn = QtWidgets.QPushButton("Add Expression")
        self._add_r_expression_btn.setProperty(qt_constants.STYLE_PROPS.BRAND, True)
        self._add_r_expression_btn.setProperty('small-btn', True)

        """ Create Layouts """
        self._right_expression_container = QtWidgets.QHBoxLayout()
        self._right_expression_container.setContentsMargins(0, 0, 0, 0)
        self._right_expression_container.setAlignment(QtCore.Qt.AlignVCenter)
        self._right_expression_container.addWidget(self._add_r_expression_btn)
        self._right_expression_container.setAlignment(QtCore.Qt.AlignVCenter)

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setAlignment(QtCore.Qt.AlignVCenter)
        self._layout.addWidget(self._label)
        self._layout.addLayout(self._right_expression_container)
        self._layout.addStretch()
        self.setLayout(self._layout)

        ''' Create Connections '''
        self._add_r_expression_btn.clicked.connect(self._handle_add_r_expression)

    def remove_expression(self):

        if self._right_expression_widget:
            self._right_expression_widget.remove_expression()

        ExpressionWidget.remove_expression(self)

    def _handle_remove_r_expression(self):
        if self._right_expression_widget:
            self._right_expression_container.removeWidget(self._right_expression_widget)
            self._right_expression_widget = None

        self._add_r_expression_btn.show()

    def _handle_add_r_expression(self):
        # open preflight window
        expression_creator_window = ExpressionCreatorWindow(self.get_plugin(), parent=self)
        expression_creator_window.exec_()

        if expression_creator_window.cancel_expression_creation():
            return

        expression_widget_type = expression_creator_window.chosen_expression_type()
        self._right_expression_widget = expression_widget_type(self.get_plugin(), self._handle_remove_r_expression)

        self._right_expression_container.addWidget(self._right_expression_widget)
        self._add_r_expression_btn.hide()


class NotExpression(UnaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        UnaryExpression.__init__(self, plugin, remove_callback, "not", parent=parent)


class BinaryExpression(ExpressionWidget):
    def __init__(self, plugin, remove_callback, name, parent=None):
        ExpressionWidget.__init__(self, plugin, remove_callback, parent=parent)

        """ Create Widgets """
        self._label = QtWidgets.QLabel(name)
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        self._left_expression_widget = None
        self._right_expression_widget = None

        self._add_l_expression_btn = QtWidgets.QPushButton("Add Expression")
        self._add_l_expression_btn.setProperty(qt_constants.STYLE_PROPS.BRAND, True)
        self._add_l_expression_btn.setProperty('small-btn', True)

        self._add_r_expression_btn = QtWidgets.QPushButton("Add Expression")
        self._add_r_expression_btn.setProperty(qt_constants.STYLE_PROPS.BRAND, True)
        self._add_r_expression_btn.setProperty('small-btn', True)

        """ Create Layouts """
        self._left_expression_container = QtWidgets.QHBoxLayout()
        self._left_expression_container.setAlignment(QtCore.Qt.AlignVCenter)
        self._left_expression_container.setContentsMargins(0, 0, 0, 0)
        self._left_expression_container.addWidget(self._add_l_expression_btn)
        self._left_expression_container.setAlignment(QtCore.Qt.AlignVCenter)

        self._right_expression_container = QtWidgets.QHBoxLayout()
        self._right_expression_container.setAlignment(QtCore.Qt.AlignVCenter)
        self._right_expression_container.setContentsMargins(0, 0, 0, 0)
        self._right_expression_container.addWidget(self._add_r_expression_btn)
        self._right_expression_container.setAlignment(QtCore.Qt.AlignVCenter)

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setAlignment(QtCore.Qt.AlignVCenter)
        self._layout.addLayout(self._left_expression_container)
        self._layout.addWidget(self._label)
        self._layout.addLayout(self._right_expression_container)
        self._layout.addStretch()
        self.setLayout(self._layout)

        ''' Create Connections '''
        self._add_l_expression_btn.clicked.connect(self._handle_add_l_expression)
        self._add_r_expression_btn.clicked.connect(self._handle_add_r_expression)

    def remove_expression(self):
        if self._left_expression_widget:
            self._left_expression_widget.remove_expression()

        if self._right_expression_widget:
            self._right_expression_widget.remove_expression()

        ExpressionWidget.remove_expression(self)

    def _handle_remove_l_expression(self):
        if self._left_expression_widget:
            self._left_expression_container.removeWidget(self._left_expression_widget)
            self._left_expression_widget = None

        self._add_l_expression_btn.show()

    def _handle_remove_r_expression(self):
        if self._right_expression_widget:
            self._right_expression_container.removeWidget(self._right_expression_widget)
            self._right_expression_widget = None

        self._add_r_expression_btn.show()

    def _handle_add_l_expression(self):
        # open preflight window
        expression_creator_window = ExpressionCreatorWindow(self.get_plugin(), parent=self)
        expression_creator_window.exec_()

        if expression_creator_window.cancel_expression_creation():
            return

        expression_widget_type = expression_creator_window.chosen_expression_type()
        self._left_expression_widget = expression_widget_type(self.get_plugin(), self._handle_remove_l_expression)

        self._left_expression_container.addWidget(self._left_expression_widget)
        self._add_l_expression_btn.hide()

    def _handle_add_r_expression(self):
        # open preflight window
        expression_creator_window = ExpressionCreatorWindow(self.get_plugin(), parent=self)
        expression_creator_window.exec_()

        if expression_creator_window.cancel_expression_creation():
            return

        expression_widget_type = expression_creator_window.chosen_expression_type()
        self._right_expression_widget = expression_widget_type(self.get_plugin(), self._handle_remove_r_expression)

        self._right_expression_container.addWidget(self._right_expression_widget)
        self._add_r_expression_btn.hide()


class AndExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, "and", parent=parent)


class OrExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, "or", parent=parent)


class EqualExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, "==", parent=parent)


class NotEqualExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, "!=", parent=parent)


class GreaterThanExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, ">", parent=parent)


class LessThanExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, "<", parent=parent)


class GreaterThanOrEqualToExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, ">=", parent=parent)


class LessThanOrEqualToExpression(BinaryExpression):
    def __init__(self, plugin, remove_callback, parent=None):
        BinaryExpression.__init__(self, plugin, remove_callback, "<=", parent=parent)


class StatementWidget(QtWidgets.QGroupBox, QtPluginAccessor):

    __statement_widgets = []

    def __init__(self, plugin, parent=None):
        QtWidgets.QGroupBox.__init__(self, parent)
        QtPluginAccessor.__init__(self, plugin)

        self._parent_layout = None

        self._test_action = QtWidgets.QAction(u"Remove", self)
        self._test_action.triggered.connect(self.remove_statement)

        self._controls_menu = QtWidgets.QMenu()
        self._controls_menu.setToolTipsVisible(True)
        self._controls_menu.addAction(self._test_action)
        plugin.get_user_interface().set_style_sheet(self._controls_menu)

        self.setProperty(qt_constants.STYLE_PROPS.STATEMENT, True)
        self.setMouseTracking(True)
        self.setAttribute(QtCore.Qt.WA_NoMousePropagation, True)
        self.setAttribute(QtCore.Qt.WA_Hover, True)

        StatementWidget.__statement_widgets.append(self)

    def mousePressEvent(self, event):
        self._controls_menu.exec_(QtGui.QCursor.pos())
        event.setAccepted(True)

    def mouseMoveEvent(self, event):
        # Un hover all the others
        for statement_widget in StatementWidget.__statement_widgets:
            statement_widget.setProperty('hover', False)
            statement_widget.style().unpolish(statement_widget)
            statement_widget.style().polish(statement_widget)

        self.setProperty('hover', True)
        self.style().unpolish(self)
        self.style().polish(self)
        event.setAccepted(True)

    def leaveEvent(self, event):
        self.setProperty('hover', False)
        self.style().unpolish(self)
        self.style().polish(self)
        event.setAccepted(True)

    def remove_statement(self):
        self._parent_layout.removeWidget(self)
        StatementWidget.__statement_widgets.remove(self)
        self.deleteLater()

    def set_parent_layout(self, layout):
        self._parent_layout = layout


class PassStatement(StatementWidget):
    def __init__(self, plugin, parent=None):
        StatementWidget.__init__(self, plugin, parent=parent)

        """ Create Widgets """
        self._label = QtWidgets.QLabel("pass")
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._label)
        self._layout.addStretch()
        self.setLayout(self._layout)


class IfStatement(StatementWidget):
    def __init__(self, plugin, parent=None):
        StatementWidget.__init__(self, plugin, parent=parent)

        self._expression_widget = None

        """ Create Widgets """
        self._label = QtWidgets.QLabel("if")
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        self._add_expression_btn = QtWidgets.QPushButton("Add Expression")
        self._add_expression_btn.setProperty(qt_constants.STYLE_PROPS.BRAND, True)
        self._add_expression_btn.setProperty('small-btn', True)

        self._add_statement_push_btn = QtWidgets.QPushButton("Add statement +")
        self._add_statement_push_btn.setFlat(True)
        self._add_statement_push_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)
        self._add_statement_push_btn.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        """ Create Layouts """
        self._if_statement_layout = QtWidgets.QHBoxLayout()
        self._if_statement_layout.setContentsMargins(0, 0, 0, 0)
        self._if_statement_layout.addWidget(self._add_expression_btn)
        self._if_statement_layout.setAlignment(QtCore.Qt.AlignRight)

        self._if_statement_layout_container = QtWidgets.QHBoxLayout()
        self._if_statement_layout_container.setContentsMargins(0, 0, 0, 0)
        self._if_statement_layout_container.addWidget(self._label)
        self._if_statement_layout_container.addLayout(self._if_statement_layout)
        self._if_statement_layout_container.addStretch()

        self._content_layout = QtWidgets.QVBoxLayout()
        self._content_layout.setContentsMargins(0, 0, 0, 0)

        self._content_layout_container = QtWidgets.QVBoxLayout()
        self._content_layout_container.setContentsMargins(16, 0, 0, 0)
        self._content_layout_container.addLayout(self._content_layout)
        self._content_layout_container.addWidget(self._add_statement_push_btn)

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addLayout(self._if_statement_layout_container)
        self._layout.addLayout(self._content_layout_container)

        self.setLayout(self._layout)

        ''' Create Connections '''
        self._add_expression_btn.clicked.connect(self._handle_add_expression)
        self._add_statement_push_btn.clicked.connect(self._handle_add_statement)

    def remove_statement(self):
        if self._expression_widget:
            self._expression_widget.remove_expression()

        while self._content_layout.count():
            layout_item = self._content_layout.itemAt(0)

            if isinstance(layout_item.widget(), StatementWidget):
                widget = layout_item.widget()
                widget.remove_statement()

        StatementWidget.remove_statement(self)

    def _handle_remove_expression(self):
        if self._expression_widget:
            self._if_statement_layout.removeWidget(self._expression_widget)
            self._expression_widget = None

        self._add_expression_btn.show()

    def _handle_add_expression(self):
        # open preflight window
        expression_creator_window = ExpressionCreatorWindow(self.get_plugin(), parent=self)
        expression_creator_window.exec_()

        if expression_creator_window.cancel_expression_creation():
            return

        expression_widget_type = expression_creator_window.chosen_expression_type()
        self._expression_widget = expression_widget_type(self.get_plugin(), self._handle_remove_expression)

        self._if_statement_layout.addWidget(self._expression_widget)
        self._add_expression_btn.hide()

    def _handle_add_statement(self):
        # open preflight window
        statement_creator_window = StatementCreatorWindow(self.get_plugin(), parent=self)
        statement_creator_window.exec_()

        if statement_creator_window.cancel_statement_creation():
            return

        statement_widget_type = statement_creator_window.chosen_statement_type()
        statement_widget = statement_widget_type(self.get_plugin())

        self._content_layout.addWidget(statement_widget)
        statement_widget.set_parent_layout(self._content_layout)


class ReturnStatement(StatementWidget):
    def __init__(self, plugin, parent=None):
        StatementWidget.__init__(self, plugin, parent=parent)

        """ Create Widgets """
        self._label = QtWidgets.QLabel("return")
        self._label.setProperty(qt_constants.STYLE_PROPS.MONOSPACE, True)

        """ Create Layouts """
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._label)
        self._layout.addStretch()
        self.setLayout(self._layout)



class ExpressionCreatorWindow(QtWidgets.QDialog, QtPluginAccessor):

    INITIAL_WINDOW_WIDTH = 800
    INITIAL_WINDOW_HEIGHT = 700

    MINIMUM_WINDOW_WIDTH = 600
    MINIMUM_WINDOW_HEIGHT = 600

    BUTTON_SIZE = 150

    def __init__(self, plugin, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        QtPluginAccessor.__init__(self, plugin)

        self.setWindowModality(QtCore.Qt.ApplicationModal)

        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Set the window title
        self.setWindowTitle(u"GridMarkets Expression Creator")

        # resize the window to it's initial size
        self.resize(ExpressionCreatorWindow.INITIAL_WINDOW_WIDTH, ExpressionCreatorWindow.INITIAL_WINDOW_HEIGHT)

        # set minimum dimensions
        self.setMinimumWidth(StatementCreatorWindow.MINIMUM_WINDOW_WIDTH)
        self.setMinimumHeight(StatementCreatorWindow.MINIMUM_WINDOW_HEIGHT)

        self._chosen_expression_type = None
        self._cancel_expression_creation = True

        """ Create Widgets """
        self._header_layout_widget = QtWidgets.QWidget()
        self._header_layout_widget.setProperty(qt_constants.STYLE_PROPS.BACKGROUND_LEVEL_0, True)
        self._title = QtWidgets.QLabel(u"Choose a Expression Type")
        self._title.setProperty(qt_constants.STYLE_PROPS.H3, True)

        self._scroll_area = QtWidgets.QScrollArea()
        self._content_widget = QtWidgets.QWidget()
        self._content_widget.setProperty("background-level-2", True)
        self._scroll_area.setWidget(self._content_widget)
        self._scroll_area.setWidgetResizable(True)


        self._simple_value_expression_types_header = QtWidgets.QLabel("Simple Value Expression Types:")
        self._simple_value_expression_types_header.setProperty(qt_constants.STYLE_PROPS.H4, True)

        self._datetime_expression_types_header = QtWidgets.QLabel("DateTime Expression Types:")
        self._datetime_expression_types_header.setProperty(qt_constants.STYLE_PROPS.H4, True)

        self._comparision_expression_types_header = QtWidgets.QLabel("Comparision Expression Types:")
        self._comparision_expression_types_header.setProperty(qt_constants.STYLE_PROPS.H4, True)

        self._logic_expression_types_header = QtWidgets.QLabel("Logic Expression Types:")
        self._logic_expression_types_header.setProperty(qt_constants.STYLE_PROPS.H4, True)

        self._true_expression_btn = self._get_expression_button("True", "True", TrueExpression)
        self._false_expression_btn = self._get_expression_button("False", "False", FalseExpression)
        self._integer_number_expression_btn = self._get_expression_button("integer\nnumber", "A integer number value such as '42'.", IntegerExpression)
        self._decimal_number_expression_btn = self._get_expression_button("decimal\nnumber", "A decimal number such as '2.25'.", DecimalExpression)

        self._date_time_expression_btn = self._get_expression_button("Date time", "A date and time value.", DateTimeExpression)
        self._date_expression_btn = self._get_expression_button("Date", "A date value.", DateExpression)
        self._time_expression_btn = self._get_expression_button("Time", "A time value.", TimeExpression)
        self._now_expression_btn = self._get_expression_button("Now()", "A date time value represeting the current date and time at the time of evaluation.", NowExpression)
        self._today_expression_btn = self._get_expression_button("Today()", "A date value represeting the current date at the time of evaluation.", TodayExpression)

        self._equal_expression_btn = self._get_expression_button("==", "Equal", EqualExpression)
        self._not_equal_expression_btn = self._get_expression_button("!=", "Not equal", NotEqualExpression)
        self._greater_than_expression_btn = self._get_expression_button(">", "Greater than", GreaterThanExpression)
        self._less_than_expression_btn = self._get_expression_button("<", "less than", LessThanExpression)
        self._greater_than_or_equal_to_expression_btn = self._get_expression_button(">=", "Greater than or equal to", GreaterThanOrEqualToExpression)
        self._less_than_or_equal_to_expression_btn = self._get_expression_button("<=", "Less than or equal to", LessThanOrEqualToExpression)
        self._and_expression_btn = self._get_expression_button("and", "And", AndExpression)
        self._or_expression_btn = self._get_expression_button("or", "Or", OrExpression)
        self._not_expression_btn = self._get_expression_button("not", "Not", NotExpression)
        self._cancel_btn = QtWidgets.QPushButton(u"Cancel")

        """ Create Layouts """
        header_layout = QtWidgets.QVBoxLayout()
        header_layout.addWidget(self._title)
        self._header_layout_widget.setLayout(header_layout)

        controls_layout = QtWidgets.QHBoxLayout()
        controls_layout.setContentsMargins(0, 0, 0, 0)
        controls_layout.setAlignment(QtCore.Qt.AlignRight)
        controls_layout.addWidget(self._cancel_btn)

        # Simple Value Expressions
        simple_value_expressions_layout = QtWidgets.QGridLayout()
        simple_value_expressions_layout.setSpacing(10)
        simple_value_expressions_layout.addWidget(self._true_expression_btn, 0, 0)
        simple_value_expressions_layout.addWidget(self._false_expression_btn, 0, 1)
        simple_value_expressions_layout.addWidget(self._integer_number_expression_btn, 0, 2)
        simple_value_expressions_layout.addWidget(self._decimal_number_expression_btn, 0, 3)

        simple_value_expressions_layout_container = QtWidgets.QHBoxLayout()
        simple_value_expressions_layout_container.setContentsMargins(0, 0, 0, 0)
        simple_value_expressions_layout_container.setAlignment(QtCore.Qt.AlignRight)
        simple_value_expressions_layout_container.addLayout(simple_value_expressions_layout)
        simple_value_expressions_layout_container.addStretch()

        # Date and Time Expressions
        datetime_expressions_layout = QtWidgets.QGridLayout()
        datetime_expressions_layout.setSpacing(10)
        datetime_expressions_layout.addWidget(self._date_time_expression_btn, 0, 0)
        datetime_expressions_layout.addWidget(self._date_expression_btn, 0, 1)
        datetime_expressions_layout.addWidget(self._time_expression_btn, 0, 2)
        datetime_expressions_layout.addWidget(self._now_expression_btn, 1, 0)
        datetime_expressions_layout.addWidget(self._today_expression_btn, 1, 1)

        datetime_expressions_layout_container = QtWidgets.QHBoxLayout()
        datetime_expressions_layout_container.setContentsMargins(0, 0, 0, 0)
        datetime_expressions_layout_container.setAlignment(QtCore.Qt.AlignRight)
        datetime_expressions_layout_container.addLayout(datetime_expressions_layout)
        datetime_expressions_layout_container.addStretch()

        # Comparision Expressions
        comparision_expressions_layout = QtWidgets.QGridLayout()
        comparision_expressions_layout.setSpacing(10)
        comparision_expressions_layout.addWidget(self._equal_expression_btn, 0, 0)
        comparision_expressions_layout.addWidget(self._not_equal_expression_btn, 0, 1)
        comparision_expressions_layout.addWidget(self._greater_than_expression_btn, 0, 2)
        comparision_expressions_layout.addWidget(self._less_than_expression_btn, 0, 3)
        comparision_expressions_layout.addWidget(self._greater_than_or_equal_to_expression_btn, 1, 0)
        comparision_expressions_layout.addWidget(self._less_than_or_equal_to_expression_btn, 1, 1)

        comparision_expressions_layout_container = QtWidgets.QHBoxLayout()
        comparision_expressions_layout_container.setContentsMargins(0, 0, 0, 0)
        comparision_expressions_layout_container.setAlignment(QtCore.Qt.AlignRight)
        comparision_expressions_layout_container.addLayout(comparision_expressions_layout)
        comparision_expressions_layout_container.addStretch()

        # logic Expressions
        logic_expressions_layout = QtWidgets.QGridLayout()
        logic_expressions_layout.setSpacing(10)
        logic_expressions_layout.addWidget(self._and_expression_btn, 0, 0)
        logic_expressions_layout.addWidget(self._or_expression_btn, 0, 1)
        logic_expressions_layout.addWidget(self._not_expression_btn, 0, 2)

        logic_expressions_layout_container = QtWidgets.QHBoxLayout()
        logic_expressions_layout_container.setContentsMargins(0, 0, 0, 0)
        logic_expressions_layout_container.setAlignment(QtCore.Qt.AlignRight)
        logic_expressions_layout_container.addLayout(logic_expressions_layout)
        logic_expressions_layout_container.addStretch()

        content_layout = QtWidgets.QVBoxLayout(self._content_widget)
        content_layout.setContentsMargins(20, 20, 20, 20)

        content_layout.addWidget(self._simple_value_expression_types_header)
        content_layout.addLayout(simple_value_expressions_layout_container)
        content_layout.addSpacing(10)

        content_layout.addWidget(self._datetime_expression_types_header)
        content_layout.addLayout(datetime_expressions_layout_container)
        content_layout.addSpacing(10)

        content_layout.addWidget(self._comparision_expression_types_header)
        content_layout.addLayout(comparision_expressions_layout_container)
        content_layout.addSpacing(10)

        content_layout.addWidget(self._logic_expression_types_header)
        content_layout.addLayout(logic_expressions_layout_container)
        content_layout.addStretch()

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self._header_layout_widget)
        layout.addWidget(self._scroll_area)
        layout.addLayout(controls_layout)

        """ Create Connections """
        self._cancel_btn.clicked.connect(self.cancel)

    def _get_expression_button(self, label, tooltip, expression_type):
        expression_btn = QtWidgets.QPushButton(label)
        expression_btn.setFlat(True)
        expression_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)
        expression_btn.setProperty(qt_constants.STYLE_PROPS.H2, True)
        expression_btn.setFixedSize(ExpressionCreatorWindow.BUTTON_SIZE, ExpressionCreatorWindow.BUTTON_SIZE)
        expression_btn.setToolTip(qt_utils.split_tool_tip(tooltip))

        if expression_type:
            expression_btn.clicked.connect(lambda: self._handle_expression_selection(expression_type))

        return expression_btn

    def _handle_expression_selection(self, statement_type):
        self._chosen_expression_type = statement_type
        self._cancel_expression_creation = False
        self.hide()

    def chosen_expression_type(self):
        return self._chosen_expression_type

    def cancel_expression_creation(self):
        return self._cancel_expression_creation

    def cancel(self):
        self._cancel_expression_creation = True
        self.hide()


class StatementCreatorWindow(QtWidgets.QDialog, QtPluginAccessor):

    INITIAL_WINDOW_WIDTH = 800
    INITIAL_WINDOW_HEIGHT = 700

    MINIMUM_WINDOW_WIDTH = 600
    MINIMUM_WINDOW_HEIGHT = 600

    def __init__(self, plugin, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        QtPluginAccessor.__init__(self, plugin)

        self.setWindowModality(QtCore.Qt.ApplicationModal)


        # Remove the window help button that only appears on Windows OS
        self.setWindowFlags(self.windowFlags() ^ QtCore.Qt.WindowContextHelpButtonHint)

        # Set the window title
        self.setWindowTitle(u"GridMarkets Statement Creator")

        # resize the window to it's initial size
        self.resize(StatementCreatorWindow.INITIAL_WINDOW_WIDTH, StatementCreatorWindow.INITIAL_WINDOW_HEIGHT)

        # set minimum dimensions
        self.setMinimumWidth(StatementCreatorWindow.MINIMUM_WINDOW_WIDTH)
        self.setMinimumHeight(StatementCreatorWindow.MINIMUM_WINDOW_HEIGHT)

        self._chosen_statement_type = None
        self._cancel_statement_creation = True

        """ Create Widgets """
        self._header_layout_widget = QtWidgets.QWidget()
        self._header_layout_widget.setProperty(qt_constants.STYLE_PROPS.BACKGROUND_LEVEL_0, True)
        self._title = QtWidgets.QLabel(u"Choose a Statement Type")
        self._title.setProperty(qt_constants.STYLE_PROPS.H3, True)

        self._scroll_area = QtWidgets.QScrollArea()
        self._content_widget = QtWidgets.QWidget()
        self._content_widget.setProperty("background-level-2", True)
        self._scroll_area.setWidget(self._content_widget)
        self._scroll_area.setWidgetResizable(True)

        self._action_statement_types_header = QtWidgets.QLabel("Action Statement Types:")
        self._action_statement_types_header.setProperty(qt_constants.STYLE_PROPS.H4, True)

        self._logic_statement_types_header = QtWidgets.QLabel("Logic Statement Types:")
        self._logic_statement_types_header.setProperty(qt_constants.STYLE_PROPS.H4, True)

        button_size = 150

        self._send_email_statement_btn = QtWidgets.QPushButton("send\nemail")
        self._send_email_statement_btn.setFlat(True)
        self._send_email_statement_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)
        self._send_email_statement_btn.setProperty(qt_constants.STYLE_PROPS.H2, True)
        self._send_email_statement_btn.setFixedSize(button_size, button_size)

        self._if_statement_btn = QtWidgets.QPushButton("if")
        self._if_statement_btn.setFlat(True)
        self._if_statement_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)
        self._if_statement_btn.setProperty(qt_constants.STYLE_PROPS.H2, True)
        self._if_statement_btn.setFixedSize(button_size, button_size)

        self._return_statement_btn = QtWidgets.QPushButton("return")
        self._return_statement_btn.setFlat(True)
        self._return_statement_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)
        self._return_statement_btn.setProperty(qt_constants.STYLE_PROPS.H2, True)
        self._return_statement_btn.setFixedSize(button_size, button_size)

        self._pass_statement_btn = QtWidgets.QPushButton("pass")
        self._pass_statement_btn.setFlat(True)
        self._pass_statement_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)
        self._pass_statement_btn.setProperty(qt_constants.STYLE_PROPS.H2, True)
        self._pass_statement_btn.setFixedSize(button_size, button_size)

        self._cancel_btn = QtWidgets.QPushButton(u"Cancel")

        """ Create Layouts """
        header_layout = QtWidgets.QVBoxLayout()
        header_layout.addWidget(self._title)
        self._header_layout_widget.setLayout(header_layout)

        controls_layout = QtWidgets.QHBoxLayout()
        controls_layout.setContentsMargins(0, 0, 0, 0)
        controls_layout.setAlignment(QtCore.Qt.AlignRight)
        controls_layout.addWidget(self._cancel_btn)

        # action statements
        action_statements_layout = QtWidgets.QGridLayout()
        action_statements_layout.setSpacing(10)
        action_statements_layout.addWidget(self._send_email_statement_btn, 0, 0)

        action_statements_layout_container = QtWidgets.QHBoxLayout()
        action_statements_layout_container.setContentsMargins(0, 0, 0, 0)
        action_statements_layout_container.setAlignment(QtCore.Qt.AlignRight)
        action_statements_layout_container.addLayout(action_statements_layout)
        action_statements_layout_container.addStretch()

        # logic statements
        logic_statements_layout = QtWidgets.QGridLayout()
        logic_statements_layout.setSpacing(10)
        logic_statements_layout.addWidget(self._if_statement_btn, 0, 0)
        logic_statements_layout.addWidget(self._return_statement_btn, 0, 1)
        logic_statements_layout.addWidget(self._pass_statement_btn, 0, 2)

        logic_statements_layout_container = QtWidgets.QHBoxLayout()
        logic_statements_layout_container.setContentsMargins(0, 0, 0, 0)
        logic_statements_layout_container.setAlignment(QtCore.Qt.AlignRight)
        logic_statements_layout_container.addLayout(logic_statements_layout)
        logic_statements_layout_container.addStretch()

        content_layout = QtWidgets.QVBoxLayout(self._content_widget)
        content_layout.setContentsMargins(20, 20, 20, 20)
        content_layout.addWidget(self._action_statement_types_header)
        content_layout.addLayout(action_statements_layout_container)
        content_layout.addSpacing(10)
        content_layout.addWidget(self._logic_statement_types_header)
        content_layout.addLayout(logic_statements_layout_container)
        content_layout.addStretch()

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self._header_layout_widget)
        layout.addWidget(self._scroll_area)
        layout.addLayout(controls_layout)

        """ Create Connections """
        self._cancel_btn.clicked.connect(self.cancel)

        self._if_statement_btn.clicked.connect(lambda: self._handle_statement_selection(IfStatement))
        self._return_statement_btn.clicked.connect(lambda: self._handle_statement_selection(ReturnStatement))
        self._pass_statement_btn.clicked.connect(lambda: self._handle_statement_selection(PassStatement))

    def _handle_statement_selection(self, statement_type):
        self._chosen_statement_type = statement_type
        self._cancel_statement_creation = False
        self.hide()

    def chosen_statement_type(self):
        return self._chosen_statement_type

    def cancel_statement_creation(self):
        return self._cancel_statement_creation

    def cancel(self):
        self._cancel_statement_creation = True
        self.hide()
