__all__ = 'FileTreeModel'

import sys
import pathlib
from PySide2 import QtCore
from PySide2 import QtGui
from gm_qt_plugin import qt_constants
from gm_meta_plugin import utils


class TreeNode:
    """ A Simple class to represent a node in a tree for use with tree models. """

    def __init__(self, column_data, parent_node=None):
        """
        :param column_data: A List containing the column data for the node.
        :type column_data: list[any]
        :param parent_node: The parent node
        :type parent_node: TreeNode
        """

        self._column_data = column_data
        self._parent_node = parent_node
        self._child_nodes = []

    def append_child(self, child):
        self._child_nodes.append(child)

    def child(self, row):
        if row < 0 or row >= len(self._child_nodes):
            return None
        return self._child_nodes[row]

    def child_count(self):
        return len(self._child_nodes)

    def children(self):
        return self._child_nodes

    def get_descendants(self):
        descendants = []
        remaining_nodes = []
        remaining_nodes.extend(self.children())

        while remaining_nodes:
            node = remaining_nodes.pop()
            descendants.append(node)
            remaining_nodes.extend(node.children())

        return descendants

    def column_count(self):
        return len(self._column_data)

    def data(self, column):
        if column < 0 or column >= len(self._column_data):
            return None

        return self._column_data[column]

    def set_data(self, column, value):
        if column < 0 or column >= len(self._column_data):
            return False

        if column == FileTreeModel.UPLOAD_INDEX:
            self._column_data[FileTreeModel.UPLOAD_INDEX] = value

            # if upload has been checked then set all the upload flag for all children to True and iterate through
            # parents to update their upload flags accordingly
            if value:

                # Update children's upload flags
                children = self._child_nodes[:]

                # For each child
                while children:
                    # remove it from the children list
                    child = children.pop(0)

                    # set it's upload flag to True
                    child._column_data[FileTreeModel.UPLOAD_INDEX] = True

                    # add all its children to the children list so that we can recurse through the tree
                    children.extend(child._child_nodes)

                # Update parent's upload flags
                child_node = self
                while child_node.parent_item():
                    parent_node = child_node.parent_item()

                    # if there is a sibling with its upload flag set to false
                    if any(filter(lambda x: not x._column_data[FileTreeModel.UPLOAD_INDEX], child_node.siblings())):
                        # then don't update the parent and break the recursion
                        break
                    else:
                        # if all the siblings have their upload flag set to True then set the paretns upload flag to
                        # True and recurse up the tree
                        parent_node._column_data[FileTreeModel.UPLOAD_INDEX] = True
                        child_node = parent_node

                return True

            # if the upload flag has been un-checked then unset the upload flag for all children and all parents
            # Update children's upload flags
            children = self._child_nodes[:]

            # For each child
            while children:
                # remove it from the children list
                child = children.pop(0)

                # set it's upload flag to False
                child._column_data[FileTreeModel.UPLOAD_INDEX] = False

                # add all its children to the children list so that we can recurse through the tree
                children.extend(child._child_nodes)

            # Update parent's upload flags
            child_node = self
            while child_node.parent_item():
                parent_node = child_node.parent_item()

                if not parent_node._column_data[FileTreeModel.UPLOAD_INDEX]:
                    break

                parent_node._column_data[FileTreeModel.UPLOAD_INDEX] = False
                child_node = parent_node
            return True

        else:
            self._column_data[column] = value

        return True

    def row(self):
        if self._parent_node:
            return self._parent_node._child_nodes.index(self)
        return 0

    def parent_item(self):
        return self._parent_node

    def siblings(self):
        if self._parent_node:
            # get all the parents children
            siblings = self._parent_node._child_nodes[:]

            # get index of node in question
            row = siblings.index(self)

            # remove self from the parents children to get all siblings
            siblings.pop(row)

            return siblings

        return []

    def parents(self):
        parents = []

        node = self
        while node.parent_item():
            parents.append(node.parent_item())
            node = node.parent_item()

        return parents

    def get_path(self):
        return self.data(FileTreeModel.PATH_INDEX)

    def upload(self):
        return self.data(FileTreeModel.UPLOAD_INDEX)


class FileTreeModelHeader:
    def __init__(self, display_text, tool_tip):
        self._display_text = display_text
        self._tool_tip = tool_tip

    def get_display_text(self):
        return self._display_text

    def get_tool_tip(self):
        return self._tool_tip


class FileTreeModel(QtCore.QAbstractItemModel):
    PATH_HEADER = FileTreeModelHeader(u"Local Path", u"The local path of the file or directory on your file system.")
    PATH_INDEX = 0

    SIZE_HEADER = FileTreeModelHeader(u"Size", u"The size of the file.")
    SIZE_INDEX = 1

    UPLOAD_HEADER = FileTreeModelHeader(u"Upload", u"Select to upload the file to the remote folder.")
    UPLOAD_INDEX = 2

    FileTreeModelHeaders = [PATH_HEADER, SIZE_HEADER, UPLOAD_HEADER]

    def __init__(self, packed_project, parent=None):
        QtCore.QAbstractItemModel.__init__(self, parent=parent)

        # Get the main local root folder for the project
        project_root_path = packed_project.get_root_dir()

        # Create the root node
        self.root_node = TreeNode([None, None, None])

        # Get the project files
        project_files = packed_project.get_files()

        self.project_root = TreeNode([project_root_path, None, bool(project_files)], parent_node=self.root_node)
        self.root_node.append_child(self.project_root)

        self._build_model(self.project_root, project_files)

        # Check for files external to the main local root folder
        for project_file in list(project_files):
            if project_root_path not in project_file.parents:
                # If a external file is found insert it into the model
                self.insert_file_into_model(project_file)

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if orientation == QtCore.Qt.Horizontal:

            header = FileTreeModel.FileTreeModelHeaders[section]

            if role == QtCore.Qt.DisplayRole:
                return header.get_display_text()

            if role == QtCore.Qt.ToolTipRole or role == QtCore.Qt.WhatsThisRole:
                return header.get_tool_tip()

        return None

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if not index.isValid():
            return None

        item = index.internalPointer()
        column = index.column()

        path = item.data(FileTreeModel.PATH_INDEX)
        size = item.data(FileTreeModel.SIZE_INDEX)
        upload = item.data(FileTreeModel.UPLOAD_INDEX)

        if column == FileTreeModel.PATH_INDEX:

            if role == QtCore.Qt.DisplayRole:
                # convert pathlib.Path data into a unicode strings that can be displayed
                if item == self.project_root:
                    # display the full path for the project root (the top most tree node that is displayed)
                    return str(path)
                else:
                    # otherwise just display the last part of the path
                    return path.parts[-1]

            if role == QtCore.Qt.ToolTipRole or role == QtCore.Qt.WhatsThisRole:
                return str(path)

            if role == QtCore.Qt.DecorationRole:
                if path.is_dir():
                    return qt_constants.DIRECTORY_ICON_WHITE
                if path.is_file():
                    return qt_constants.FILE_ICON_WHITE

        elif column == FileTreeModel.SIZE_INDEX:
            if role == QtCore.Qt.DisplayRole:
                if size is not None:
                    return utils.bytes_to_human(size)

        elif column == FileTreeModel.UPLOAD_INDEX:
            if role == QtCore.Qt.CheckStateRole:
                return QtCore.Qt.Checked if upload else QtCore.Qt.Unchecked

        if column != FileTreeModel.UPLOAD_INDEX and role == QtCore.Qt.ForegroundRole:

            if upload:
                return QtGui.QBrush(QtGui.QColor(255, 255 ,255))
            else:
                return QtGui.QBrush(QtGui.QColor(150, 150, 150))

        return None

    def setData(self, index, value, role):
        if not index.isValid():
            return False

        item = index.internalPointer()
        column = index.column()

        # toggle the upload checkbox
        if column == FileTreeModel.UPLOAD_INDEX:
            if role == QtCore.Qt.CheckStateRole:
                # get the checked state for the current item
                is_checked = item.data(FileTreeModel.UPLOAD_INDEX)

                # toggle the check state for the current the item
                item.set_data(FileTreeModel.UPLOAD_INDEX, not is_checked)

                # emitting dataChanged for items with different parents is undefined in qt tree models so just update
                # the whole data structure
                self.dataChanged.emit(QtCore.QModelIndex(), QtCore.QModelIndex())

        return QtCore.QAbstractItemModel.setData(self, index, value, role)

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags

        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsUserCheckable

    def index(self, row, column, parent=None):

        if parent is None:
            parent = QtCore.QModelIndex()

        # validate the index exists
        if not self.hasIndex(row, column, parent):
            return QtCore.QModelIndex()

        if not parent.isValid():
            parent_item = self.root_node
        else:
            parent_item = parent.internalPointer()

        child_item = parent_item.child(row)
        if child_item:
            return self.createIndex(row, column, child_item)
        return QtCore.QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QtCore.QModelIndex()

        child_item = index.internalPointer()
        parent_item = child_item.parent_item()

        if parent_item == self.root_node:
            return QtCore.QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parent_item = self.root_node
        else:
            parent_item = parent.internalPointer()

        return parent_item.child_count()

    def columnCount(self, parent=QtCore.QModelIndex()):

        if parent.isValid():
            return parent.internalPointer().column_count()

        return self.root_node.column_count()

    def _build_model(self, node, project_files):

        # get the path for the node
        path = node.data(self.PATH_INDEX)

        # recursively iterate through the files and directories in this path
        for item in path.iterdir():

            # if the item is a file or a directory that contains at lest one file
            if item.is_file() or item.is_dir() and utils.get_files_in_directory(item):

                size = item.stat().st_size if item.is_file() else None

                # create a new node for the child path
                child = TreeNode([item, size, False], parent_node=node)

                # append it to the parent node
                node.append_child(child)

                # select the file if it is in the project files
                if item.is_file() and item in project_files:
                    child.set_data(FileTreeModel.UPLOAD_INDEX, True)

                # finally recurse on directories
                if item.is_dir():
                    self._build_model(child, project_files)

    """
    Given a file path this method creates TreeNodes for every part of the file path (should the TreeNode not already 
    exist). It does not add every file / directory along the way, only the parts required to add this file. 
    """
    def insert_file_into_model(self, file_path):
        current_parent_node = self.root_node
        parts = file_path.parts

        # Iterate through the path parts for the new file and create any TreeNodes which do not already exist
        for part in parts:

            # Get the pathlib.Path of this part of the path
            path_part = pathlib.Path(parts[0])
            for _part in parts:
                path_part /= pathlib.Path(_part)
                if (part == _part):
                    break

            for child_node in current_parent_node.children():
                if child_node.get_path() == path_part:
                    current_parent_node = child_node
                    break
            else:
                # If the path part is not already a Tree node attached to the parent then create it
                new_child_node = TreeNode([path_part, None, True], parent_node=current_parent_node)
                current_parent_node.append_child(new_child_node)

                current_parent_node = new_child_node

    def get_files_to_upload(self):
        files_to_upload = []

        descendants = self.root_node.get_descendants()

        while descendants:
            descendant = descendants.pop(0)

            descendant_path = descendant.data(FileTreeModel.PATH_INDEX)
            descendant_upload_flag = descendant.data(FileTreeModel.UPLOAD_INDEX)

            if descendant_path.is_file() and descendant_upload_flag:
                files_to_upload.append(descendant_path)

        return files_to_upload

    def get_project_nodes(self):
        project_nodes = self.root_node.get_descendants()
        return project_nodes

    def set_default_expanded_state(self, tree_view):
        tree_view.collapseAll()

        project_nodes = self.get_project_nodes()

        indexes_to_expand = []

        def add_node_to_expanded_list(node):
            index = self.createIndex(node.row(), 0, node)
            indexes_to_expand.append(index)

        for project_node in project_nodes:

            if project_node.upload():
                # append the node to the expanded nodes list
                add_node_to_expanded_list(project_node)

                # expand all parents as well
                for parent in project_node.parents():
                    if parent:
                        add_node_to_expanded_list(parent)

        for index in indexes_to_expand:
            tree_view.expand(index)
