__all__ = "WatchPatternTable"

from PySide2 import QtCore


from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.watch_pattern import WatchPattern
from gm_meta_plugin.utils import get_desktop_path


class WatchPatternModelHeader:
    def __init__(self, display_text, tool_tip):
        self._display_text = display_text
        self._tool_tip = tool_tip

    def get_display_text(self):
        return self._display_text

    def get_tool_tip(self):
        return self._tool_tip


class WatchPatternTable(QtCore.QAbstractTableModel):

    PATTERN_HEADER = WatchPatternModelHeader("Watch Pattern", "The regex pattern that will be used to match files created by your job.")
    PATTERN_INDEX = 0

    DOWNLOAD_PATH_HEADER = WatchPatternModelHeader("Download Path", "The local path that the matched files should be downloaded to.")
    DOWNLOAD_PATH_INDEX = 1

    JOB_NAME_HEADER = WatchPatternModelHeader("Job", "The job the watch pattern belongs to.")
    JOB_NAME_INDEX = 2

    WatchPatternModelHeaders = [PATTERN_HEADER, DOWNLOAD_PATH_HEADER, JOB_NAME_HEADER]

    def __init__(self, packed_project, job_preset, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent=parent)

        self._packed_project = packed_project
        self._job_preset = job_preset

        self._watch_patterns = packed_project.get_watch_patterns()

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if orientation == QtCore.Qt.Horizontal:
            header = WatchPatternTable.WatchPatternModelHeaders[section]

            if role == QtCore.Qt.DisplayRole:
                return header.get_display_text()

            if role == QtCore.Qt.ToolTipRole or role == QtCore.Qt.WhatsThisRole:
                return header.get_tool_tip()

        return QtCore.QAbstractTableModel.headerData(self, section, orientation, role=role)

    def rowCount(self, parent=QtCore.QModelIndex()):

        #if parent.isValid():
        #    return parent.internalPointer().column_count()

        return len(self._watch_patterns)

    def columnCount(self, parent=QtCore.QModelIndex()):

        #if parent.isValid():
        #    return parent.internalPointer().column_count()

        return 3

    def get_watch_patterns(self):
        return self._watch_patterns

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.DisplayRole:

            row = index.row()
            column = index.column()


            if column < 2:
                watch_file = self._watch_patterns[row]

                if column == 0:
                    return watch_file.get_remote_watch_pattern()

                if column == 1:
                    return watch_file.get_local_download_path()

            if column == WatchPatternTable.JOB_NAME_INDEX:
                return self._job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.JOB_NAME).get_value()

    def add_row(self):
        self._watch_patterns.append(WatchPattern("/{}/.+".format(self._packed_project.get_name()),
                                                 get_desktop_path().replace('\\', '/')))
        self.layoutChanged.emit()

    def removeRows(self, position, rows=1, index=QtCore.QModelIndex()):
        self.beginRemoveRows(QtCore.QModelIndex(), position, position + rows - 1)

        i = 0
        while i < rows:
            del self._watch_patterns[position]
            i = i + 1

        self.endRemoveRows()

        return True

    def setData(self, modelIndex, value, role=QtCore.Qt.EditRole):
        if not modelIndex.isValid():
            return False

        row = modelIndex.row()
        column = modelIndex.column()

        watch_pattern = self._watch_patterns[row]
        if column == WatchPatternTable.PATTERN_INDEX:
            watch_pattern.set_remote_watch_pattern(value)

        if column == WatchPatternTable.DOWNLOAD_PATH_INDEX:
            watch_pattern.set_local_download_path(value.replace('\\', '/'))

        return True

    def flags(self, index):
        if not index.isValid(): return 0

        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable