__all__ = 'FrameRangeInput'

from .input_widget import InputWidget
from PySide2 import QtWidgets
from PySide2 import QtCore
from gm_qt_plugin import qt_constants
from gm_qt_plugin.widgets.input_widgets.spin_box import SpinBox
from gm_qt_plugin.widgets.input_widgets.check_box import CheckBox
from gm_qt_plugin.widgets.elided_label import ElidedLabel
from gm_meta_plugin.utils import do_frame_ranges_overlap


class FrameRangeInput(QtWidgets.QWidget, InputWidget):
    # Don't make these too large or the spin boxes refuse to shrink (even when empty)
    MAX_FRAME_VALUE = 99999
    MIN_FRAME_VALUE = -99999

    def __init__(self, plugin, parent=None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        InputWidget.__init__(self, plugin)

        """ Create Widgets """
        self.group_box = QtWidgets.QGroupBox()
        self.group_box.setProperty("background-level-2", True)

        self.list = QtWidgets.QListWidget()
        self.list.setProperty("background-level-2", True)
        self.list.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.list.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.list.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.list.setAlternatingRowColors(True)

        self._add_frame_range_btn = QtWidgets.QPushButton(u"+")
        self._add_frame_range_btn.setFlat(True)
        self._add_frame_range_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)

        self._remove_frame_range_btn = QtWidgets.QPushButton(u"-")
        self._remove_frame_range_btn.setFlat(True)
        self._remove_frame_range_btn.setProperty(qt_constants.STYLE_PROPS.BORDERED, True)

        self._warning_label = QtWidgets.QWidget()
        self._warning_label_text = ElidedLabel("Warning: frame ranges overlap. You may be charged for overlapping " 
                                               "frames but will receive only one output frame for each conflict.")
        self._warning_label_text.setColor('grey')

        self._warning_label_icon = QtWidgets.QLabel()
        self._warning_label_icon.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self._warning_label_icon.setPixmap(qt_constants.WARNING_ICON_OFF_WHITE.pixmap(24))
        self._warning_label.setVisible(False)  # starts invisible

        """ Create Layout """
        self.vlayout = QtWidgets.QVBoxLayout()
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.vlayout.setAlignment(QtCore.Qt.AlignTop)
        self.vlayout.addWidget(self._add_frame_range_btn)
        self.vlayout.addWidget(self._remove_frame_range_btn)

        self.hlayout = QtWidgets.QHBoxLayout(self.group_box)
        self.hlayout.setContentsMargins(0, 0, 0, 0)
        self.hlayout.addWidget(self.list)
        self.hlayout.addLayout(self.vlayout)

        self._warning_label_layout = QtWidgets.QHBoxLayout(self._warning_label)
        self._warning_label_layout.setContentsMargins(0, 0, 0, 0)
        self._warning_label_layout.addWidget(self._warning_label_icon)
        self._warning_label_layout.addWidget(self._warning_label_text)

        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.group_box)
        self.layout.addWidget(self._warning_label)

        """ Create Connections """
        self._add_frame_range_btn.clicked.connect(lambda: self._add_frame_range(1, 1, 1))
        self._remove_frame_range_btn.clicked.connect(self._remove_frame_range)
        self.input_value_changed.connect(self._update_warning_label)

    def _update_warning_label(self):
        frame_ranges = []

        count = self.list.count()
        i = 0
        while i < count:
            item = self.list.item(i)
            item_widget = self.list.itemWidget(item)

            if item_widget.enabled():
                start = item_widget.start()
                end = item_widget.end()
                step = item_widget.step()
                frame_ranges.append((start, end, step))

            i = i + 1

        self._warning_label.setVisible(do_frame_ranges_overlap(frame_ranges))

    def _update_list_height(self):
        item_count = self.list.count()

        base_height = -35
        item_height = 40

        if item_count <= 1:
            height = base_height + (item_height * 2)
        elif item_count == 2:
            height = base_height + (item_height * 3)
        elif item_count == 3:
            height = base_height + (item_height * 4)
        elif item_count == 4:
            height = base_height + (item_height * 5)
        else:
            height = base_height + (item_height * 6)

        self.list.setFixedHeight(height)

    def _remove_frame_range(self):

        if self.list.count() <= 1:
            return

        self.list.takeItem(self.list.count() - 1)
        self._remove_frame_range_btn.setEnabled(self.list.count() > 1)
        self._update_warning_label()
        self._update_list_height()

    def _add_frame_range(self, frame_start, frame_end, frame_step):
        frame_range_item = QtWidgets.QListWidgetItem()

        """ Create Widgets """
        frame_range_widget = QtWidgets.QWidget()

        # Check box for enabling the frame range
        check_box = CheckBox(self.get_plugin())
        check_box.set_value(True)
        check_box.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        # Create spin boxes and labels
        start_label = QtWidgets.QLabel("Start:")
        start_label.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
        start_input = SpinBox(self.get_plugin())
        start_input.setMinimum(FrameRangeInput.MIN_FRAME_VALUE)
        start_input.setMaximum(FrameRangeInput.MAX_FRAME_VALUE)
        start_input.set_value(frame_start)

        QtWidgets.QLabel("End")
        end_label = QtWidgets.QLabel("End:")
        end_label.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
        end_input = SpinBox(self.get_plugin())
        end_input.setMinimum(FrameRangeInput.MIN_FRAME_VALUE)
        end_input.setMaximum(FrameRangeInput.MAX_FRAME_VALUE)
        end_input.set_value(frame_end)

        step_label = QtWidgets.QLabel("Step:")
        step_label.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
        step_input = SpinBox(self.get_plugin())
        step_input.setMinimum(1)
        step_input.setMaximum(FrameRangeInput.MAX_FRAME_VALUE)
        step_input.set_value(frame_step)

        # let the spin boxes shrink
        check_box.setMinimumSize(0, 0)
        start_input.setMinimumSize(0, 0)
        end_input.setMinimumSize(0, 0)
        step_input.setMinimumSize(0, 0)

        """ Create Layout """
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 20)

        layout.addWidget(check_box)

        layout.addWidget(start_label)
        layout.addWidget(start_input)
        layout.addSpacing(5)
        layout.addWidget(end_label)
        layout.addWidget(end_input)
        layout.addSpacing(5)
        layout.addWidget(step_label)
        layout.addWidget(step_input)

        frame_range_widget.setLayout(layout)
        frame_range_widget.setMinimumSize(41, 60)
        frame_range_item.setSizeHint(frame_range_widget.sizeHint())

        """ Create Connections """

        def handleCheckBoxToggle(enabled):

            # Prevent disable if all other frame ranges are disabled
            if not check_box.isChecked():
                count = self.list.count()
                i = 0
                while True:
                    item = self.list.item(i)
                    item_widget = self.list.itemWidget(item)

                    if item_widget is not frame_range_widget:
                        if item_widget.enabled():
                            # Found another enabled frame range
                            break

                    i = i + 1
                    if i >= count:
                        # Could not find another enabled frame range
                        check_box.setChecked(True)

                        QtWidgets.QMessageBox.information(self,
                                                          u"",
                                                          u"You must have at least one enabled frame range.",
                                                          QtWidgets.QMessageBox.Ok,
                                                          defaultButton=QtWidgets.QMessageBox.Ok)

                        return

            start_input.setEnabled(enabled)
            end_input.setEnabled(enabled)
            step_input.setEnabled(enabled)
            self._update_warning_label()

        check_box.toggled.connect(handleCheckBoxToggle)

        def _catch_start_value_change():
            # Start input must always be smaller than end input
            if start_input.get_value() > end_input.get_value():
                end_input.set_value(start_input.get_value())

            self._catch_value_change()

        def _catch_end_value_change():
            # End input must always be greater than start input
            if end_input.get_value() < start_input.get_value():
                start_input.set_value(end_input.get_value())

            self._catch_value_change()

        start_input.valueChanged.connect(_catch_start_value_change)
        end_input.valueChanged.connect(_catch_end_value_change)
        step_input.valueChanged.connect(self._catch_value_change)

        frame_range_widget.enabled = lambda: check_box.isChecked()
        frame_range_widget.start = lambda: start_input.value()
        frame_range_widget.end = lambda: end_input.value()
        frame_range_widget.step = lambda: step_input.value()
        frame_range_widget.value = lambda: "{0} {1} {2}".format(start_input.value(), end_input.value(),
                                                                step_input.value())

        # Add widget to QListWidget funList
        self.list.addItem(frame_range_item)
        self.list.setItemWidget(frame_range_item, frame_range_widget)
        self._update_warning_label()
        self._remove_frame_range_btn.setEnabled(self.list.count() > 1)
        self._update_list_height()

    def minimumSizeHint(self):
        return QtCore.QSize(0, 0)

    def get_value(self):
        return self.to_string()

    def set_value(self, value):
        frame_ranges = FrameRangeInput.parse_frame_ranges(value)

        # There must always be at least one frame range
        if not frame_ranges:
            frame_ranges = [(1, 1, 1)]

        self.list.clear()
        for frame_range in frame_ranges:
            self._add_frame_range(frame_range[0], frame_range[1], frame_range[2])

    def to_string(self):
        string = ''
        count = self.list.count()
        i = 0

        while i < count:
            item = self.list.item(i)
            item_widget = self.list.itemWidget(item)

            if item_widget.enabled():
                string = string + item_widget.value()

                if i < count - 1:
                    string = string + ','

            i = i + 1

        return string

    @staticmethod
    def parse_frame_ranges(frame_ranges_str):
        return list(map(lambda x: FrameRangeInput.parse_frame_range(x), frame_ranges_str.split(',')))

    @staticmethod
    def parse_frame_range(frame_range):
        values = frame_range.split(' ')

        try:
            start = int(values[0])
            end = int(values[1])
            step = int(values[2])
        except Exception:
            start = 0
            end = 0
            step = 0

        return start, end, step
