__all__ = 'Job Preset'

import gm_meta_plugin
from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_qt_plugin.widgets.job_preset_attribute import JobPresetAttribute

from gm_qt_plugin import qt_constants, qt_utils
from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

from gm_qt_plugin.widgets.qhline import QHLine
from gm_qt_plugin.widgets.input_widgets.check_box import CheckBox
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor


class JobPresetAttributeGroup(QtWidgets.QWidget, QtPluginAccessor):

    def __init__(self, plugin, job_preset, name, parent=None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        QtPluginAccessor.__init__(self, plugin)

        self._job_preset = job_preset
        self._name = name

        ''' Create Widgets '''
        self._header = QtWidgets.QLabel(name)
        self._header.setProperty('h4', True)

        self._separator = QHLine()

        self._attribute_group_box = QtWidgets.QGroupBox()
        self._attribute_group_box.setAlignment(QtCore.Qt.AlignTop)

        ''' Create Layouts '''
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(0,0,0,0)

        self._grid_layout = QtWidgets.QGridLayout()
        self._grid_layout.setAlignment(QtCore.Qt.AlignTop)
        self._grid_layout.setSpacing(8)
        self._grid_layout.setContentsMargins(0, 0, 0, 0)

        group_box_v_layout = QtWidgets.QVBoxLayout()
        group_box_v_layout.setAlignment(QtCore.Qt.AlignTop)
        group_box_v_layout.setContentsMargins(10, 7, 10, 7)
        group_box_v_layout.addWidget(self._header)
        group_box_v_layout.addWidget(self._separator)
        group_box_v_layout.addSpacing(10)
        group_box_v_layout.addLayout(self._grid_layout)

        self._attribute_group_box.setLayout(group_box_v_layout)
        self._layout.addWidget(self._attribute_group_box)

        self._active_job_preset_attributes = []

        # Job Attribute groups are not visible until redrawn and explicitly shown
        self.setVisible(False)

    def redraw(self):
        old_active_job_preset_attributes = self._active_job_preset_attributes
        self._active_job_preset_attributes = self.get_active_job_preset_attributes()

        if old_active_job_preset_attributes == self._active_job_preset_attributes:
            return

        qt_utils.remove_widgets_from_layout(self._grid_layout)
        show_hidden_job_preset_attributes = JobPreset.show_hidden_job_preset_attributes()

        is_enabled_column = 0
        display_name_label_column = 1
        key_column = 2
        main_widget_column = 3 if show_hidden_job_preset_attributes else 2
        inference_source_column = 4 if show_hidden_job_preset_attributes else 3

        # add widgets
        current_row = 0
        widgets_to_show = []
        for i, job_preset_attribute in enumerate(self._active_job_preset_attributes):

            job_attribute = job_preset_attribute.get_job_attribute()

            # Always make sure the job preset attribute is initialised
            job_preset_attribute.init()

            # If hidden job preset attributes are not visible
            if not show_hidden_job_preset_attributes:
                inferences_sources = job_attribute.get_inference_sources()

                should_hide = True
                for inferences_source in inferences_sources:
                    if inferences_source == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
                        should_hide = False
                        break

                    if inferences_source == gm_meta_plugin.InferenceSource.get_application_inference_source():
                        should_hide = False
                        break

                if should_hide:
                    continue

            current_row = current_row + 1

            # Is enabled
            is_enabled_widget = job_preset_attribute.get_is_enabled_widget()
            vboxlayout = QtWidgets.QVBoxLayout()
            vboxlayout.setContentsMargins(0, 4, 0, 0)
            vboxlayout.addWidget(is_enabled_widget)
            self._grid_layout.addLayout(vboxlayout, current_row, is_enabled_column, alignment=QtCore.Qt.AlignTop)

            # display name
            display_name_label = job_preset_attribute.get_display_name_label()
            display_name_label.setAlignment(QtCore.Qt.AlignTop)
            vboxlayout = QtWidgets.QVBoxLayout()
            vboxlayout.setContentsMargins(0, 6, 0, 0)
            vboxlayout.addWidget(display_name_label)
            self._grid_layout.addLayout(vboxlayout, current_row, display_name_label_column)

            # key
            if show_hidden_job_preset_attributes:
                self._grid_layout.addWidget(job_preset_attribute.get_key_label(), current_row, key_column)

            # main widget
            job_preset_attribute.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
            self._grid_layout.addWidget(job_preset_attribute, current_row, main_widget_column)

            # inference source checkbox
            inference_source_widget = job_preset_attribute.get_inference_source_widget()
            vboxlayout = QtWidgets.QVBoxLayout()
            vboxlayout.setContentsMargins(0, 6, 0, 0)
            vboxlayout.addWidget(inference_source_widget)
            self._grid_layout.addLayout(vboxlayout, current_row, inference_source_column, alignment=QtCore.Qt.AlignTop)

            display_name_label.show()

            is_enabled_widget.setVisible(job_attribute.is_optional() and (job_attribute.get_optional_parent() is None))
            if show_hidden_job_preset_attributes:
                widgets_to_show.append(job_preset_attribute.get_key_label())

            widgets_to_show.append(job_preset_attribute)
            widgets_to_show.append(inference_source_widget)

        # set the column stretch factors
        self._grid_layout.setColumnStretch(is_enabled_column, 0)
        self._grid_layout.setColumnStretch(display_name_label_column, 0)
        if show_hidden_job_preset_attributes:
            self._grid_layout.setColumnStretch(key_column, 0)
        self._grid_layout.setColumnStretch(main_widget_column, 3)
        self._grid_layout.setColumnStretch(inference_source_column, 0)

        for widget_to_show in widgets_to_show:
            widget_to_show.show()

        # Only show the group is it contains widgets to show
        self.setVisible(bool(widgets_to_show))

    def get_job_preset(self):
        return self._job_preset

    def get_name(self):
        return self._name

    def get_job_preset_attributes(self):
        return self.get_job_preset().get_job_preset_attributes_in_group(self.get_name())

    def get_active_job_preset_attributes(self):
        active_job_preset_attributes = []
        for job_preset_attribute in self.get_job_preset_attributes():
            if job_preset_attribute.is_active():
                active_job_preset_attributes.append(job_preset_attribute)
        return active_job_preset_attributes

    @QtCore.Slot()
    def onJobPresetAttributeValueChanged(self):
        self.redraw()


class JobPreset(QtWidgets.QWidget, gm_meta_plugin.DynamicJobPreset):

    job_preset_attribute_value_changed = QtCore.Signal(str)

    _show_hidden_job_preset_attributes = CheckBox(None)
    _show_hidden_job_preset_attributes_label = QtWidgets.QLabel("Show Hidden Job Preset Attributes")

    def __init__(self, plugin, name, id, job_definition, toggle_lock_callback, is_locked=False, parent=None):
        QtWidgets.QWidget.__init__(self, parent=parent)

        self._id = id
        job_preset_attributes = list(map(lambda job_attribute: JobPresetAttribute(plugin, self, job_attribute),
                                         job_definition.get_attributes()))

        gm_meta_plugin.DynamicJobPreset.__init__(self, plugin, name, id, job_definition, job_preset_attributes, is_locked=is_locked)

        self._toggle_lock_callback = toggle_lock_callback
        self._column_count = 1

        ''' Create Widgets '''
        self._job_preset_attribute_group_widgets = []
        for attribute_group_index, attribute_group in enumerate(self.get_job_definition().get_attribute_groups()):
            self._job_preset_attribute_group_widgets.append(JobPresetAttributeGroup(self.get_plugin(),
                                                                                    self,
                                                                                    attribute_group))

        ''' Create Layouts '''
        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(0,0,0,0)
        self._layout.setSpacing(qt_constants.MARGIN_SIZE)
        self._column_widgets = []
        self._redraw_columns()

    def _redraw_columns(self):
        # hide all existing column widgets
        for column_widget in self._column_widgets:
            column_widget.hide()

        # create column widgets
        for i in range(self._column_count):
            # if the column_widget does not already exist then create one
            if len(self._column_widgets) < i + 1:
                column_widget = QtWidgets.QWidget()

                column_layout = QtWidgets.QVBoxLayout()
                column_layout.setContentsMargins(0, 0, 0, 0)
                column_layout.setAlignment(QtCore.Qt.AlignTop)
                column_layout.setSpacing(qt_constants.MARGIN_SIZE / 2)

                # add a stretch that all other widgets should be inserted after
                column_layout.addStretch()

                # set direction to bottom to top so the stretch stays at the bottom and all groups can be easily added
                # "before" the stretch with 'addWidget'
                column_layout.setDirection(QtWidgets.QBoxLayout.BottomToTop)

                column_widget.setLayout(column_layout)
                self._column_widgets.append(column_widget)

            self._layout.addWidget(self._column_widgets[i])
            self._column_widgets[i].show()

        job_preset_attribute_group_count = len(self.get_job_definition().get_attribute_groups())
        groups_per_column = job_preset_attribute_group_count / self._column_count
        current_column = 0

        for index, jp_attr_group_widget in enumerate(self._job_preset_attribute_group_widgets):
            if (index - (current_column * groups_per_column)) / groups_per_column >= 1 and current_column < self._column_count - 1:
                current_column = current_column + 1

            self._column_widgets[current_column].layout().insertWidget(1, jp_attr_group_widget)
            jp_attr_group_widget.redraw()

    def update_app_inferred_values(self):
        for job_preset_attribute in self.get_job_preset_attributes():
            job_preset_attribute.update_app_inferred_values()

    def toggle_locked_state(self):
        gm_meta_plugin.JobPreset.toggle_locked_state(self)
        self._toggle_lock_callback(self)

    def set_locked_state(self, is_locked):
        gm_meta_plugin.JobPreset.set_locked_state(self, is_locked)

    def show_hidden_job_preset_attributes_slot(self, state):
        pass

    def get_icon(self):
        job_definition = self.get_job_definition()
        job_definition_app_attribute = job_definition.get_attribute_with_id(api_constants.API_KEYS.APP)
        job_definition_app = job_definition_app_attribute.get_attribute().get_default_value()
        return qt_utils.get_product_icon(job_definition_app)

    def get_columns(self):
        if self.size().width() < 1250:
            return 1
        else:
            return 2

    def resizeEvent(self, event):
        # If old_width is not -1:
        new_column_count = self.get_columns()

        if self._column_count != new_column_count:
            self._column_count = new_column_count
            self._redraw_columns()

    @QtCore.Slot(str)
    def onJobPresetAttributeValueChanged(self, job_attribute_id: str):
        self.job_preset_attribute_value_changed.emit(job_attribute_id)

        for jp_attr_group_widget in self._job_preset_attribute_group_widgets:
            old_active_jp_attrs = jp_attr_group_widget._active_job_preset_attributes
            jp_attr_group_widget.redraw()
            new_active_jp_attrs = jp_attr_group_widget._active_job_preset_attributes

            activated_jp_attrs = []
            for new_active_jp_attr in new_active_jp_attrs:
                if new_active_jp_attr not in old_active_jp_attrs:

                    new_id = new_active_jp_attr.get_job_attribute().get_id()

                    # Do not double emit value changed signal for the same job preset attribute
                    if new_id != job_attribute_id:
                        self.job_preset_attribute_value_changed.emit(new_id)
                        activated_jp_attrs.append(new_active_jp_attr)

            # Flash all widgets that have just become active
            user_interface = self.get_plugin().get_user_interface()
            theme = user_interface.get_theme()
            original_color = QtGui.QColor(theme.FOREGROUND_COLOR)
            flash_color = QtGui.QColor(theme.BRAND_COLOR)
            for activated_jp_attr in activated_jp_attrs:
                qt_utils.flash_widget_color(activated_jp_attr.get_current_widget(), original_color, flash_color)

    @staticmethod
    def show_hidden_job_preset_attributes():
        return JobPreset._show_hidden_job_preset_attributes.checkState() == QtCore.Qt.Checked
