__all__ = 'JobPresetAttribute'

from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

import gm_meta_plugin
from gm_qt_plugin.widgets.input_widgets import Constant, CheckBox, ComboBox, ListRadioInput
from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin import qt_utils, qt_constants
from gm_qt_plugin.widgets.elided_label import ElidedLabel
from gm_qt_plugin.widgets.resizable_stacked_widget import ResizableStackedWidget


class JobPresetAttribute(QtWidgets.QWidget, gm_meta_plugin.DynamicJobPresetAttribute):

    def __init__(self, plugin, parent_preset, job_attribute, inference_source=None, value=None, parent=None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        gm_meta_plugin.DynamicJobPresetAttribute.__init__(self,
                                                          plugin,
                                                          parent_preset,
                                                          job_attribute,
                                                          inference_source=inference_source,
                                                          value=value)

        job_attribute = self.get_job_attribute()
        attribute = job_attribute.get_attribute()
        prop_id = job_attribute.get_id()
        inferences_sources = self.get_job_attribute().get_inference_sources()

        self._initialised = False

        ''' Create Widgets '''

        # display name label
        self._display_name_label = ElidedLabel(attribute.get_display_name())
        self._display_name_label.setMinimumWidth(130)
        self._display_name_label.setProperty('h5', True)

        # attribute key label
        self._attribute_key_label = ElidedLabel(self.get_key())
        self._attribute_key_label.setMinimumWidth(125)

        # various input widgets
        self._widgets = {}
        qt_utils.add_widgets_for_attribute(self.get_plugin(),
                                           job_attribute.get_attribute(),
                                           self._widgets,
                                           prop_id,
                                           job_preset=parent_preset)

        # create the inference source widget
        self._inference_source_widget = CheckBox(self.get_plugin())
        self._inference_source_widget_always_disabled = False

        # Create the is enabled widget
        self._is_enabled_widget = CheckBox(self.get_plugin())
        self._is_enabled_widget.set_value(job_attribute.default_enabled_state())
        self._is_enabled_widget.setToolTip(qt_utils.split_tool_tip('Optional attributes can be enabled or disabled. Disabling an attribute will mean no explicit value is passed for this attribute when submitting.'))

        # create a stacked widget to hold the different input widgets
        self._stacked_widget = ResizableStackedWidget()

        # validate the inference source options ( this are temporary checks for v1)
        is_application_inferred = gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources
        is_python_inferred = gm_meta_plugin.InferenceSource.get_python_inference_source() in inferences_sources
        is_user_inferred = gm_meta_plugin.InferenceSource.get_user_defined_inference_source() in inferences_sources
        is_constant_inferred = gm_meta_plugin.InferenceSource.get_constant_inference_source() in inferences_sources
        is_project_inferred = gm_meta_plugin.InferenceSource.get_project_inference_source() in inferences_sources

        if is_constant_inferred | is_project_inferred:
            if is_application_inferred | is_user_inferred:
                assert False

        if is_application_inferred and is_python_inferred:
            assert False

        if inferences_sources[0] == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
            self._inference_source_widget.setCheckState(QtCore.Qt.Checked)

        # if the job preset attribute is not either app inferred or user inferred or python inferred then don't show the
        # input
        if not (is_user_inferred and (is_python_inferred or is_application_inferred)):
            self._inference_source_widget.setProperty('hidden', True)
            self._inference_source_widget.setEnabled(False)
            self._inference_source_widget_always_disabled = True

        self._inference_source_widget.input_value_changed.connect(self.inference_source_option_changed)
        self._is_enabled_widget.input_value_changed.connect(self.is_enabled_changed)

        if gm_meta_plugin.InferenceSource.get_user_defined_inference_source() in inferences_sources:
            self._user_input_widget = self._widgets[prop_id]
            self._user_input_widget.input_value_changed.connect(
                lambda: self.get_parent_preset().onJobPresetAttributeValueChanged(job_attribute.get_id()))

            self._stacked_widget.addWidget(self._user_input_widget)

        if gm_meta_plugin.InferenceSource.get_constant_inference_source() in inferences_sources:
            self._constant_widget = Constant(plugin, job_attribute.get_attribute().get_default_value())
            self._constant_widget.setToolTip(qt_utils.split_tool_tip(attribute.get_description()))
            self._stacked_widget.addWidget(self._constant_widget)

        if gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources:
            # deferred until requested as it requires eval which cannot be used in the constructor
            self._app_inferred_value = None

        if gm_meta_plugin.InferenceSource.get_python_inference_source() in inferences_sources:
            self._python_widget = Constant(plugin, self.get_python_defined_value())
            self._python_widget.setToolTip(qt_utils.split_tool_tip(attribute.get_description()))
            self._python_widget.clicked.connect(self._handle_input_widget_clicked)
            self._python_widget.input_value_changed.connect(
                lambda: self.get_parent_preset().onJobPresetAttributeValueChanged(self.get_job_attribute().get_id()))

            self._stacked_widget.addWidget(self._python_widget)

        if len(inferences_sources) == 1:
            self._inference_source_widget.setEnabled(False)
            self._inference_source_widget_always_disabled = True

        # --- Connect attribute value changed slot
        self._inference_source_widget.input_value_changed.connect(lambda: self.get_parent_preset().onJobPresetAttributeValueChanged(job_attribute.get_id()))

        """ Create Layouts """
        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(0,0,0,0)
        self._layout.addWidget(self._stacked_widget)

    def init(self):
        """ Call before adding to a layout """
        if not self._initialised:

            job_attribute = self.get_job_attribute()

            # add the app input widget to that stacked widget
            inferences_sources = job_attribute.get_inference_sources()
            if gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources:
                self._stacked_widget.addWidget(self.get_app_inferred_widget())

            # set the starting widget
            current_widget = self.get_current_widget()

            # Check that the widget is in the stacked widget before setting to avoid warnings
            if self._stacked_widget.indexOf(current_widget) >= 0:
                self._stacked_widget.setCurrentWidget(current_widget)

            if job_attribute.is_optional():
                self.is_enabled_changed()

            # If the attribute has an optional parent then it's enabled state should be set based on that attributes
            # enabled changed signal
            if job_attribute.get_optional_parent():
                self.get_optional_parent()._is_enabled_widget.input_value_changed.connect(self.is_enabled_changed)

            self._initialised = True

    def update_app_inferred_values(self):
        current_widget = self._stacked_widget.currentWidget()

        # required to update combo box and radio items
        if isinstance(current_widget, ComboBox) or isinstance(current_widget, ListRadioInput):
            current_widget.update_items()

        is_app_inferred = self.get_inference_source() == gm_meta_plugin.InferenceSource.get_application_inference_source()
        is_python_inferred = self.get_inference_source() == gm_meta_plugin.InferenceSource.get_python_inference_source()

        if is_app_inferred or is_python_inferred:

            if current_widget is None:
                assert False, "current_widget is none for {}".format(self.get_id())

            old_value = current_widget.get_value()
            new_value = str(self.eval())

            if is_app_inferred:
                self.set_application_defined_value(new_value)
            else:
                self.set_python_defined_value(new_value)

            current_widget.set_value(new_value)

            if old_value != new_value:
                self.get_parent_preset().onJobPresetAttributeValueChanged(self.get_job_attribute().get_id())
                user_interface = self.get_plugin().get_user_interface()
                theme = user_interface.get_theme()
                original_color = QtGui.QColor(theme.FOREGROUND_COLOR)
                flash_color = QtGui.QColor(theme.BRAND_COLOR)
                qt_utils.flash_widget_color(current_widget, original_color, flash_color)

    @QtCore.Slot(int)
    def inference_source_option_changed(self):
        if self._inference_source_widget.isChecked():
            self.set_inference_source(gm_meta_plugin.InferenceSource.get_user_defined_inference_source())
            self._stacked_widget.setCurrentWidget(self.get_user_input_widget())
        else:
            inferences_sources = self.get_job_attribute().get_inference_sources()
            is_application_inferred = gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources
            is_python_inferred = gm_meta_plugin.InferenceSource.get_python_inference_source() in inferences_sources

            if is_application_inferred:
                self.set_inference_source(gm_meta_plugin.InferenceSource.get_application_inference_source())
                self._stacked_widget.setCurrentWidget(self.get_app_inferred_widget())
            elif is_python_inferred:
                self.set_inference_source(gm_meta_plugin.InferenceSource.get_python_inference_source())
                self._stacked_widget.setCurrentWidget(self.get_python_inferred_widget())
            else:
                assert False

        self.update_app_inferred_values()

    def is_enabled(self) -> bool:
        if self.get_job_attribute().get_optional_parent():
            is_enabled = self.get_optional_parent()._is_enabled_widget.get_value()
        else:
            is_enabled = self._is_enabled_widget.get_value()
        return is_enabled

    @QtCore.Slot(int)
    def is_enabled_changed(self):
        is_enabled = self.is_enabled()
        self.get_display_name_label().setEnabled(is_enabled)
        self.get_display_name_label().setColor('white' if is_enabled else 'grey')
        self.setEnabled(is_enabled)

        # If the inference source widget is not set to always be disabled then set it's enabled state
        if not self._inference_source_widget_always_disabled:
            inference_source_widget = self.get_inference_source_widget()
            inference_source_widget.setEnabled(is_enabled)

    def get_display_name_label(self):
        return self._display_name_label

    def get_key_label(self):
        return self._attribute_key_label

    def get_constant_widget(self):
        return self._constant_widget

    def get_user_input_widget(self):
        return self._user_input_widget

    def _handle_input_widget_clicked(self):
        inferences_sources = self.get_job_attribute().get_inference_sources()
        is_user_inferred = gm_meta_plugin.InferenceSource.get_user_defined_inference_source() in inferences_sources

        if is_user_inferred:
            # get the current app / python inferred value
            inferred_value = str(self.eval())

            self.get_user_input_widget().set_value(inferred_value)
            self._inference_source_widget.set_value(True)

    def get_app_inferred_widget(self):

        assert hasattr(self, '_app_inferred_value'), "JobPresetAttribute '{}' does not have an app inferred " \
                                                     "value".format(self.get_job_attribute().get_id())

        # we defer the creation of this widget as to set its value we need to use 'eval' which cannot be done in the
        # constructor
        if self._app_inferred_value is None:
            attribute = self.get_job_attribute().get_attribute()
            self._app_inferred_value = Constant(self.get_plugin(), u"")
            value = str(self.eval())
            self._app_inferred_value.set_value(value)
            self._app_inferred_value.setToolTip(qt_utils.split_tool_tip(attribute.get_description()))
            self._app_inferred_value.clicked.connect(self._handle_input_widget_clicked)
            self._app_inferred_value.input_value_changed.connect(lambda: self.get_parent_preset().onJobPresetAttributeValueChanged(self.get_job_attribute().get_id()))

        return self._app_inferred_value

    def get_python_inferred_widget(self):
        return self._python_widget

    def get_is_enabled_widget(self):
        return self._is_enabled_widget

    def get_inference_source_widget(self):
        return self._inference_source_widget

    def get_current_widget(self, project_inferred_value="Project inferred value"):
        inference_source = self.get_inference_source()

        if inference_source == gm_meta_plugin.InferenceSource.get_constant_inference_source():
            widget = self.get_constant_widget()

        elif inference_source == gm_meta_plugin.InferenceSource.get_application_inference_source():
            widget = self.get_app_inferred_widget()

        elif inference_source == gm_meta_plugin.InferenceSource.get_python_inference_source():
            widget = self.get_python_inferred_widget()
                
        elif inference_source == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
            widget = self.get_user_input_widget()

        elif inference_source == gm_meta_plugin.InferenceSource.get_project_inference_source():
            widget = Constant(self.get_plugin(), project_inferred_value)
            widget.setToolTip(qt_utils.split_tool_tip(self.get_job_attribute().get_attribute().get_description()))
        else:
            raise ValueError("Unsupported inference source type: " + inference_source.get_display_name())

        return widget

    """
    def get_user_defined_value(self) -> any:
        return self.get_user_input_widget().get_value()

    def set_user_defined_value(self, value: any) -> None:
        gm_meta_plugin.JobPresetAttribute.set_user_defined_value(value)
        self.get_user_input_widget().set_value(value)
    """

    def eval_user_defined_value(self, fail_on_eval_lock=True) -> any:
        if self._is_eval_locked:
            if fail_on_eval_lock:
                raise ValueError("Attempted to evaluate job preset attribute ('{}') value while evaluation lock is "
                                 "enabled".format(self.get_id()))
            return self.get_user_defined_value()

        user_defined_value = self.get_user_input_widget().get_value()
        self.set_user_defined_value(user_defined_value)
        return self.get_user_defined_value()

    def set_python_defined_value(self, value: any) -> None:
        try:
            self.get_python_inferred_widget().set_value(value)
        except AttributeError as e:
            # The widget may not exist yet since set value is called by the super class
            pass

        gm_meta_plugin.DynamicJobPresetAttribute.set_python_defined_value(self, value)

    """
    def get_value(self):
        inference_source = self.get_inference_source()

        if inference_source == gm_meta_plugin.InferenceSource.get_constant_inference_source():
            value = self.get_job_attribute().get_attribute().get_default_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_application_inference_source():
            value = self.get_app_inferred_widget().get_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_user_defined_inference_source():
            value = self.get_user_input_widget().get_value()

        elif inference_source == gm_meta_plugin.InferenceSource.get_project_inference_source():
            raise NotImplementedError
        else:
            raise ValueError("Unsupported inference source type: " + inference_source.get_display_name())

        return value

    def set_value(self, value):
        self.get_user_input_widget().set_value(value)
    """

    def get_inference_source(self):
        """
        index = self.get_inference_source_widget().currentIndex()
        inference_source = self.get_job_attribute().get_inference_sources()[index]
        return inference_source
        """
        inference_source_widget = self.get_inference_source_widget()
        inferences_sources = self.get_job_attribute().get_inference_sources()

        if gm_meta_plugin.InferenceSource.get_user_defined_inference_source() in inferences_sources and \
                (gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources or
                gm_meta_plugin.InferenceSource.get_python_inference_source() in inferences_sources):

            if inference_source_widget.isChecked():
                return gm_meta_plugin.InferenceSource.get_user_defined_inference_source()
            else:
                is_application_inferred = gm_meta_plugin.InferenceSource.get_application_inference_source() in inferences_sources
                is_python_inferred = gm_meta_plugin.InferenceSource.get_python_inference_source() in inferences_sources

                if is_application_inferred:
                    return gm_meta_plugin.InferenceSource.get_application_inference_source()
                elif is_python_inferred:
                    return gm_meta_plugin.InferenceSource.get_python_inference_source()
                else:
                    assert False

        return inferences_sources[0]

    def should_skip(self) -> bool:

        if self.get_job_attribute().should_skip():
            return True

        if self.get_job_attribute().is_optional():
            return not self._is_enabled_widget.get_value()

        return False
