__all__ = 'UploadPackedProjectWidget'

import pathlib

from PySide2 import QtCore
from PySide2 import QtWidgets


from gm_meta_plugin import PackedProject, RemoteProject, utils
from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.errors import *

from gm_qt_plugin.qt_plugin_accessor import QtPluginAccessor
from gm_qt_plugin.widgets.qt_project_attribute_instance import QtProjectAttributeInstance
from gm_qt_plugin.widgets import input_widgets
from gm_qt_plugin import qt_utils


class UploadPackedProjectWidget(QtWidgets.QWidget, QtPluginAccessor):

    FIRST_DYNAMIC_FORM_LAYOUT_WIDGET_INDEX = 4

    def __init__(self, plugin, parent=None):
        QtWidgets.QWidget.__init__(self, parent=parent)
        QtPluginAccessor.__init__(self, plugin)

        """ Create Widgets """
        self.group_box = QtWidgets.QGroupBox()
        self.group_box.setTitle("Upload Packed Project")

        self.description_label = QtWidgets.QLabel("Upload a project which has already had all it's external assets packed into one folder.")
        self.description_label.setEnabled(False)

        self.root_directory_label = QtWidgets.QLabel("Root Directory")
        self.root_directory_input = input_widgets.PathInput(plugin, file_mode=QtWidgets.QFileDialog.Directory)

        self.upload_everything_label = QtWidgets.QLabel("Upload Everything")
        self.upload_everything_input = QtWidgets.QCheckBox()

        cached_api_schema = plugin.get_api_client().get_cached_api_schema()
        self.project_attribute_widgets = cached_api_schema.create_project_attribute_widgets()

        for key, widget in self.project_attribute_widgets.iteritems():
            if hasattr(widget, 'input_value_changed'):
                widget.input_value_changed.connect(self.redraw)

        self.upload_btn = QtWidgets.QPushButton("Upload Packed Project")
        self.upload_btn.clicked.connect(self.execute)

        """ Create Layouts """
        self.form_layout = QtWidgets.QFormLayout()
        self.form_layout.addRow(self.root_directory_label, self.root_directory_input)
        self.form_layout.addRow(self.upload_everything_label, self.upload_everything_input)

        self.container_layout = QtWidgets.QVBoxLayout()
        self.container_layout.addWidget(self.description_label)
        self.container_layout.addLayout(self.form_layout)
        self.container_layout.addWidget(self.upload_btn)

        self.group_box.setLayout(self.container_layout)

        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(0,0,0,0)
        self.layout.setAlignment(QtCore.Qt.AlignTop)

        self.layout.addWidget(self.group_box)

        self.redraw()

    def get_root_dir(self):
        return pathlib.Path(self.root_directory_input.get_value())

    def is_upload_everything_checked(self):
        return self.upload_everything_input.isChecked()

    def redraw(self):

        # hide all widgets in form layout after the fixed widgets that should never change
        while True:
            row = self.form_layout.takeAt(self.FIRST_DYNAMIC_FORM_LAYOUT_WIDGET_INDEX)

            if row is None:
                break

            widget = row.widget()
            widget.hide()

        root = self.get_plugin().get_api_client().get_cached_api_schema().get_project_definition_schema().get_root_project_attribute()
        enabled = not QtProjectAttributeInstance.draw_project_attributes(self.form_layout,
                                                                 root,
                                                                 self.project_attribute_widgets,
                                                                 remote_source=False)

        self.upload_btn.setEnabled(enabled)

    def execute(self):
        plugin = self.get_plugin()
        api_client = plugin.get_api_client()
        user_interface = plugin.get_user_interface()
        logging_coordinator = plugin.get_logging_coordinator()
        logger = logging_coordinator.get_logger(__name__)

        # validate that the project attributes make up a valid project
        try:
            # do not force transition as all inputs should be valid
            attributes = qt_utils.get_project_attributes(plugin, self.project_attribute_widgets)
        except RejectedTransitionInputError as e:
            # if a transition failed then one of the attributes is invalid
            logger.error(e.user_message)
            return

        # validate that the selected root directory exists
        root_dir = self.get_root_dir()
        if not root_dir.exists():
            logger.error(u"Root Directory does not exist.")
            return

        # collect the files included in the packed project
        upload_everything = self.is_upload_everything_checked()
        project_files = set(qt_utils.get_project_files(plugin, self.project_attribute_widgets))
        if upload_everything:
            project_files |= utils.get_files_in_directory(root_dir)

        # check that all the files are under the root directory
        for project_file in project_files:

            # make the file path absolute / resolve symlinks
            project_file.resolve()

            if root_dir not in project_file.parents:
                logger.error(u"Project file \"" + str(project_file) + u"\" is outside of the root directory.")
                return

        project_name = attributes[api_constants.API_KEYS.PROJECT_NAME]

        # if project already exists warn the user
        root_directories = api_client.get_root_directories(ignore_cache=True)
        if project_name in root_directories:
            msb_box = QtWidgets.QMessageBox.warning(self,
                                                    u"Project Already Exists",
                                                    u"A project already exists with that name.\nWould you like to "
                                                    "submit anyway? Any project files with the same name will be "
                                                    "overwritten.",
                                                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                    defaultButton=QtWidgets.QMessageBox.Yes)

            if msb_box == QtWidgets.QMessageBox.Yes:
                pass
            else:
                return

        packed_project = PackedProject(project_name,
                                       root_dir,
                                       project_files,
                                       attributes)

        logger.info(u"Uploading packed project...")

        worker_thread = user_interface.get_worker_thread(u"Uploading packed project...")

        ################################################################################################################

        target = plugin.get_api_client().upload_project

        args = (
            packed_project,
            upload_everything
        )

        kwargs = {
            "delete_local_files_after_upload": False,
        }

        worker_thread.process(target, args, kwargs, self._handle_result_slot)

    def _handle_result_slot(self, result):

        if isinstance(result, RemoteProject):
            # add remote project to container
            plugin = self.get_plugin()
            plugin.get_remote_project_container().append(result)

            # change view to remote projects view
            main_window = self.get_plugin().get_user_interface().get_main_window()
            main_window.get_main_ui_widget().get_main_ui_switcher().display_remote_project_container()
            return True

        elif type(result) == NotSignedInError:
            msg = u"Could not upload your current scene. You are not signed in."
            self.error(msg)
            return True

        elif type(result) == APIClientError:
            msg = u"Could not upload your current scene. " + result.user_message
            self.error(msg)
            return True

        return False
