import datetime
import errno
import json
import logging
import os
import socket
import sys
import typing

if typing.TYPE_CHECKING:
    from submit2gm.worker_job import WorkerJob


REQUEST_TIMEOUT = 1

# Return codes
RETURN_CODE_SUCCESS = 0
RETURN_CODE_FAILED = 1

SUBMIT2GM_ALREADY_RUNNING_MSG = "{0} is already running, redirecting request to existing server..."
SUBMIT2GM_NOT_ALREADY_RUNNING_MSG = "{0} is not already running, creating a new {0} server..."
SUBMIT2GM_NOT_ALREADY_RUNNING_MSG_2 = "{0} is not already running, processing command..."

# Create logging folder if id does not already exist
LOGGING_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logs')
LOGGING_FILE = os.path.join(LOGGING_DIR, datetime.datetime.now().strftime("%Y_%m_%d-%I_%M_%S_%p.log"))

if not os.path.exists(LOGGING_DIR):
    os.makedirs(LOGGING_DIR, exist_ok=True)

logging.basicConfig(filename=LOGGING_FILE,
                    level=logging.DEBUG,
                    format='%(asctime)s - p%(process)s - %(threadName)s - {%(pathname)s:%(lineno)d} - %(levelname)s - %(message)s',
                    datefmt='%m/%d/%Y %I:%M:%S %p')

SUBMIT2GM_PWY_LOGGER = logging.getLogger('submit2gm-pyw')
SUBMIT2GM_PWY_LOGGER.setLevel(logging.INFO)

# set up logging to stdout
__stream_handler = logging.StreamHandler(sys.stdout)
__stream_handler.setLevel(logging.INFO)
__stream_handler.setFormatter(logging.Formatter('%(message)s'))
SUBMIT2GM_PWY_LOGGER.addHandler(__stream_handler)


def _setup_libs():
    """ Make sure the submit2gm module and all lib dependencies are available for import and correctly added to sys.path """

    if getattr(sys, 'frozen', False):
        # we are running in a bundle
        bundle_dir = sys._MEIPASS
    else:
        # we are running in a normal Python environment
        bundle_dir = os.path.dirname(os.path.abspath(__file__))

    # Set sys.path so modules can be imported
    parent_folder = bundle_dir
    libs_folder = os.path.join(parent_folder, 'libs')

    ### - Submit2gm multiple versions edit
    latest_libs = os.path.join(os.path.dirname(parent_folder), 'libs')
    if latest_libs not in sys.path:
        sys.path.insert(0, latest_libs)

    # Add the parent folder so that the submit2gm module can be found
    if parent_folder not in sys.path:
        sys.path.insert(0, parent_folder)

    # Add the libs folder so that any modules in libs can be found (PySide2, shiboken2, etc...)
    if libs_folder not in sys.path:
        sys.path.insert(0, libs_folder)


def _is_submit2gm_server_running() -> bool:
    from submit2gm import constants

    if os.name == 'nt':
        # The server uses named pipes, check if the named pipe is open
        import ctypes
        pipe = ctypes.windll.kernel32.CreateFileA(constants.WIN_PIPE_NAME,
                                                  constants.WIN_PIPE_GENERIC_READ | constants.WIN_PIPE_GENERIC_WRITE,
                                                  0,
                                                  None,
                                                  constants.WIN_PIPE_OPEN_EXISTING,
                                                  0,
                                                  None)

        if pipe == constants.WIN_PIPE_INVALID_HANDLE_VALUE:
            return False

        ctypes.windll.kernel32.CloseHandle(pipe)
        return True

    else:
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        try:
            # Attempt to connect to the socket file as a client
            s.connect(constants.UNIX_FILE_SOCKET_ADDRESS)
        except socket.error as e:

            if e.errno == errno.EADDRINUSE:
                # If the socket file already exists
                return True
            elif e.errno == errno.ENOENT:
                # The socket file does not exist
                return False
            elif e.errno == errno.ECONNREFUSED:
                # If the socket file exists but is not open to connections
                return False
            else:
                print("Unhandled socket error:", str(e))
                return True
        finally:
            s.close()

    return True


def _get_windows_pipe_handle(retries=5):
    # The server uses named pipes, check if the named pipe is open
    import ctypes
    import time

    retry = 0
    while 1:
        logging.debug("Attempting to create Windows file for use as pipe (retry: {})".format(retry))
        pipe_handle = ctypes.windll.kernel32.CreateFileA(constants.WIN_PIPE_NAME,
                                                  constants.WIN_PIPE_GENERIC_READ | constants.WIN_PIPE_GENERIC_WRITE,
                                                  0,
                                                  None,
                                                  constants.WIN_PIPE_OPEN_EXISTING,
                                                  0,
                                                  None)

        # If the pipe is not invalid then exit the loop
        if pipe_handle != constants.WIN_PIPE_INVALID_HANDLE_VALUE:
            logging.debug("Windows file pipe created successfully")
            break

        # Otherwise handle the error
        last_error_msg = ctypes.FormatError(ctypes.windll.kernel32.GetLastError())
        logging.debug("Error message: {}".format(last_error_msg))

        if ctypes.windll.kernel32.GetLastError() != constants.WIN_PIPE_ERROR_PIPE_BUSY:
            logging.warning("Could not create windows file (Pipe busy)")
            if retry >= retries:
                return None

        if ctypes.windll.kernel32.WaitNamedPipeA(constants.WIN_PIPE_NAME, 20000) == 0:
            logging.warning("Could not create windows file after waiting.")
            if retry >= retries:
                return None

        # Wait longer after each retry to give some extra cool off time
        retry_time = 0.1 + (1 * retry)

        logging.debug("Retrying in {} seconds.".format(str(retry_time)))
        time.sleep(retry_time)
        retry += 1

    return pipe_handle


def _get_writable_window_file_handle(retries=5):
    import ctypes
    import time

    dwMode = ctypes.c_ulong(constants.WIN_PIPE_PIPE_READMODE_MESSAGE)

    retry = 0
    while retry < retries:
        pipe = _get_windows_pipe_handle()

        if pipe is None:
            break

        ok = ctypes.windll.kernel32.SetNamedPipeHandleState(pipe, ctypes.byref(dwMode), None, None)

        if ok != 0:
            return pipe

        last_error = ctypes.windll.kernel32.GetLastError()
        logging.warning("Unable to SetNamedPipeHandleState: ({}) {}".format(last_error,
                                                                            ctypes.FormatError(last_error)))

        # Cleanup the pipe
        if not ctypes.windll.kernel32.FlushFileBuffers(pipe):
            logging.debug("Unable to flush file buffers: {}".format(
                ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

        if not ctypes.windll.kernel32.DisconnectNamedPipe(pipe):
            logging.debug("Unable to disconnect named pipe: {}".format(
                ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

        if not ctypes.windll.kernel32.CloseHandle(pipe):
            logging.debug("Unable to close handle: {}".format(ctypes.FormatError(
                ctypes.windll.kernel32.GetLastError())))

        # Wait longer after each retry to give some extra cool off time
        retry_time = 0.1 + (1 * retry)

        logging.debug("Retrying in {} seconds.".format(str(retry_time)))
        time.sleep(retry_time)
        retry += 1

    return None


def _use_existing_submit2gm_server() -> int:
    """
    Sends the current submit2gm arguments to the existing submit2gm server to be handled. Uses Python's sockets module to
    speed up the post.
    """
    logging.debug("Attempting to use existing Submit2GM server to process arguments:")
    logging.debug(json.dumps(sys.argv, indent=3))

    try:
        message = json.dumps(sys.argv).encode('utf8') + constants.WIN_PIPE_MESSAGE_SEPARATOR
    except TypeError as e:
        logging.error(str(e))
        return RETURN_CODE_FAILED

    if os.name == 'nt':
        # The server uses named pipes, check if the named pipe is open
        import ctypes
        import time

        retry = 0
        should_retry = False

        while True:
            pipe = _get_writable_window_file_handle()

            if pipe is None:
                return RETURN_CODE_FAILED

            cbWritten = ctypes.c_ulong(0)

            # todo is this really necassary?
            #time.sleep(0.5)

            message_char_pointer = ctypes.c_char_p(message)
            message_length = len(message)
            cb_written_byte_ref = ctypes.byref(cbWritten)

            logging.debug("Writing message to Windows file pipe...")
            ok = ctypes.windll.kernel32.WriteFile(pipe, message_char_pointer, message_length, cb_written_byte_ref, None)
            logging.debug("Message status: {}".format(ok))
            logging.debug("{} bytes out of {} written".format(cbWritten.value, len(message)))

            if (not ok) or (message_length != cbWritten.value):

                if not ok:
                    logging.warning("Unable to write to windows pipe (Not ok)")
                else:
                    logging.warning("Unable to write to windows pipe (only {} bytes out of {} written)".format(cbWritten.value, len(message)))

                last_error = ctypes.windll.kernel32.GetLastError()
                logging.debug("Write File failed: ({}) {}".format(last_error, ctypes.FormatError(last_error)))
                should_retry = True

            # Cleanup
            if not ctypes.windll.kernel32.FlushFileBuffers(pipe):
                last_error = ctypes.windll.kernel32.GetLastError()

                if last_error == constants.WIN_PIPE_ERROR_PIPE_NOT_CONNECTED:
                    should_retry = True

                logging.debug("Unable to flush file buffers: ({}) {}".format(last_error, ctypes.FormatError(last_error)))

            if not ctypes.windll.kernel32.DisconnectNamedPipe(pipe):
                last_error = ctypes.windll.kernel32.GetLastError()
                logging.debug("Unable to disconnect named pipe: ({}) {}".format(last_error, ctypes.FormatError(last_error)))

            if not ctypes.windll.kernel32.CloseHandle(pipe):
                last_error = ctypes.windll.kernel32.GetLastError()
                logging.debug("Unable to close handle: ({}) {}".format(last_error, ctypes.FormatError(last_error)))

            if should_retry:
                should_retry = False

                if retry < 5:
                    # Wait longer after each retry to give some extra cool off time
                    retry_time = 0.1 + (1 * retry)

                    logging.debug("Retrying in {} seconds.".format(str(retry_time)))
                    time.sleep(retry_time)
                    retry += 1

                    continue
                else:
                    return RETURN_CODE_FAILED

            logging.debug("Existing Submit2gm instance messaged successfully (Win).")
            return RETURN_CODE_SUCCESS
    else:
        logging.debug("Creating unix file socket...")
        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        retries = 5
        while retries:

            # connect to the client
            logging.debug("Connecting unix file socket to '{}'".format(constants.UNIX_FILE_SOCKET_ADDRESS))
            error_code = client.connect_ex(constants.UNIX_FILE_SOCKET_ADDRESS)

            logging.debug("Connection error code: {}".format(error_code))

            if error_code == 0:

                try:
                    # Send the request
                    client.send(message)
                    client.close()
                except Exception as e:
                    logging.debug(str(e))

                logging.debug("Existing Submit2gm instance messaged successfully (other).")
                return RETURN_CODE_SUCCESS

            logging.debug("Retrying in 1 second.")
            import time
            time.sleep(1)
            retries = retries - 1

    logging.error("Unable to use existing Submit2GM server")
    return RETURN_CODE_FAILED


def _create_application_instance():
    from PySide2 import QtCore
    from PySide2 import QtWidgets

    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

    app = QtWidgets.QApplication([])
    app.setOrganizationName("GridMarkets")
    app.setOrganizationDomain("gridmarkets.com")
    app.setApplicationName("Submit2gm")
    return app


def _main_event_loop() -> int:
    import queue
    job_queue = queue.Queue()

    # Run the http server in another thread
    from submit2gm.submit2gm_server_thread import Submit2gmServerThread
    thread = Submit2gmServerThread(job_queue)
    thread.setDaemon(True)  # Set as a daemon so it will be killed once the main thread is dead.
    thread.start()

    from submit2gm import constants
    from submit2gm.job_processor import JobProcessor
    from PySide2 import QtWidgets
    import time

    # Create / Get QApplication
    if not QtWidgets.QApplication.instance():
        app = _create_application_instance()
    else:
        app = QtWidgets.QApplication.instance()

    app.setApplicationName(constants.PROGRAM_NAME)

    # Create the job processor
    job_processor = JobProcessor()

    # Create system tray icon widget
    from gm_qt_plugin.qt_constants import GM_ICON_BLACK_AND_WHITE

    should_daemon_stop = False

    def _quit_function():
        logging.debug("System try quit action triggered")
        nonlocal should_daemon_stop
        should_daemon_stop = True

    quit_action = QtWidgets.QAction(u"Stop {}".format(constants.PROGRAM_NAME))
    quit_action.triggered.connect(_quit_function)

    system_tray_menu = QtWidgets.QMenu()
    system_tray_menu.setToolTipsVisible(True)
    system_tray_menu.addAction(quit_action)

    def _on_system_tray_activated(reason):
        if reason == QtWidgets.QSystemTrayIcon.Trigger:
            job_processor.raise_window([], {})

    system_tray_icon = QtWidgets.QSystemTrayIcon(GM_ICON_BLACK_AND_WHITE, parent=app)
    system_tray_icon.setToolTip(constants.PROGRAM_NAME)
    system_tray_icon.setContextMenu(system_tray_menu)
    system_tray_icon.activated.connect(_on_system_tray_activated)
    system_tray_icon.show()

    try:
        while thread.is_alive() and not should_daemon_stop:

            try:
                job: 'WorkerJob' = job_queue.get(block=False, timeout=1)
                job_processor.process(job)
            except queue.Empty:
                pass

            app.processEvents()
            time.sleep(0.02)

    except KeyboardInterrupt:
        pass

    logging.debug("Stopped job processor")

    # Explicitly remove the system try icon on shutdown
    # Without this the system try icon doesn't disappear until you hover your mouse over it
    system_tray_icon.hide()
    system_tray_icon.deleteLater()

    job_processor.clean_up()

    if thread.is_alive():
        logging.debug("Terminating Submit2gm Server thread...")
        thread.terminate()

    # Tell the app to quit
    # Without this any unprocessed signals will not get handled which may cause unexpected behaviour
    logging.debug("Shutting down Qt QApplication")
    try:
        logging.debug("   Sending remaining posted events...")
        app.sendPostedEvents()

        logging.debug("   Processing remaining events...")
        app.processEvents()

        logging.debug("   Quitting QApplication...")
        app.quit()

        logging.debug("Checking thread err")
        if thread.err:
            SUBMIT2GM_PWY_LOGGER.info(str(thread.err))
            SUBMIT2GM_PWY_LOGGER.info("Server stopped.")
            return RETURN_CODE_FAILED

    except Exception as e:
        logging.debug(str(e))
        return RETURN_CODE_FAILED

    SUBMIT2GM_PWY_LOGGER.info("Server stopped gracefully.")
    return RETURN_CODE_SUCCESS


def no_daemon_main():
    """
    submit2gm.pyw list-job-definitions --no-daemon
    submit2gm.pyw submit-job --no-daemon
    """
    from submit2gm import constants
    from submit2gm.job_processor import JobProcessor
    from submit2gm.worker_job import WorkerJob

    # Add a stream handler when not using the daemon
    stream_logging_handler = logging.StreamHandler()
    stream_log_format = '%(message)s'
    stream_formatter = logging.Formatter(stream_log_format)
    stream_logging_handler.setFormatter(stream_formatter)
    stream_logging_handler.setLevel(logging.INFO)
    stream_logging_handler.addFilter(logging.Filter(constants.CLI_LOGGER_ID))
    logging.getLogger().addHandler(stream_logging_handler)

    from PySide2 import QtWidgets
    if not QtWidgets.QApplication.instance():
        _create_application_instance()

    job_processor = JobProcessor()
    job = WorkerJob(sys.argv)
    job_processor.process(job)

    return RETURN_CODE_SUCCESS


if __name__ == '__main__':
    _setup_libs()

    from submit2gm import constants

    if constants.FORCE_USE_EXISTING_DAEMON in sys.argv:
        sys.exit(_use_existing_submit2gm_server())

    if constants.USE_DAEMON_FLAG in sys.argv:
        SUBMIT2GM_PWY_LOGGER.info("Checking if {} is already running...".format(constants.PROGRAM_NAME))
        if _is_submit2gm_server_running():
            SUBMIT2GM_PWY_LOGGER.info(SUBMIT2GM_ALREADY_RUNNING_MSG.format(constants.PROGRAM_NAME))
            sys.exit(_use_existing_submit2gm_server())
        else:
            SUBMIT2GM_PWY_LOGGER.info(SUBMIT2GM_NOT_ALREADY_RUNNING_MSG.format(constants.PROGRAM_NAME))
            sys.exit(_main_event_loop())

    sys.exit(no_daemon_main())
