"""
command_port_server_utils.py

This is the entry point for most plug-ins when submitting through submit2gm. It is designed to be directly imported by
external applications and provides logic for running a command port server. Since this file is designed to be imported
by the external application it was written to be compatible with Python 2.7 and Python 3 (unlike most of submit2gm which
is pure Python 3).
"""

__all__ = 'start_command_port_server'

import datetime
import errno
import logging
import subprocess
import json
import os
import socket
import sys
import threading
import time

if sys.version_info[0] < 3:
    import Queue as queue
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
else:
    import queue
    from http.server import HTTPServer, BaseHTTPRequestHandler

# Globals
__global_command_port_server = None
__global_cps_thread = None
__global_submit2gm_thread = None

# Constants
VERSION_STR = "1.0.2"
EXISTING_SUBMIT2GM_SERVER_POLL_INTERVAL = 10  # seconds
NEW_SERVER_RETRY_COUNT = 3
SUBPROCESS_NAME = "Submit2GM"
SUBMIT2GM_PATH = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'submit2gm.pyw')

WIN_PIPE_NAME = b"\\\\.\\pipe\\submit2gm"
WIN_PIPE_GENERIC_READ = 0x80000000
WIN_PIPE_GENERIC_WRITE = 0x40000000
WIN_PIPE_OPEN_EXISTING = 0x3
WIN_PIPE_INVALID_HANDLE_VALUE = -1

UNIX_FILE_SOCKET_ADDRESS = '/tmp/submit2gm_unix_file_socket.sock'

COMMAND_PORT_SERVER_TIMEOUT_KEEP_ALIVE_TOKEN = b'_'

# Keep in sync with submit2gm
SUBMIT2GM_NAME = "Submit2GM"
SUBMIT2GM_ALREADY_RUNNING_MSG_PART = "{0} is already running, redirecting request to existing server...".format(SUBMIT2GM_NAME)
SUBMIT2GM_NOT_ALREADY_RUNNING_MSG = "{0} is not already running, creating a new {0} server...".format(SUBMIT2GM_NAME)

# Return codes
RETURN_CODE_SUCCESS = 0
RETURN_CODE_FAILED = 1

USE_DAEMON_FLAG = '--use-daemon'
FORCE_USE_EXISTING_DAEMON = '--force-use-existing-daemon'
ENVIRONMENT_VARIABLES = os.environ.copy()

# Python 2 does not allow you to pass unicode environment variables to subprocess.Popen so this encodes them
if sys.version_info[0] < 3:
    for env_key in ENVIRONMENT_VARIABLES:
        value = ENVIRONMENT_VARIABLES[env_key]
        if isinstance(env_key, unicode) or isinstance(value, unicode):
            new_key = env_key.encode('ascii','ignore')
            new_value = value.encode('ascii','ignore')
            del ENVIRONMENT_VARIABLES[env_key]
            ENVIRONMENT_VARIABLES[new_key] = new_value

# Remove LD_LIBRARY_PATH  from the environment variables. Some apps set this and pythons.subprocess's default behaviour
# is to inherit the parent processes environment variables. This wouldn't be an issue except PySide2 has a bug in Centos
# and Fedora where it gives precedence to the system environment variables.
ENVIRONMENT_VARIABLES.pop('LD_LIBRARY_PATH', None)


class Job():
    """
    Wrapper class that is used to pass data between threads via qt signals and slots
    """

    STATUS_QUEUED = "QUEUED"
    STATUS_COMPLETED = "COMPLETED"
    STATUS_FAILED = "FAILED"

    def __init__(self, code, return_variables):
        self._code = code
        self._return_variables = return_variables
        self._response = b""
        self._status = Job.STATUS_QUEUED

    def get_code(self):
        return self._code

    def get_return_variables(self):
        return self._return_variables

    def is_finished(self):
        return self.is_completed() or self.is_failed()

    def is_completed(self):
        return self._status == Job.STATUS_COMPLETED

    def is_failed(self):
        return self._status == Job.STATUS_FAILED

    def mark_as_complete(self):
        self._status = Job.STATUS_COMPLETED

    def mark_as_failed(self):
        self._status = Job.STATUS_FAILED

    def get_response(self):
        return self._response

    def set_response(self, byte_str):
        self._response = byte_str


class JobProcessor():

    def __init__(self, logger, evaluate_code_cb):
        self._logger = logger
        self._evaluate_code_cb = evaluate_code_cb
        self._job_queue = queue.Queue()

    def put_job(self, job):
        self._job_queue.put(job, block=False)

    def process_jobs(self):
        while True:
            try:
                job = self._job_queue.get(block=False)
            except queue.Empty:
                break

            self.process_job(job)

    def process_job(self, job):
        """
        Global slot for processing jobs. Guaranteed to emit the jobs complete signal.

        :param job: Job object
        :return: None
        """

        def _set_response(value):
            if sys.version_info[0] < 3:
                job.set_response(bytes(json.dumps(value)))
            else:
                job.set_response(bytes(json.dumps(value), 'utf8'))

        # Try to evaluate the code contained in the job
        try:
            results_dictionary = self._evaluate_code_cb(job.get_code(), job.get_return_variables())
            failed = False
        except Exception as e:
            # If an exception occurs while evaluating the code then return the exception
            try:
                exc_msg = str(e)
            except Exception as e:
                exc_msg = "Unknown evaluation exception"

            self._logger.error(exc_msg)
            results_dictionary = {'error': exc_msg}

            failed = True

        _set_response(results_dictionary)

        if failed:
            job.mark_as_failed()
        else:
            job.mark_as_complete()


class HTTPRequestHandler(BaseHTTPRequestHandler):

    protocol_version = 'HTTP/1.1'

    def __init__(self, request, client_address, server):
        BaseHTTPRequestHandler.__init__(self, request, client_address, server)

    def do_POST(self):
        logger = self.server.get_logger()
        logger.debug("Handling POST request")

        try:
            if self.path == "/eval":
                return_variables = json.loads(self.headers['Return-Variables'])
                content_length = int(self.headers['Content-Length'])
                code = self.rfile.read(content_length)

                job = Job(code, return_variables)

                def _send_chunk(data):
                    to_send = b'%X\r\n%s\r\n' % (len(data), data)
                    self.wfile.write(to_send)

                self.server.get_job_processor().put_job(job)

                self.send_response(200)
                self.send_header('Transfer-Encoding', 'chunked')
                self.send_header("Content-type", "text/plain")
                self.end_headers()

                last_keep_alive_chunk_msg = time.time()
                while not job.is_finished():
                    if time.time() - last_keep_alive_chunk_msg > 2:
                        _send_chunk(COMMAND_PORT_SERVER_TIMEOUT_KEEP_ALIVE_TOKEN)

                    # Yield the current thread
                    time.sleep(0.0001)

                response = job.get_response()
                _send_chunk(response)

                # Last chunk must be empty
                _send_chunk(b'')

            elif self.path == "/shutdown":
                self.send_response(200)
                self.send_header("Content-Length", 0)
                self.send_header("Keep-Alive", "timeout=0, max=0")
                self.end_headers()

                self.server.server_close()

        except Exception as e:
            logger.error(str(e))
            self.server.server_close()

    def log_message(self, format, *args):
        logger = self.server.get_logger()
        logger.debug("%s - - [%s] %s\n" % (self.address_string(),self.log_date_time_string(), format%args))


class CommandPortServer(HTTPServer):

    def __init__(self, server_address, logger, evaluate_code_cb):
        # We only want one instance of submit2gm running at a time so make sure the port is not already in use
        self.allow_reuse_address = False
        self._running = False
        self._should_stop = False
        self.timeout = 0.1

        self._job_processor = JobProcessor(logger, evaluate_code_cb)
        self._logger = logger

        HTTPServer.__init__(self, server_address, HTTPRequestHandler)

    def get_job_processor(self):
        return self._job_processor

    def get_logger(self):
        return self._logger

    def serve_until_stopped(self):
        self._should_stop = False
        self._running = True

        try:
            while not self._should_stop:
                self.handle_request()
        except Exception as e:
            raise e
        finally:
            self._running = False

    def isRunning(self):
        return self._running

    def server_close(self):
        self._logger.debug("Closing command port server.")

        self._should_stop = True

        # Closes the underlying tcp socket
        HTTPServer.server_close(self)


def _is_submit2gm_server_running(logger):
    if os.name == 'nt':
        # The server uses named pipes, check if the named pipe is open
        import ctypes
        pipe = ctypes.windll.kernel32.CreateFileA(WIN_PIPE_NAME,
                                                  WIN_PIPE_GENERIC_READ | WIN_PIPE_GENERIC_WRITE,
                                                  0,
                                                  None,
                                                  WIN_PIPE_OPEN_EXISTING,
                                                  0,
                                                  None)

        if pipe == WIN_PIPE_INVALID_HANDLE_VALUE:
            create_file_failed_error = ctypes.windll.kernel32.GetLastError()
            logging.debug("Unable to close handle: ({}) {}".format(create_file_failed_error,
                                                                   ctypes.FormatError(create_file_failed_error)))
            return False

        ctypes.windll.kernel32.CloseHandle(pipe)
        return True

    else:
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        try:
            s.bind(UNIX_FILE_SOCKET_ADDRESS)
        except socket.error as e:
            if e.errno == errno.EADDRINUSE:
                return True
            elif e.errno == errno.ENOENT:  # The file does not exist
                return False
            elif e.errno == errno.ECONNREFUSED:
                # If the socket file exists but is not open to connections
                return False
            else:
                print("Unhandled socket error:", str(e))
                return True
        finally:
            s.close()

    return True


class OpenSubmissionPopupThread(threading.Thread):

    def __init__(self, command_port_server, envoy_python_path, product, meta_info):
        self._command_port_server = command_port_server
        self._envoy_python_path = envoy_python_path
        self._product = product
        self._meta_info = meta_info
        self.err = None
        threading.Thread.__init__(self)

    def _open_submit2gm_submission_popup(self):
        logger = self._command_port_server.get_logger()

        # Wait for the cps to start running
        logger.info("Waiting for CPS to start...")
        while not self._command_port_server.isRunning():
            time.sleep(0.01)

        args = [self._envoy_python_path, SUBMIT2GM_PATH,
                'open-submission-popup',
                '--product', self._product,
                '--cps-address', 'http://127.0.0.1:{}'.format(str(self._command_port_server.server_port)),
                '--meta-info', self._meta_info,
                USE_DAEMON_FLAG]

        logger.info("Attempting to open submit2gm submission popup")
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stdin=subprocess.PIPE, env=ENVIRONMENT_VARIABLES)
        p.name = SUBPROCESS_NAME

        used_existing_server = None

        # Read stout / stderr until process exits
        retries = NEW_SERVER_RETRY_COUNT
        stdout_buffer = ""
        while 1:
            stdout, stderr = p.communicate()
            stdout_buffer = stdout_buffer + str(stdout)

            if SUBMIT2GM_ALREADY_RUNNING_MSG_PART in stdout_buffer:
                logger.info("Existing submit2gm server found")
                used_existing_server = True

            if SUBMIT2GM_NOT_ALREADY_RUNNING_MSG in stdout_buffer:
                logger.info("submit2gm not already running")
                used_existing_server = False

            if p.returncode is not None:
                logger.debug("subprocess return code: {}".format(str(p.returncode)))

                if used_existing_server is False:
                    if p.returncode == 0:
                        break
                    elif "Server stopped gracefully." in stdout_buffer:
                        # In theory simply relying on the return code being zero should be enough but it is possible for
                        # the return code to be other values (120, -1073741819) for unexplained reasons or for the
                        # server to crash unexpectedly due to a bug later on in it's lifecycle. In either case we do not
                        # want to try and reopen submit2gm automatically.
                        logger.warning("Submit2gm stopped gracefully with non-zero exit code '{}'".format(p.returncode))
                        break
                    else:
                        logger.debug(stdout_buffer)
                        if retries:
                            # If we tried to open a new instance of submit2gm but it failed then retry, if two instances
                            # of submit2gm open at the same time then one will fail
                            logger.info("New submit2gm server failed to start. Retrying (retries left: {})".format(retries))

                            # If the command port server stops running (either due to an error or Submit2gm stopping it)
                            # then there is no point continuing
                            if not self._command_port_server.isRunning():
                                err_msg = "Command port server has stopped, aborting open Submit2gm submission popup retry attempt"
                                logger.warning(err_msg)
                                raise RuntimeError(err_msg)

                            retries = retries - 1

                            used_existing_server = None
                            stdout_buffer = ""

                            p = subprocess.Popen(args, stdout=subprocess.PIPE, stdin=subprocess.PIPE, env=ENVIRONMENT_VARIABLES)
                            p.name = SUBPROCESS_NAME
                        else:
                            err_msg = "Failed to start new submit2gm server after {} retries".format(NEW_SERVER_RETRY_COUNT)
                            logger.warning(err_msg)
                            raise RuntimeError(err_msg)
                else:
                    break

            # If not using an existing process then that means the submit2gm server is running in
            # process p. Here we limit the polling of p to see if it is still running
            if used_existing_server is False:
                    logger.debug("sleeping")
                    time.sleep(2)

        logger.info("Open submission popup process has finished with exit code '{}'".format(str(p.returncode)))

        if used_existing_server is None:
            # in this case we do not know if we need to stop the cps or not. It is better to stop it and throw
            # and exception because otherwise 3ds could be left in a frozen state
            e_msg = "unable to determine if submit2gm was able to use an existing server or not."
            logger.warning(e_msg)
            logger.debug(stdout_buffer)
            self._command_port_server.server_close()
            raise Exception(e_msg)

        elif used_existing_server:
            logger.info("Submission popup is using an existing submit2gm server.")

            # If submit2gm tried to use an existing server and succeeded then poll to check the existing server is still
            # running
            if p.returncode == 0:

                # While the port is in use then assume submit2gm is still running
                polling_time_start = time.time()
                while 1:
                    current_time = time.time()

                    if current_time - polling_time_start > EXISTING_SUBMIT2GM_SERVER_POLL_INTERVAL:
                        polling_time_start = time.time()
                        if not _is_submit2gm_server_running(logger):
                            logger.debug("External submit2gm server appears to have stopped")
                            break

                    if not self._command_port_server.isRunning():
                        logger.debug("Command port server is no longer running.")
                        break

                    time.sleep(0.5)
            else:
                err_msg = "Unexpected return code for existing server: {}".format(p.returncode)
                logger.error(err_msg)
                raise ValueError(err_msg)

        else:
            # If a new server was started and the process has finished then we can stop the command port server
            # as well.
            logger.info("New submit2gm server stopped.")

        # Stop the command port server once the submit2gm is no longer running
        self._command_port_server.server_close()

    def run(self):
        logger = self._command_port_server.get_logger()

        try:
            self._open_submit2gm_submission_popup()
        except Exception as e:
            logger.error(str(e))
            self.err = e

        logger.debug("OpenSubmissionPopupThread finished")


def raise_window(envoy_python_path, logger):
    args = [envoy_python_path, SUBMIT2GM_PATH, 'raise-window', FORCE_USE_EXISTING_DAEMON]
    logger.info("Attempting to raise submit2gm window")
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stdin=subprocess.PIPE, env=ENVIRONMENT_VARIABLES)
    p.name = SUBPROCESS_NAME

    stdout_buffer = ""
    while 1:
        stdout, stderr = p.communicate()
        stdout_buffer = stdout_buffer + str(stdout)

        if p.returncode is not None:
            logger.info("raise submit2gm window returned: {}".format(str(p.returncode)))
            logger.debug(stdout_buffer)
            break


def disconnect_from_submit2gm(envoy_python_path, logger=None):
    """
    Called when the host app is shutting down
    """
    global __global_command_port_server

    if __global_command_port_server:
        cps_address = 'http://' + ':'.join(map(lambda x: str(x), __global_command_port_server.server_address))
        args = [envoy_python_path,
                SUBMIT2GM_PATH,
                'disconnect-cps',
                '--cps-address', cps_address,
                FORCE_USE_EXISTING_DAEMON]

        if logger:
            logger.info("Attempting to disconnect cps '{}' from submit2gm".format(cps_address))

        p = subprocess.Popen(args, stdout=subprocess.PIPE, stdin=subprocess.PIPE, env=ENVIRONMENT_VARIABLES)
        p.name = SUBPROCESS_NAME

        stdout_buffer = ""
        while 1:
            stdout, stderr = p.communicate()
            stdout_buffer = stdout_buffer + str(stdout)

            if p.returncode is not None:
                if logger:
                    logger.debug("Disconnect cps returned: {}".format(str(p.returncode)))
                    logger.debug(stdout_buffer)
                break


def get_logger(logger_name, debug_mode):
    logger = logging.getLogger(logger_name)
    logger.propagate = False

    # logging.getLogger is a singleton, it will return the same logger on multiple calls with the same name. Therefore
    # we must clean up the handlers added to the logger last time.
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    logs_dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'logs')
    if not os.path.exists(logs_dir_path):
        os.mkdir(logs_dir_path)

    log_file_dir_path = os.path.join(logs_dir_path, logger_name)
    if not os.path.exists(log_file_dir_path):
        os.mkdir(log_file_dir_path)

    log_file_path = os.path.join(log_file_dir_path, datetime.datetime.now().strftime("%Y_%m_%d-%I_%M_%S_%p.log"))

    # create formatters and add it to the handlers
    stream_logging_handler = logging.StreamHandler()
    stream_log_format = '%(asctime)s - p%(process)s - %(thread)d - {%(pathname)s:%(lineno)d} - %(levelname)s - %(message)s'
    stream_formatter = logging.Formatter(stream_log_format)
    stream_formatter.datefmt = '%m/%d/%Y %I:%M:%S %p'
    stream_logging_handler.setFormatter(stream_formatter)

    file_logging_handler = logging.FileHandler(log_file_path)
    file_logging_handler.setLevel(logging.DEBUG)  # File logging is alway in debug mode
    file_log_format = '%(asctime)s - p%(process)s - %(thread)d - {%(pathname)s:%(lineno)d} - %(levelname)s - %(message)s'
    file_formatter = logging.Formatter(file_log_format)
    file_formatter.datefmt = '%m/%d/%Y %I:%M:%S %p'
    file_logging_handler.setFormatter(file_formatter)

    # Only add the stream handler if in debug mode
    if debug_mode:
        stream_logging_handler.setLevel(logging.DEBUG)
        logger.addHandler(stream_logging_handler)
        logger.setLevel(logging.DEBUG)

    # File handler is always added for logging purposes
    logger.addHandler(file_logging_handler)

    return logger


def process_events():
    global __global_command_port_server
    if __global_command_port_server:
        __global_command_port_server.get_job_processor().process_jobs()


def _debug_status():
    global __global_command_port_server
    global __global_cps_thread
    global __global_submit2gm_thread

    print("is cps running: " + str(__global_command_port_server))
    if __global_command_port_server:
        print("is cps running: " + str(__global_command_port_server.isRunning()))

    print("cps thread: " + str(__global_cps_thread))
    if __global_cps_thread:
        print("is cps thread alive: " + str(__global_cps_thread.is_alive()))

    print("submit2gm thread: " + str(__global_submit2gm_thread))
    if __global_submit2gm_thread:
        print("is submit2gm thread alive: " + str(__global_submit2gm_thread.is_alive()))


def start_command_port_server(envoy_python_path, logger_name, product, meta_info, evaluate_code_cb, debug_mode):
    logger = get_logger(logger_name, debug_mode)
    logger.info("Creating command port server...")

    global __global_command_port_server
    global __global_cps_thread
    global __global_submit2gm_thread

    if not os.path.exists(envoy_python_path):
        logger.error("Envoy distributed Python not found ('{}')".format(envoy_python_path))
        return

    if not os.path.isfile(envoy_python_path):
        logger.error("Envoy distributed Python is not a file ('{}')".format(envoy_python_path))
        return

    if not os.access(envoy_python_path, os.X_OK):
        logger.error("Execute permissions missing on Envoy distributed Python ('{}')".format(envoy_python_path))
        return

    # The globals should either be all defined or all None. If this is not the case then a value error is raised
    if isinstance(__global_command_port_server, CommandPortServer) and \
            isinstance(__global_cps_thread, threading.Thread) and \
            isinstance(__global_submit2gm_thread, OpenSubmissionPopupThread):

        logger.debug("is cps running: " + str(__global_command_port_server.isRunning()))
        logger.debug("is cps thread alive: " + str(__global_cps_thread.is_alive()))
        logger.debug("is submit2gm thread alive: " + str(__global_submit2gm_thread.is_alive()))

        # If only the submit2gm thread is still running but the CPS has stopped then one of two things could be true
        # 1) The submit2gm thread is still running because it spwned to primary submit2gm process which has not finished
        #    yet. Possibly because another application has connected to it.
        # 2) The submit2gm thread is in the process of stopping but just hasn't quite yet.
        if not __global_command_port_server.isRunning() and \
                not __global_cps_thread.is_alive() and \
                __global_submit2gm_thread.is_alive():

            # Assume that we are in case 1). Attempt to open the submission popup like normal but without changing the
            # value of __global_submit2gm_thread

            # cps
            __global_command_port_server.server_close()
            __global_command_port_server = CommandPortServer(('127.0.0.1', 0), logger, evaluate_code_cb)

            # cps thread
            __global_cps_thread = threading.Thread(target=__global_command_port_server.serve_until_stopped, args=())
            __global_cps_thread.setDaemon(True)

            # submit2gm thread
            temp_submit2gm_thread = OpenSubmissionPopupThread(__global_command_port_server,
                                                                  envoy_python_path,
                                                                  product,
                                                                  meta_info)
            temp_submit2gm_thread.setDaemon(True)

            # Request submit2gm opens a submission popup in another thread so that we don't block the cps starting
            temp_submit2gm_thread.start()
            logger.info("Handling command requests at {}".format(__global_command_port_server.server_address))
            __global_cps_thread.start()
            return

        # if everything is still running then we just need to raise the existing window
        if __global_command_port_server.isRunning() and \
                __global_cps_thread.is_alive() and \
                __global_submit2gm_thread.is_alive():
            raise_window(envoy_python_path, logger)
            return

        # if nothing is still running then recreate the threads
        if not __global_command_port_server.isRunning() and \
                not __global_cps_thread.is_alive() and \
                not __global_submit2gm_thread.is_alive():
            __global_command_port_server.server_close()
            __global_command_port_server = None
            __global_submit2gm_thread = None
            __global_cps_thread = None

        # Else if the cps and thread are still running but not submit2gm then stop the existing cps / thread and
        # recreate
        elif __global_command_port_server.isRunning() and \
                __global_cps_thread.is_alive() and \
                not __global_submit2gm_thread.is_alive():

            __global_command_port_server.server_close()

            __global_command_port_server = None
            __global_submit2gm_thread = None
            __global_cps_thread = None

    elif __global_command_port_server is None and __global_cps_thread is None and __global_submit2gm_thread is None:
        pass
    else:
        raise ValueError("Unhandled Exception")

    if __global_command_port_server is None:
        # cps
        __global_command_port_server = CommandPortServer(('127.0.0.1', 0), logger, evaluate_code_cb)

        # submit2gm thread
        __global_submit2gm_thread = OpenSubmissionPopupThread(__global_command_port_server,
                                                              envoy_python_path,
                                                              product,
                                                              meta_info)
        __global_submit2gm_thread.setDaemon(True)

        # cps thread
        __global_cps_thread = threading.Thread(target=__global_command_port_server.serve_until_stopped, args=())
        __global_cps_thread.setDaemon(True)

    # Request submit2gm opens a submission popup in another thread so that we don't block the cps starting
    __global_submit2gm_thread.start()
    logger.info("Handling command requests at {}".format(__global_command_port_server.server_address))

    __global_cps_thread.start()
