__all__ = 'Job Processor'

import argparse
import json
import logging
import pathlib
import sys
import typing

from . import constants

if typing.TYPE_CHECKING:
    from .submit2gm_plugin import Submit2gmPlugin
    from .worker_job import WorkerJob


main_window = None
PLUGIN: typing.Optional['Submit2gmPlugin'] = None


class JobProcessor:

    SUBMISSION_POPUP_ALREADY_OPEN = 1

    def __init__(self):
        self.active_job : typing.Optional['WorkerJob'] = None

        self.parser = argparse.ArgumentParser(prog=constants.PROGRAM_NAME,
                                              description=constants.APP_DESCRIPTION)
        sub_parsers = self.parser.add_subparsers()

        # open submission popup
        parser_submission_popup = sub_parsers.add_parser('open-submission-popup', help='Submit a project')
        parser_submission_popup.add_argument('--product',
                                             type=str,
                                             dest='product',
                                             required=True,
                                             help="A GridMarkets product code.")

        parser_submission_popup.add_argument('--cps-address',
                                             type=str,
                                             dest='cps_address',
                                             required=False,
                                             help="The address of the command port server.")

        parser_submission_popup.add_argument('--auto-sign-in',
                                             type=bool,
                                             dest='auto_sign_in',
                                             default=True,
                                             required=False,
                                             help="Should {} automatically attempt to sign in via Envoy on startup.".format(constants.PROGRAM_NAME))

        parser_submission_popup.add_argument('--meta-info',
                                             type=str,
                                             dest='meta_info',
                                             default="",
                                             required=False,
                                             help="Meta data to regarding the CPS plug-in.")

        parser_submission_popup.set_defaults(func=self.open_submission_popup)

        # raise window
        parser_raise_window = sub_parsers.add_parser('raise-window', help='Open the UI if it has already been created but has been hidden.')
        parser_raise_window.set_defaults(func=self.raise_window)

        # close window
        parser_close_window = sub_parsers.add_parser('close-window', help='Close the UI if it has already been created and is open.')
        parser_close_window.set_defaults(func=self.close_window)

        # reset
        parser_reset = sub_parsers.add_parser('reset', help='Return Submit2gm to it\'s inital state. I.e sign out the user, close the UI and shutdown any command port servers.')
        parser_reset.set_defaults(func=self.reset)

        # reset
        parser_close_cps = sub_parsers.add_parser('disconnect-cps', help='Shutsdown and removes the command port server with the given address. If the CPS is active then reset Submit2gm.')
        parser_close_cps.add_argument('--cps-address',
                                      type=str,
                                      dest='cps_address',
                                      required=True,
                                      help="The address of the command port server to disconnect.")
        parser_close_cps.set_defaults(func=self.disconnect_cps)

        # reset
        parser_version = sub_parsers.add_parser('version', help='Print the submit2gm version number.')
        parser_version.set_defaults(func=self.version)

        # list job definitions
        parser_list_job_definitions = sub_parsers.add_parser('list-job-definitions', help='List all the job definition types that can be used for job submissions.')
        parser_list_job_definitions.set_defaults(func=self.list_job_definitions)

        # show job definition
        parser_show_job_definition = sub_parsers.add_parser('show-job-definition', help='Displays the list of attributes required for a job definition.')
        parser_show_job_definition.add_argument('--job-definition-id',
                                             type=str,
                                             dest='job_definition_id',
                                             default="",
                                             required=True,
                                             help="The ID of the job definition to show.")
        parser_show_job_definition.set_defaults(func=self.show_job_definition)

        # get product versions
        parser_get_product_versions = sub_parsers.add_parser('get-product-versions',
                                                             help='Displays the list of available app versions for a specfied app.')

        parser_get_product_versions.add_argument('--app',
                                                 type=str,
                                                 dest='app',
                                                 required=True,
                                                 help="The app id.")

        parser_get_product_versions.set_defaults(func=self.get_product_versions)

        # get machines
        parser_get_machines = sub_parsers.add_parser('get-machines',
                                                     help='Displays the list of available machines for an app based on the specified operation, machine category and plug-ins.')

        parser_get_machines.add_argument('--app',
                                         type=str,
                                         dest='app',
                                         required=True,
                                         help="The app id.")

        parser_get_machines.add_argument('--operation',
                                         type=str,
                                         default='',
                                         dest='operation',
                                         required=True,
                                         help="The operation type.")

        parser_get_machines.add_argument('--plugins',
                                         type=str,
                                         dest='plugins',
                                         default='',
                                         help="Comma separated list of plug-ins.")

        parser_get_machines.set_defaults(func=self.get_machines)

        # get user info
        parser_get_user_info = sub_parsers.add_parser('get-user-info', help='Displays all user level info.')
        parser_get_user_info.set_defaults(func=self.get_user_info)

        # submit job
        parser_submit_job = sub_parsers.add_parser('submit-job', help='Submit a job based on a job definition')

        parser_submit_job.add_argument('--project-dir',
                                         type=str,
                                         dest='project_dir',
                                         required=True,
                                         help="If --skip-upload is not specified then this is the root folder of your project on your local file system. If --skip-upload is specified then this is the name of the existing remote project as seen within Envoy (e.g. \"example_project\").")

        parser_submit_job.add_argument('--job-definition-id',
                                                type=str,
                                                dest='job_definition_id',
                                                default="",
                                                required=True,
                                                help="The ID of the job definition to show.")

        parser_submit_job.add_argument('--skip-upload',
                                       dest='skip_upload',
                                       action='store_true',
                                       help="Determines if file upload will be skipped for the submission.")

        parser_submit_job.add_argument('--show-ui',
                                       dest='show_ui',
                                       action='store_true',
                                       help="Determines if Submit2GM will display a user interface for the user to confirm their job and project settings before submission.")

        parser_submit_job.add_argument('--enable-cli-prompt',
                                       dest='enable_cli_prompt',
                                       action='store_true',
                                       help="Determines if Submit2GM will prompt the user on the cli for missing attributes. Has no effect when --show-ui is enabled.")

        parser_submit_job.set_defaults(func=self.submit_job)

    def _sign_out_callback(self, args, unknown_args) -> None:
        from .plugin_fetcher import PluginFetcher
        global main_window
        global PLUGIN

        args.auto_sign_in = False

        # Close the active cps session (if there is one) otherwise the cps will never accept another connection
        # Only close the session and don't stop beacuse the same cps may still be needed if the user signs back in
        # using the same app session.
        try:
            PLUGIN.get_active_command_port_server().close_session()
        except ValueError as e:
            pass

        # main_window.destroyed.connect(reopen_submision_popup)
        main_window.close()
        main_window.deleteLater()
        main_window = None

        PluginFetcher.delete_cached_plugin()
        PLUGIN = None

        self.open_submission_popup(args, unknown_args)

    def _clean_up_main_window(self):
        from .plugin_fetcher import PluginFetcher
        global main_window
        global PLUGIN

        if main_window:
            main_window.close()
            main_window.deleteLater()
            main_window = None

        if PLUGIN:
            PLUGIN.shutdown_command_port_servers()
            PluginFetcher.delete_cached_plugin()
            PLUGIN = None

    def re_open_submission_popup(self, args, unknown_args) -> None:
        """
        Called if the main window and plugin already exist
        """
        global main_window
        global PLUGIN
        logger = PLUGIN.get_logging_coordinator().get_logger(__file__)
        logger.debug("re_open_submission_popup")

        # Reset the windows reset callback since the arguments might have changed
        main_window.set_reset_callback(lambda: self._sign_out_callback(args, unknown_args))

        # Get the active cps address (if one exists)
        active_cps_address = None
        try:
            active_cps_address = PLUGIN.get_active_command_port_server().get_address()
        except ValueError:
            pass

        # If the main window already exists and the user is signed in then raise the current window (even if the
        # active cps is no longer running)
        if PLUGIN.get_api_client().is_user_signed_in():
            logger.debug("user is signed in")

            # Remove all old job presets
            job_preset_container = PLUGIN.get_preferences_container().get_job_preset_container()
            job_preset_container.reset()

            # Remove the old job preset from the ui
            PLUGIN.get_user_interface().get_main_switcher_ui().get_submission_view_widget().redraw_job_preset()

            # If we're reopening with a new CPS then make sure to clean up and close the old one
            if active_cps_address != args.cps_address:
                PLUGIN.shutdown_command_port_servers()
                PLUGIN.add_command_port_server(args.product, args.cps_address, meta_info=args.meta_info)

            # Create a new job preset to use / display
            PLUGIN.create_default_job_presets()
        else:
            logger.debug("user is signed out")

            # If we're reopening with a new CPS then make sure to clean up and close the old one
            if active_cps_address != args.cps_address:
                PLUGIN.shutdown_command_port_servers()
                PLUGIN.add_command_port_server(args.product, args.cps_address, meta_info=args.meta_info)

            # If the main window already exists but the user is not signed in then call the sign out call back to
            # just to reset the state of submit2gm
            self._sign_out_callback(args, unknown_args)

        # Show the window
        main_window.show()
        main_window.raise_()
        main_window.activateWindow()

    def open_submission_popup(self, args, unknown_args) -> None:
        self.active_job.mark_as_complete()

        from .plugin_fetcher import PluginFetcher
        global main_window
        global PLUGIN

        # create the plugin
        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(__file__)

        from PySide2 import QtWidgets, QtCore
        if not QtWidgets.QApplication.instance():
            app = QtWidgets.QApplication([])
        else:
            app = QtWidgets.QApplication.instance()
        app.setWindowIcon(PLUGIN.get_user_interface().get_icon())

        from gm_qt_plugin.qt_dialog_main_window import MainWindow

        if main_window:
            return self.re_open_submission_popup(args, unknown_args)
        else:
            # create the main plugin window
            def sign_out_callback():
                self._sign_out_callback(args, unknown_args)

            main_window = MainWindow(PLUGIN, sign_out_callback, parent=None)
            main_window.setWindowModality(QtCore.Qt.ApplicationModal)

        if args.cps_address:
            logger.debug("adding command port server")
            PLUGIN.add_command_port_server(args.product, args.cps_address, meta_info=args.meta_info)

        if args.auto_sign_in and not PLUGIN.get_api_client().is_user_signed_in():
            main_window.get_sign_in_widget().sign_in_envoy_slot()

        #main_window.finished.connect(lambda res: PLUGIN.shutdown_command_port_servers())
        main_window.show()
        main_window.raise_()
        main_window.activateWindow()

    def reload_scene_settings(self, args, unknown_args) -> None:
        global main_window
        global PLUGIN
        if main_window and PLUGIN:
            if PLUGIN.get_api_client().is_user_signed_in():
                main_window.get_main_ui_widget().get_main_ui_switcher().get_submission_view_widget().reload_scene_settings()

    def raise_window(self, args, unknown_args) -> None:
        global main_window
        global PLUGIN
        if main_window:

            # Do nothing if the user is not signed in and the active cps is not alive
            if not PLUGIN.get_api_client().is_user_signed_in():
                try:
                    cps = PLUGIN.get_active_command_port_server()
                    if not cps.is_still_running():
                        return
                except ValueError:
                    return

            main_window.show()
            main_window.raise_()
            main_window.activateWindow()
            self.reload_scene_settings(args, unknown_args)

    def close_window(self, args, unknown_args) -> None:
        global main_window
        global PLUGIN
        if main_window and main_window.isVisible():

            try:
                PLUGIN.get_active_command_port_server().close_session()
            except ValueError as e:
                pass

            main_window.hide()

    def reset(self, args, unknown_args) -> None:
        self._clean_up_main_window()

    def disconnect_cps(self, args, unknown_args) -> None:

        cps_address = args.cps_address

        global PLUGIN
        if PLUGIN:
            logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

            # If the cps is the active cps then we need to reset the UI so that no connection issues arise.
            try:
                active_cps = PLUGIN.get_active_command_port_server()

                if active_cps.get_address() == cps_address:
                    logger.debug("Disconnecting active cps")
                    return self.reset(args, unknown_args)
            except ValueError:
                pass

            logger.debug("Disconnecting cps")
            try:
                PLUGIN.shutdown_command_port_server(cps_address)
            except ValueError as e:
                logger.debug(str(e))
                return

    def version(self, args, unknown_args) -> None:
        from .version import VERSION
        version_str = '{} Version: '.format(constants.PROGRAM_NAME) + '.'.join(map(lambda x: str(x), VERSION))

        global PLUGIN
        if PLUGIN:
            logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)
            logger.info(version_str)
        else:
            print(version_str)

    def list_job_definitions(self, args, unknown_args):
        import gm_meta_plugin
        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except gm_meta_plugin.errors.APIClientError as e:
                logger.info(str(e))
                return

        logger.info('\nListing job definitions:')
        api_schema = PLUGIN.get_api_client().get_api_schema()
        job_definitions = api_schema.get_job_definitions()

        table = [['Name'],['ID'],['Supports CMD submission']]
        for job_definition in job_definitions:
            table[0].append(job_definition.get_display_name())
            table[1].append(job_definition.get_definition_id())

            supports_cmd_submission = True
            argv_inference_source = gm_meta_plugin.InferenceSource.get_argv_inference_source()
            job_attributes = job_definition.get_attributes()

            for job_attribute in job_attributes:
                if argv_inference_source not in job_attribute.get_inference_sources():
                    supports_cmd_submission = False
                    break

            table[2].append(str(supports_cmd_submission))

        table_str = gm_meta_plugin.utils.format_table_str(table)
        logger.info(table_str)

    def show_job_definition(self, args, unknown_args):
        import gm_meta_plugin
        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except gm_meta_plugin.errors.APIClientError as e:
                logger.info(str(e))
                return

        job_definition_id = args.job_definition_id
        api_schema = PLUGIN.get_api_client().get_api_schema()

        try:
            job_definition = api_schema.get_job_definition_with_id(job_definition_id)
        except ValueError as e:
            logger.info("No job definition with ID '{}'".format(job_definition_id))
            return

        logger.info('Name: {}'.format(job_definition.get_display_name()))
        logger.info('ID: {}'.format(job_definition.get_definition_id()))

        logger.info('\nAttribute Groups:')
        attribute_groups = job_definition.get_attribute_groups()
        for attribute_group in attribute_groups:
            logger.info(attribute_group)

        logger.info('\nJob Attributes:')
        job_attributes = job_definition.get_attributes()
        for job_attribute in job_attributes:
            attribute = job_attribute.get_attribute()
            logger.info('Name: {}'.format(attribute.get_display_name()))
            logger.info('Description: {}'.format(attribute.get_description()))
            logger.info('Key: {}'.format(attribute.get_key()))
            logger.info('Type: {}'.format(attribute.get_type()))
            logger.info('Default value: {}'.format(attribute.get_default_value()))
            logger.info('Subtype Kargs: {}'.format(attribute.get_subtype_kwargs()))
            logger.info('Inference sources: {}'.format(', '.join(map(lambda x: str(x), job_attribute.get_inference_sources()))))
            logger.info('Dependencies: {}'.format(','.join(job_attribute.get_dependencies_ids())))
            logger.info('')

    def get_product_versions(self, args, unknown_args):
        import gm_meta_plugin
        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except gm_meta_plugin.errors.APIClientError as e:
                logger.info(str(e))
                return

        app = args.app
        products = PLUGIN.get_api_client().get_product_versions(app)

        if len(products) > 0:
            for product in products:
                logger.info("{}".format(product))
        else:
            logger.info("No versions found")

    def get_machines(self, args, unknown_args):
        import gm_meta_plugin
        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except gm_meta_plugin.errors.APIClientError as e:
                logger.info(str(e))
                return

        app = args.app
        operation = args.operation
        plugins = (args.plugins).split(',')

        machines = PLUGIN.get_api_client().get_machines(app, operation, plugins)

        logger.info('Found {} machine options'.format(len(machines)))

        if machines:
            max_length_name = max(map(lambda x: len(x.get_name()), machines))
            logger.info(('\n{0:{1}}  {2}').format("Name", max_length_name, "ID"))
            for machine in machines:
                logger.info(('{0:{1}}  {2}').format(machine.get_name(),
                                                    max_length_name,
                                                    machine.get_id()))

    def get_user_info(self, args, unknown_args):
        import gm_meta_plugin
        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except gm_meta_plugin.errors.APIClientError as e:
                logger.info(str(e))
                return

        logger.info('\nUser Information:')
        serialised_user_info = PLUGIN.get_api_client().get_user_info().serialise()

        table = [['Key'], ['Value']]
        for key, value in serialised_user_info.items():
            table[0].append(key)
            table[1].append(str(value))

        table_str = gm_meta_plugin.utils.format_table_str(table)
        logger.info(table_str)

    def submit_job(self, args, unknown_args):
        import gm_meta_plugin
        from gm_meta_plugin.gridmarkets import constants as api_constants
        from .plugin_fetcher import PluginFetcher
        global PLUGIN

        if not PLUGIN:
            PLUGIN = PluginFetcher.get_plugin()

        logger = PLUGIN.get_logging_coordinator().get_logger(constants.CLI_LOGGER_ID)

        if not PLUGIN.get_api_client().is_user_signed_in():
            try:
                logger.info('Signing in with Envoy...')
                PLUGIN.get_api_client().sign_in_with_envoy()
            except gm_meta_plugin.errors.APIClientError as e:
                logger.info(str(e))
                return
            except ValueError as e:
                logger.info(str(e))
                return

        project_dir = pathlib.Path(args.project_dir)
        job_definition_id = args.job_definition_id
        skip_upload = args.skip_upload

        api_client = PLUGIN.get_api_client()
        api_schema = api_client.get_api_schema()

        try:
            job_definition = api_schema.get_job_definition_with_id(job_definition_id)
        except ValueError as e:
            logger.info("No job definition with ID '{}'".format(job_definition_id))
            return

        job_attributes = job_definition.get_attributes()
        dynamic_job_preset = gm_meta_plugin.dynamic_job_preset.DynamicJobPreset(PLUGIN, 'temp-job-preset', 'temp-job-preset', job_definition)

        show_ui = args.show_ui
        enable_cli_prompt = args.enable_cli_prompt

        if show_ui:
            raise NotImplementedError

        # Set all job preset attributes to read from sys.argv
        project_inference_source = gm_meta_plugin.inference_source.InferenceSource.get_project_inference_source()
        argv_inference_source = gm_meta_plugin.inference_source.InferenceSource.get_argv_inference_source()
        for job_attribute in job_attributes:
            attribute = job_attribute.get_attribute()

            job_attribute_id = job_attribute.get_id()

            job_preset_attribute = dynamic_job_preset.get_job_preset_attribute_with_job_attribute_id(job_attribute_id)

            inference_sources = job_preset_attribute.get_job_attribute().get_inference_sources()
            if project_inference_source in inference_sources:
                job_preset_attribute.set_inference_source(project_inference_source)
            else:
                try:
                    job_preset_attribute.set_inference_source(argv_inference_source)
                except ValueError:
                    logger.info("Job Definition '{}' does not support command line submissions.\n Job attribute '{}' "
                                "does not support the {} inference source.".format(job_definition_id,
                                                                                   job_attribute_id,
                                                                                   argv_inference_source.get_id()))
                    return

            """
            if not attribute.has_argv_value() and not job_preset_attribute.get_job_attribute().is_optional():
                logger.info("Missing argument '{}' for Job attribute '{}' ".format('--' + key.replace('_', '-'),
                                                                                   job_definition_id))
                return
            """

        """
        unprocessed_job_preset_attributes = set(job_preset.get_job_preset_attributes())
        processed_job_preset_attributes = set()
        while True:
            for job_preset_attribute in unprocessed_job_preset_attributes:

                # If all of the dependencies have already been processed
                if set(job_preset_attribute.get_dependencies()).issubset(processed_job_preset_attributes):

                    # And the job preset attribute is "active"
                    if job_preset_attribute.is_active():
        """
        try:
            if skip_upload:
                remote_project = gm_meta_plugin.RemoteProject(project_dir, '/{}'.format(project_dir), [], dict())
                api_client.submit_to_remote_project(remote_project,
                                                    dynamic_job_preset.get_static_job_preset(remote_project))
            else:
                # get a list of all the files associated with the scene
                project_files = gm_meta_plugin.utils.get_files_in_directory(project_dir)

                cached_api_schema = api_client.get_cached_api_schema()
                project_definition_schema = cached_api_schema.get_project_definition_schema()
                project_instance_builder_factory = PLUGIN.get_factory_collection().get_project_instance_builder_factory()
                project_instance_builder = project_instance_builder_factory.get_project_instance_builder(project_definition_schema)

                project_instance_builder.update_states(infer_from_cmd=True)
                project_attributes = project_instance_builder.get_project_attributes(force_transition=False)

                packed_project = gm_meta_plugin.PackedProject(project_attributes[api_constants.API_KEYS.PROJECT_NAME],
                                                              project_dir, project_files, project_attributes)

                api_client.submit_new_project(packed_project, dynamic_job_preset)
        except gm_meta_plugin.errors.CLIPromptNotEnabledError as e:
            print(e)

    def process(self, job: 'WorkerJob') -> None:

        self.active_job = job

        command_line_arguments = job.get_argv()[1:] # [1:] to strip the program's path
        try:
            # Remove all the special unknown cli flags from the command line arguments
            for unknown_cli_flag in constants.UNKNOWN_CLI_ARGS:
                if unknown_cli_flag in command_line_arguments:
                    command_line_arguments.remove(unknown_cli_flag)

            # If after removing the unknown cli flag from the arguments there are no arguments left then print
            # help info
            if not len(command_line_arguments):
                if constants.USE_DAEMON_FLAG in job.get_argv()[1:]:
                    command_line_arguments = ['version']
                else:
                    command_line_arguments = ['-h']

            if constants.USE_DAEMON_FLAG in sys.argv:
                logging.info(json.dumps(command_line_arguments, indent=3))

            # Parse the args without emitting any errors
            args, unknown_args = self.parser.parse_known_args(args=command_line_arguments)

            func = None
            try:
                func = args.func
            except AttributeError:
                pass

            if func:
                func(args, unknown_args)
            else:
                args, unknown_args = self.parser.parse_known_args(args=['-h'])
                args.func(args, unknown_args)

        except SystemExit:
            exc = sys.exc_info()[1]
            #print("SystemExit:", exc)

        job.mark_as_complete()

    def clean_up(self) -> None:
        self._clean_up_main_window()
