__all__ = 'BlenderJobPresetValidator'

import pathlib
import typing

from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin import JobPresetValidator
from . import blender_constants


class BlenderJobPresetValidator(JobPresetValidator):

    def validate(self, job_preset) -> typing.Tuple[bool, str]:
        is_valid = True
        validation_message = u""

        # Validate Blender renderer
        renderer_attribute = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.RENDER_ENGINE)
        renderer = renderer_attribute.eval()
        if renderer not in blender_constants.SUPPORTED_RENDER_ENGINES:
            renderer_name = renderer

            if renderer == blender_constants.RENDER_ENGINE_CYCLES:
                renderer_name = "Cycles"
            elif renderer == blender_constants.RENDER_ENGINE_EEVEE:
                renderer_name = "Eevee"
            elif renderer == blender_constants.RENDER_ENGINE_BLENDER_INTERNAL:
                renderer_name = "Blender Internal"
            elif renderer == blender_constants.RENDER_ENGINE_BLENDER_WORKBENCH:
                renderer_name = "Blender Workbench"
            elif renderer == blender_constants.RENDER_ENGINE_VRAY_RT:
                renderer_name = "V-Ray"

            return False, "Render engine '%s' is not currently supported." % renderer_name

        # Validate the output prefix
        output_prefix_attribute = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.OUTPUT_PREFIX)

        if output_prefix_attribute.is_active() and output_prefix_attribute.is_enabled():
            output_prefix: str = output_prefix_attribute.eval()
            error_message_suffix = "in the Submit2gm submission interface."
        else:
            output_prefix = output_prefix_attribute.eval_application_defined_value()
            error_message_suffix = "in 'Output Properties > Output > Output Path' in Blender."

        if not output_prefix:
            return False, "No value set for your output path.\n" \
                          "Please enter an output path value {}".format(error_message_suffix)

        # Check that the output path is relative
        if not output_prefix.startswith('//'):
            return False, "Output path must be a relative file path.\n" \
                          "Please set a relative value (i.e. starting with '//') {}".format(error_message_suffix)

        # Check that the output path is within the project folder
        if output_prefix.startswith('//..'):
            return False, "Output path must not start with '//..' otherwise results may be written outside of the project folder.\n" \
                          "Please modify the value {}".format(error_message_suffix)

        return is_valid, validation_message
