__all__ = 'C4DAttributeSource'

import os
import pathlib
import typing

import gm_meta_plugin
from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.errors import ApplicationAttributeNotFound
from gm_meta_plugin.attribute_types import EnumItem
from submit2gm import constants
from . import c4d_constants, c4d_cpc


class ArnoldAttributeSource(gm_meta_plugin.ApplicationAttributeSource):
    """ A simple attribute source for arnold specific settings """

    DEVICE_CPU = 0
    DEVICE_GPU = 1

    def get_enum_items(self, app, app_version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.MACHINE_TYPE:

            # Older versions of MtoA do not have a render device attribute and only support cpu
            try:
                device = c4d_cpc.get_arnold_render_device(cps)
            except:
                device = ArnoldAttributeSource.DEVICE_CPU

            if device == ArnoldAttributeSource.DEVICE_CPU:
                use_gpu_default = False
            else:
                use_gpu_default = True

            plugins = [api_constants.PLUGINS.ARNOLD]

            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.C4D,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=use_gpu_default,
                                                                filter_cpu=use_gpu_default,
                                                                filter_gpu=not use_gpu_default)

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, version: str, key: str) -> any:
        raise ApplicationAttributeNotFound(app, version, key)


class RedshiftAttributeSource(gm_meta_plugin.ApplicationAttributeSource):
    """ A simple attribute source for redshift specific settings """

    def get_enum_items(self, app, app_version, key):
        plugin = self.get_plugin()

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            plugins = [api_constants.PLUGINS.REDSHIFT]

            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.C4D,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=True,
                                                                filter_cpu=True,
                                                                filter_gpu=False)

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, version: str, key: str) -> any:
        raise ApplicationAttributeNotFound(app, version, key)


class VrayAttributeSource(gm_meta_plugin.ApplicationAttributeSource):
    """ A simple attribute source for redshift specific settings """

    def get_enum_items(self, app, app_version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            try:
                requires_gpu = c4d_cpc.c4d_vray_requires_gpu(cps)
            except:
                requires_gpu = False

            plugins = [api_constants.PLUGINS.VRAY]

            plugin.get_logging_coordinator().get_logger().info("vray requires_gpu: {}".format(str(requires_gpu)))

            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.C4D,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=requires_gpu,
                                                                filter_cpu=requires_gpu,
                                                                filter_gpu=not requires_gpu)

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, version: str, key: str) -> any:
        raise ApplicationAttributeNotFound(app, version, key)


class C4DAttributeSource(gm_meta_plugin.ApplicationAttributeSource):

    def __init__(self, plugin):
        gm_meta_plugin.ApplicationAttributeSource.__init__(self, plugin)

        self._arnold_as = ArnoldAttributeSource(plugin)
        self._redshift_as = RedshiftAttributeSource(plugin)
        self._vray_as = VrayAttributeSource(plugin)

        # create a dictionary of attribute sources for the different renderers.
        self._renderer_attribute_source_dict: typing.Dict[str, gm_meta_plugin.ApplicationAttributeSource] = {
            'arnold': self._arnold_as,
            'redshift': self._redshift_as,
            'vray': self._vray_as,
        }

    def get_enum_items(self, app: str, app_version: str, key: str):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        current_renderer = self.get_attribute_value(app, app_version, api_constants.API_KEYS.RENDERER)

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            for renderer_key, attrib_source in self._renderer_attribute_source_dict.items():
                if renderer_key in current_renderer:
                    try:
                        return attrib_source.get_enum_items(app, app_version, key)
                    except ApplicationAttributeNotFound as e:
                        pass

        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass

        if key == api_constants.API_KEYS.APP_VERSION:
            items = []
            product_versions = self.get_plugin().get_api_client().get_product_versions(app)
            for product_version in product_versions:
                items.append(EnumItem(product_version, product_version, u""))
            return items, self._get_enum_index(app, app_version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.C4D,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                use_gpu_default=True)

        if key == api_constants.API_KEYS.TAKE:
            takes = c4d_cpc.get_takes(cps)
            items = list(map(lambda x: EnumItem(x['name'], x['name'], u""), takes))
            return items, self._get_enum_index(app, app_version, key, items)

        if key == api_constants.API_KEYS.TAKES:
            takes = c4d_cpc.get_takes(cps)
            items = list(map(lambda x: EnumItem(x['name'], x['name'], u""), takes))
            return items, 0

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, app_version: str, key: str):
        cps = self.get_plugin().get_active_command_port_server()

        current_renderer = c4d_cpc.get_render_engine(cps)

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            for renderer_key, attrib_source in self._renderer_attribute_source_dict.items():
                if renderer_key in current_renderer:
                    try:
                        return attrib_source.get_enum_items(app, app_version, key)
                    except ApplicationAttributeNotFound as e:
                        pass

        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass

        if app != api_constants.PRODUCTS.C4D:
            raise ValueError(str(app))

        if key == api_constants.API_KEYS.IS_SAVED:
            return len(c4d_cpc.get_document_file_path(cps)) > 0

        if key == api_constants.API_KEYS.PROJECT_NAME:
            return pathlib.Path(c4d_cpc.get_document_name(cps)).stem

        if key == api_constants.API_KEYS.SUBMISSION_NAME:
            return c4d_cpc.get_document_name(cps)

        if key == api_constants.API_KEYS.JOB_NAME:
            job_name_parts = []

            take_name = c4d_cpc.get_current_take_name(cps)
            if take_name:
                job_name_parts.append(take_name)

            if current_renderer:
                job_name_parts.append(current_renderer.capitalize())

            return ' '.join(list(map(lambda x: "[{}]".format(x), job_name_parts)))

        if key == api_constants.API_KEYS.APP:
            return api_constants.PRODUCTS.C4D

        if key == api_constants.API_KEYS.APP_VERSION:
            c4d_version_str = c4d_cpc.get_c4d_version_str(cps)
            return self.get_plugin().get_api_client().get_closest_matching_product_version(
                api_constants.PRODUCTS.C4D,
                c4d_version_str)

        if key == api_constants.API_KEYS.SCENE_FILE_PATH_UNUSED:
            return c4d_cpc.get_document_file_path(cps)

        if key == api_constants.API_KEYS.RENDERER:
            return current_renderer

        if key == api_constants.API_KEYS.OUTPUT_WIDTH:
            return c4d_cpc.get_output_width(cps)

        if key == api_constants.API_KEYS.OUTPUT_HEIGHT:
            return c4d_cpc.get_output_height(cps)

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            return c4d_cpc.get_output_format(cps)

        if key == api_constants.API_KEYS.OUTPUT_PREFIX:
            return c4d_cpc.get_render_path(cps)

        if key == api_constants.API_KEYS.OUTPUT_MULTIPASS_FORMAT:
            return c4d_cpc.get_output_multipass_format(cps)

        if key == api_constants.API_KEYS.OUTPUT_MULTIPASS_PREFIX:
            return c4d_cpc.get_multipass_render_path(cps)

        if key == api_constants.API_KEYS.FRAMES:
            return c4d_cpc.get_frame_range_str(cps)

        if key == api_constants.API_KEYS.TAKE:
            takes = c4d_cpc.get_takes(cps)
            for take in takes:
                if take['is_checked']:
                    return take['name']
            return 'Main'

        if key == api_constants.API_KEYS.TAKES:
            takes = c4d_cpc.get_takes(cps)
            result = []
            for take in takes:
                if take['is_checked']:
                    result.append(take['name'])
            return result

        raise ApplicationAttributeNotFound(app, app_version, key)
