__all__ = 'BlenderJobPresetValidator'

import pathlib
import typing

from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin import JobPresetValidator
from . import c4d_constants, c4d_cpc


class C4DJobPresetValidator(JobPresetValidator):

    def validate(self, job_preset) -> typing.Tuple[bool, str]:
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        is_valid = True
        validation_message = u""

        global_save = c4d_cpc.get_global_save(cps)
        is_global_multipass_enabled = c4d_cpc.is_global_multipass_enabled(cps)

        save_regular_image = c4d_cpc.should_save_regular_image(cps)
        is_multipass_save_enabled = c4d_cpc.is_multipass_save_enabled(cps)

        # Validate the output prefix
        output_prefix_attribute = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.OUTPUT_PREFIX)

        if not global_save:
            return False, "File saving is globally disabled for your scene.\n" \
                          "Please enable 'Save' in your Render Settings otherwise your results will not be saved."

        if (not save_regular_image) and (not is_global_multipass_enabled or not is_multipass_save_enabled):
            return False, "File saving is disabled for both regular and multi-pass outputs.\n" \
                          "Please enable either your regular image output or multi-pass output otherwise your results will not be saved."

        """
        C4D still saves the image even if the regular output or global output render options are disabled in the scenes 
        render settings when -oimage is provided on the command line so there is no need to validate that these are set.
        
        If we are not explicitly passing an output_prefix value then -oimage will not be set on the command line. This 
        means we need to validate the value defined in the scene file as well as whether it is enabled.
        """
        eval_output_prefix = output_prefix_attribute.is_active() and output_prefix_attribute.is_enabled()
        if eval_output_prefix or save_regular_image:

            if eval_output_prefix:
                output_prefix: str = output_prefix_attribute.eval()
                error_message_suffix = "in the Submit2gm submission interface for '{}'.".format(output_prefix_attribute.get_job_attribute().get_attribute().get_display_name())
            else:
                output_prefix = output_prefix_attribute.eval_application_defined_value()
                error_message_suffix = "in 'Render Settings > Save > Regular Image > File...'"

            if not output_prefix:
                return False, "No output file path set for regular image.\n" \
                              "Please set a value {}".format(error_message_suffix)

            # Check that the output prefix is relative
            if pathlib.Path(output_prefix).is_absolute() or output_prefix.startswith('/'):
                return False, "Output path must be a relative file path.\n" \
                              "Please set a relative value {}".format(error_message_suffix)

            # Check that the output prefix is within the project folder
            if output_prefix.startswith('..'):
                return False, "Output path must not start with '..' otherwise results may be written outside of the project folder.\n" \
                              "Please modify the value {}".format(error_message_suffix)

        # Validate the multi-pass output prefix
        multipass_output_prefix_attribute = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.OUTPUT_MULTIPASS_PREFIX)

        eval_multi_pass = multipass_output_prefix_attribute.is_active() and multipass_output_prefix_attribute.is_enabled()
        scene_value_multi_pass = is_global_multipass_enabled or is_multipass_save_enabled
        if eval_multi_pass or scene_value_multi_pass:
            if eval_multi_pass:
                multipass_output_prefix = multipass_output_prefix_attribute.eval()
                error_message_suffix = "in the Submit2gm submission interface for '{}'.".format(multipass_output_prefix_attribute.get_job_attribute().get_attribute().get_display_name())
            else:
                multipass_output_prefix = multipass_output_prefix_attribute.eval_application_defined_value()
                error_message_suffix = "in 'Render Settings > Save > Multi-pass Image > File...'"

            # If there the multi-pass output prefix is an empty string then the api will simply ignore it
            if multipass_output_prefix:
                # Check that the multi-pass output prefix is relative
                if pathlib.Path(multipass_output_prefix).is_absolute() or multipass_output_prefix.startswith('/'):
                    return False, "Multi-pass Output path must be a relative file path.\n" \
                                  "Please set a relative value {}".format(error_message_suffix)

                # Check that the multi-pass output prefix is within the project folder
                if multipass_output_prefix.startswith('..'):
                    return False, "Multi-pass Output path must not start with '..' otherwise results may be written outside of the project folder.\n" \
                                  "Please modify the value {}".format(error_message_suffix)

        return is_valid, validation_message


