__all__ = 'C4DPackedSceneBuilder'

import os
import pathlib
import typing
import tempfile

import gm_meta_plugin

from gm_meta_plugin.errors import FilePackingError, ProjectRequiresSavingError
from gm_meta_plugin.gridmarkets import constants as api_constants
from . import c4d_constants, c4d_cpc
from .c4d_scene_exporter import C4DSceneExporter

if typing.TYPE_CHECKING:
    from gm_meta_plugin import PackedProject
    from gm_qt_plugin.widgets.job_preset import JobPreset


class C4DPackedSceneBuilder(gm_meta_plugin.PackedSceneBuilder):

    def __init__(self, plugin):
        self._plugin = plugin

    def get_plugin(self):
        return self._plugin

    def _build_scene_internal(self,
                              project_name: str,
                              product: str,
                              product_version: str,
                              save_project_if_required: bool,
                              attributes: typing.Dict[str, any],
                              job_preset: typing.Optional['JobPreset'] = None):

        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        # Is the scene saved
        if not c4d_cpc.is_document_saved(cps):
            raise FilePackingError(u"Scene file not saved.")

        # Has the scene file been modified
        if c4d_cpc.is_document_modified(cps):
            if save_project_if_required:
                c4d_cpc.save_project(cps)
            else:

                def _get_current_scene():
                    return self.get_current_scene(project_name, product, product_version, True, attributes,
                                               job_preset=job_preset), job_preset

                raise ProjectRequiresSavingError(_get_current_scene, u"Scene file has been modified.")

        pre_packed_local_root = pathlib.Path(c4d_cpc.get_document_file_path(cps)).parent

        temp_dir = tempfile.mkdtemp(prefix='gm-temp-dir-')

        # Resolve the path because Max OS likes to return a symlink path instead of the real path sometimes for temp
        # files
        temp_dir_path = pathlib.Path(temp_dir).resolve()

        if os.name == 'nt':
            # Windows sometimes returns short name versions of the path so expand the path to the full name
            temp_dir_path = gm_meta_plugin.utils.expand_windows_short_name_path(temp_dir_path)

        if product == api_constants.PRODUCTS.C4D:
            root_dir = C4DSceneExporter(self.get_plugin()).export(temp_dir_path)
        else:
            raise RuntimeError("Unknown project type")

        # get a list of all the files associated with the scene
        project_files = gm_meta_plugin.utils.get_files_in_directory(root_dir)

        # set the path attribute
        attributes[api_constants.API_KEYS.PATH] = str(root_dir / c4d_cpc.get_document_name(cps))

        packed_project = gm_meta_plugin.PackedProject(project_name,
                                                      root_dir,
                                                      project_files,
                                                      attributes,
                                                      pre_packed_local_root=pre_packed_local_root)

        return packed_project

    def get_project_type_id(self):
        return api_constants.PROJECT_TYPE_IDS.C4D_PROJECT

    def get_product_type(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, api_constants.API_KEYS.APP)

    def get_product_version(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, api_constants.API_KEYS.APP_VERSION)
