__all__ = 'MaxAttributeSource'

import pathlib
import typing

import gm_meta_plugin
from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.errors import ApplicationAttributeNotFound
from gm_meta_plugin.attribute_types import EnumItem
from submit2gm import constants
from . import max_constants, max_cpc


class RedshiftAttributeSource(gm_meta_plugin.ApplicationAttributeSource):
    """ A simple attribute source for redshift specific settings """

    def get_enum_items(self, app, app_version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            plugins = [api_constants.PLUGINS.REDSHIFT]
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAX,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=True)

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, app_version: str, key: str) -> any:
        raise ApplicationAttributeNotFound(app, app_version, key)


class VrayAttributeSource(gm_meta_plugin.ApplicationAttributeSource):
    """ A simple attribute source for vray specific settings """

    def get_enum_items(self, app, app_version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()
        application_attribute_source = plugin.get_application_attribute_source()
        current_renderer = application_attribute_source.get_attribute_value(app,
                                                                            app_version,
                                                                            api_constants.API_KEYS.RENDERER)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            plugins = [api_constants.PLUGINS.VRAY]
            gpu = 'gpu' in MaxAttributeSource.sanitise_renderer(current_renderer).casefold()

            if gpu:
                filter_cpu = True
                filter_gpu = False
                use_gpu_default = True
            else:
                filter_cpu = False
                filter_gpu = True
                use_gpu_default = False

            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAX,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                filter_cpu=filter_cpu,
                                                                filter_gpu=filter_gpu,
                                                                use_gpu_default=use_gpu_default)

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            native_items = list(map(lambda x: EnumItem(x[0], x[1], x[2]), max_constants.NATIVE_OUTPUT_FORMATS))
            vray_items = list(map(lambda x: EnumItem(x[0], x[1], x[2]), max_constants.VRAY_OUTPUT_FORMATS))
            items = native_items + vray_items
            return items, application_attribute_source._get_enum_index(app, app_version, key, items)

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, version: str, key: str) -> any:
        raise ApplicationAttributeNotFound(app, version, key)


class ArnoldAttributeSource(gm_meta_plugin.ApplicationAttributeSource):
    """ A simple attribute source for arnold specific settings """

    def get_enum_items(self, app, app_version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            plugins = [api_constants.PLUGINS.ARNOLD]

            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAX,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins)

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, version: str, key: str) -> any:
        raise ApplicationAttributeNotFound(app, version, key)


class MaxAttributeSource(gm_meta_plugin.ApplicationAttributeSource):

    def __init__(self, plugin):
        gm_meta_plugin.ApplicationAttributeSource.__init__(self, plugin)

        self._redshift_as = RedshiftAttributeSource(plugin)
        self._vray_as = VrayAttributeSource(plugin)
        self._arnold_as = ArnoldAttributeSource(plugin)

        # create a dictionary of attribute sources for the different renderers.
        self._renderer_attribute_source_dict: typing.Dict[str, gm_meta_plugin.ApplicationAttributeSource] = {
            'redshift': self._redshift_as,
            'vray': self._vray_as,
            'arnold': self._arnold_as,
        }

    @staticmethod
    def sanitise_renderer(renderer: str) -> str:
        return renderer.replace('_', '').casefold()

    def get_enum_items(self, app: str, app_version: str, key: str):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        current_renderer = self.get_attribute_value(app, app_version, api_constants.API_KEYS.RENDERER)

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            current_renderer_sanitised = MaxAttributeSource.sanitise_renderer(current_renderer)
            for renderer_key, attrib_source in self._renderer_attribute_source_dict.items():
                if renderer_key in current_renderer_sanitised:
                    try:
                        return attrib_source.get_enum_items(app, app_version, key)
                    except ApplicationAttributeNotFound as e:
                        pass

        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass

        if key == api_constants.API_KEYS.APP_VERSION:
            items = []

            current_max_version = self.get_attribute_value(app, app_version, api_constants.API_KEYS.APP_VERSION)
            items.append(EnumItem(current_max_version, current_max_version, u""))

            return items, 0

        if key == api_constants.API_KEYS.CAMERA:
            items = []

            for camera_name in max_cpc.get_camera_names(cps):
                items.append(EnumItem(camera_name, camera_name, u""))

            return items, self._get_enum_index(app, app_version, key, items)

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            items = list(map(lambda x: EnumItem(x[0], x[1], x[2]), max_constants.NATIVE_OUTPUT_FORMATS))
            return items, self._get_enum_index(app, app_version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAX,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                use_gpu_default=False)

        raise ApplicationAttributeNotFound(app, app_version, key)

    def get_attribute_value(self, app: str, app_version: str, key: str):
        cps = self.get_plugin().get_active_command_port_server()

        current_renderer = max_cpc.get_max_current_renderer(cps)
        if ':' in current_renderer:
            current_renderer = current_renderer.split(':')[0]

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            current_renderer_sanitised = MaxAttributeSource.sanitise_renderer(current_renderer)
            for renderer_key, attrib_source in self._renderer_attribute_source_dict.items():
                if renderer_key in current_renderer_sanitised:
                    try:
                        return attrib_source.get_attribute_value(app, app_version, key)
                    except ApplicationAttributeNotFound as e:
                        pass

        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass

        if key == api_constants.API_KEYS.IS_SAVED:
            return len(max_cpc.get_max_file_name(cps)) > 0

        if key == api_constants.API_KEYS.PROJECT_NAME:
            return pathlib.Path(max_cpc.get_current_project_folder(cps)).stem

        if key == api_constants.API_KEYS.SUBMISSION_NAME:
            scene_name = pathlib.Path(max_cpc.get_max_file_name(cps)).name

            if not scene_name:
                return "Untitled"

            return scene_name

        if key == api_constants.API_KEYS.JOB_NAME:
            job_name_parts = []

            camera_name = self.get_attribute_value(app, app_version, api_constants.API_KEYS.CAMERA)
            if camera_name:
                job_name_parts.append(camera_name)

            if current_renderer:
                job_name_parts.append(current_renderer.capitalize())

            return ' '.join(list(map(lambda x: "[{}]".format(x), job_name_parts)))

        if key == api_constants.API_KEYS.APP:
            return api_constants.PRODUCTS.MAX

        if key == api_constants.API_KEYS.APP_VERSION:
            max_release_version: str = max_cpc.get_release_version(cps)

            if max_release_version.startswith('19'):
                return '2017'
            if max_release_version.startswith('20'):
                return '2018'
            if max_release_version.startswith('21'):
                return '2019'
            if max_release_version.startswith('22'):
                return '2020'
            if max_release_version.startswith('23'):
                return '2021'

            # Guessing these two since they are not released yet
            if max_release_version.startswith('24'):
                return '2022'
            if max_release_version.startswith('25'):
                return '2023'

            return '2021'

        if key == api_constants.API_KEYS.SCENE_FILE_PATH_UNUSED:
            file_name = max_cpc.get_max_file_name(cps)

            if not file_name:
                return ''

            return (pathlib.Path(max_cpc.get_max_file_path(cps)) / max_cpc.get_max_file_name(cps)).as_posix()

        if key == api_constants.API_KEYS.RENDERER:
            return current_renderer

        if key == api_constants.API_KEYS.IN_GAMMA:
            return max_cpc.get_file_in_gamma(cps)

        if key == api_constants.API_KEYS.OUT_GAMMA:
            return max_cpc.get_file_out_gamma(cps)

        if key == api_constants.API_KEYS.FRAMES:
            return max_cpc.get_frame_range_str(cps)

        if key == api_constants.API_KEYS.CAMERA:
            camera_names = max_cpc.get_camera_names(cps)

            if len(camera_names):
                return camera_names[0]

            return ""

        if key == api_constants.API_KEYS.OUTPUT_WIDTH:
            return max_cpc.get_output_width(cps)

        if key == api_constants.API_KEYS.OUTPUT_HEIGHT:
            return max_cpc.get_output_height(cps)

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            render_output_path = max_cpc.get_render_output_path(cps)
            default_output_format = 'png'

            if not render_output_path:
                return default_output_format

            file_suffix = pathlib.Path(render_output_path).suffix

            if not file_suffix:
                return default_output_format

            if file_suffix in list(map(lambda x: '.' + x[0], max_constants.NATIVE_OUTPUT_FORMATS)):
                return file_suffix[1:]

            return default_output_format

        if key == api_constants.API_KEYS.OUTPUT_PREFIX:
            local_project_path = pathlib.Path(max_cpc.get_current_project_folder(cps))

            # 3ds can write your render results to the project defined render output folder or to the file path defined
            # in your render settings. The render settings path takes precedence.
            render_settings_output_path = max_cpc.get_render_output_path(cps)

            if render_settings_output_path:
                output_file_name = pathlib.Path(render_settings_output_path).stem

                # Remove the file extension from the output path
                output_path = pathlib.Path(render_settings_output_path).parent / output_file_name

                try:
                    # Make sure the path is relative to the project directory.
                    output_path = output_path.relative_to(local_project_path).as_posix()

                    # Make sure the output path is a directory
                    if not (output_path.endswith('/') or output_path.endswith('\\')):
                        output_path = output_path + '/'

                    return output_path
                except ValueError:
                    # If the render settings output path is not relative to the local project path then we have a
                    # problem as we cannot validate the path at this point in the logic and without absolute file
                    # support the users requested path will not work. In this case we fall back to the project defined
                    # output path.
                    pass

            project_output_path = max_cpc.get_render_output(cps)

            if project_output_path:
                try:
                    output_path = pathlib.Path(project_output_path).relative_to(local_project_path).as_posix()

                    # Make sure the output path is a directory
                    if not (output_path.endswith('/') or output_path.endswith('\\')):
                        output_path = output_path + '/'

                    return output_path
                except ValueError:
                    # If even the project output path is not relative to the project folder then we have to fallback to
                    # an arbitrary default output path
                    pass

            return 'renderoutput/'

        raise ApplicationAttributeNotFound(app, app_version, key)
