__all__ = 'MayaAttributeSource'

import re
import pathlib

import gm_meta_plugin
from gm_meta_plugin.gridmarkets import constants as api_constants
from gm_meta_plugin.errors import ApplicationAttributeNotFound
from gm_meta_plugin.attribute_types import EnumItem
from . import maya_constants, maya_cpc


class ThirdPartyRendererAttributeSource(gm_meta_plugin.ApplicationAttributeSource):

    def __init__(self, plugin):
        gm_meta_plugin.ApplicationAttributeSource.__init__(self, plugin)
        self._is_setup: bool = False

    def is_setup(self) -> bool:
        return self._is_setup

    def setup(self) -> None:
        self._is_setup = True


class ArnoldAttributeSource(ThirdPartyRendererAttributeSource):
    """ A simple attribute source for mtoa specific settings """

    DEVICE_CPU = 0
    DEVICE_GPU = 1

    def __init__(self, plugin):
        ThirdPartyRendererAttributeSource.__init__(self, plugin)

    def setup(self):
        ThirdPartyRendererAttributeSource.setup(self)

    def get_enum_items(self, app, version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.MACHINE_TYPE:

            # Older versions of MtoA do not have a render device attribute and only support cpu
            try:
                device = maya_cpc.get_arnold_render_device(cps)
            except:
                device = ArnoldAttributeSource.DEVICE_CPU

            if device == ArnoldAttributeSource.DEVICE_CPU:
                filter_cpu = False
                filter_gpu = True
            else:
                filter_cpu = True
                filter_gpu = False

            plugins = [maya_constants.RENDERERS.MAYA_ARNOLD_RENDERER.get_commandline_name()]
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAYA,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=False,
                                                                filter_cpu=filter_cpu,
                                                                filter_gpu=filter_gpu)

        elif key == api_constants.API_KEYS.OUTPUT_FORMAT:
            arnold_translator_names = maya_cpc.get_arnold_translator_names(cps)
            items = list(map(lambda x: EnumItem(x, x, u""), arnold_translator_names))
            return items, self._get_enum_index(app, version, key, items)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        cps = self.get_plugin().get_active_command_port_server()

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_ARNOLD_RENDERER.get_commandline_name()

        elif key == api_constants.API_KEYS.OUTPUT_FORMAT:
            return maya_cpc.get_arnold_output_format(cps)

        raise ApplicationAttributeNotFound(app, version, key)


class VrayAttributeSource(ThirdPartyRendererAttributeSource):
    """
    A simple attribute source for vray specific settings
    See C:\Program Files\Autodesk\Maya2020\vray\scripts\vray for vray python api files
    """

    def setup(self):
        ThirdPartyRendererAttributeSource.setup(self)

    def _vray_output_format_label_to_value(self, label):
        return label.split(' ')[0]

    def get_enum_items(self, app, version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            image_format_keys = list(maya_constants.VRAY_IMAGE_FORMATS.keys())
            items = list(map(lambda x: EnumItem(maya_constants.VRAY_IMAGE_FORMATS[x], x, ''), image_format_keys))
            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            # Detect which engine is set
            engine = maya_cpc.get_vray_production_engine(cps)

            # For vray 5
            # CPU = 0
            # GPU = 2

            # For vray 4.3
            # CPU = 0
            # GPU = 1
            # RTX = 2

            if engine == 0:
                filter_cpu = False
                filter_gpu = True
                use_gpu_default = False
            else:
                filter_cpu = True
                filter_gpu = False
                use_gpu_default = True

            plugins = [maya_constants.RENDERERS.MAYA_VRAY_RENDERER.get_commandline_name()]
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAYA,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=use_gpu_default,
                                                                filter_cpu=filter_cpu,
                                                                filter_gpu=filter_gpu)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        cps = self.get_plugin().get_active_command_port_server()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            vray_output_format_label = maya_cpc.get_vray_output_format_label(cps)
            return self._vray_output_format_label_to_value(vray_output_format_label)

        if key == api_constants.API_KEYS.OUTPUT_PREFIX:
            output_prefix = maya_cpc.get_vray_output_prefix(cps)

            if not output_prefix:
                output_prefix = '<Scene>'

            return output_prefix

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_VRAY_RENDERER.get_commandline_name()

        raise ApplicationAttributeNotFound(app, version, key)


class RedshiftAttributeSource(ThirdPartyRendererAttributeSource):
    """ A simple attribute source for redshift specific settings """

    def setup(self):
        ThirdPartyRendererAttributeSource.setup(self)

    def get_enum_items(self, app, version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:

            items = []

            for rs_image_format_id, rs_image_format in maya_constants.REDSHIFT_IMAGE_FORMATS.items():
                items.append(EnumItem(rs_image_format[0], rs_image_format[1], u""))

            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            plugins = [maya_constants.RENDERERS.MAYA_REDSHIFT_RENDERER.get_commandline_name()]
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAYA,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=True)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        cps = self.get_plugin().get_active_command_port_server()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            index = maya_cpc.get_rs_image_format(cps)

            if index not in maya_constants.REDSHIFT_IMAGE_FORMATS:
                index = 0

            return maya_constants.REDSHIFT_IMAGE_FORMATS[index][0]

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_REDSHIFT_RENDERER.get_commandline_name()

        raise ApplicationAttributeNotFound(app, version, key)


class RendermanAttributeSource(ThirdPartyRendererAttributeSource):
    """ A simple attribute source for rman specific settings """

    def __init__(self, plugin):
        ThirdPartyRendererAttributeSource.__init__(self, plugin)

    def setup(self):
        ThirdPartyRendererAttributeSource.setup(self)

    def get_enum_items(self, app, version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            rman_img_exts = maya_cpc.get_prman_image_exts(cps)
            items = list(map(lambda x: EnumItem(x, x, u""), rman_img_exts))
            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            plugins = [maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_commandline_name()]
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAYA,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                plugins=plugins,
                                                                use_gpu_default=False)

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, version, key):
        cps = self.get_plugin().get_active_command_port_server()

        if key == api_constants.API_KEYS.RENDERER:
            return maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_commandline_name()

        elif key == api_constants.API_KEYS.OUTPUT_FORMAT:
            return maya_cpc.get_prman_output_format(cps)

        raise ApplicationAttributeNotFound(app, version, key)


class MayaAttributeSource(gm_meta_plugin.ApplicationAttributeSource):

    # A regular expression that detects image formats of the form
    # "<image format display name> (<image format command line name>)"
    # so that we can extract the value within the brackets
    bracketed_image_format_regex = re.compile("(.*\(.*\))")

    def __init__(self, plugin):
        gm_meta_plugin.ApplicationAttributeSource.__init__(self, plugin)

        self._arnold_as = ArnoldAttributeSource(plugin)
        self._vray_as = VrayAttributeSource(plugin)
        self._redshift_as = RedshiftAttributeSource(plugin)
        self._renderman_as = RendermanAttributeSource(plugin)

        # create a dictionary of attribute sources for the different renderers.
        self._renderer_attribute_source_dict = {
            maya_constants.RENDERERS.MAYA_ARNOLD_RENDERER.get_internal_name(): self._arnold_as,
            maya_constants.RENDERERS.MAYA_VRAY_RENDERER.get_internal_name(): self._vray_as,
            maya_constants.RENDERERS.MAYA_REDSHIFT_RENDERER.get_internal_name(): self._redshift_as,
            maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_internal_name(): self._renderman_as
        }

    def get_enum_items(self, app, version, key):
        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()

        current_renderer = self.get_attribute_value(app, version, api_constants.API_KEYS.RENDERER)

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            renderer_attribute_source = self._renderer_attribute_source_dict[current_renderer]

            if not renderer_attribute_source.is_setup():
                renderer_attribute_source.setup()

            try:
                return renderer_attribute_source.get_enum_items(app, version, key)
            except ApplicationAttributeNotFound as e:
                pass
        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            image_formats = list(maya_constants.IMAGE_FORMATS.values())
            image_formats.sort(key=lambda x: x[1], reverse=True)
            items = list(map(lambda x: EnumItem(x[0], x[1], ''), image_formats))
            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.APP_VERSION:
            items = []

            current_maya_version = maya_cpc.get_maya_version(cps)
            items.append(EnumItem(current_maya_version, current_maya_version, u""))

            return items, 0

        if key == api_constants.API_KEYS.CAMERA:
            camera_options = maya_cpc.get_camera_options(cps)
            items = list(map(lambda x: EnumItem(x, x, u""), camera_options))
            return items, self._get_enum_index(app, version, key, items)

        if key == api_constants.API_KEYS.MACHINE_TYPE:
            return plugin.get_api_client().get_machines_options(api_constants.PRODUCTS.MAYA,
                                                                api_constants.OPERATION_TYPES.RENDER,
                                                                use_gpu_default=False)

        if key == api_constants.API_KEYS.RENDER_LAYERS:
            render_layers = maya_cpc.get_render_layers(cps)
            items = list(map(lambda x: EnumItem(x['render_layer_name'], x['render_layer_name'], u""), render_layers))
            return items, 0

        raise ApplicationAttributeNotFound(app, version, key)

    def get_attribute_value(self, app, app_version, key):

        cps = self.get_plugin().get_active_command_port_server()

        current_renderer = maya_cpc.get_current_renderer(cps)

        # if the request can be fulfilled by one of the renderer attribute sources then use that attribute source
        # instead
        try:
            renderer_attribute_source = self._renderer_attribute_source_dict[current_renderer]
            try:
                return renderer_attribute_source.get_attribute_value(app, app_version, key)
            except ApplicationAttributeNotFound as e:
                # if the renderer's attribute source exists but does not support this attribute key pass
                pass
        except KeyError as e:
            # if the renderer does not match a renderer attribute source in the attribute source dictionary then pass
            pass

        if key == api_constants.API_KEYS.IS_SAVED:
            return maya_cpc.get_maya_scene_name(cps) and maya_cpc.does_scene_exist(cps)

        if key == api_constants.API_KEYS.PROJECT_NAME:
            return pathlib.Path(maya_cpc.get_project_directory(cps)).stem

        if key == api_constants.API_KEYS.SUBMISSION_NAME:
            scene_name = pathlib.Path(maya_cpc.get_maya_scene_name(cps)).name

            if not scene_name:
                return "Untitled"

            return scene_name

        if key == api_constants.API_KEYS.JOB_NAME:
            job_name_parts = []

            camera_name = self.get_attribute_value(app, app_version, api_constants.API_KEYS.CAMERA)
            if camera_name:
                job_name_parts.append(camera_name)

            if current_renderer:
                job_name_parts.append(current_renderer.capitalize())

            return ' '.join(list(map(lambda x: "[{}]".format(x), job_name_parts)))

        if key == api_constants.API_KEYS.APP:
            return api_constants.PRODUCTS.MAYA

        if key == api_constants.API_KEYS.APP_VERSION:
            return maya_cpc.get_maya_version(cps)

        if key == api_constants.API_KEYS.SCENE_FILE_PATH_UNUSED:
            return maya_cpc.get_maya_scene_name(cps)

        if key == api_constants.API_KEYS.RENDERER:

            if current_renderer == maya_constants.RENDERERS.MAYA_SOFTWARE_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_SOFTWARE_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_HARDWARE_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_HARDWARE_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_HARDWARE_2_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_HARDWARE_2_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_VECTOR_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_VECTOR_RENDERER.get_commandline_name()

            elif current_renderer == maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_internal_name():
                return maya_constants.RENDERERS.MAYA_RENDERMAN_RENDERER.get_commandline_name()

            else:
                return maya_constants.UNSUPPORTED_RENDERER + u" ({0})".format(current_renderer)

        if key == api_constants.API_KEYS.OUTPUT_FORMAT:
            image_format = maya_cpc.get_output_format(cps)

            if image_format in maya_constants.IMAGE_FORMATS:
                return maya_constants.IMAGE_FORMATS.get(image_format)[0]

            return 'png'

        if key == api_constants.API_KEYS.FRAMES:
            return maya_cpc.get_frame_range_str(cps)

        if key == api_constants.API_KEYS.CAMERA:
            # return the first camera in the list camera list
            # Maya by default always has at least four cameras, persp, top, front & side which can not be deleted
            return maya_cpc.get_default_camera(cps)

        if key == api_constants.API_KEYS.OUTPUT_WIDTH:
            return maya_cpc.get_output_width(cps)

        if key == api_constants.API_KEYS.OUTPUT_HEIGHT:
            return maya_cpc.get_output_height(cps)

        if key == api_constants.API_KEYS.OUTPUT_PATH:
            return maya_cpc.get_images_path(cps)

        if key == api_constants.API_KEYS.OUTPUT_PREFIX:
            output_prefix = maya_cpc.get_output_prefix(cps)

            if not output_prefix:
                output_prefix = '<Scene>'

            return output_prefix

        if key == api_constants.API_KEYS.RENDER_LAYERS:
            render_layers = maya_cpc.get_render_layers(cps)
            result = []
            for render_layer in render_layers:
                if render_layer['is_renderable']:
                    result.append(render_layer['render_layer_name'])
            return result

        raise ApplicationAttributeNotFound(app, app_version, key)
