import typing

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import CommandPortServer


def create_render_settings_menu(cps: 'CommandPortServer'):
    return cps.eval("""
def gm_process_idle_events(counter_limit=20):
    # This list contains all the current events/commands in Maya's idle queue
    queued_cmds = maya.cmds.evalDeferred(list=True)

    # The counter here is for safety
    counter = 0

    # Idle events might spawn more idle events so we need to loop until the list is empty
    while queued_cmds:

        # if a counter limit is provided and counter has reached the limit then break
        if counter_limit is not None and counter > counter_limit:
            break

        maya.utils.processIdleEvents()
        queued_cmds = maya.cmds.evalDeferred(list=True)
        counter = counter + 1

if not maya.cmds.optionMenuGrp("imageMenuMayaSW", exists=True):
    # open the maya renderer settings window to initialise it
    maya.mel.eval("unifiedRenderGlobalsWindow;")

    # process idle events to give the render settings window a chance to load. Not required but is good practice
    # Remember the plug-in may be set to auto-load so there will be alot of other deferred events at startup.
    gm_process_idle_events(counter_limit=300)

    # hide the render settings window
    #maya.mel.eval("workspaceControl - e - vis 0 unifiedRenderGlobalsWindow;")
""")


def get_current_renderer(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
renderer = maya.cmds.getAttr("defaultRenderGlobals.currentRenderer")
if renderer == 'renderman':
    renderer = 'rman'
""", return_variable_name='renderer')


def get_project_directory(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
project_dir = maya.cmds.workspace(query=True, rd=True)
""", return_variable_name='project_dir')


def get_maya_version(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
maya_version = maya.cmds.about(version=True)
""", return_variable_name='maya_version')


def get_maya_short_scene_name(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
scene_name = maya.cmds.file(query=True, sceneName=True, shortName=True)
""", return_variable_name='scene_name')


def get_output_format(cps: 'CommandPortServer') -> int:
    return cps.eval_cmd("""
output_format = maya.cmds.getAttr("defaultRenderGlobals.imageFormat")
""", return_variable_name='output_format')


def get_frame_range_str(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
frames = None
start_frame = int(maya.cmds.getAttr("defaultRenderGlobals.startFrame"))
end_frame = int(maya.cmds.getAttr("defaultRenderGlobals.endFrame"))
frame_step = int(maya.cmds.getAttr("defaultRenderGlobals.byFrame"))
frames = "{0} {1} {2}".format(start_frame, end_frame, frame_step)
""", return_variable_name='frames')


def get_default_camera(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
cameras = maya.cmds.listCameras()
default_camera = cameras[0] if cameras else ''
""", return_variable_name='default_camera')


def get_output_width(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
output_width = maya.cmds.getAttr("defaultResolution.width")
""", return_variable_name='output_width')


def get_output_height(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
output_height = maya.cmds.getAttr("defaultResolution.height")
""", return_variable_name='output_height')


def get_output_prefix(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
output_prefix = maya.cmds.getAttr('defaultRenderGlobals.imageFilePrefix')
""", return_variable_name='output_prefix')


def get_images_path(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
images_path = maya.cmds.workspace(fileRuleEntry="images")
""", return_variable_name='images_path')


def get_camera_options(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
camera_options = maya.cmds.listCameras()
""", return_variable_name='camera_options')


def get_output_format_labels(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
labels = []
menu_items = maya.cmds.optionMenuGrp("imageMenuMayaSW", query=True, itemListShort=True)

for menu_item in menu_items:

    # if the menu_item is a divider then skip it
    if maya.cmds.menuItem(menu_item, query=True, divider=True):
        continue

    label = maya.cmds.menuItem(menu_item, query=True, label=True)
    labels.append(label)
""", return_variable_name='labels')


def get_maya_scene_name(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
scene_name = maya.cmds.file(query=True, sceneName=True)
""", return_variable_name='scene_name')


def get_render_layers(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
render_layers_serialised = []

if maya.cmds.mayaHasRenderSetup():
    render_setup = maya.app.renderSetup.model.renderSetup.instance()
    render_layers = render_setup.getRenderLayers()
    
    # Add the default render layer
    default_render_layer = render_setup.getDefaultRenderLayer()
    render_layers_serialised.append({"render_layer_name": default_render_layer.name(), 
        "is_renderable": default_render_layer.isRenderable()})
    
    # Add the additional render layers
    for render_layer in render_layers:
        # Without "rs_" prefix
        render_layer_name = render_layer.name()
        is_renderable = render_layer.isRenderable()
        render_layers_serialised.append({"render_layer_name": render_layer_name, "is_renderable": is_renderable})
else:
    render_layers = maya.cmds.ls(type='renderLayer')
    
    # Move the default render layer to the begining of the list
    default_render_layer = render_layers.pop()
    render_layers.insert(0, default_render_layer)
    
    for render_layer in render_layers:
        try:
            is_renderable = maya.cmds.getAttr('{}.renderable'.format(render_layer))
        except ValueError:
            is_renderable = False
        
        render_layers_serialised.append({"render_layer_name": render_layer, "is_renderable": is_renderable})

""", return_variable_name='render_layers_serialised')


def does_scene_exist(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
scene_exists = maya.cmds.file(query=True, exists=True)
""", return_variable_name='scene_exists')


def has_scene_been_modified(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
any_modified = maya.cmds.file(query=True, anyModified=True)
""", return_variable_name='any_modified')


def save_scene(cps: 'CommandPortServer'):
    return cps.eval("""
maya.cmds.file(force=True, save=True)   
""")


def get_reference_nodes(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
reference_nodes = maya.cmds.ls(type='reference')
""", return_variable_name='reference_nodes')


def is_reference_loaded(cps: 'CommandPortServer', reference_node: str):
    return cps.eval_cmd("""
is_reference_loaded = maya.cmds.referenceQuery('{}', isLoaded=True)
""".format(reference_node), return_variable_name='is_reference_loaded')


def get_file_list(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
file_list = maya.cmds.file(query=True, list=True, withoutCopyNumber=True)   
""", return_variable_name='file_list')


def get_workspace_fullname(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
workspace_fullname = maya.cmds.workspace(query=True, fullName=True)
""", return_variable_name='workspace_fullname')


def get_directories_list(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
directories_list = maya.cmds.filePathEditor(query=True, listDirectories='', unresolved=True)
""", return_variable_name='directories_list')


def get_file_list_with_attr(cps: 'CommandPortServer', directory: str):
    directory = directory.replace('\\', '\\\\')

    return cps.eval_cmd("""
file_list_with_attr = maya.cmds.filePathEditor(query=True, listFiles='{}', withAttribute=True)
""".format(directory), return_variable_name='file_list_with_attr')


def get_arnold_render_device(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
arnold_render_device = maya.cmds.getAttr("defaultArnoldRenderOptions.renderDevice")
""", return_variable_name='arnold_render_device')


def get_arnold_output_format(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
# The nodes holding Arnold's settings are created at runtime by mtoa. Before we access any of mtoa's settings they must 
# be initialised by calling the mtoa.core.createOptions() method. This will do nothing if the options already exist.
import mtoa
mtoa.core.createOptions()

arnold_output_format = maya.cmds.getAttr('defaultArnoldDriver.ai_translator')
""", return_variable_name='arnold_output_format')

def get_arnold_translator_names(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import mtoa
translator_names = []
for transName, arnoldNode in mtoa.core.listTranslators("aiAOVDriver"):
    translator_names.append(transName)
""", return_variable_name='translator_names')


def get_arnold_texture_files(cps: 'CommandPortServer') -> typing.List[str]:
    return cps.eval_cmd("""
import mtoa
mtoa.core.createOptions()

mtoa_version = tuple(map(lambda x: int(x), maya.cmds.pluginInfo('mtoa', query=True, version=True ).split('.')))

arnold_texture_files = []

if mtoa_version >= (4,1,0):
    scanned_files = mtoa.txManager.lib.get_scanned_files(mtoa.txManager.lib.scene_default_texture_scan)
    arnold_texture_files = list(map(lambda x: x.get('txpath', ''), scanned_files.values()))
else:
    texture_files = []
    files_count = []
    
    #total files
    files_count.append(0)
    #missing files
    files_count.append(0)
    
    mtoa.txManager.GetTxList(texture_files, files_count)
    
    use_existing_tx_textures = maya.cmds.getAttr('defaultArnoldRenderOptions.use_existing_tiled_textures')
    if use_existing_tx_textures:
        for texture_file in texture_files:
            full_path_arnold_texture = texture_file[0]
            arnold_texture_files.append(full_path_arnold_texture)
            
            arnold_tx_file = os.path.splitext(full_path_arnold_texture)[0]+".tx"
    
            if os.path.isfile(arnold_tx_file):
                arnold_texture_files.append(arnold_tx_file)

# Filter any None values, possible if the texture does not have a txpath
arnold_texture_files = list(filter(None, arnold_texture_files))

""", return_variable_name='arnold_texture_files')


def get_vray_output_prefix(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
maya.mel.eval('vrayCreateVRaySettingsNode')
vray_output_prefix = maya.cmds.getAttr("vraySettings.fileNamePrefix")
""", return_variable_name='vray_output_prefix')


def get_vray_production_engine(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
maya.mel.eval('vrayCreateVRaySettingsNode')
vray_production_engine = maya.cmds.getAttr("vraySettings.productionEngine")
""", return_variable_name='vray_production_engine')


def get_vray_output_format_label(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
maya.mel.eval('vrayCreateVRaySettingsNode')
vray_output_format_label = maya.cmds.getAttr('vraySettings.imageFormatStr')
""", return_variable_name='vray_output_format_label')


def get_vray_scene_name_ascii(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
from vray import vray_cloud_rendering
vray_scene_name_ascii = vray_cloud_rendering.vrayGetSceneNameASCII()
""", return_variable_name='vray_scene_name_ascii')


def archive_vrscene_file(cps: 'CommandPortServer', output_dir: str, camera: str, zip: bool = False):
    output_dir = output_dir.replace('\\', '\\\\')
    camera = camera.replace('\\', '\\\\')
    return cps.eval("""
from vray import vray_cloud_rendering
vray_cloud_rendering.vrayArchiveVrsceneCmd('{}', '{}', {})
""".format(output_dir, camera, zip))


def get_rs_image_format_labels(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
rs_image_format_labels = maya.mel.eval("global string $g_redshiftImageFormatLabels[];$temp_var = $g_redshiftImageFormatLabels;")
""", return_variable_name='rs_image_format_labels')


def get_rs_image_format_values(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
rs_image_format_values = maya.mel.eval("global string $g_redshiftImageFormatLabels[];$temp_var = $g_redshiftImageFormatExtensions;")
""", return_variable_name='rs_image_format_values')


def get_rs_image_format(cps: 'CommandPortServer') -> int:
    return cps.eval_cmd("""    
try:
    maya.cmds.listAttr("redshiftOptions")
except ValueError as e:
    maya.cmds.createNode('RedshiftOptions')

rs_image_format = maya.cmds.getAttr("redshiftOptions.imageFormat")
""", return_variable_name='rs_image_format')


def get_prman_image_exts(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import rfm2
rman_img_exts = rfm2.constants.IMG_EXTS
""", return_variable_name='rman_img_exts')


def get_prman_output_format(cps: 'CommandPortServer'):
    return cps.eval_cmd("""
import rfm2
rman_output_format = None

# see Pixar\\RenderManForMaya-23.2\\scripts\\rfm2\\api\\strings.py set_display
driver = None
dspy_node = rfm2.config.DEFAULT_DISPLAY_NODE
dspy_node = dspy_node.replace('_variance', '')
if maya.cmds.objExists(dspy_node):
    # connectionInfo() was failing: use listConnections() instead !
    try:
        tmp = cmds.listConnections('%s.displayType' % dspy_node)
    except:
        pass
    else:
        driver = None if not tmp else tmp[0]
if driver:
    rman_output_format = rfm2.config.cfg().displayDriverExtensions[maya.cmds.nodeType(driver)]
else:
    rman_output_format = 'exr'
""", return_variable_name='rman_output_format')

