__all__ = 'MayaPackedSceneBuilder'

import pathlib
import os
import typing

import gm_meta_plugin

from gm_meta_plugin.errors import FilePackingError, ProjectRequiresSavingError
from gm_meta_plugin.gridmarkets import constants as api_constants
from . import maya_constants, maya_cpc

if typing.TYPE_CHECKING:
    from gm_meta_plugin.job_preset import JobPreset


class MayaPackedSceneBuilder(gm_meta_plugin.PackedSceneBuilder):

    def __init__(self, plugin):
        self._plugin = plugin

    def get_plugin(self):
        return self._plugin

    def _build_vray_scene_internal(self,
                                   project_name: str,
                                   product: str,
                                   product_version: str,
                                   file_name: str,
                                   root_dir: pathlib.Path,
                                   attributes: typing.Dict[str, any],
                                   job_preset: 'JobPreset'):
        cps = self.get_plugin().get_active_command_port_server()

        output_prefix = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.OUTPUT_PREFIX).eval()

        if '<' in output_prefix or '>' in output_prefix:
            raise FilePackingError("V-Ray standalone does not support Maya keywords."
                                   "Please remove all Maya keywords (<Scene>, <Layer>, etc) from your output prefix.")

        output_dir = root_dir / maya_constants.VRAY_OUTPUT_DIR_NAME

        # Get the camera name
        camera = job_preset.get_job_preset_attribute_with_job_attribute_id(api_constants.API_KEYS.CAMERA).eval()

        # Output the current maya scene as a .vrscene file using V-Ray's own api logic
        maya_cpc.archive_vrscene_file(cps, str(output_dir), camera)

        # Wait until no files in the output directory have been modified for at least 2 seconds. This is required
        # because V-rays archiving logic happens in a separarate thread with no callback so we don't know when it has
        # Finished.
        import time
        start = time.time()
        max_wait_time = 40
        most_recent_modification = start
        project_files = []
        while time.time() - start < max_wait_time:
            # Get all files in the vray output directory
            project_files = gm_meta_plugin.utils.get_files_in_directory(output_dir)

            for project_file in project_files:
                file_modification_time = project_file.stat().st_mtime
                if file_modification_time > most_recent_modification:
                    most_recent_modification = file_modification_time

            if time.time() - most_recent_modification > 2:
                break

            time.sleep(0.5)

        # Get the vray scene name
        vray_scene_name_ascii: str = maya_cpc.get_vray_scene_name_ascii(cps)

        main_vrscene_file_path = output_dir / vray_scene_name_ascii / (vray_scene_name_ascii.lower() + '.vrscene')

        # set the path attribute
        attributes[api_constants.API_KEYS.PATH] = main_vrscene_file_path

        packed_project = gm_meta_plugin.PackedProject(project_name, root_dir, project_files, attributes)
        return packed_project

    def _build_scene_internal(self,
                              project_name,
                              product,
                              product_version,
                              save_project_if_required: bool,
                              attributes,
                              job_preset: typing.Optional['JobPreset'] = None):

        if job_preset is None:
            raise ValueError("No job preset passed to project packer")

        plugin = self.get_plugin()
        cps = plugin.get_active_command_port_server()
        logger = self.get_plugin().get_logging_coordinator().get_logger()

        fileName = maya_cpc.get_maya_scene_name(cps)

        # If the scene has not been saved
        if fileName == "":
            raise FilePackingError(u"Scene file not saved.")

        # If the scene has been created, saved and then erased from the disk
        if maya_cpc.does_scene_exist(cps) == 0:
            raise FilePackingError(u"Scene file does not exist on filesystem.")

        # If the scene has been modified then save the scene
        if maya_cpc.has_scene_been_modified(cps):
            if save_project_if_required:
                maya_cpc.save_scene(cps)
            else:

                def _get_current_scene():
                    return self.get_current_scene(project_name, product, product_version, True, attributes,
                                               job_preset=job_preset), job_preset

                raise ProjectRequiresSavingError(_get_current_scene, u"Scene file has been modified.")

        # get the project root dir
        root_dir = pathlib.Path(maya_cpc.get_project_directory(cps))

        # Check if we should create .vrscene files instead
        vray_standalone_job_preset_attribute = job_preset.get_job_preset_attribute_with_job_attribute_id(api_constants.API_KEYS.USE_VRAY_STANDALONE)
        if vray_standalone_job_preset_attribute.get_value() and vray_standalone_job_preset_attribute.is_active():
            return self._build_vray_scene_internal(project_name, product, product_version, fileName, root_dir, attributes, job_preset)

        # get a list of all the files associated with the scene
        project_files = self.get_scene_files()

        # Find all referenced files of the current scene. For any unloaded reference, load them first, get file list
        # that should be archived and then restore its unloaded status. Use ls() instead of file(q=True, r=True) because
        # it also gets nested references.
        reference_nodes = maya_cpc.get_reference_nodes(cps)
        for ref_node in reference_nodes:
            if ref_node.find('sharedReferenceNode') == -1:
                try:
                    is_loaded = maya_cpc.is_reference_loaded(cps, ref_node)
                except Exception:
                    # Maya has a bug where some file references exist but don't reference anything. The throw an
                    # exception if you try to query them along the lines of.
                    # "Reference node '<name>' is not associated with a reference file."
                    continue

                # Skip unloaded references
                if not is_loaded:
                    continue

                # Get all external files related to this reference
                files_of_this_ref = maya_cpc.get_file_list(cps)

                for file_of_this_ref in files_of_this_ref:
                    project_files.append(file_of_this_ref)

        # Get arnold specific files
        current_renderer = job_preset.get_job_preset_attribute_by_key(api_constants.API_KEYS.RENDERER).get_value()
        if current_renderer == maya_constants.RENDERERS.MAYA_ARNOLD_RENDERER.get_internal_name():
            arnold_texture_files = maya_cpc.get_arnold_texture_files(cps)
            project_files.extend(arnold_texture_files)

        # add the project workspace.mel file
        workspacePath = maya_cpc.get_workspace_fullname(cps) + '/workspace.mel'
        project_files.append(workspacePath)

        # IO boundary check. Encode file paths based on system file encoding
        project_files = list(map(lambda x: str(pathlib.Path(x)), project_files))

        # remove empty file references (e.g. "")
        project_files = list(filter(lambda x: bool(x), project_files))

        # convert to a set of pathlib.Path objects
        project_files = set(map(lambda f: pathlib.Path(f), project_files))

        # if a file is not under the root directory then raise an error
        external_files = []
        for project_file in list(project_files):
            if root_dir not in project_file.parents:
                external_files.append(project_file)

        if external_files:
            for external_file in external_files:
                logger.error("External file - '{}'".format(str(external_file)))

            msg = u"Warning - Your scene contains external files.\n" \
                  u"Please move the following files into your project directory:\n"

            counter = 0
            while external_files and counter < 5:
                external_file = external_files.pop()
                counter = counter + 1

                msg = msg + u"   " + str(external_file) + "\n"

            if external_files:
                msg = msg + u"   ({0} files not shown)\n".format(len(external_files))

            msg = msg + u"\nYour project directory is:\n   {0}\n\n" \
                        u"Either move the file into your Maya project folder or set your Maya project to a different " \
                        u"location using 'File -> Set Project...'".format(root_dir)

            raise FilePackingError(msg)

        # set the path attribute
        attributes[api_constants.API_KEYS.PATH] = fileName

        packed_project = gm_meta_plugin.PackedProject(project_name, root_dir, project_files, attributes)

        return packed_project

    def get_scene_files(self):
        cps = self.get_plugin().get_active_command_port_server()

        # Get all directories detected by the file path editor
        directories = maya_cpc.get_directories_list(cps)

        # Directories can be none if there are no file paths in the editor so make it an empty list in that case
        directories = directories if directories else []
        files = []

        # Iterate through each directory and get the files for that directory that are used in the scene
        for directory in directories:
            list_file_elem = maya_cpc.get_file_list_with_attr(cps, directory)

            # combine every other item to get a list of types of the form (file name, type)
            it = iter(list_file_elem)
            list_tuple_file = zip(it, it)

            for (file_name, file_type) in list_tuple_file:
                files.append(os.path.normpath(os.path.join(directory, file_name)))

        def _filter_function(file_path):
            try:
                return pathlib.Path(file_path).is_file()
            except WindowsError as e:  # If the path is not a valid windows path then this is thrown
                return False

        # Union the files list with the list of files returned by maya.cmds.file
        files = files + maya_cpc.get_file_list(cps)

        # Filter any files which don't exist from the file path editor files
        files = list(filter(_filter_function, files))

        return files

    def get_project_type_id(self):
        return api_constants.PROJECT_TYPE_IDS.MAYA_PROJECT

    def get_product_type(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, api_constants.API_KEYS.APP)

    def get_product_version(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, api_constants.API_KEYS.APP_VERSION)
