__all__ = 'PackedSceneBuilder'

import typing
import pathlib

import gm_meta_plugin
from gm_meta_plugin.errors import FilePackingError, ProjectRequiresSavingError
from gm_meta_plugin.gridmarkets import constants as api_constants
from . import nuke_cpc

if typing.TYPE_CHECKING:
    from submit2gm.submit2gm_plugin import CommandPortServer
    from gm_meta_plugin import PackedProject
    from gm_qt_plugin.widgets.job_preset import JobPreset



class NukePackedSceneBuilder(gm_meta_plugin.PackedSceneBuilder):

    def __init__(self, plugin):
        self._plugin = plugin

    def get_plugin(self):
        return self._plugin

    def _build_scene_internal(self,
                              project_name: str,
                              product: str,
                              product_version: str,
                              save_project_if_required: bool,
                              attributes: typing.Dict[str, any],
                              job_preset: typing.Optional['JobPreset'] = None):

        cps: 'CommandPortServer' = self.get_plugin().get_active_command_port_server()

        # get the file name of the script
        file_name = pathlib.Path(nuke_cpc.get_script_name(cps))

        if not file_name:
            raise FilePackingError("Nuke script not saved. Please save and retry.")

        # If the scene has been created, saved and then erased from the disk
        if not file_name.exists():
            raise FilePackingError("Script file '" + str(file_name) + "' does not exist on filesystem.")

        # Save the script if the script has been modified
        if nuke_cpc.is_script_dirty(cps):
            if save_project_if_required:
                nuke_cpc.save_script(cps)
            else:

                def _get_current_scene():
                    return self.get_current_scene(project_name, product, product_version, True, attributes,
                                               job_preset=job_preset), job_preset

                raise ProjectRequiresSavingError(_get_current_scene, u"Scene file has been modified.")

        # get the project root dir
        root_dir = pathlib.Path(nuke_cpc.get_project_dir(cps))

        read_and_write_paths = nuke_cpc.get_read_and_write_paths(cps)
        read_paths = read_and_write_paths['nuke_read_paths']
        write_paths = read_and_write_paths['nuke_write_paths']

        # Convert to pathlib paths
        read_paths = list(map(lambda x: pathlib.Path(x), read_paths))
        write_paths = list(map(lambda x: pathlib.Path(x), write_paths))

        project_files = [file_name] + read_paths

        # set the path attribute
        attributes[api_constants.API_KEYS.PATH] = file_name

        # remap project location and write paths
        remote_project_root = pathlib.Path('/data/input/{0}'.format(project_name))

        # remap the scripts project directory
        remapped_paths = {root_dir.as_posix(): remote_project_root.as_posix()}

        # if a file is not under the root directory then raise an error
        external_files = []
        for project_file in list(project_files):
            if root_dir not in project_file.parents:
                external_files.append(project_file)

        # Remap the drive any external files are on
        external_file_drives = set()

        # Get all the drives which contain external files
        for external_file in external_files:
            if external_file.drive:
                external_file_drives.add(external_file.drive)
        for external_file_drive in external_file_drives:
            remapped_remote_drive = (remote_project_root / api_constants.EXTERNAL_FILES_DIR_NAME) / external_file_drive.replace(':', '', 1)
            remapped_paths[pathlib.Path(external_file_drive).as_posix()] = remapped_remote_drive.as_posix()

        # Check if create_directories is set to true for all write nodes
        write_nodes_without_create_dir_enabled = nuke_cpc.get_write_nodes_without_create_dir_enabled(cps)
        if write_nodes_without_create_dir_enabled:
            raise FilePackingError("Write node '{0}' does not have 'create directories' enabled.\n"
                                   "This is required otherwise the node may not be able to write it's results on our "
                                   "system.\n\nPlease enable 'create directories' for ""'{0}' and any other write "
                                   "nodes".format(write_nodes_without_create_dir_enabled[0]))

        # remap all absolute write paths
        for write_path in write_paths:
            skip_remap = False

            # convert relative paths to absolute paths
            if not write_path.is_absolute():
                # It is not possible to remap relative paths in Nuke so skip the remapping stage. This is only an issue
                # for relative files which are outside the project directory
                skip_remap = True

                # make path absolute
                write_path = root_dir / write_path

            # if the write path is under the root directory
            if root_dir in write_path.parents:
                relative_write_path = write_path.relative_to(root_dir)
                is_internal_file = True
            else:
                # if the write path is an external file remap to an external files directory under the remote
                # project directory
                write_path_str = str(write_path)
                write_path_str = write_path_str.replace(':', '') # replace ':' character used in drive name on Windows

                # if the write path is a posix absolute path
                if write_path_str.startswith("/"):
                    # remove the first '/' to make the path relative
                    write_path_str = write_path_str[1:]

                relative_write_path = pathlib.Path(api_constants.EXTERNAL_FILES_DIR_NAME) / pathlib.Path(write_path_str)
                is_internal_file = False

            remapped_path = remote_project_root / relative_write_path

            if skip_remap:
                pass
            else:
                # Don't remap internal write files as they will already be remapped
                if not is_internal_file:
                    # Remap the output directory not the file itself
                    output_dir_str = str(write_path.parent.as_posix())

                    if output_dir_str not in remapped_paths:
                        remapped_paths[output_dir_str] = str(remapped_path.parent.as_posix())

        # Convert the remap dictionary to the format nuke expects (comma separated string)
        remapped_paths_list = []
        for local_path, remote_path in remapped_paths.items():
            remapped_paths_list.append("{},{}".format(local_path, remote_path))
        remap_arg = '"' + ",".join(remapped_paths_list) + '"'

        attributes[api_constants.API_KEYS.REMAP] = remap_arg

        packed_project = gm_meta_plugin.PackedProject(project_name, root_dir, project_files, attributes)

        return packed_project

    def get_project_type_id(self):
        return api_constants.PROJECT_TYPE_IDS.NUKE_PROJECT

    def get_product_type(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, api_constants.API_KEYS.APP)

    def get_product_version(self):
        application_attribute_source = self.get_plugin().get_application_attribute_source()
        return application_attribute_source.get_attribute_value(None, None, api_constants.API_KEYS.APP_VERSION)
