import gm_meta_plugin

from gm_qt_plugin import QtPlugin, QtPreferencesContainer, QtAPISchemaFactory, QtProjectAttributeInstanceFactory, \
    QtProjectInstanceBuilderFactory
from gm_qt_plugin.widgets import JobPresetContainer, LogHistoryContainer, RemoteProjectContainer, JobPreset

from .api_client import APIClient
from .plugin_utils import PluginUtils
from .submit2gm_attribute_source import Submit2gmAttributeSource
from . import constants
from .version import VERSION
from .submit2gm_packed_scene_builder import Submit2GMPackedSceneBuilder
from .sumbit2gm_user_interface import Submit2GMUserInterface
from .submit2gm_job_preset_validator import Submit2gmJobPresetValidator
from .logging_coordinator import LoggingCoordinator

import json
import requests
import typing


class CommandPortServer:

    def __init__(self, product: str, address: str, meta_info: typing.Optional[str] = None):
        self._product = product
        self._address = address
        self._session = requests.Session()

        if meta_info:
            self._meta_info = meta_info
        else:
            self._meta_info = ""

    def get_product(self) -> str:
        return self._product

    def get_address(self) -> str:
        return self._address

    def get_meta_info(self) -> str:
        return self._meta_info

    def is_still_running(self) -> bool:
        from urllib.parse import urlparse
        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        url = urlparse(self._address)
        netloc = url.netloc
        address, port = netloc.split(':')

        can_bind_to_socket = True

        try:
            s.bind((address, int(port)))
        except socket.error as e:
            can_bind_to_socket = False
        finally:
            try:
                s.close()
            except Exception:
                pass

        is_cps_still_running = not can_bind_to_socket
        return is_cps_still_running

    def eval(self, code: str, return_variables: typing.List[str] = None) -> typing.Dict['str', 'any']:

        if return_variables is None:
            return_variables = []

        url = '{}/eval'.format(self.get_address())
        payload = code.encode('utf8')
        headers = {
            'Content-Type': 'text/plain',
            'Return-Variables': json.dumps(return_variables)
        }

        try:
            response = self._session.post(url,
                                          data=payload,
                                          headers=headers,
                                          timeout=constants.COMMAND_PORT_SERVER_TIMEOUT)

        except requests.exceptions.ConnectTimeout as e:
            raise gm_meta_plugin.errors.CommandPortServerError(
                "Connection error to GridMarkets {} plug-in. Connect timeout.".format(self._product))
        except requests.exceptions.ReadTimeout as e:
            raise gm_meta_plugin.errors.CommandPortServerError(
                "Connection error to GridMarkets {} plug-in. Read timeout.".format(self._product))
        except Exception as e:
            raise gm_meta_plugin.errors.CommandPortServerError(
                "Connection error to GridMarkets {} plug-in.\n{}\n{}".format(self._product, str(type(e)), str(e)))

        binary_data = response.content

        # Keep alive tokens are sent as part of the response so that the timeout can tell if the cps has crashed or if
        # it is just taking a long time
        binary_data = binary_data.strip(constants.COMMAND_PORT_SERVER_TIMEOUT_KEEP_ALIVE_TOKEN)

        try:
            utf8_data = binary_data.decode('utf8')
        except UnicodeDecodeError as e:
            raise gm_meta_plugin.errors.CommandPortServerError("Unable to decode response: {}".format(str(e)))

        try:
            json_result = json.loads(utf8_data)
        except json.decoder.JSONDecodeError as e:
            # If the response is not valid json then there is a problem
            raise gm_meta_plugin.errors.CommandPortServerError(
                "Invalid response from command port server: JSON error\n" + response.text)

        if 'error' in json_result:
            exc_msg = json_result['error']

            raise gm_meta_plugin.errors.CommandPortServerError(
                "Unable to evaluate request. Code:\n{}\n\nException:{}".format(code[0:20], exc_msg))

        return json_result

    def eval_cmd(self, command: str, return_variable_name: typing.Optional[str] = None) -> any:
        if not return_variable_name:
            return_variable_name = command.split()[0]

        results = self.eval(command, [return_variable_name])
        return results[return_variable_name]

    def shutdown(self) -> None:
        url = '{}/shutdown'.format(self.get_address())
        self._session.post(url, timeout=5)

    def close_session(self) -> None:
        self._session.close()


class Submit2gmPlugin(QtPlugin):

    def _get_factory_collection(self):
        api_schema_factory = QtAPISchemaFactory()
        project_attribute_factory = gm_meta_plugin.ProjectAttributeFactory(self)
        project_attribute_instance_factory = QtProjectAttributeInstanceFactory(self)
        project_instance_builder_factory = QtProjectInstanceBuilderFactory(self)

        return gm_meta_plugin.FactoryCollection(api_schema_factory,
                                                project_attribute_factory,
                                                project_attribute_instance_factory,
                                                project_instance_builder_factory)

    def __init__(self):
        version = VERSION

        main_window = None # QtWidgets.QMainWindow() #maya_utils.maya_main_window()

        self._version = gm_meta_plugin.PluginVersion(version[0], version[1], version[2])

        self._factory_collection = self._get_factory_collection()

        log_history_container = LogHistoryContainer(self, [])
        self._logging_coordinator = LoggingCoordinator(self, log_history_container)

        user_container = gm_meta_plugin.UserContainer(self, [])
        job_preset_container = JobPresetContainer(self, [], parent=main_window)
        self._preferences_container = QtPreferencesContainer(user_container, job_preset_container)

        self._api_client = APIClient(self, self._logging_coordinator, self._factory_collection)
        self._user_interface = Submit2GMUserInterface(self)
        self._remote_project_container = RemoteProjectContainer(self, [], parent=main_window)
        self._plugin_utils = PluginUtils()
        self._command_port_servers = []
        self._application_attribute_source = Submit2gmAttributeSource(self)
        self._packed_scene_builder = Submit2GMPackedSceneBuilder(self)
        self._job_preset_validator = Submit2gmJobPresetValidator(self)

    def get_name(self):
        return constants.PROGRAM_NAME

    def get_version(self):
        return self._version

    def get_logging_coordinator(self):
        return self._logging_coordinator

    def get_preferences_container(self):
        return self._preferences_container

    def get_api_client(self) -> 'APIClient':
        return self._api_client

    def get_user_interface(self):
        return self._user_interface

    def get_remote_project_container(self):
        return self._remote_project_container

    def add_command_port_server(self, product: str, address: str, meta_info: typing.Optional[str] = None):

        if meta_info is None:
            meta_info = ""

        cps = CommandPortServer(product, address, meta_info=meta_info)
        self._command_port_servers.append(cps)

    def get_active_command_port_server(self) -> CommandPortServer:

        if self._command_port_servers:
            return self._command_port_servers[0]

        raise ValueError("No command port servers")

    def shutdown_command_port_server(self, cps_address: str) -> CommandPortServer:

        for cps in self._command_port_servers:
            if cps.get_address() == cps_address:
                command_port_server = cps
                break
        else:
            raise ValueError("No command port server with address '{}'".format(cps_address))

        try:
            command_port_server.shutdown()
        except Exception as e:
            pass

        self._command_port_servers.remove(command_port_server)
        return command_port_server

    def shutdown_command_port_servers(self) -> None:

        for cps in self._command_port_servers:
            try:
                cps.shutdown()
            except Exception as e:
                self.get_logging_coordinator().get_logger().error(str(e))

        self._command_port_servers = []

    def get_application_attribute_source(self):
        return self._application_attribute_source

    def get_packed_scene_builder(self):
        return self._packed_scene_builder

    def get_factory_collection(self):
        return self._factory_collection

    def get_plugin_utils(self):
        return self._plugin_utils

    def create_default_job_presets(self):
        api_schema = self.get_api_client().get_cached_api_schema()
        job_preset_container = self._preferences_container.get_job_preset_container()
        job_preset_container.reset()

        project_type_id = self.get_packed_scene_builder().get_project_type_id()
        project_attribute = api_schema.get_project_definition_schema().get_project_attribute_with_id(project_type_id)
        compatible_job_definitions = project_attribute.get_compatible_job_definitions()

        if not compatible_job_definitions:
            raise ValueError("no compatible job definitions found for project type '{}'".format(project_type_id))

        job_definition = compatible_job_definitions[0]
        job_preset_container.add_job_preset_slot(job_definition.get_definition_id())

    def get_job_preset_validator(self) -> 'Submit2gmJobPresetValidator':
        return self._job_preset_validator
