__all__ = 'Submit2gmServerThread'

import logging
import json
import os
import threading
import typing
import sys

from submit2gm import constants

if typing.TYPE_CHECKING:
    import queue
    from submit2gm.worker_job import WorkerJob


class Submit2gmServerThread(threading.Thread):

    def __init__(self, job_queue: 'queue.Queue[WorkerJob]'):
        self.err = None
        self._job_queue = job_queue
        self._force_stop = False
        threading.Thread.__init__(self)

    def _pipe_worker(self, pipe):
        from submit2gm.worker_job import WorkerJob
        import ctypes
        import time

        chBuf = ctypes.create_string_buffer(constants.WIN_PIPE_SERVER_BUFSIZE)
        cbRead = ctypes.c_ulong(0)

        try:
            while True:
                current_message = b''

                ok = ctypes.windll.kernel32.ReadFile(pipe, chBuf, constants.WIN_PIPE_SERVER_BUFSIZE,
                                                           ctypes.byref(cbRead), None)

                # Break the while loop once either the pipe returns an invalid state
                if ok == 0:
                    last_error = ctypes.windll.kernel32.GetLastError()
                    last_error_msg = ctypes.FormatError(last_error)
                    logging.error("Unable to read data from pipe worker - error: {} {}".format(last_error, last_error_msg))

                    # Wait for the other process to open the other end of the pipe.
                    if last_error == constants.WIN_PIPE_ERROR_PIPE_LISTENING:
                        continue

                    break

                # Break if the client passes no data then break
                if cbRead.value == 0:
                    # todo just log for now, not sure if I need to break here (I was before)
                    logging.debug("No data read from pipe.")
                    #break

                split_message = chBuf.value.split(constants.WIN_PIPE_MESSAGE_SEPARATOR)

                for index, value in enumerate(split_message):
                    current_message = current_message + value

                    # The last item in the split message is a special case.
                    if index >= len(split_message) - 1:
                        continue

                    decoded_message = current_message.decode('utf8')

                    try:
                        argv = json.loads(decoded_message)
                        current_message = b''
                        job = WorkerJob(argv)
                        self._job_queue.put(job)

                        while not job.is_completed():
                            time.sleep(0.25)
                    except json.decoder.JSONDecodeError as e:
                        logging.error("Invalid JSON message: '{}'".format(decoded_message))

        except Exception as e:
            logging.warning("Unhandled Exception: {}".format(str(e)))
        finally:
            # Cleanup
            if not ctypes.windll.kernel32.FlushFileBuffers(pipe):
                logging.debug("Unable to flush file buffers: {}".format(
                    ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

            if not ctypes.windll.kernel32.DisconnectNamedPipe(pipe):
                logging.debug("Unable to disconnect named pipe: {}".format(
                    ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

            if not ctypes.windll.kernel32.CloseHandle(pipe):
                logging.debug("Unable to close handle: {}".format(
                    ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

        logging.debug("Windows pipe worker gracefully finished")
        return 0

    def _run_submit2gm_server_windows(self) -> None:
        import ctypes

        THREADFUNC = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int)
        thread_func = THREADFUNC(self._pipe_worker)

        # Attempt to delete the named pipe if it already exists
        try:
            if os.path.exists(constants.WIN_PIPE_NAME):
                os.remove(constants.WIN_PIPE_NAME)
        except OSError as e:
            logging.error(str(e))

        while not self._force_stop:
            logging.debug("Creating windows named pipe...")
            pipe = ctypes.windll.kernel32.CreateNamedPipeA(constants.WIN_PIPE_NAME,
                                                           constants.WIN_PIPE_PIPE_ACCESS_DUPLEX,
                                                           constants.WIN_PIPE_PIPE_TYPE_MESSAGE | constants.WIN_PIPE_PIPE_READMODE_MESSAGE | constants.WIN_PIPE_PIPE_WAIT,
                                                           constants.WIN_PIPE_PIPE_UNLIMITED_INSTANCES,
                                                           constants.WIN_PIPE_SERVER_BUFSIZE,
                                                           constants.WIN_PIPE_SERVER_BUFSIZE,
                                                           constants.WIN_PIPE_NMPWAIT_USE_DEFAULT_WAIT,
                                                           None)

            if pipe == constants.WIN_PIPE_INVALID_HANDLE_VALUE:
                logging.debug("Not able to create named pipe: {}".format(
                    ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))
                continue

            if not ctypes.windll.kernel32.DisconnectNamedPipe(pipe):
                logging.debug("Unable to disconnect previously connected clients from named pipe: {}".format(
                    ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

            logging.debug("Connecting to named pipe...")
            ok = ctypes.windll.kernel32.ConnectNamedPipe(pipe, None)
            logging.debug("Connected status is: {}".format(str(ok)))

            if ok == 0:
                logging.debug("Not able to connect named pipe: ({}) {}".format(ok,
                    ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

                # If a client connects before ConnectNamedPipe is called, it returns zero and GetLastError returns
                # ERROR_PIPE_CONNECTED. This can happen if a client connects in the interval between the call to
                # CreateNamedPipe and the call to ConnectNamedPipe. In this situation, there is a good connection
                # between client and server, even though the function returns zero.
                if ctypes.windll.kernel32.GetLastError() == constants.WIN_PIPE_ERROR_PIPE_CONNECTED:
                    ok = 1

            if ok != 0:
                dwThreadId = ctypes.c_ulong(0)
                thread = ctypes.windll.kernel32.CreateThread(None, 0, thread_func, pipe, 0, ctypes.byref(dwThreadId))
                if thread == -1:
                    logging.warning("Create windows pipe thread failed")
                    return 0
                else:
                    # Close the thread handle
                    if not ctypes.windll.kernel32.CloseHandle(thread):
                        logging.debug("Unable to close handle: {}".format(
                            ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

                logging.debug("Threading logic fin")
            else:
                logging.warning("Unable to connect windows pipe.")

                if not ctypes.windll.kernel32.CloseHandle(pipe):
                    logging.debug("Unable to close handle: {}".format(
                        ctypes.FormatError(ctypes.windll.kernel32.GetLastError())))

    def _run_submit2gm_server_unix(self) -> None:
        import os
        from submit2gm.submit2gm_unix_server import ThreadedUnixStreamServer

        # Remove the old file socket
        try:
            os.unlink(constants.UNIX_FILE_SOCKET_ADDRESS)
        except FileNotFoundError:
            pass

        with ThreadedUnixStreamServer(constants.UNIX_FILE_SOCKET_ADDRESS, self._job_queue) as server:
            self._unix_server = server
            server.serve_forever()

    def _run_submit2gm_server(self) -> None:
        from submit2gm.worker_job import WorkerJob

        # Add the current commandline args to the job queue.
        if len(sys.argv) > 1:
            job = WorkerJob(sys.argv)
            self._job_queue.put(job)

        if os.name == 'nt':
            self._run_submit2gm_server_windows()
        else:
            self._run_submit2gm_server_unix()

    def run(self):
        try:
            self._run_submit2gm_server()
        except Exception as e:
            self.err = e

    def terminate(self):
        logging.debug("Attempting to terminate Submit2gm server thread...")
        self._force_stop = True

        try:
            if os.name == 'nt':
                # Open the pipe so that pipe server stops blocking
                import ctypes

                logging.debug("Send stop blocking signal")
                pipe = ctypes.windll.kernel32.CreateFileA(constants.WIN_PIPE_NAME,
                                                          constants.WIN_PIPE_GENERIC_READ | constants.WIN_PIPE_GENERIC_WRITE,
                                                          0,
                                                          None,
                                                          constants.WIN_PIPE_OPEN_EXISTING,
                                                          0,
                                                          None)

                logging.debug("Cleaning up windows pipe...")
                ctypes.windll.kernel32.FlushFileBuffers(pipe)
                ctypes.windll.kernel32.DisconnectNamedPipe(pipe)
                ctypes.windll.kernel32.CloseHandle(pipe)
                logging.debug("Finished cleaning up windows pipe")

            else:
                if hasattr(self, '_unix_server') and self._unix_server:
                    self._unix_server.shutdown()
                    self._unix_server.socket.close()
        except Exception as e:
            logging.debug(str(e))

        logging.debug("Submit2gm server thread termination finished")
